{
        Program Name : Allocation_Cluster_Manager
        Package Name : Subroutine Library Layer 2
        Purpose      : Manage a chain of allocation clusters
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Oct-2015
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Manages a chain of allocation clusters on a store.
}

unit Allocation_Cluster_Manager ;

interface

{$I EDEFINES.INC}

uses // Subroutine Library - Layer 0...
     COMInter, // TCommon_COM_Interface
     MStores, // TCOM_Managed_Store64
     TypeDefs, { Max_Memory }
     _Types, { Ansistring }
{$IFDEF CC_Track_Resources}
     ResMon,
{$ENDIF}
{$IFDEF CC_Regression_Test}
     _TestInt, // TTest_Interface
{$ENDIF}

     // Subroutine Library - Layer 1...
     Heaps, // TCOM_Heap64
     _Stores, // TStore_Address64
     _UE ; // TUnified_Exception

const ACM_Error_Cannot_Allocate_Cluster = 1 ;
      ACM_Error_Corrupted_File_System_Structure = 2 ;
      ACM_Error_Data_Expansion_Failed = 3 ;
      ACM_Error_Data_Truncation_Failed = 4 ;
      ACM_Error_Data_Insufficient_Memory = 5 ;
      ACM_Error_Data_No_Room_On_Store = 6 ;

type TAllocation_Record64 = array[ 0..Max_Structure_Size div sizeof( TStore_Address64 ) ] of TStore_Address64 ;
     PAllocation_Record64 = ^TAllocation_Record64 ;
     { An allocation record is two or more TStore_Addresses.  The last address
       in the record is the pointer to the next allocation record in the list. }

type TCOM_Allocation_Cluster_Manager64 = class( TCommon_COM_Interface )
                                             public // Constructors and destructors...
                                                 constructor Create ;
                                                 destructor Destroy ; override ;

                                             private // Instance data...
                                                 _Buffer : PAllocation_Record64 ; // Work buffer
                                                 _Buffer_Size : longint ; // Size of memory pointed to by _Buffer
                                                 _Current_AC : longint ; // Current Allocation Cluster in _Buffer
                                                 _Current_AC_Pointer : TStore_Address64 ; // Address of data for _Current_AC
                                                 _Clustersize : TStore_Size64 ; // Size of allocation clusters
                                                 _Max_Index : int64 ; // Maximum index in a cluster
                                                 _Root : TStore_Address64 ; // Root of allocation clusters (first allocation cluster)
                                                 _Size : TStore_Size64 ; // Size of data represented by allocation clusters
                                                 _Store : TCOM_Managed_Store64 ; // Store upon which the allocation clusters are stored
                                                 _Turns : int64 ;
{$IFDEF _64BIT}
                                                 _Heap : TCOM_Heap64 ;
{$ELSE}
                                                 _Heap : TCOM_Heap ;
{$ENDIF}

                                             protected // Internal utility routines...
                                                 function Buffer : PAllocation_Record64 ;
                                                 procedure Pull( AC : longint ) ;
                                                 procedure Read( P : TStore_Address64 ) ;
                                                 procedure Write( P : TStore_Address64 ) ;
                                                 function Allocate_Allocation_Cluster : TStore_Address64 ;
                                                 procedure Deallocate_Allocation_Cluster( P : TStore_Address64 ) ;

                                             public // API...
                                                 function Is_Class( Name : PAnsiChar ) : boolean ;
                                                     override ;

                                                 function Get_Clustersize : TStore_Size64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 procedure Set_Clustersize( Value : TStore_Size64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
{$IFDEF _64BIT}
                                                 function Get_Heap : TCOM_Heap64 ;
                                                     virtual ;
  {$IFDEF Use_STDCALL}
                                                     stdcall ;
  {$ENDIF}
                                                 procedure Set_Heap( Value : TCOM_Heap64 ) ;
                                                     virtual ;
  {$IFDEF Use_STDCALL}
                                                     stdcall ;
  {$ENDIF}
{$ELSE}
                                                 function Get_Heap : TCOM_Heap ;
                                                     virtual ;
  {$IFDEF Use_STDCALL}
                                                     stdcall ;
  {$ENDIF}
                                                 procedure Set_Heap( Value : TCOM_Heap ) ;
                                                     virtual ;
  {$IFDEF Use_STDCALL}
                                                     stdcall ;
  {$ENDIF}
{$ENDIF}
                                                 function Get_Root : TStore_Address64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 procedure Set_Root( Value : TStore_Address64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 procedure Set_Root_And_Size( R : TStore_Address64 ;
                                                     S : TStore_Size64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 procedure Set_Root_And_Cluster_Count( R : TStore_Address64 ;
                                                     S : TStore_Size64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 function Get_Size : TStore_Size64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                procedure Set_Size( Value : TStore_Size64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 function Get_Cluster_Count : TStore_Size64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                procedure Set_Cluster_Count( Value : TStore_Size64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                procedure Prepare_Size( Value : TStore_Size64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 function Get_Store : TCOM_Managed_Store64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 procedure Set_Store( Value : TCOM_Managed_Store64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}

                                                 function Offset_To_Pointer( Offset : TStore_Size64 ) : TStore_Address64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}

                                                 function Get_Allocation_Cluster( I : longint ) : TStore_Address64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}

                                                 procedure Set_Allocation_Cluster( I : longint ;
                                                     Cluster : TStore_Address64 ) ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 function Next_ACM_Cluster( P : int64 ) : int64 ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}
                                                 function Rebuild : boolean ;
                                                     virtual ;
{$IFDEF Use_STDCALL}
                                                     stdcall ;
{$ENDIF}

                                             public // Properties...
                                                 property Cluster_Count : TStore_Size64
                                                     read Get_Cluster_Count
                                                     write Set_Cluster_Count ;
                                                     
                                                 property Clustersize : TStore_Size64
                                                     read Get_Clustersize
                                                     write Set_Clustersize ;

{$IFDEF _64BIT}
                                                 property Heap : TCOM_Heap64
                                                     read Get_Heap
                                                     write Set_Heap ;
{$ELSE}
                                                 property Heap : TCOM_Heap
                                                     read Get_Heap
                                                     write Set_Heap ;
{$ENDIF}

                                                 property Size : TStore_Size64
                                                     read Get_Size
                                                     write Set_Size ;

                                                 property Store : TCOM_Managed_Store64
                                                     read Get_Store
                                                     write Set_Store ;

                                                 property Turns : int64
                                                     read _Turns
                                                     write _Turns ;
                                         end ; // TCOM_Allocation_Cluster_Manager64

{$IFDEF CC_Regression_Test}
procedure Test_Unit( T : TTest_Interface ) ;
{$ENDIF}

var Total_Turns : int64 = 0 ; // Number of allocation cluster turns


implementation

uses // Pascal...
     SysUtils, // Lowercase
{$IFDEF CC_Debug}
     Classes, // TList
{$ENDIF}

     // Subroutine Library - Layer 1...
     UE, // TStandard_Error_Interface

     // Subroutine Library - Layer 2...
     CCStandard
{$IFDEF UOS}
     ,UOS_Util // TInteger_List64
{$ENDIF}

{$IFDEF CC_Regression_Test}
     ,Forms
{$ENDIF}
     ;

{$IFDEF CC_Debug}
var Allocated_CAs : TList = nil ;
    Other_Allocated : TList = nil ;
    Check_Allocated_CAs : boolean = False ;
{$ENDIF}

{$IFDEF CC_Regression_Test}
var ACM : TCOM_Allocation_Cluster_Manager64 ;
    F, FI : textfile ;
    Buff : array[ 0..511 ] of byte ;
    Buffer : array[ 0..63 ] of int64 ;
    Buffer_Size : integer ;
    H : TCOM_Heap ;
    Old_P, P, P1 : int64 ;
    S64 : TCOM_Managed_Store64 ;

procedure Show_ACs( Force : boolean = false ) ;

var C, I : longint ;
    P : TStore_Address64 ;

begin
    P := ACM.Get_Root ; // Start with first allocation cluster
    if( P = 0 ) then // Nothing allocated yet
    begin
        writeln( f, '  No allocation clusters' ) ;
        exit ;
    end ;
    C := 0 ;
    while( P <> 0 ) do // While more clusters
    begin
{$IFDEF CC_Debug}
        if( Check_Allocated_CAs ) then
        begin
            if( Allocated_CAs.Indexof( pointer( P ) ) = -1 ) then
            begin
                halt ;
            end ;
        end ;
{$ENDIF}
        S64.Read( P, Buffer_Size, Buffer ) ; // Get allocation cluster
        writeln( F, '  Cluster at ', P, ':' ) ;
        for I := 0 to Buffer_Size div 8 - 2 do // For each pointer in this cluster
        begin
            writeln( F, '    ', I, ' (', C, '): ', Buffer[ I ] ) ;
            inc( C ) ;
        end ;
        P := Buffer[ Buffer_Size div 8 - 1 ] ; // Follow link to next allocation cluster
        writeln( F, '    next cluster=', P ) ;
    end ; // while( P <> 0 )
    closefile( F ) ;
    append( F ) ;
end ; // Show_ACs


procedure Validate( Root, Current_AC, Current_AC_Pointer, Buffer_Size : int64 ;
    Store : TCOM_Managed_Store64 ) ;

var AC : longint ;
    Buff : array[ 0..512 ] of int64 ;
    _Root : int64 ;
    _Max : longint ;

begin
    if( Current_AC = -1 ) then
    begin
        exit ;
    end ;
    _Max := ( Buffer_Size div 8 ) - 1 ;
    _Root := Root ;
    if( _Root = 0 ) then
    begin
        if( Current_AC <> -1 ) then
        begin
            halt ; // No root requires Current_AC = -1
        end ;
        exit ;
    end ;
    AC := 0 ;
    while( _Root <> 0 ) do
    begin
        if( Current_AC_Pointer = _Root ) then
        begin
            if( AC = Current_AC ) then
            begin
                exit ; // Match found, everything okay
            end ;
            Show_ACs ;
            halt ; // Mismatch between Current_AC and Current_AC_Pointer
        end ;
        Store.Read( _Root, Buffer_Size, Buff ) ;
        _Root := Buff[ _Max ] ;
        inc( AC ) ;
    end ;
    Show_ACs ;
    halt ; // Current_AC_Pointer not found
end ;
{$ENDIF}


function ERT( C : longint ) : string ;

begin
    case C of
        ACM_Error_Cannot_Allocate_Cluster : Result := 'Cannot allocate cluster' ;
        ACM_Error_Corrupted_File_System_Structure : Result := 'Corrupt file system structure' ;
        ACM_Error_Data_Expansion_Failed : Result := 'Data expansion failed' ;
        ACM_Error_Data_Truncation_Failed : Result := 'Data truncation failed' ;
        else Result := 'Unknown error' ;
    end ;
end ;


type tAllocated_Cluster_Manager_Unified_Exception = class( TStandard_Error_Interface )
                                                        private
                                                            _Code : longint ;
                                                            _Text : Ansistring ;
                                                            _Previous : tUnified_Exception ;

                                                        public
                                                            function Get_Facility : longint ; override ;
                                                            function Get_Facility_Version : longint ;
                                                                override ;
                                                            function Version : longint ; override ;
                                                            function Severity : longint ; override ;
                                                            function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                                                override ;
                                                            function Get_Error : longint ; override ;
                                                            function Get_Previous : tUnified_Exception ;
                                                                override ;
                                                            procedure Terminate ; override ;
                                                    end ;


function tAllocated_Cluster_Manager_Unified_Exception.Get_Facility : longint ;

begin
    Result := 133 ;
end ;


function tAllocated_Cluster_Manager_Unified_Exception.Get_Facility_Version : longint ;

begin
    Result := 10 ;
end ;


function tAllocated_Cluster_Manager_Unified_Exception.Version : longint ;

begin
    Result := 10 ;
end ;


function tAllocated_Cluster_Manager_Unified_Exception.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function tAllocated_Cluster_Manager_Unified_Exception.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

begin
    _Text := ERT( _Code ) ;
    Size := length( _Text ) ;
    Typ := 0 ;
    Result := PAnsiChar( _Text ) ;
end ;


function tAllocated_Cluster_Manager_Unified_Exception.Get_Error : longint ;

begin
    Result := _Code ;
end ;


function tAllocated_Cluster_Manager_Unified_Exception.Get_Previous : tUnified_Exception ;

begin
    Result := _Previous ;
end ;


procedure tAllocated_Cluster_Manager_Unified_Exception.Terminate ;

begin
    Free ;
end ;



function Create_Exception( C : longint ; E : TUnified_Exception ) : tAllocated_Cluster_Manager_Unified_Exception ;

begin
    Result := tAllocated_Cluster_Manager_Unified_Exception.Create ;
    Result._Code := C ;
    Result._Previous := E ;
end ;



// TCOM_Allocation_Cluster_Manager64 nethods...

// Constructors and destructors...

constructor TCOM_Allocation_Cluster_Manager64.Create ;

begin
    inherited Create ;

    _Current_AC := -1 ;
end ;


destructor TCOM_Allocation_Cluster_Manager64.Destroy ;

begin
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    freemem( _Buffer ) ;
    Set_Heap( nil ) ;

    inherited Destroy ;
end ;


// Internal utility routines...

function TCOM_Allocation_Cluster_Manager64.Buffer : PAllocation_Record64 ;

begin
    Result := nil ; // To get rid of compiler warnings
    if( _Buffer_Size = 0 ) then // Buffer not already created
    begin
        _Buffer_Size := ( _Max_Index + 1 ) * sizeof( TStore_Address64 ) ;
        _Buffer := allocmem( _Buffer_Size ) ;
{$IFDEF CC_Track_Resources}
        Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        if( _Buffer = nil ) then // Couldn't allocate memory
        begin
            Set_Last_Error( Create_Exception( ACM_Error_Data_Insufficient_Memory, nil ) ) ;
            exit ;
        end ;
        fillchar( _Buffer^, _Buffer_Size, 0 ) ;
    end ;
    Result := _Buffer ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Read( P : TStore_Address64 ) ;

begin
{$IFDEF CC_Debug}
    if( Check_Allocated_CAs ) then
    begin
        if( Allocated_CAs.Indexof( pointer( P ) ) = -1 ) then
        begin
            halt ;
        end ;
    end ;
{$ENDIF}
    _Store.Read( P, _Buffer_Size, Buffer^ ) ;
    if( _Store.Last_Error <> nil ) then // Couldn't allocate memory
    begin
        Set_Last_Error( _Store.Last_Error ) ;
        exit ;
    end ;
    if( _Current_AC_Pointer <> P ) then
    begin
        _Current_AC := -1 ;
    end ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Write( P : TStore_Address64 ) ;

begin
{$IFDEF CC_Debug}
    if( Check_Allocated_CAs ) then
    begin
        if( Allocated_CAs.Indexof( pointer( P ) ) = -1 ) then
        begin
            halt ;
        end ;
    end ;
{$ENDIF}
    _Store.Write( P, _Buffer_Size, Buffer^ ) ;
    if( _Store.Last_Error <> nil ) then // Couldn't write to store
    begin
        Set_Last_Error( _Store.Last_Error ) ;
        exit ;
    end ;
    if( _Current_AC_Pointer <> P ) then
    begin
        _Current_AC := -1 ;
    end ;
end ;


function TCOM_Allocation_Cluster_Manager64.Allocate_Allocation_Cluster : TStore_Address64 ;

begin
    Result := _Store.Allocate( _Buffer_Size ) ;
    if( Result <> 0 ) then
    begin
{$IFDEF CC_Debug}
        if( Check_Allocated_CAs ) then
        begin
            if( Allocated_CAs.Indexof( pointer( Result ) ) <> -1 ) then
            begin
                halt ;
            end ;
            Allocated_CAs.Add( pointer( Result ) ) ;
        end ;
{$ENDIF}
        _Store.Fill( Result, _Buffer_Size, 0 ) ;
    end else
    begin
        Set_Last_Error( Create_Exception( ACM_Error_Data_No_Room_On_Store, nil ) ) ;
    end ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Deallocate_Allocation_Cluster( P : TStore_Address64 ) ;

begin
{$IFDEF CC_Debug}
    if( Check_Allocated_CAs ) then
    begin
        if( Allocated_CAs.Indexof( pointer( P ) ) = -1 ) then
        begin
            halt ;
        end ;
        Allocated_CAs.Remove( pointer( P ) ) ;
    end ;
{$ENDIF}
    _Store.Deallocate( P, _Buffer_Size ) ;
    if( _Store.Last_Error <> nil ) then
    begin
        Set_Last_Error( _Store.Last_Error ) ;
    end ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Pull( AC : longint ) ;

var Original, P : TStore_Address64 ;

begin
    Buffer ; // Make sure buffer is allocated
    if( _Last_Error <> nil ) then
    begin
        exit ;
    end ;
{$IFDEF CC_Regression_Test}
    Validate( _Root, _Current_AC, _Current_AC_Pointer, _Buffer_Size, _Store ) ;
{$ENDIF}
    if( AC < 0 ) then
    begin
        exit ; // Not a valid cluster index
    end ;
    if( _Current_AC = AC ) then
    begin
        exit ; // Already have it
    end ;
    Original := AC ;

    if( ( _Current_AC >= 0 ) and ( AC > _Current_AC ) ) then // After our current AC
    begin
        AC := AC - _Current_AC - 1 ; // How much further to go
        P := Buffer^[ _Max_Index ] ;
    end else
    begin
        P := _Root ; // Start from the beginning
    end ;
    while( AC >= 0 ) do
    begin
        if( P = 0 ) then
        begin
            Set_Last_Error( Create_Exception( ACM_Error_Corrupted_File_System_Structure, nil ) ) ;
            exit ;
        end ;
        dec( AC ) ;
        Read( P ) ;
        _Current_AC_Pointer := P ;
        P := Buffer^[ _Max_Index ] ;
        Turns := Turns + 1 ;
        inc( Total_Turns ) ;
    end ;
    _Current_AC := Original ;
{$IFDEF CC_Regression_Text}
    if( _Current_AC_Pointer = 0 ) then
    begin
        halt ;
    end ;
    Validate( _Root, _Current_AC, _Current_AC_Pointer, _Buffer_Size, _Store ) ;
{$ENDIF}
end ; // TCOM_Allocation_Cluster_Manager64.Pull


// API...

function TCOM_Allocation_Cluster_Manager64.Is_Class( Name : PAnsiChar ) : boolean ;

var S : string ;

begin
    S := Name ;
    Result := lowercase( S ) = 'tcom_allocation_cluster_manager64' ;
end ;


{$IFDEF _64BIT}
function TCOM_Allocation_Cluster_Manager64.Get_Heap : TCOM_Heap64 ;
{$ELSE}
function TCOM_Allocation_Cluster_Manager64.Get_Heap : TCOM_Heap ;
{$ENDIF}

begin
    Set_Last_Error( nil ) ;
    if( _Heap = nil ) then
    begin
{$IFDEF _64BIT}
        Set_Heap( TCOM_Heap64.Create ) ;
{$ELSE}
        Set_Heap( TCOM_Heap.Create ) ;
{$ENDIF}
    end ;
    Result := _Heap ;
end ;


{$IFDEF _64BIT}
procedure TCOM_Allocation_Cluster_Manager64.Set_Heap( Value : TCOM_Heap64 ) ;
{$ELSE}
procedure TCOM_Allocation_Cluster_Manager64.Set_Heap( Value : TCOM_Heap ) ;
{$ENDIF}

begin
    Set_Last_Error( nil ) ;
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Heap <> nil ) then
    begin
{$IFDEF CC_Track_Resources}
        Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        freemem( _Buffer ) ;
        _Buffer := nil ;
        _Buffer_Size := 0 ;
        _Heap.Detach ;
    end ;
    _Heap := Value ;
end ;


function TCOM_Allocation_Cluster_Manager64.Get_Clustersize : TStore_Size64 ;

begin
    Set_Last_Error( nil ) ;
    Result := _Clustersize ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Set_Clustersize( Value : TStore_Size64 ) ;

begin
    Set_Last_Error( nil ) ;
    if( _Clustersize <> Value ) then // If anything changed
    begin
        _Clustersize := Value ;
        Set_Root( _Root ) ; // Recalculate size
    end ;
end ;


function TCOM_Allocation_Cluster_Manager64.Get_Cluster_Count : TStore_Size64 ;

begin
    Result := Size div Clustersize ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Set_Cluster_Count( Value : TStore_Size64 ) ;

begin
    Size := Value * Clustersize ;
end ;


function TCOM_Allocation_Cluster_Manager64.Get_Root : TStore_Address64 ;

begin
    Set_Last_Error( nil ) ;
    Result := _Root ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Set_Root( Value : TStore_Address64 ) ;

var I : integer ;

begin
    Set_Last_Error( nil ) ;
    if( _Store = nil ) then // Makes no sense to set the root if there is no store
    begin
        exit ;
    end ;

    // Set root and recalculate size...
    _Root := Value ;
    _Size := 0 ;
    Buffer ; // Make sure the buffer is allocated
    while( Value <> 0 ) do
    begin
        Read( Value ) ; // Get allocation cluster
        if( _Last_Error <> nil ) then
        begin
            exit ;
        end ;
        for I := 0 to _Max_Index - 1 do
        begin
            if( Buffer^[ I ] = 0 ) then // Hit last allocated cluster
            begin
                break ; // We're done
            end ;
            _Size := _Size + _Clustersize ; // Add in another clustersize worth of bytes
        end ; // for I := 0 to Max_Index - 1
        Value := Buffer^[ _Max_Index ] ; // Follow link to next allocation cluster
        if( Value <> 0 ) then
        begin
            Turns := Turns + 1 ;
            inc( Total_Turns ) ;
        end ;
    end ; // while( Value <> 0 )
    _Current_AC := -1 ;
end ; // TCOM_Allocation_Cluster_Manager64.Set_Root


procedure TCOM_Allocation_Cluster_Manager64.Set_Root_And_Size( R : TStore_Address64 ;
    S : TStore_Size64 ) ;

begin
    Set_Last_Error( nil ) ;
    _Root := R ;
    _Size := S ;
    _Current_AC := -1 ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Set_Root_And_Cluster_Count( R : TStore_Address64 ;
    S : TStore_Size64 ) ;

begin
    Set_Root_And_Size( R, S * Clustersize ) ;
end ;


function TCOM_Allocation_Cluster_Manager64.Get_Size : TStore_Size64 ;

begin
    Set_Last_Error( nil ) ;
    Result := _Size ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Prepare_Size( Value : TStore_Size64 ) ;

var I : integer ;
    New_Size : TStore_Size64 ;
    P : TStore_Address64 ;

begin
    Set_Last_Error( nil ) ;

    // Determine max cluster for new size...
    Value := ( Value + _Clustersize - 1 ) div _Clustersize ; // Determine number of clusters required for the specified size
    New_Size := Value * _Clustersize ; // Total new size, if we succeed
    if( New_Size = _Size ) then // No size change
    begin
        exit ;
    end ;

    // Update allocation clusters for the new size...
    Buffer ; // Make sure buffer is allocated
    if( _Last_Error <> nil ) then
    begin
        exit ;
    end ;
    if( New_Size > _Size ) then
    begin
        // Expanding...
        Pull( ( _Size - 1 ) div _Clustersize div _Max_Index ) ; // Get last allocation cluster
        if( _Last_Error <> nil ) then
        begin
            exit ;
        end ;
        Value := Value - _Current_AC * _Max_Index ;
        P := _Current_AC_Pointer ;
        while( Value > 0 ) do
        begin
            for I := 0 to _Max_Index - 1 do // For each pointer in this cluster
            begin
                dec( Value ) ; // One less cluster to account for
                if( Value = 0 ) then
                begin
                    break ; // No need to process any more of the pointers in this allocation cluster
                end ;
            end ; // for I := 0 to Max_Index - 1
            if( ( Value > 0 ) and ( Buffer^[ _Max_Index ] = 0 ) ) then // Need to extend the allocation chain
            begin
                Buffer^[ _Max_Index ] := Allocate_Allocation_Cluster ;
                if( Buffer^[ _Max_Index ] = 0 ) then
                begin
                    Set_Last_Error( Create_Exception( ACM_Error_Data_Expansion_Failed, _Store.Last_Error ) ) ;
                    exit ; // Failure {UNTESTED}
                end ;
            end ;
            Write( P ) ;
            if( _Last_Error <> nil ) then
            begin
                exit ;
            end ;
            if( Value > 0 ) then
            begin
                P := Buffer^[ _Max_Index ] ;
                if( P <> 0 ) then
                begin
                    Fillchar( Buffer^, _Buffer_Size, 0 ) ;
                end ;
            end ;
        end ; // while( Value > 0 )
        _Current_AC := ( New_Size - 1 ) div _Clustersize div _Max_Index ;
{$IFDEF CC_Regression_Test}
        if( P = 0 ) then
        begin
            halt ;
        end ;
{$ENDIF}
        _Current_AC_Pointer := P ;
    end ;
end ; // TCOM_Allocation_Cluster_Manager64.Prepare_Size


procedure TCOM_Allocation_Cluster_Manager64.Set_Size( Value : TStore_Size64 ) ;

var I : longint ; // Loop indexes
    Last, P : TStore_Address64 ;
    New_Size : TStore_Size64 ;
    List : TInteger64_List ;

begin
    Set_Last_Error( nil ) ;

    // Determine max cluster for new size...
    Value := ( Value + _Clustersize - 1 ) div _Clustersize ; // Determine number of clusters required for the specified size
    New_Size := Value * _Clustersize ; // Total new size, if we succeed
    if( New_Size = _Size ) then // No size change
    begin
        exit ;
    end ;

    // Update allocation clusters for the new size...
    Buffer ; // Make sure buffer is allocated
    if( _Last_Error <> nil ) then
    begin
        exit ;
    end ;
    Last := 0 ;
    if( _Root = 0 ) then // Nothing allocated yet
    begin
        _Root := Allocate_Allocation_Cluster ;
        _Current_AC := -1 ;
        if( _Root = 0 ) then
        begin
            exit ; // Failure {UNTESTED}
        end ;
    end ;
    if( New_Size > _Size ) then
    begin
        // Expanding...
        Pull( ( _Size - 1 ) div _Clustersize div _Max_Index ) ; // Get last allocation cluster
        if( _Last_Error <> nil ) then
        begin
            exit ;
        end ;
        Value := Value - _Current_AC * _Max_Index ;
        P := _Current_AC_Pointer ;
        while( Value > 0 ) do
        begin
            for I := 0 to _Max_Index - 1 do // For each pointer in this cluster
            begin
                if( Buffer^[ I ] = 0 ) then // Unallocated cluster
                begin
                    Buffer^[ I ] := _Store.Allocate( _Clustersize ) ;
                    if( Buffer^[ I ] = 0 ) then
                    begin
                        Set_Last_Error( Create_Exception( ACM_Error_Data_Expansion_Failed, _Store.Last_Error ) ) ;
                        exit ; // Failure {UNTESTED}
                    end ;
{$IFDEF CC_Debug}
Other_Allocated.Add( pointer( Buffer^[ I ] ) ) ;
{$ENDIF}
                end ;
                dec( Value ) ; // One less cluster to account for
                if( Value = 0 ) then
                begin
                    break ; // No need to process any more of the pointers in this allocation cluster
                end ;
            end ; // for I := 0 to Max_Index - 1
            if( ( Value > 0 ) and ( Buffer^[ _Max_Index ] = 0 ) ) then // Need to extend the allocation chain
            begin
                Buffer^[ _Max_Index ] := Allocate_Allocation_Cluster ;
                if( Buffer^[ _Max_Index ] = 0 ) then
                begin
                    Set_Last_Error( Create_Exception( ACM_Error_Data_Expansion_Failed, _Store.Last_Error ) ) ;
                    exit ; // Failure {UNTESTED}
                end ;
            end ;
            Write( P ) ;
            if( _Last_Error <> nil ) then
            begin
                exit ;
            end ;
            if( Value > 0 ) then
            begin
                P := Buffer^[ _Max_Index ] ;
                if( P <> 0 ) then
                begin
                    Fillchar( Buffer^, _Buffer_Size, 0 ) ;
                end ;
            end ;
        end ; // while( Value > 0 )
        _Current_AC := ( New_Size - 1 ) div _Clustersize div _Max_Index ;
{$IFDEF CC_Regression_Test}
        if( P = 0 ) then
        begin
            halt ;
        end ;
{$ENDIF}
        _Current_AC_Pointer := P ;
    end else
    begin
        // Truncating...
        List := TInteger64_List.Create ;
        try
            Pull( ( Value - 1 ) div _Max_Index ) ; // Get new last allocation cluster
            if( _Last_Error <> nil ) then
            begin
                exit ;
            end ;
            Value := Value - _Current_AC * _Max_Index ;
            P := _Current_AC_Pointer ;
            while( P <> 0 ) do // Until remaining allocation clusters are dealt with
            begin
                for I := 0 to _Max_Index - 1 do // For each pointer in this cluster
                begin
                    if( Buffer^[ I ] = 0 ) then // Unallocated cluster
                    begin
                        break ; // No need to go through the rest of the pointers once we hit a 0
                    end ;
                    if( Value > 0 ) then // Until we reach the
                    begin
                        dec( Value ) ;
                    end else
                    begin
                        List.Add( Buffer^[ I ] ) ; // Mark the cluster for release
                        Buffer^[ I ] := 0 ;
                    end ; // if( Value > 0 )
                end ; // for I := 0 to _Max_Index - 1

                Last := Buffer^[ _Max_Index ] ;
                if( Buffer^[ 0 ] = 0 ) then // This allocation cluster is now unused
                begin
                    Deallocate_Allocation_Cluster( P ) ; // Remove allocation cluster
                    if( _Last_Error <> nil ) then
                    begin
                        exit ;
                    end ;
                end else
                if( Value <= 0 ) then
                begin
                    Buffer^[ _Max_Index ] := 0 ; // No more allocation clusters are needed for this file
                    Write( P ) ; // Update on-store cluster immediately for safety
                    if( _Last_Error <> nil ) then
                    begin
                        exit ;
                    end ;
                end ;

                // Deallocate the data clusters...
                for I := 0 to List.Count - 1 do
                begin
                    if( List[ I ] <> 0 ) then
                    begin
                        _Store.Deallocate( List[ I ], _Clustersize ) ;
                        if( _Store.Last_Error <> nil ) then
                        begin
                            Set_Last_Error( Create_Exception( ACM_Error_Data_Truncation_Failed, _Store.Last_Error ) ) ;
                            exit ;
                        end ;
                    end ;
{$IFDEF CC_Debug}
if( Other_Allocated.Indexof( pointer( List[ I ] ) ) = -1 ) then
begin
    halt ;
end ;
Other_Allocated.Remove( pointer( List[ I ] ) ) ;
{$ENDIF}
                end ; // for I := 0 to List.Count - 1
                List.Clear ;

                _Size := New_Size ;
                // Once we update even one of the allocation clusters, that means
                // our new size is valid, even if we haven't finished removing all other ACs

                // Move to next allocation cluster in chain...
                if( ( Value <= 0 ) and ( Last = 0 ) ) then // Done with the truncation
                begin
                    Last := P ;
                    break ;
                end ;
                P := Last ;
                if( P <> 0 ) then
                begin
                    Turns := Turns + 1 ;
                    inc( Total_Turns ) ;
                    Read( P ) ;
                end ;
            end ; // while( P <> 0 )
        finally
            List.Free ;
        end ;
        if( Buffer^[ 0 ] = 0 ) then // This allocation cluster is now unused or full of random data...
        begin
            _Current_AC := -1 ; // Reset buffer cache state
        end else
        begin
            _Current_AC := ( New_Size - 1 ) div _Clustersize div _Max_Index ;
{$IFDEF CC_Regression_Test}
            if( Last = 0 ) then
            begin
                halt ;
            end ;
{$ENDIF}
            _Current_AC_Pointer := Last ;
        end ;
    end ; // if( New_Size > _Size )
    _Size := New_Size ;
    if( New_Size = 0 ) then
    begin
        _Root := 0 ;
        _Current_AC := -1 ;
    end ;
end ; // TCOM_Allocation_Cluster_Manager64.Set_Size


function TCOM_Allocation_Cluster_Manager64.Get_Store : TCOM_Managed_Store64 ;

begin
    Result := _Store ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Set_Store( Value : TCOM_Managed_Store64 ) ;

begin
    Set_Last_Error( nil ) ;
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Store <> nil ) then
    begin
        _Store.Detach ;
    end ;
    _Store := Value ;
    _Size := 0 ;
    _Root := 0 ;
    _Current_AC := -1 ;
    if( Value <> nil ) then
    begin
        // Default the allocation cluster and file cluster size to the store clustersize...
        _Clustersize := Value.Min_Storage ;
        if( _Clustersize < 128 ) then {ELSE UNTESTED}
        begin
            _Clustersize := 128 ;
        end ;
        _Max_Index := ( _Clustersize div sizeof( TStore_Address64 ) ) - 1 ;
    end ;
end ;


// API...

function TCOM_Allocation_Cluster_Manager64.Offset_To_Pointer( Offset : TStore_Size64 ) : TStore_Address64 ;

var AC, C : longint ;

begin
    // Sanity check...
    Result := 0 ; // Assume failure
    Set_Last_Error( nil ) ;
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        exit ;
    end ;

    // Determine cluster for offset...
    C := Offset div _Clustersize ; // Determine cluster for offset
    AC := C div _Max_Index ; // Allocation cluster index
    Offset := C mod _Max_Index ; // Pointer index within allocation cluster index

    // Read the allocation cluster...
    Pull( AC ) ;
    if( _Last_Error <> nil ) then
    begin
        exit ;
    end ;

    // Return the disk pointer...
    Result := Buffer^[ Offset ] ;
{$IFDEF CC_Debug}
    if( Check_Allocated_CAs ) then
    begin
        if( Allocated_CAs.Indexof( pointer( Result ) ) <> -1 ) then
        begin
            halt ;
        end ;
    end ;
{$ENDIF}
end ; // TCOM_Allocation_Cluster_Manager64.Offset_To_Pointer


function TCOM_Allocation_Cluster_Manager64.Get_Allocation_Cluster( I : longint ) : TStore_Address64 ;

var AC, Offset : cardinal ;

begin
    Set_Last_Error( nil ) ;
    Result := 0 ;
    if( I < 0 ) then
    begin
        exit ;
    end ;
    if( Size < I * _Clustersize ) then
    begin
        exit ;
    end ;
    AC := I div _Max_Index ; // Allocation cluster index
    Offset := I mod _Max_Index ; // Pointer index within allocation cluster index
    Pull( AC ) ;
    if( _Last_Error <> nil ) then
    begin
        exit ;
    end ;
    Result := Buffer^[ Offset ] ;
end ;


procedure TCOM_Allocation_Cluster_Manager64.Set_Allocation_Cluster( I : longint ;
     Cluster : TStore_Address64 ) ;

var AC, Offset : cardinal ;

begin
    Set_Last_Error( nil ) ;
    if( I < 0 ) then
    begin
        exit ;
    end ;
    if( Size < I * _Clustersize ) then
    begin
        Size := I * _Clustersize ;
    end ;
    AC := I div _Max_Index ; // Allocation cluster index
    Offset := I mod _Max_Index ; // Pointer index within allocation cluster index
    Pull( AC ) ;
    if( _Last_Error <> nil ) then
    begin
        exit ;
    end ;
    Buffer^[ Offset ] := Cluster ;
    Write( _Current_AC_Pointer ) ;
end ;


function TCOM_Allocation_Cluster_Manager64.Next_ACM_Cluster( P : int64 ) : int64 ;

begin
    Result := 0 ;
    Set_Last_Error( nil ) ;
    if( P < 0 ) then
    begin
        exit ;
    end ;
    _Current_AC_Pointer := -1 ;
    Read( P ) ;
    Result := Buffer^[ _Max_Index ] ;
end ;


function TCOM_Allocation_Cluster_Manager64.Rebuild : boolean ;

var Cluster, P : int64 ;

    function Valid_Pointer( P : int64 ) : boolean ;

    begin
        Result := True ; // Assume it is good
        if( P = 0 ) then // Nil is always valid
        begin
            exit ;
        end ;
        if(
            ( P < Store.Min_Storage ) // Too small
            or
            ( P > Store.Max_Storage )
            or
            ( ( P mod Store.Min_Storage ) <> 0 ) // not even
          ) then
        begin
            Result := False ;
        end ;
    end ;


    procedure Set_Corrupted ;

    begin
        Result := True ; // For now, just return true
    end ;

begin // TCOM_Allocation_Cluster_Manager64.Rebuild
    Result := False ;

    // Allocate the ACM blocks...
    P := _Root ;
    while( P <> 0 ) do
    begin
        if( not Valid_Pointer( P ) ) then
        begin
            Set_Corrupted ;
            Result := True ;
            exit ;
        end ;
        if( not Store.Allocate_At( P, Clustersize ) ) then
        begin
            // Cluster already allocated.  Something has corrupted the file structure
            Result := True ;
        end ;
        P := Next_ACM_Cluster( P ) ;
    end ;

    // Allocate the clusters for the chain...
    Cluster := 0 ;
    P := Get_Allocation_Cluster( Cluster ) ;
    while( P > 0 ) do // Allocate for each cluster of the chain
    begin
        if( not Valid_Pointer( P ) ) then
        begin
            Set_Corrupted ;
            Result := True ;
        end else
        if( not Store.Allocate_At( P, Clustersize ) ) then
        begin
            // Cluster already allocated.  Something has corrupted the file structure
            Result := True ;
        end ;
        inc( Cluster ) ;
        P := Get_Allocation_Cluster( Cluster ) ;
    end ;
end ; // TCOM_Allocation_Cluster_Manager64.Rebuild


// Testing support...

{$IFDEF CC_Regression_Test}
procedure error( s : string ) ;

begin
    halt ;
end ;


var OpCount : int64 = 0 ;

procedure Test_Unit( T : TTest_Interface ) ;

    function Status( const S : string ) : boolean ;

    begin
        Result := T.Log_Operation( S ) ;
        writeln( f, s ) ;
        closefile( f ) ;
        append( f ) ;
    end ;


    function Do_Size( P : int64 ) : boolean ;

    begin
        Result := Status( 'S' + inttostr( P ) ) ;
        ACM.Set_Size( P ) ;
        if( P = 0 ) then
        begin
  {$IFDEF CC_Debug}
            if( Check_Allocated_CAs ) then
            begin
                if( Allocated_CAs.Count + Other_Allocated.Count > 0 ) then
                begin
                    halt ;
                end ;
            end ;
  {$ENDIF}
            rewrite( f ) ;
        end ;
        Show_ACs ;
    end ; // Do_Size


    function Do_Read( P : int64 ) : boolean ;

    var Index : longint ;
        P1 : int64 ;

    begin
        Result := Status( 'R' + inttostr( P ) ) ;
        P1 := ACM.Offset_To_Pointer( P * 512 ) ;
        if( P1 <> 0 ) then
        begin
            S64.Read( P1, 512, Buff ) ;
            for Index := 0 to 511 do
            begin
                if( Buff[ Index ] <> P ) then
                begin
                    Show_ACs( true ) ;
                    ACM.Offset_To_Pointer( P * 512 ) ;
                    halt ;
                end ;
            end ;
        end ; // if( P1 <> 0 )
    end ; // Do_Read


var Dummy : integer ;
  {$IFDEF CC_Debug}
    CA : integer ;
  {$ENDIF}
    S : string ;

begin
    // Coverage test...
    Check_Allocated_CAs := True ;
    ACM := TCOM_Allocation_Cluster_Manager64.Create ;
    S64 := TCOM_Heap64.Create ;
    if( ACM.Get_Store <> nil ) then
    begin
        error( 'Invalid store' ) ;
    end ;
    ACM.Set_Store( S64 ) ;
    if( ACM.Get_Store <> S64 ) then
    begin
        error( 'Invalid store' ) ;
    end ;
    H := TCOM_Heap.Create ;
    ACM.Set_Heap( H ) ;
    if( ACM.Get_Clustersize <> 128 ) then
    begin
        error( 'Invalid clustersize' ) ;
    end ;
    if( ACM.Get_Heap <> H ) then
    begin
        error( 'Invalid heap' ) ;
    end ;
    if( ACM.Get_Size <> 0 ) then
    begin
        error( 'Invalid size' ) ;
    end ;
    if( ACM.Get_Root <> 0 ) then
    begin
        error( 'Invalid root' ) ;
    end ;
    ACM.Set_Clustersize( 512 ) ;
    if( ACM.Get_Clustersize <> 512 ) then
    begin
        error( 'Invalid clustersize' ) ;
    end ;
    if( ACM.Get_Size <> 0 ) then
    begin
        error( 'Invalid size' ) ;
    end ;
    ACM.Set_Size( 65536 ) ;
    if( ACM.Get_Size <> 65536 ) then
    begin
        error( 'Invalid size' ) ;
    end ;
    ACM.Set_Size( 65536 ) ;
    if( ACM.Get_Size <> 65536 ) then
    begin
        error( 'Invalid size' ) ;
    end ;
    ACM.Set_Size( 256 ) ;
    if( ACM.Get_Size <> 512 ) then
    begin
        error( 'Invalid size' ) ;
    end ;
    if( ACM.Offset_To_Pointer( 1024000 ) <> 0 ) then
    begin
        error( 'Invalid translation' ) ;
    end ;
    ACM.Set_Size( 65536 ) ;
    P := ACM.Offset_To_Pointer( 32768 ) ;
    if( P = 0 ) then
    begin
        error( 'Invalid translation' ) ;
    end ;
    fillchar( Buff, 512, 1 ) ;
    S64.Write( P, 512, Buff ) ;
    P := ACM.Offset_To_Pointer( 65500 ) ;
    if( P = 0 ) then
    begin
        error( 'Invalid translation' ) ;
    end ;
    fillchar( Buff, 512, 2 ) ;
    S64.Write( P, 512, Buff ) ;
    P := ACM.Offset_To_Pointer( 32768 ) ;
    S64.Read( P, 512, Buff ) ;
    if( Buff[ 0 ] <> 1 ) then
    begin
        error( 'Invalid data' ) ;
    end ;
    P := ACM.Offset_To_Pointer( 32769 ) ;
    if( P = 0 ) then
    begin
        error( 'Offset_To_Pointer error' ) ;
    end ;
    P := ACM.Offset_To_Pointer( 65500 ) ;
    S64.Read( P, 512, Buff ) ;
    if( Buff[ 0 ] <> 2 ) then
    begin
        error( 'Invalid data' ) ;
    end ;
    if( ACM.Get_Root = 0 ) then
    begin
        error( 'Invalid root' ) ;
    end ;
    ACM.Set_Size( 2*65536 ) ;
    ACM.Set_Size( 0 ) ;

    ACM.Set_Heap( nil ) ;
    ACM.Set_Store( nil ) ;
    ACM.Free ;
    ACM := TCOM_Allocation_Cluster_Manager64.Create ;
    ACM.Free ;

    // Stress test...
  {$IFDEF CC_Debug}
    Allocated_CAs.Clear ;
    Other_Allocated.Clear ;
  {$ENDIF}
    randomize ;
    ACM := TCOM_Allocation_Cluster_Manager64.Create ;
    S64 := TCOM_Heap64.Create ;
    ACM.Set_Store( S64 ) ;
    H := TCOM_Heap.Create ;
    ACM.Set_Heap( H ) ;
    ACM.Set_Clustersize( 512 ) ;
    Buffer_Size := H.Min_Storage ;
    if( Buffer_Size < 128 ) then
    begin
        Buffer_Size := 128 ;
    end ;

    assignfile( F, 'e:\work\acmtest.txt' ) ;
    rewrite( f ) ;
    assignfile( FI, 'e:\work\acmtest.tst' ) ;
    if( fileexists( 'e:\work\acmtest.tst' ) ) then
    begin
        reset( fi ) ;
        while( not eof( fi ) ) do
        begin
            readln( fi, s ) ;
            if( s = '' ) then
            begin
                continue ;
            end ;
            case s[ 1 ] of
                'S' : begin
                          Dummy := pos( ' ', S + ' ' ) ;
                          S := copy( S, 1, Dummy - 1 ) ;
                          P := strtoint( copy( S, 2, length( S ) ) ) ;
                          if( Do_Size( P ) ) then
                          begin
                              exit ;
                          end ;
                      end ;
                'R' : begin
                          Dummy := pos( ' ', S + ' ' ) ;
                          S := copy( S, 1, Dummy - 1 ) ;
                          P := strtoint( copy( S, 2, length( S ) ) ) ;
                          if( Do_Read( P ) ) then
                          begin
                              exit ;
                          end ;
                      end ;
                'W' : begin
                          Dummy := pos( ' ', S + ' ' ) ;
                          S := copy( S, 1, Dummy - 1 ) ;
                          Old_P := strtoint( copy( S, 2, length( S ) ) ) ;
                          P1 := ACM.Offset_To_Pointer( Old_P ) ;
                          if( Status( 'W' + inttostr( Old_P ) + ' ' + inttostr( Old_P div 512 ) + ' ' + inttostr( P1 ) ) ) then
                          begin
                              exit ;
                          end ;
if( P1 = 0 ) then
begin
                          P1 := ACM.Offset_To_Pointer( Old_P ) ;
end ;
  {$IFDEF CC_Debug}
                          if( Other_Allocated.Indexof( pointer( P1 ) ) = -1 ) then
                          begin
                              halt ;
                          end ;
                          for Dummy := 0 to Allocated_CAs.Count - 1 do
                          begin
                              CA := integer( Allocated_CAs[ Dummy ] ) ;
                              if(
                                  ( ( P1 >= CA ) and ( P1 <= CA + Buffer_Size ) )
                                  or
                                  ( ( P1 < CA ) and ( P1 + 512 >= CA ) )
                                ) then
                              begin
                                  halt ;
                              end ;
                          end ;
  {$ENDIF}
                          S64.Fill( P1, 512, Old_P div 512 ) ;
                      end ;
            end ; // case s[ 1 ]
        end ; // while( not eof( fi ) )
        closefile( fi ) ;
    end ;

    try
        while( true ) do
        begin
            case random( 101 ) of
                0..50 : // Set size
                    begin
                        Old_P := ACM.Get_Size ;
                        P := random( 256 ) * 512 ;
                        if( Do_Size( P ) ) then
                        begin
                            exit ;
                        end ;
                        if( P > Old_P ) then // Expand
                        begin
                            if( Do_Size( P ) ) then
                            begin
                                exit ;
                            end ;
                            while( Old_P < P ) do
                            begin
                                P1 := ACM.Offset_To_Pointer( Old_P ) ;
                                if( Status( 'W' + inttostr( Old_P ) + ' ' + inttostr( Old_P div 512 ) + ' ' + inttostr( P1 ) ) ) then
                                begin
                                    exit ;
                                end ;
if( P1 = 0 ) then
begin
                                P1 := ACM.Offset_To_Pointer( Old_P ) ;
end ;
  {$IFDEF CC_Debug}
                                if( Other_Allocated.Indexof( pointer( P1 ) ) = -1 ) then
                                begin
                                    halt ;
                                end ;
                                for Dummy := 0 to Allocated_CAs.Count - 1 do
                                begin
                                    CA := integer( Allocated_CAs[ Dummy ] ) ;
                                    if(
                                        ( ( P1 >= CA ) and ( P1 <= CA + Buffer_Size ) )
                                        or
                                        ( ( P1 < CA ) and ( P1 + 512 >= CA ) )
                                      ) then
                                    begin
                                        halt ;
                                    end ;
                                end ;
  {$ENDIF}
                                S64.Fill( P1, 512, Old_P div 512 ) ;
                                Old_P := Old_P + 512 ;
                            end ; // while( Old_P < P )
                        end ; // if( P > Old_P )
                    end ;
                51..99 : // Read
                    begin
                        P := random( 256 ) ;
                        if( Do_Read( P ) ) then
                        begin
                            exit ;
                        end ;
                    end ;
                100 : // Zero
                    begin
                        if( Do_Size( P ) ) then
                        begin
                            exit ;
                        end ;
  {$IFDEF CC_Debug}
                        if( Allocated_CAs.Count + Other_Allocated.Count > 0 ) then
                        begin
                            halt ;
                        end ;
  {$ENDIF}
                    end ;
            end ;
        end ; // while( true )
    except
    end ;
    ACM.Free ;
    closefile( F ) ;
end ;
{$ENDIF}


begin
{$IFDEF CC_Debug}
    Allocated_CAs := TList.Create ;
    Other_Allocated := TList.Create ;
{$ENDIF}
end.

