{
        Program Name : Behavior
        Package Name : Subroutine Library - Layer 1
        Purpose      : Base behavior class
        Institution  : Conroy & Conroy Co.
        Date Written : 18-July-2004
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2004 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TBehavior is an abstract class defining an interface to a behavior.
        Behaviors are routines which affect the behavior of other objects, which
        interface with the behavior through events.
}

unit Behavior ;

interface

{$I EDEFINES.INC}

uses
{$IFDEF FPC_OR_Delphi}
     { Borland... }
     Classes,
{$ENDIF}
     { Subroutine Library - Layer 0... }
     CommBase, { TObject }

     { Subroutine Library - Layer 1... }
     _Events ; { PSGeneric_Event }

type TSBehavior = object( TBase_Object )
                      public { API... }
                          { Called before default processing of event.  Allows
                            behavior to indicate whether or not to allow the
                            event.  If False is returned, the event is ignored. }
                          function Allow_Normal_Processing( Event : PSGeneric_Event ) : boolean ;
                              virtual ;

                          { Called to execute the event.  Default processing may
                            have happened prior to this call. }
                          procedure Execute( Event : PSGeneric_Event ) ; virtual ;
                  end ;
     PSBehavior = ^TSBehavior ;

{$IFDEF DYNAMIC_OBJECTS}
type TBehavior = class( TBase_Delphi_Object )
                     public { API... }
                         { Called before default processing of event.  Allows
                           behavior to indicate whether or not to allow the
                           event.  If False is returned, the event is ignored. }
                         function Allow_Normal_Processing( Event : TGeneric_Event ) : boolean ;
                             virtual ; abstract ;

                         { Called to execute the event.  Default processing may
                           have happened prior to this call. }
                         procedure Execute( Event : TGeneric_Event ) ;
                             virtual ; abstract ;
                 end ;
{$ENDIF}

implementation

{ TSBehavior methods... }

{ API... }

function TSBehavior.Allow_Normal_Processing( Event : PSGeneric_Event ) : boolean ;

begin
    Allow_Normal_Processing := True ;
end ;


procedure TSBehavior.Execute( Event : PSGeneric_Event ) ;

begin
    { This method intentionally left blank }
end ;



end.


