{
        Program Name : Button_2D
        Package Name : Subroutine Library - Layer 1
        Purpose      : 2-D GUI Button
        Institution  : Conroy & Conroy Co.
        Date Written : 13-Nov-2006
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2006 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          Implements a GUI button in 2D.  The main purpose of this class is to
        virtualize access to the Glyph property of TSpeedButton.
}

unit Button_2D ;

interface

uses // Borland...
     Buttons, // TSpeedButton
     Classes, // TComponent
     Graphics, // TBitmap
     SysUtils ; // Exception

type T2D_Button = class( TSpeedButton )
                    public // Constructors and destructors...
                        constructor Create( AOwner : TComponent ) ;
                            override ;
                        destructor Destroy ; override ;

{$IFDEF Development}
                    private // Instance data...
                        _Sanity : string[ 10 ] ;
{$ENDIF}

                    protected // Property handlers...
                        function Get_Glyph : TBitmap ; virtual ;
                        // Always returns a copy of the glyph

                        function Get_NumGlyphs : integer ; virtual ;
                        procedure Set_Glyph( Value : TBitmap ) ; virtual ;
                        procedure Set_NumGlyphs( Value : integer ) ; virtual ;

                    public // API...
                        property Glyph : TBitmap
                            read Get_Glyph
                            write Set_Glyph ;

                        function Glyph_Height : integer ; virtual ;

                        function Glyph_Width : integer ; virtual ;

                        property NumGlyphs : integer
                                     read Get_NumGlyphs
                                     write Set_NumGlyphs ;

                        procedure Paint ; override ;
                end ; // T2D_Button

type TXButton = class( TSpeedButton )
                    public // Constructors and destructors...
                        constructor Create( AOwner : TComponent ) ; override ;

                    private // Instance data...
                        _Color : TColor ;

                    protected // Property handlers...
                        function Get_Color : TColor ;
                        procedure Set_Color( Value : TColor ) ;

                    public // Overrides...
                        procedure Paint ; override ;

                    public // API...
                        property Color : TColor
                            read Get_Color
                            write Set_Color ;
                end ;

implementation

uses // Borland...
     Controls, // cs_*
     Windows ; // SetBkColor

// T2D_Button methods...

{$IFDEF Development}
procedure Check_2D_Button( O : TObject ) ;

begin
    if( O = nil ) then
    begin
        raise Exception.Create( 'Call to method of nil object' ) ;
    end ;
    if( not ( O is T2D_Button ) ) then
    begin
        raise Exception.Create( 'Invalid type-cast' ) ;
    end ;
    if( T2D_Button( O )._Sanity <> 'T2D_Button' ) then
    begin
        raise Exception.Create( 'Call to method of deleted instance' ) ;
    end ;
end ;
{$ENDIF}

// Constructors and destructors...

constructor T2D_Button.Create( AOwner : TComponent ) ;

begin
    inherited Create( AOwner ) ;

{$IFDEF Development}
    _Sanity := 'T2D_Button' ;
{$ENDIF}
    ControlStyle := [csCaptureMouse, csDoubleClicks] ;
    ParentFont := True ;
    Color := clBtnFace ;
    Spacing := 4 ;
    Margin := -1 ;
    Layout := blGlyphLeft ;
    Transparent := True ;
end ;


destructor T2D_Button.Destroy ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
    _Sanity := '          ' ;
{$ENDIF}

    inherited Destroy ;
end ;


// Property handlers...

function T2D_Button.Get_Glyph : Graphics.TBitmap ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
{$ENDIF}
    Result := Graphics.TBitmap.Create ;
    Result.Assign( inherited Glyph ) ;
end ;


function T2D_Button.Get_NumGlyphs : integer ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
{$ENDIF}
    Result := inherited NumGlyphs ;
end ;


procedure T2D_Button.Set_Glyph( Value : Graphics.TBitmap ) ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
{$ENDIF}
    inherited Glyph := Value ;
end ;


procedure T2D_Button.Set_NumGlyphs( Value : integer ) ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
{$ENDIF}
    inherited NumGlyphs := Value ;
end ;


// API...

function T2D_Button.Glyph_Height : integer ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
{$ENDIF}
    Result := inherited Glyph.Height ;
end ;


function T2D_Button.Glyph_Width : integer ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
{$ENDIF}
    Result := inherited Glyph.Width ;
end ;


procedure T2D_Button.Paint ;

begin
{$IFDEF Development}
    Check_2D_Button( self ) ;
{$ENDIF}
    inherited Paint ;
end ;



// TXButton methods...

// Constructors and destructors...

constructor TXButton.Create( AOwner : TComponent ) ;

begin
    inherited Create( AOwner ) ;

    Color := clBtnFace ;
end ;


// Property handlers...

function TXButton.Get_Color : TColor ;

begin
    Result := _Color ;
end ;


procedure TXButton.Set_Color( Value : TColor ) ;

begin
    _Color := Value ;
end ;


// Overrides...

procedure TXButton.Paint ;

var DC : HDC ;
    X, Y : integer ;

begin
    Canvas.Brush.Color := Color ;
    Canvas.FillRect( ClientRect ) ;
    if( Caption <> '' ) then
    begin
        X := ( ClientWidth - Canvas.TextWidth( Caption ) ) div 2 ;
        Y := ( ClientHeight - Canvas.TextHeight( Caption ) ) div 2 ;
        if( X < 0 ) then
        begin
            X := 0 ;
        end ;
        if( Y < 0 ) then
        begin
            Y := 0 ;
        end ;
        if( Down or ( FState = bsDown ) ) then
        begin
            inc( X ) ;
            inc( Y ) ;
        end ;
        Canvas.Pen.Color := Font.Color ;
        DC := Canvas.Handle ;
        SetBkMode( DC, Windows.TRANSPARENT ) ;
        SetTextColor( DC, Font.Color ) ;
        Canvas.TextOut( X, Y, Caption ) ;
    end ;
    Canvas.MoveTo( ClientWidth, 0 ) ;
    if( Down or ( FState = bsDown ) ) then
    begin
        Canvas.Pen.Color := clBlack ;
        Canvas.LineTo( 0, 0 ) ;
        Canvas.LineTo( 0, ClientHeight - 1 ) ;
        Canvas.Pen.Color := clWhite ;
        Canvas.LineTo( ClientWidth - 1, ClientHeight - 1 ) ;
        Canvas.LineTo( ClientWidth - 1, 0 ) ;
    end else
    begin
        Canvas.Pen.Color := clWhite ;
        Canvas.LineTo( 0, 0 ) ;
        Canvas.LineTo( 0, ClientHeight - 1 ) ;
        Canvas.Pen.Color := clBlack ;
        Canvas.LineTo( ClientWidth - 1, ClientHeight - 1 ) ;
        Canvas.LineTo( ClientWidth - 1, 0 ) ;
    end ;
end ; // TXButton.Paint



end.
