/*
        Program Name : TCommon_COM_Interface
        Package Name : LB
        Purpose      : Abstract virtual common COM interface class
        Institution  : Conroy & Conroy Co.
        Date Written : 21-March-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TCommon_COM_Interface is a base class definition for COM objects
        used in the Application Foundation.
*/

#include "DebugInt.h" // TDebug_Interface
#include "_ue.h" // TUnified_Exception

#if !defined(__Common_COM_Interface)
  #define __Common_COM_Interface
class TCommon_COM_Interface
{
     public: // API...
         // Initialize the object.  Returns result of initialization attempt.
         virtual TUnified_Exception __stdcall Initialize() = 0;

         // Terminate the use of the object.  Returns result of termination.
         // This generally should not be used - instead, use the Detach method.
		 virtual TUnified_Exception __stdcall Terminate() = 0;

         // Increments the reference count for the object.
         virtual void __stdcall Attach() = 0;

         // Decrements the reference count for the object.  When count reaches
         // 0, the object is destroyed.
         virtual void __stdcall Detach() = 0;

         // Version of the COM interface for this object.
         virtual int __stdcall Interface_Version() = 0;

         // Facility ID for the facility represented by this object.  Returns -1
         // if no facility assigned. }
         virtual int __stdcall Facility_ID() = 0;

         // Facility version.  Result only has meaning if Facility_ID doesn't
         // return -1. }
         virtual int __stdcall Facility_Version() = 0;

         // Returns a debugging interface for the object.  Returns nil if not s
         // upported.
         virtual TDebug_Interface* __stdcall Debugger() = 0;

         // Returns a pointer to an object with extended common COM methods for
         // object.  Always returns nil for now. }
         virtual void* __stdcall Extension() = 0;
} ; // TCommon_COM_Interface

#endif
