{
        Program Name : CRCs
        Package Name : Subroutine library - Layer 4
        Purpose      : Calculate checksums
        Institution  : Conroy & Conroy Co.
        Date Written :
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2015 by Alan Conroy.  All rights reserved.
        Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        This unit contains default tool set for FMU.
}

unit CRCs ;

interface

{$I EDEFINES.INC}

type Checksum_Type = ( CT_Add, // Simple add
                       CT_CRC // Cyclic Redundancy check
                     ) ;

{ Generate a checksum for passed string Value.  CT is the type of checksum to
  to generate.  Width is the width of the checksum value.  Init is the initial
  value (or value passed in from previous partial string).  XOR_Out indicates
  the value to XOR with the result.  Polynomial is the polynomial for CRCs.
  Reflect_In indicates to reflect input value (for CRC only).  Reflect_Out
  indicates to reflect output value (CRC only). }
{$IFDEF INT64}
function Checksum( CT : Checksum_Type ; Value : PChar ; Length : int64 ;
    Width : longint ; Init, XOR_Out, Polynomial : int64 ;
    Reflect_In, Reflect_Out : boolean ) : int64 ;
function Simple_Checksum( Width : integer ; Buffer : string ) : int64 ;
{$ELSE}
function Checksum( CT : Checksum_Type ; Value : PChar ; Length : longint ;
    Width : longint ; Init, XOR_Out, Polynomial : longint ;
    Reflect_In, Reflect_Out : boolean ) : longint ;
function Simple_Checksum( Width : integer ; Buffer : string ) : longint ;
{$ENDIF}

{ Calculate Cyclic Redundancy Codes }
function CRC16( Buffer : string ) : word ;
function CRC16_CITT( Buffer : string ) : word ;
function CRC16_XModem( Buffer : string ) : word ;
function CRC16_ARC( Buffer : string ) : word ;
function CRC32( Buffer : string ) : longint ;

{ Continue previous CRC, where Init is the value from the previous CRC. }
function CRC16_Next( Buffer : string ; Init : longint ) : word ;
function CRC16_CITT_Next( Buffer : string ; Init : longint ) : word ;
function CRC16_XModem_Next( Buffer : string ; Init : longint ) : word ;
function CRC16_ARC_Next( Buffer : string ; Init : longint ) : word ;
function CRC32_Next( Buffer : string ; Init : longint ) : longint ;


implementation

uses { Subroutine Library - Layer 3... }
     _Standard ; { XBit_Values }

function CRC16( Buffer : string ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, 0, 0,
        $8005, True, True ) and $FFFF ;
end ;


function CRC16_Next( Buffer : string ; Init : longint ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, Init, 0,
        $8005, True, True ) and $FFFF ;
end ;


function CRC16_CITT( Buffer : string ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, $FFFF, 0,
        $1021, False, False ) and $FFFF ;
end ;


function CRC16_CITT_Next( Buffer : string ; Init : longint ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, Init, 0,
        $1021, False, False ) and $FFFF ;
end ;


function CRC16_XModem( Buffer : string ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, 0, 0,
        $8408, True, True ) and $FFFF ;
end ;


function CRC16_XModem_Next( Buffer : string ; Init : longint ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, Init, 0,
        $8408, True, True ) and $FFFF ;
end ;


function CRC16_ARC( Buffer : string ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, 0, 0,
        $8005, True, True ) and $FFFF ;
end ;


function CRC16_ARC_Next( Buffer : string ; Init : longint ) : word ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 16, Init, 0,
        $8005, True, True ) and $FFFF ;
end ;


function CRC32( Buffer : string ) : longint ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 32, $FFFFFFFF, $FFFFFFFF,
        $4C11DB7, True, True ) and $FFFFFFFF ;
end ;


function CRC32_Next( Buffer : string ; Init : longint ) : longint ;

begin
    Result := Checksum( CT_CRC, PChar( Buffer ), length( Buffer ), 32, Init, $FFFFFFFF,
        $4C11DB7, True, True ) and $FFFFFFFF ;
end ;


{$IFDEF INT64}
function Checksum( CT : Checksum_Type ; Value : PChar ; Length : int64 ;
    Width : longint ; Init, XOR_Out, Polynomial : int64 ;
    Reflect_In, Reflect_Out : boolean ) : int64 ;

var Check_Bit, R, Temp : int64 ;
{$ELSE}
function Checksum( CT : Checksum_Type ; Value : PChar ; Length : longint ;
    Width : longint ; Init, XOR_Out, Polynomial : longint ;
    Reflect_In, Reflect_Out : boolean ) : longint ;

var Check_Bit, R, Temp : longint ;
{$ENDIF}
    Loop : integer ;

begin
    if( CT = CT_Add ) then
    begin
        while( Length > 0 ) do
        begin
            dec( Length ) ;
            Init := Init + ord( Value[ 0 ] ) ;
            inc( Value ) ;
        end ;
        CheckSum := Init ;
    end else // CRC
    begin
        // Note: This could be improved, performance-wise
        Check_Bit := XBit_Values[ Width ] ;
        while( Length > 0 ) do
        begin
            R := ord( Value[ 0 ] ) ; // Get next byte
            inc( Value ) ;
            dec( Length ) ;
            for Loop := 0 to 7 do
            begin
                if( Reflect_In ) then
                begin
                    Init := ( Init shl 1 ) or ( R and 1 ) ;
                    R := R shr 1 ;
                end else
                begin
                    Init := Init shl 1 ;
                    if( ( R and 128 ) <> 0 ) then
                    begin
                        R := R or 1 ;
                    end ;
                    R := R shl 1 ;
                end ;
                if( ( Init and Check_Bit ) <> 0 ) then
                begin
                    Init := Init xor Polynomial ;
                end ;
            end ; // for Loop := 0 to 7
        end ; // while( Length > 0 )

        // Process augment
        for Loop := 1 to Width do
        begin
            Init := Init shl 1 ;
            if( ( Init and Check_Bit ) <> 0 ) then
            begin
                Init := Init xor Polynomial ;
            end ;
        end ;

        // Final result processing...
        Init := ( Init xor XOR_Out ) and ( XBit_Values[ Width ] - 1 ) ;
        if( Reflect_Out ) then
        begin
            Temp := 0 ;
            for Loop := 1 to Width do
            begin
                Temp := ( Temp shl 1 ) or ( Init and 1 ) ;
                Init := Init shr 1 ;
            end ;
            Checksum := Temp ;
        end else
        begin
            Checksum := Init ;
        end ;
    end ; // if
end ; // Checksum


{$IFDEF INT64}
function Simple_Checksum( Width : integer ; Buffer : string ) : int64 ;
{$ELSE}
function Simple_Checksum( Width : integer ; Buffer : string ) : longint ;
{$ENDIF}

begin
    Result := Checksum( CT_Add, PChar( Buffer ), length( Buffer ), Width, 0, 0,
        0, False, False ) ;
end ; // Simple_Checksum


{$IFDEF CC_Regression_Test}
procedure Test_Unit ;

begin
    if( CRC16( '123456789' ) <> $BB3D ) then
    begin
        writeln( 'CRC16 failure' ) ;
    end ;
    if( CRC32( '123456789' ) <> longint( $CBF43926 ) ) then
    begin
        writeln( 'CRC32 failure' ) ;
    end ;
end ;


begin
//    Test_Unit ;
{$ENDIF}
end.

