{
        Program Name : Deferred
        Package Name : Subroutine Library - Layer 1
        Purpose      : Deferred execution manager
        Institution  : Conroy & Conroy Co.
        Date Written : 20-Jun-2015
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit Deferred ;

interface

uses // Borland...
     Classes, // TList

     // Subroutine Library - Layer 1
     _Deferred ; // TDeferred_Manager


type TStandard_Deferred_Manager = class( TDeferred_Manager )
                                      public // Constructors and destructors...
                                          constructor Create ;
                                          destructor Destroy ; override ;

                                      private // Instance data...
                                          Items : TList ;
                                          _Concurrent : boolean ;
                                          Contexts : TList ;
                                          In_Execute : integer ;

                                      public // API...
                                          procedure Add( Item : TDeferred_Execution ;
                                              Context : pointer ) ; override ;
                                          procedure Remove( Item : TDeferred_Execution ;
                                              Context : pointer ) ;
                                              override ;
                                          procedure Remove_All( Item : TDeferred_Execution ) ;
                                              override ;
                                          procedure Execute_Next ;
                                              override ;
                                          function Index_Of( Item : TDeferred_Execution ;
                                              Context : pointer ) : integer ;
                                              override ;
                                          function Index_Of_First( Item : TDeferred_Execution ) : integer ;
                                              override ;
                                          function Empty : boolean ; override ;

                                      published
                                          property Concurrent : boolean
                                              read _Concurrent
                                              write _Concurrent ;
                                  end ; // TStandard_Deferred_Manager

implementation

// TStandard_Deferred_Manager methods...

// Constructors and destructors...

constructor TStandard_Deferred_Manager.Create ;

begin
    inherited Create ;

    Items := TList.Create ;
    Contexts := TList.Create ;
end ;


destructor TStandard_Deferred_Manager.Destroy ;

begin
    Items.Free ;
    Items := nil ;
    Contexts.Free ;
    Contexts := nil ;

    inherited Destroy ;
end ;


// API...

procedure TStandard_Deferred_Manager.Add( Item : TDeferred_Execution ;
    Context : pointer ) ;

begin
    if( Item = nil ) then
    begin
	exit ;
    end ;
    Items.Add( Item ) ;
    Contexts.Add( Contexts ) ;
end ;


procedure TStandard_Deferred_Manager.Remove( Item : TDeferred_Execution ;
    Context : pointer ) ;

var Loop : integer ;

begin
    for Loop := 0 to Items.Count - 1 do
    begin
        if( ( Items[ Loop ] = Item ) and ( Contexts[ Loop ] = Context ) ) then
        begin
            Items.Delete( Loop ) ;
            Contexts.Delete( Loop ) ;
            exit ;
        end ;
    end ;
end ;


procedure TStandard_Deferred_Manager.Remove_All( Item : TDeferred_Execution ) ;

var Loop : integer ;

begin
    for Loop := Items.Count - 1 downto 0 do
    begin
        if( Items[ Loop ] = Item ) then
        begin
            Items.Delete( Loop ) ;
            Contexts.Delete( Loop ) ;
        end ;
    end ;
end ;


procedure TStandard_Deferred_Manager.Execute_Next ;

var Context : pointer ;
    Item : TDeferred_Execution ;

begin
    if( In_Execute > 0 ) then
    begin
        if( not Concurrent ) then
        begin
            exit ;
        end ;
    end ;

    if( Items.Count > 0 ) then
    begin
        try
            inc( In_Execute ) ;
            Item := Items[ 0 ] ;
            Context := Contexts[ 0 ] ;

            Items.Delete( 0 ) ;
            Contexts.Delete( 0 ) ;

            Item.Execute( Context ) ;
        finally
            dec( In_Execute ) ;
        end ;
    end ;
end ;


function TStandard_Deferred_Manager.Index_Of( Item : TDeferred_Execution ;
    Context : pointer ) : integer ;

var Loop : integer ;

begin
    Result := -1 ;
    for Loop := 0 to Items.Count - 1 do
    begin
        if( ( Items[ Loop ] = Item ) and ( Contexts[ Loop ] = Context ) ) then
        begin
            Result := Loop ;
            exit ;
        end ;
    end ;
end ;


function TStandard_Deferred_Manager.Index_Of_First( Item : TDeferred_Execution ) : integer ;

var Loop : integer ;

begin
    Result := -1 ;
    for Loop := 0 to Items.Count - 1 do
    begin
        if( Items[ Loop ] = Item ) then
        begin
            Result := Loop ;
            exit ;
        end ;
    end ;
end ;


function TStandard_Deferred_Manager.Empty : boolean ;

begin
    Result := Items.Count = 0 ;
end ;


initialization
    Deferred_Manager := TStandard_Deferred_Manager.Create ;

finalization
    Deferred_Manager.Free ;
    Deferred_Manager := nil ;
end.

