{
        Program Name : DebugInt
        Package Name : Subroutine Library - Layer 1
        Purpose      : Debugger Interface definition
        Institution  : Conroy & Conroy Co.
        Date Written : 11-Jan-2001
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2001-2015 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides an implementation of some debug interfaces.
}

unit DebugInt ;

{ TDebug_Interface is used when debugging code by providing a way of examining
  the internals of the software component at run-time (for production
  versions of software, this is often not supported).  The interface assumes a
  hierarchical form with the root of a component returning children
  representing internal objects.  Each of those objects may have children, and
  so forth. }

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

uses { Borland... }
     Classes, { TStringList } //TODO: Support older versions of Pascal

     { Subroutine Library... }
{$IFDEF DYNAMIC_OBJECTS}
     _DebugIn, { TDebug_Interface }
{$ELSE}
     _DebugIn, { TDebug_Interface }
     Strings,
{$ENDIF}
     Compatib ; // PAnsiChar

{$IFDEF DYNAMIC_OBJECTS}
type TCOM_Debug_Iterator = class( TCOM_Debug_Interface )
                               protected { Instance data... }
                                   _Children : TStringList ;
                                   _Context : longint ;
                                   Reference_Count : integer ;

                               public { API... }
                                   constructor Create ;

                                   destructor Destroy ; override ;

                                   function Count : longint ; override ;

                                   function Child( Ordinal : longint ) : TCOM_Debug_Interface ;
                                       override ;

                                   function Get_Title : PAnsiChar ; override ;

                                   function Has_Children : boolean ; override ;

                                   procedure Set_Title( Value : PAnsiChar ) ;
                                       override ;

                                   procedure Activate ; override ;

                                   function Can_Change : boolean ; override ;

                                   procedure Kill ; override ;

                                   function Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ;
                                       override ;

                                   function Get_Context : longint ; override ;

                                   procedure Set_Context( Value : longint ) ;
                                       override ;

                                   function Recursive : boolean ; override ;

                                   function Arbitrary_Length : boolean ;
                                       override ;

                                   procedure Attach ; virtual ;

                                   procedure Detach ; virtual ;

                                   procedure Build_Children ; virtual ; stdcall ;
                           end ;

     // TText_Debugger provides a single one-line textual item for a debug interface
type TText_Debugger = class( TCOM_Debug_Iterator )
                                private { Instance data... }
                                    _Title : Ansistring ;

                                public { API... }
                                    function Has_Children : boolean ;
                                        override ;

				                    function Can_Change : boolean ; override ;

                                    function Child( Ordinal : longint ) : TCOM_Debug_Interface ;
                                        override ;

                                    function Count : longint ; override ;

                                    function Get_Title : PAnsiChar ; override ;

                                    procedure Set_Title( Value : PAnsiChar ) ;
                                        override ;

                                    procedure Activate ; override ;

                                    procedure Kill ; override ;

                                    function Modify( Data : PAnsiChar ; Size : integer ) : PAnsichar ;
                                        override ;

                                    function Recursive : boolean ; override ;

                                    function Arbitrary_Length : boolean ;
                                        override ;
                      end ; // TText_Debugger
     PText_Debugger = TText_Debugger ;
{$ENDIF}

type TSStandard_Debug_Interface = object( TSDebug_Interface )
                                      public { API... }
                                          function Has_Children : boolean ;
                                              virtual ;
                                  end ; { TSStandard_Debug_Interface }
     PSStandard_Debug_Interface = ^TSStandard_Debug_Interface ;

     PSText_Debugger = ^TSText_Debugger ;
     TSText_Debugger = object( TSStandard_Debug_Interface )
                           public { API... }
                               _Title : string ;

                               constructor Init ;

                               destructor Done ; virtual ;

                               function Count : longint ; virtual ;

                               function Get_Title : PAnsichar ; virtual ;

                               procedure Set_Title( Value : PAnsichar ) ; virtual ;

                               procedure Activate ; virtual ;

		                       function Can_Change : boolean ; virtual ;

                               function Child( Ordinal : longint ) : PSDebug_Interface ;
                                   virtual ;

                               procedure Kill ; virtual ;

		                       function Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ;
                                   virtual ;

                               function Recursive : boolean ; virtual ;

                               function Arbitrary_Length : boolean ; virtual ;
                       end ;

type PSDebug_Iterator = ^TSDebug_Iterator ;
     TSDebug_Iterator = object( TSDebug_Interface )
                            protected { Instance data... }
                                _Children : TStringList ;

                            public { API... }
                                constructor Init ;

                                destructor Done ; virtual ;

                                function Count : longint ; virtual ;

                                function Child( Ordinal : longint ) : PSDebug_Interface ;
                                    virtual ;

                                function Recursive : boolean ; virtual ;

                                function Arbitrary_Length : boolean ; virtual ;
                         end ;

function Create_Text_Debugger : PSText_Debugger ;
{$IFDEF DYNAMIC_OBJECTS}
function Create_Delphi_Text_Debugger : TText_Debugger ;
function Static_Debug_Wrapper( DI : TCOM_Debug_Interface ) : PSDebug_Interface ;
function Delphi_Debug_Wrapper( DI : PSDebug_Interface ) : TCOM_Debug_Interface ;
{$ENDIF}

{$IFDEF CC_Regresion_Test}
procedure Test_Unit ;
{$ENDIF}

implementation

{$IFDEF DYNAMIC_OBJECTS}
uses { Borland... }
     SysUtils ; { strpas }

{ TCOM_Debug_Iterator methods... }

{ API... }

constructor TCOM_Debug_Iterator.Create ;

begin
    inherited Create ;

    _Children := TStringList.Create ;
end ;


destructor TCOM_Debug_Iterator.Destroy ;

begin
    _Children := TStringList.Create ;
    _Children := nil ;

    inherited Destroy ;
end ;


function TCOM_Debug_Iterator.Count : longint ;

begin
    Result := 0 ;
    if( _Children <> nil ) then
    begin
        if( _Children.Count = 0 ) then
        begin
            Build_Children ;
        end ;
        Result := _Children.Count ;
    end ;
end ;


function TCOM_Debug_Iterator.Child( Ordinal : longint ) : TCOM_Debug_Interface ;

begin
    Result := nil ;
    if( _Children <> nil ) then
    begin
        if( _Children.Count = 0 ) then
        begin
            Build_Children ;
        end ;
        _Children.Sort ;
        Result := TCOM_Debug_Interface( _Children.Objects[ Ordinal ] ) ;
    end ;
end ;


procedure TCOM_Debug_Iterator.Build_Children ;

begin
    // This routine left intentionally blank
end ;


function TCOM_Debug_Iterator.Get_Title : PAnsichar ;

begin
    Result := nil ;
end ;


function TCOM_Debug_Iterator.Has_Children : boolean ;

begin
    Result := False ;
    if( _Children.Count = 0 ) then
    begin
        Build_Children ;
        Result := _Children.Count > 0 ;
    end ;
end ;


procedure TCOM_Debug_Iterator.Set_Title( Value : PAnsichar ) ;

begin
end ;


procedure TCOM_Debug_Iterator.Activate ;

begin
end ;


function TCOM_Debug_Iterator.Can_Change : boolean ;

begin
    Result := False ;
end ;


procedure TCOM_Debug_Iterator.Kill ;

begin
end ;


function TCOM_Debug_Iterator.Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ;

begin
    Result := nil ;
end ;


function TCOM_Debug_Iterator.Get_Context : longint ;

begin
    Result := _Context ;
end ;


procedure TCOM_Debug_Iterator.Set_Context( Value : longint ) ;

begin
    _Context := Value ;
end ;


function TCOM_Debug_Iterator.Recursive : boolean ;

begin
    Result := False ;
end ;


function TCOM_Debug_Iterator.Arbitrary_Length : boolean ;

begin
    Result := False ;
end ;


procedure TCOM_Debug_Iterator.Attach ;

begin
    inc( Reference_Count ) ;
end ;


procedure TCOM_Debug_Iterator.Detach ;

begin
    dec( Reference_Count ) ;
    if( Reference_Count < 1 ) then
    begin
         Free ;
    end ;
end ;



type TSStatic_Debug_Wrapper = object( TSStandard_Debug_Interface )
                                  private { Instance data... }
                                      Debug_Interface : TCOM_Debug_Interface ;

                                  public { API... }
                                      function Count : longint ;
                                          virtual ;

                                      function Get_Title : PAnsichar ;
                                          virtual ;

                                      function Has_Children : boolean ;
                                          virtual ;

                                      procedure Set_Title( Value : PAnsichar ) ;
                                          virtual ;

                                      procedure Activate ;
                                          virtual ;

                                      function Can_Change : boolean ; virtual ;

                                      function Child( Ordinal : longint ) : PSDebug_Interface ;
                                          virtual ;

                                      procedure Kill ; virtual ;

                                      function Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ;
                                           virtual ;

                                      function Get_Context : longint ; virtual ;

                                      procedure Set_Context( Value : longint ) ;
                                          virtual ;
                              end ;
     PSStatic_Debug_Wrapper = ^TSStatic_Debug_Wrapper ;

function TSStatic_Debug_Wrapper.Count : longint ;

begin
    Count := Debug_Interface.Count ;
end ;


function TSStatic_Debug_Wrapper.Get_Title : PAnsichar ;

begin
    Get_Title := Debug_Interface.Get_Title ;
end ;


function TSStatic_Debug_Wrapper.Has_Children : boolean ;

begin
    has_Children := Debug_Interface.Has_Children ;
end ;


procedure TSStatic_Debug_Wrapper.Set_Title( Value : PAnsichar ) ;

begin
    Debug_Interface.Set_Title( Value ) ;
end ;


procedure TSStatic_Debug_Wrapper.Activate ;

begin
    Debug_Interface.Activate ;
end ;


function TSStatic_Debug_Wrapper.Can_Change : boolean ;

begin
    Can_Change := Debug_Interface.Can_Change
end ;


function TSStatic_Debug_Wrapper.Child( Ordinal : longint ) : PSDebug_Interface ;

begin
    Child := Static_Debug_Wrapper( Debug_Interface.Child( Ordinal ) ) ;
end ;


procedure TSStatic_Debug_Wrapper.Kill ;

begin
    Debug_Interface.Kill ;
//    dispose( @Self, Done ) ;
end ;


function TSStatic_Debug_Wrapper.Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ;

begin
    Modify := Debug_Interface.Modify( Data, Size ) ;
end ;


function TSStatic_Debug_Wrapper.Get_Context : longint ;

begin
    Get_Context := Debug_Interface.Get_Context ;
end ;


procedure TSStatic_Debug_Wrapper.Set_Context( Value : longint ) ;

begin
    Debug_Interface.Set_Context( Value ) ;
end ;
{$ENDIF}


{ TSDebug_Iterator methods... }

{ API... }

constructor TSDebug_Iterator.Init ;

begin
    inherited Init ;

    _Children := TStringList.Create ;
end ;


destructor TSDebug_Iterator.Done ;

begin
    _Children.Free ;
    _Children := nil ;

    inherited Done ;
end ;


function TSDebug_Iterator.Count : longint ;

begin
    Count := 0 ;
    if( _Children <> nil ) then
    begin
        _Children.Sort ;
        Count := _Children.Count ;
    end ;
end ;


function TSDebug_Iterator.Child( Ordinal : longint ) : PSDebug_Interface ;

begin
    Result := PSDebug_Interface( _Children[ Ordinal ] ) ;
end ;


function TSDebug_Iterator.Recursive : boolean ;

begin
    Recursive := False ;
end ;


function TSDebug_Iterator.Arbitrary_Length : boolean ;

begin
    Arbitrary_Length := False ;
end ;



{ Unit API... }

function Create_Text_Debugger : PSText_Debugger ;

begin
    Create_Text_Debugger := new( PSText_Debugger, Init ) ;
end ;


{$IFDEF DYNAMIC_OBJECTS}
function Create_Delphi_Text_Debugger : TText_Debugger ;

begin
    Create_Delphi_Text_Debugger := TText_Debugger.Create ;
end ;


function Static_Debug_Wrapper( DI : TCOM_Debug_Interface ) : PSDebug_Interface ;

begin
    if( DI = nil ) then
    begin
        Result := nil ;
        exit ;
    end ;

    Result := new( PSStatic_Debug_Wrapper, Init ) ;
    PSStatic_Debug_Wrapper( Result )^.Debug_Interface := DI ;
end ;


type TDynamic_Debug_Wrapper = class( TCOM_Debug_Interface )
                                  private { Instance data... }
                                      Debug_Interface : PSDebug_Interface ;

                                  public { API... }
                                      function Count : longint ; override ;

                                      function Get_Title : PAnsichar ; override ;

                                      function Has_Children : boolean ;
                                          override ;

                                      procedure Set_Title( Value : PAnsichar ) ;
                                          override ;

                                      procedure Activate ; override ;

                                      function Can_Change : boolean ; override ;

                                      function Child( Index : longint ) : TCOM_Debug_Interface ;
                                          override ;

                                      procedure Kill ; override ;

                                      function Modify( Data : PAnsichar ;
                                          Size : integer ) : PAnsichar ; override ;

                                      function Get_Context : longint ; override ;

                                      procedure Set_Context( Value : longint ) ;
                                          override ;

                                      function Recursive : boolean ; override ;

                                      function Arbitrary_Length : boolean ;
                                          override ;
                              end ;

{ API... }

function TDynamic_Debug_Wrapper.Count : longint ; stdcall ;

begin
    Result := Debug_Interface^.Count ;
end ;


function TDynamic_Debug_Wrapper.Get_Title : PAnsichar ; stdcall ;

begin
    Result := Debug_Interface^.Get_Title ;
end ;


function TDynamic_Debug_Wrapper.Has_Children : boolean ; stdcall ;

begin
    Result := Debug_Interface^.Has_Children ;
end ;


procedure TDynamic_Debug_Wrapper.Set_Title( Value : PAnsichar ) ; stdcall ;

begin
    Debug_Interface^.Set_Title( Value ) ;
end ;


procedure TDynamic_Debug_Wrapper.Activate ; stdcall ;

begin
    Debug_Interface^.Activate ;
end ;


function TDynamic_Debug_Wrapper.Can_Change : boolean ; stdcall ;

begin
    Result := Debug_Interface^.Can_Change ;
end ;


function TDynamic_Debug_Wrapper.Child( Index : longint ) : TCOM_Debug_Interface ; stdcall ;

begin
    Result := Delphi_Debug_Wrapper( Debug_Interface^.Child( Index ) ) ;
end ;


procedure TDynamic_Debug_Wrapper.Kill ; stdcall ;

begin
    Debug_Interface^.Kill ;
end ;


function TDynamic_Debug_Wrapper.Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ; stdcall ;

begin
    Result := Debug_Interface^.Modify( Data, Size ) ;
end ;


function TDynamic_Debug_Wrapper.Get_Context : longint ; stdcall ;

begin
    Result := Debug_Interface^.Get_Context ;
end ;


procedure TDynamic_Debug_Wrapper.Set_Context( Value : longint ) ; stdcall ;

begin
    Debug_Interface^.Set_Context( Value ) ;
end ;


function TDynamic_Debug_Wrapper.Recursive : boolean ;

begin
    Result := Debug_Interface^.Recursive ;
end ;


function TDynamic_Debug_Wrapper.Arbitrary_Length : boolean ;

begin
    Result := Debug_Interface^.Arbitrary_Length ;
end ;



function Delphi_Debug_Wrapper( DI : PSDebug_Interface ) : TCOM_Debug_Interface ;

begin
    if( DI = nil ) then
    begin
        Result := nil ;
        exit ;
    end ;

    Result := TDynamic_Debug_Wrapper.Create ;
    TDynamic_Debug_Wrapper( Result ).Debug_Interface := DI ;
end ;
{$ENDIF}


{ TDebug_Interface methods... }

{$IFDEF DYNAMIC_OBJECTS}
{ TText_Debugger methods... }

{ API... }

function TText_Debugger.Has_Children : boolean ; stdcall ;

begin
    Result := ( Count <> 0 ) ;
end ;


function TText_Debugger.Can_Change : boolean ; stdcall ;

begin
    Can_Change := False ;
end ;


function TText_Debugger.Child( Ordinal : longint ) : TCOM_Debug_Interface ; stdcall ;

begin
    Child := nil ;
end ;


function TText_Debugger.Count : longint ; stdcall ;

begin
    Count := 0 ;
end ;


function TText_Debugger.Get_Title : PAnsichar ; stdcall ;

begin
    Get_Title := PAnsiChar( _Title ) ;
end ;


procedure TText_Debugger.Set_Title( Value : PAnsichar ) ; stdcall ;

begin
    _Title := Value ;
end ;


procedure TText_Debugger.Activate ; stdcall ;

begin
    // This space intentionally left blank
end ;


procedure TText_Debugger.Kill ; stdcall ;

begin
    TObject( self ).Free ;
end ;


function TText_Debugger.Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ; stdcall ;

begin
    Modify := nil ;
end ;


function TText_Debugger.Recursive : boolean ;

begin
    Result := False ;
end ;


function TText_Debugger.Arbitrary_Length : boolean ;

begin
    Result := False ;
end ;
{$ENDIF}



{ TSStandard_Debug_Interface methods... }

{ API... }

function TSStandard_Debug_Interface.Has_Children : boolean ;

begin
    Has_Children := ( Count > 0 ) ;
end ;


{ TSText_Debugger methods... }

{ API... }

constructor TSText_Debugger.Init ;

begin
    inherited Init ;
    
{$IFNDEF DYNAMIC_STRING_INITIALIZATION}
    initialize( _Title ) ;
{$ENDIF}
end ;


destructor TSText_Debugger.Done ;

begin
{$IFNDEF DYNAMIC_STRING_INITIALIZATION}
    finalize( _Title ) ;
{$ENDIF}

    inherited Done ;
end ;


function TSText_Debugger.Count : longint ;

begin
    Count := 0 ;
end ;


function TSText_Debugger.Get_Title : PAnsichar ;

begin
    if( ( _Title[ length( _Title ) ] <> #0 ) or ( length( _Title ) = 0 ) ) then
    begin
        _Title := _Title + #0 ;
    end ;
    Get_Title := @_Title[ 1 ] ;
end ;


procedure TSText_Debugger.Set_Title( Value : PAnsichar ) ;

begin
    if( Value = nil ) then
    begin
        _Title := '' ;
    end else
    begin
        _Title := strpas( Value ) ;
    end ;
end ;


procedure TSText_Debugger.Activate ;

begin
    { This procedure left intentionally blank... }
end ;


function TSText_Debugger.Can_Change : boolean ;

begin
    Can_Change := False ;
end ;


function TSText_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

begin
    Child := nil ;
end ;


procedure TSText_Debugger.Kill ;

begin
    dispose( PSText_Debugger( @self ), Done ) ;
end ;


function TSText_Debugger.Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ;

begin
    Modify := nil ;
end ;


function TSText_Debugger.Recursive : boolean ;

begin
    Recursive := False ;
end ;


function TSText_Debugger.Arbitrary_Length : boolean ;

begin
    Arbitrary_Length := False ;
end ;



{$IFDEF CC_Regression_Test}
  {$IFDEF DYNAMIC_OBJECTS}
type Test_COM_Debug_Interface = class( TCOM_Debug_Interface )
                                    public
                                        Last_Routine : string ;

                                        function Count : longint ; override ;

                                        function Get_Title : PAnsichar ;
                                            override ;

                                        function Has_Children : boolean ;
                                            override ;

                                        procedure Set_Title( Value : PAnsichar ) ;
                                            override ;

                                        procedure Activate ; override ;

                                        function Can_Change : boolean ;
                                            override ;

                                        function Child( Ordinal : longint ) : PDebug_Interface ;
                                            override ;

                                        procedure Kill ; override ;

                                        function Modify( Data : PAnsichar ;
                                            Size : integer ) : PAnsichar ;
                                             override ;

                                        function Get_Context : longint ;
                                            override ;

                                        procedure Set_Context( Value : longint ) ;
                                            override ;
                                end ;

function Test_COM_Debug_Interface.Count : longint ;

begin
    Last_Routine := 'Count' ;
    Result := 0 ;
end ;


function Test_COM_Debug_Interface.Get_Title : PAnsichar ;

begin
    Last_Routine := 'Get_Title' ;
    Result := PAnsiChar( Last_Routine ) ;
end ;


function Test_COM_Debug_Interface.Has_Children : boolean ;

begin
    Last_Routine := 'Has_Children' ;
    Result := True ;
end ;


procedure Test_COM_Debug_Interface.Set_Title( Value : PAnsichar ) ;

begin
    Last_Routine := 'Set_Title' ;
end ;


procedure Test_COM_Debug_Interface.Activate ;

begin
    Last_Routine := 'Activate' ;
end ;


function Test_COM_Debug_Interface.Can_Change : boolean ;

begin
    Last_Routine := 'Can_Change' ;
    Result := True ;
end ;


function Test_COM_Debug_Interface.Child( Ordinal : longint ) : PDebug_Interface ;

begin
    Last_Routine := 'Child' ;
    Result := nil ;
end ;


procedure Test_COM_Debug_Interface.Kill ;

begin
    Last_Routine := 'Kill' ;
end ;


function Test_COM_Debug_Interface.Modify( Data : PAnsichar ;
    Size : integer ) : PAnsichar ;

begin
    Last_Routine := 'Modify' ;
    Result := nil ;
end ;


function Test_COM_Debug_Interface.Get_Context : longint ;

begin
    Last_Routine := 'Get_Context' ;
    Result := 1 ;
end ;


procedure Test_COM_Debug_Interface.Set_Context( Value : longint ) ;

begin
    Last_Routine := 'Set_Context' ;
end ;


type Test_Debug_Interface = object( TSStandard_Debug_Interface )
                                public
                                    Last_Routine : string ;

                                    function Count : longint ; virtual ;

                                    function Get_Title : PAnsichar ; virtual ;

                                    function Has_Children : boolean ;
                                        virtual ;

                                    procedure Set_Title( Value : PAnsichar ) ;
                                        virtual ;

                                    procedure Activate ; virtual ;

                                    function Can_Change : boolean ; virtual ;

                                    function Child( Ordinal : longint ) : PSDebug_Interface ;
                                        virtual ;

                                    procedure Kill ; virtual ;

                                    function Modify( Data : PAnsichar ;
                                        Size : integer ) : PAnsichar ;  virtual ;

                                    function Get_Context : longint ; virtual ;

                                    procedure Set_Context( Value : longint ) ;
                                        virtual ;
                            end ;
     PTest_Debug_Interface = ^Test_Debug_Interface ;

function Test_Debug_Interface.Count : longint ;

begin
    Last_Routine := 'Count' ;
    Result := 0 ;
end ;


function Test_Debug_Interface.Get_Title : PAnsichar ;

begin
    Last_Routine := 'Get_Title' ;
    Result := PAnsichar( Last_Routine ) ;
end ;


function Test_Debug_Interface.Has_Children : boolean ;

begin
    Last_Routine := 'Has_Children' ;
    Result := True ;
end ;


procedure Test_Debug_Interface.Set_Title( Value : PAnsichar ) ;

begin
    Last_Routine := 'Set_Title' ;
end ;


procedure Test_Debug_Interface.Activate ;

begin
    Last_Routine := 'Activate' ;
end ;


function Test_Debug_Interface.Can_Change : boolean ;

begin
    Last_Routine := 'Can_Change' ;
    Result := True ;
end ;


function Test_Debug_Interface.Child( Ordinal : longint ) : PSDebug_Interface ;

begin
    Last_Routine := 'Child' ;
    Result := nil ;
end ;


procedure Test_Debug_Interface.Kill ;

begin
    Last_Routine := 'Kill' ;
end ;


function Test_Debug_Interface.Modify( Data : PAnsichar ; Size : integer ) : PAnsichar ;

begin
    Last_Routine := 'Modify' ;
    Result := nil ;
end ;


function Test_Debug_Interface.Get_Context : longint ;

begin
    Last_Routine := 'Get_Context' ;
    Result := 1 ;
end ;


procedure Test_Debug_Interface.Set_Context( Value : longint ) ;

begin
    Last_Routine := 'Set_Context' ;
end ;
  {$ENDIF}


procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_Text_Debugger ;

  {$IFDEF DYNAMIC_OBJECTS}
var Test : TText_Debugger ;
  {$ENDIF}
var STest : PSText_Debugger ;

begin
  {$IFDEF DYNAMIC_OBJECTS}
    Test := Create_Delphi_Text_Debugger ;
    Test.Title := 'Test' ;
    if( Test.Title <> 'Test' ) then
    begin
        Error ;
    end ;
    Test.Kill ;
  {$ENDIF}

    STest := Create_Text_Debugger ;
    STest^.Title := 'Test' ;
    if( STest^.Title <> 'Test' ) then
    begin
        Error ;
    end ;
    STest^.Kill ;
end ;


  {$IFDEF DYNAMIC_OBJECTS}
procedure Test_Wrappers ;

var Test_COM : Test_COM_Debug_Interface ;
    Test_Static : PTest_Debug_Interface ;
    Test : TDebug_Interface ;
    STest : PSDebug_Interface ;

begin
    Test_COM := Test_COM_Debug_Interface.Create ;
    Test_Static := new( PTest_Debug_Interface, Init ) ;

    STest := Static_Debug_Wrapper( Test_COM ) ;
    if( Delphi_Debug_Wrapper( nil ) <> nil ) then
    begin
        Error ;
    end ;
    Test := Delphi_Debug_Wrapper( Test_Static ) ;

    if( STest.Count <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Count' ) then
    begin
        Error ;
    end ;
    if( STest.Title <> 'Get_Title' ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_Title' ) then
    begin
        Error ;
    end ;
    if( not STest.Has_Children ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Has_Children' ) then
    begin
        Error ;
    end ;
    STest.Title := 'Test' ;
    if( Test_COM.Last_Routine <> 'Set_Title' ) then
    begin
        Error ;
    end ;
    STest.Activate ;
    if( Test_COM.Last_Routine <> 'Activate' ) then
    begin
        Error ;
    end ;
    if( not STest.Can_Change ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Can_Change' ) then
    begin
        Error ;
    end ;
    if( STest.Child( 0 ) <> nil ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Child' ) then
    begin
        Error ;
    end ;
    if( STest.Modify( '', 0 ) <> nil ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Modify' ) then
    begin
        Error ;
    end ;
    if( STest.Get_Context <> 1 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Last_Routine <> 'Get_Context' ) then
    begin
        Error ;
    end ;
    STest.Set_Context( 11 ) ;
    if( Test_COM.Last_Routine <> 'Set_Context' ) then
    begin
        Error ;
    end ;

    if( Test.Count <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Count' ) then
    begin
        Error ;
    end ;
    if( Test.Title <> 'Get_Title' ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Get_Title' ) then
    begin
        Error ;
    end ;
    if( not Test.Has_Children ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Has_Children' ) then
    begin
        Error ;
    end ;
    Test.Title := 'Test' ;
    if( Test_Static.Last_Routine <> 'Set_Title' ) then
    begin
        Error ;
    end ;
    Test.Activate ;
    if( Test_Static.Last_Routine <> 'Activate' ) then
    begin
        Error ;
    end ;
    if( not Test.Can_Change ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Can_Change' ) then
    begin
        Error ;
    end ;
    if( Test.Child( 0 ) <> nil ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Child' ) then
    begin
        Error ;
    end ;
    if( Test.Modify( '', 0 ) <> nil ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Modify' ) then
    begin
        Error ;
    end ;
    if( Test.Context <> 1 ) then
    begin
        Error ;
    end ;
    if( Test_Static.Last_Routine <> 'Get_Context' ) then
    begin
        Error ;
    end ;
    Test.Context := 11 ;
    if( Test_Static.Last_Routine <> 'Set_Context' ) then
    begin
        Error ;
    end ;

    Test.Kill ;
    STest.Kill ;
end ;
 {$ENDIF}


procedure Test_Unit ;

begin
    Test_Text_Debugger ;
  {$IFDEF DYNAMIC_OBJECTS}
    Test_Wrappers ;
  {$ENDIF}
end ;
{$ENDIF}



end.
