{
        Program Name : MP3s
        Package Name : LB
        Purpose      : MP3 file management
        Institution  : Conroy & Conroy Co.
        Date Written :
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2012 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit MP3s ;

interface

uses { C&C Subroutine Library... }
     Media ; { TMP3_ID3 }

type TMP3 = class
                public { Constructors and destructors... }
                    constructor Create( _Filename : string ) ;

                private { Instance data... }
                    Filename : string ;

                public { Low-level API... }
                    function Get_ID3( var Info : TMP3_ID3 ) : boolean ;
                    { Return MP3_ID3 record.  Returns False if none. }

                    procedure Set_ID3( var Info : TMP3_ID3 ) ;
                    { Set's file's ID3 record. }

                    function Get_Extended_ID3( var Info : TExtended_MP3_ID3 ) : boolean ;
                    { Return extended MP3_ID3 record.  Returns False if none. }

                    procedure Set_Extended_ID3( var Info : TExtended_MP3_ID3 ) ;
                    { Set's file's extended ID3 record. }

                    function Get_ID3v2( var Info : TID3v2 ) : boolean ;
                    { Return ID3v2 record.  Returns False if none. }

                    procedure Set_ID3v2( var Info : TID3v2 ) ;
                    { Set's file's ID3v2 record. }

                public { High-level API... }
                    function Get_ID3_Title : string ;

                    procedure Set_ID3_Title( Value : string ) ;
            end ;

implementation

uses { Borland... }
     Sysutils, { fmOpenRead }

     { C&C Subroutine Library - layer 0... }
     ASCIIDef ; { NUL }

{ TMP3 methods... }

{ Constructors and destructors... }

constructor TMP3.Create( _Filename : string ) ;

begin
    inherited Create ;

    Filename := _Filename ;
end ;


function TMP3.Get_ID3( var Info : TMP3_ID3 ) : boolean ;

var F : file ;
    Saved : integer ;

begin
    assignfile( F, Filename ) ;
    Saved := FileMode ;
    FileMode := fmOpenRead ;
    reset( F, 1 ) ;
    FileMode := Saved ;
    seek( F, filesize( F ) - 128 ) ;
    blockread( F, Info, 128 ) ;
    closefile( F ) ;
    Get_ID3 := ( Info.Tag = 'TAG' ) ;
end ;


procedure TMP3.Set_ID3( var Info : TMP3_ID3 ) ;

var F : file ;
    Temp : TMP3_ID3 ;

begin
    assignfile( F, Filename ) ;
    reset( F, 1 ) ;
    seek( F, filesize( F ) - 128 ) ;
    blockread( F, Temp, 128 ) ;
    if( Temp.Tag = 'TAG' ) then
    begin
        seek( F, filesize( F ) - 128 ) ;
    end else
    begin
        seek( F, filesize( F ) ) ;
    end ;
    blockwrite( F, Info, 128 ) ;
    closefile( F ) ;
end ;


function TMP3.Get_Extended_ID3( var Info : TExtended_MP3_ID3 ) : boolean ;

var F : file ;
    Saved : integer ;

begin
    assignfile( F, Filename ) ;
    Saved := FileMode ;
    FileMode := fmOpenRead ;
    reset( F, 1 ) ;
    FileMode := Saved ;
    seek( F, filesize( F ) - 128 - 227 ) ;
    blockread( F, Info, 227 ) ;
    closefile( F ) ;
    Get_Extended_ID3 := ( Info.Tag = 'TAG+' ) ;
end ;


procedure TMP3.Set_Extended_ID3( var Info : TExtended_MP3_ID3 ) ;

var F : file ;
    Saved : integer ;
    Temp : TMP3_ID3 ;
    Temp1 : TExtended_MP3_ID3 ;

begin
    { Open file... }
    assignfile( F, Filename ) ;
    Saved := FileMode ;
    FileMode := fmOpenRead ;
    reset( F, 1 ) ;
    FileMode := Saved ;

    { See if extended ID3 exists... }
    if( Get_ID3( Temp ) ) then
    begin
        if( Get_Extended_ID3( Temp1 ) ) then
        begin
            seek( F, filesize( F ) - 128 - 227 ) ;
        end else
        begin
            seek( F, filesize( F ) - 128 ) ;
        end ;
    end else
    begin
        seek( F, filesize( F ) ) ; { No ID3 tag }
    end ;

    { Write data and close file... }
    blockwrite( F, Info, 227 ) ;
    blockwrite( F, Temp, 128 ) ;
    closefile( F ) ;
end ; // TMP3.Set_Extended_ID3


function TMP3.Get_ID3v2( var Info : TID3v2 ) : boolean ;

var Buffer : PChar ;
    F : file ;
    Count, I, Offset, Saved : integer ;

begin
    assignfile( F, Filename ) ;
    Saved := FileMode ;
    FileMode := fmOpenRead ;
    reset( F, 1 ) ;
    FileMode := Saved ;
    getmem( Buffer, filesize( F ) ) ;
    blockread( F, Buffer^, filesize( F ), Count ) ;
    Offset := -1 ;
    for I := filesize( F ) - 10 downto 0 do
    begin
        if( ( Buffer[ I ] = 'I' ) and ( Buffer[ I + 1 ] = 'D' ) and ( Buffer[ I + 2 ] = '3' ) ) then
        begin
            Offset := I ;
        end ;
    end ;
    closefile( F ) ;
    if( Offset = -1 ) then
    begin
        Result := False ;
        exit ;
    end ;
    move( Buffer[ Offset ], Info, sizeof( Info ) ) ;
    Result := True ;
    freemem( Buffer ) ;
    Get_ID3v2 := ( Info.Tag = 'ID3' ) ;
end ; // TMP3.Get_ID3v2


procedure TMP3.Set_ID3v2( var Info : TID3v2 ) ;

begin
    // Not yet supported
end ;


function TMP3.Get_ID3_Title : string ;

var Dummy : integer ;
    MP3_ID3 : TMP3_ID3 ;
    Extended_MP3_ID3 : TExtended_MP3_ID3 ;
    S : string ;

begin
    Result := '' ;
    if( Get_ID3( MP3_ID3 ) ) then
    begin
        setlength( Result, sizeof( MP3_ID3.Title ) ) ;
        move( MP3_ID3.Title, PChar( Result )[ 0 ], sizeof( MP3_ID3.Title ) ) ;
        Dummy := pos( NUL, Result + NUL ) ;
        setlength( Result, Dummy - 1 ) ;
        if( Get_Extended_ID3( Extended_MP3_ID3 ) ) then
        begin
            setlength( S, sizeof( Extended_MP3_ID3.Title ) ) ;
            move( Extended_MP3_ID3.Title, PChar( S )[ 0 ], sizeof( Extended_MP3_ID3.Title ) ) ;
            Dummy := pos( NUL, S + NUL ) ;
            setlength( S, Dummy - 1 ) ;
            Result := Result + S ;
        end ;
    end ;
end ;


procedure TMP3.Set_ID3_Title( Value : string ) ;

var Dummy : integer ;
    MP3_ID3 : TMP3_ID3 ;
    Extended_MP3_ID3 : TExtended_MP3_ID3 ;
    Has_Extended : boolean ;

begin
    Has_Extended := False ;
    if( Get_ID3( MP3_ID3 ) ) then
    begin
        if( Get_Extended_ID3( Extended_MP3_ID3 ) ) then
        begin
            Has_Extended := True ;
        end else
        begin
            fillchar( Extended_MP3_ID3, sizeof( Extended_MP3_ID3 ), 0 ) ;
        end ;
    end else
    begin
        fillchar( MP3_ID3, sizeof( MP3_ID3 ), 0 ) ;
    end ;
    Dummy := length( Value ) ;
    if( Dummy > sizeof( MP3_ID3.Title ) ) then
    begin
        Has_Extended := True ;
        move( PChar( Value )[ 0 ], MP3_ID3.Title, sizeof( MP3_ID3.Title ) ) ;
        Value := copy( Value, sizeof( MP3_ID3.Title ) + 1, length( Value ) ) ;
        if( length( Value ) > sizeof( Extended_MP3_ID3.Title ) )  then
        begin
            setlength( Value, sizeof( Extended_MP3_ID3.Title ) ) ;
        end ;
    end else
    begin
        move( PChar( Value )[ 0 ], MP3_ID3.Title, Dummy ) ;
    end ;
    Set_ID3( MP3_ID3 ) ;
    Set_Extended_ID3( Extended_MP3_ID3 ) ;
end ; // TMP3.Set_ID3_Title


end.
