{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
unit MStores ;
{
        Program Name : MStores
        Package Name : Subroutine Library - Layer 1
        Purpose      : Managed Storage objects
        Institution  : Conroy & Conroy Co.
        Date Written : 1999
        Written By   : Alan Conroy
        Version      : 1.1B

        Copyright (C) 1999-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

           DATE        BY          REASON

         1-Nov-2002    EAC         Use TObject from Common.
        13-Nov-2002    EAC         Rename TAllocator to
                                   TManaged_Store.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TManaged_Store is the base (mostly-) abstract class of all allocation
        objects.  It descends from TStore and provides a managed store of data
        (allocations and deallocations, reuse of deallocated storage, etc).

          If your program is experiencing problems possibly related to maanaged
        stores, you can set the DEBUG compiler conditional and all managed
        store allocations, deallocations, reads, and writes will be checked for
        correctness.

          LEVEL 2 CERTIFICATION: 6-Sep-2008 by EAC
}

interface

{$I EDEFINES.INC}
{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

uses { VCL... }
{$IFDEF CC_Debug}
     Classes, { TList }
{$ENDIF}

     { Subroutine Library - Layer 0... }
     COMInter, // TCommon_COM_Interface
{$IFNDEF MOBILE}
     _DebugIn, { PSDebug_Interface }
{$ENDIF}
     _Types, { PAnsiChar }
     _UE, { TUnified_Exception }

     { Subroutine Library - Layer 1... }
{$IFDEF CC_Debug}
     CommonUt, { TInteger_List }
{$ENDIF}
     _Stores, { TStore_Address }
     Stores { TBase_Store }

{$IFDEF CC_Regression_Test}
     { Subroutine Library - Layer 3... }
     ,CCStandard  { TInteger64_List }
{$ELSE}
  {$IFDEF CC_Debug}
     ,CCStandard
  {$ENDIF}
{$ENDIF}
     ;

type TManaged_Store_Statistics = packed record
                                     Bytes_Allocated : longint ; { Sum total size of allocations }
                                     Bytes_Deallocated : longint ; { Sum total size of deallocations }
                                     Allocations : longint ; { Count of allocations }
                                     Deallocations : longint ; { Count of deallocations }
                                     Reallocations : longint ; { Count of physical reallocations }
                                     Reallocations_In_Place : longint ;
                                     { Count of reallocations not requiring physical reallocation }
                                     Root_Writes : longint ; { Count of writes to root }
                                     Root_Reads : longint ; { Count of reads from root }
                                     Copies : longint ; { Copy operation count }
                                     Bytes_Copied : longint ; { Bytes copied via Copy method }
                                     Fills : longint ; { Fill operation count }
                                     Bytes_Filled : longint ; { Bytes modified via Fill method }
                                     Extensions : longint ; { Number of extension operations }
                                     { Note: Allocations - Extensions = allocations
                                       not requiring file extensions
                                       (reallocations). }
                                 end ; { TManaged_Store_Statistics }

type TManaged_Store_Statistics64 = packed record
                                       Bytes_Allocated : int64 ; { Sum total size of allocations }
                                       Bytes_Deallocated : int64 ; { Sum total size of deallocations }
                                       Allocations : int64 ; { Count of allocations }
                                       Deallocations : int64 ; { Count of deallocations }
                                       Reallocations : int64 ; { Count of physical reallocations }
                                       Reallocations_In_Place : int64 ;
                                       { Count of reallocations not requiring physical reallocation }
                                       Root_Writes : int64 ; { Count of writes to root }
                                       Root_Reads : int64 ; { Count of reads from root }
                                       Copies : int64 ; { Copy operation count }
                                       Bytes_Copied : int64 ; { Bytes copied via Copy method }
                                       Fills : int64 ; { Fill operation count }
                                       Bytes_Filled : int64 ; { Bytes modified via Fill method }
                                       Extensions : int64 ; { Number of extension operations }
                                       { Note: Allocations - Extensions = allocations
                                         not requiring file extensions
                                         (reallocations). }
                                   end ; { TManaged_Store_Statistics64 }

type TSManaged_Store = object( TSBase_Store )
{$IFDEF CC_Debug}
                           public { Instance data... }
                              Allocations : TList ;
                              _Sizes : TInteger_List ;
                              IDs : TInteger_List ;
                              Allocation_Count : integer ;
{$ENDIF}

                           public { API... }
                               Statistics : TManaged_Store_Statistics ;
{$IFDEF CC_Debug}
                               function Check_Access( PTR, Count : TStore_Address ) : boolean ;
                               procedure Log_Allocation( PTR, Size : TStore_Address ) ;
                               procedure Log_Deallocation( PTR, Size : TStore_Address ) ;
{$ENDIF}

                           public { Constructors and destructors... }
                               constructor Init ;
                               destructor Done ; virtual ;

                           public { Overrides... }
                               function Max_Storage : TStore_Size ; virtual ;

                               function Read_Data( var Data ; Address, _Size : TStore_Address ;
                                   var UEC : PSUnified_Exception ) : TStore_Address ; virtual ;

                               function Write_Data( var Data ;
                                   Address, _Size : TStore_Address ;
                                   var UEC : PSUnified_Exception ) : TStore_Address ; virtual ;

                           public { API... }
                               function Allocate( Size : TStore_Address ) : TStore_Address ;
                                   virtual ; { Allocate space}
                               function Allocate_At( Offset, Size : TStore_Address ) : boolean ;
                                   virtual ; { Allocate space at specific location }
                               procedure Copy( Source, Destination : TStore_Address ;
                                   Count : integer ) ; virtual ; { Copy data }
                               procedure Deallocate( PTR, Size : TStore_Address ) ;
                                   virtual ;
                               { Deallocate space }

                               procedure Fill( PTR : TStore_Address ; Count : integer ;
                                   Value : byte ) ; virtual ;
                               function Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
                                   virtual ;
                               procedure Read( PTR, Size : TStore_Address ;
                                   var Buffer ) ; virtual ; { Read data }
                               procedure Write( PTR, Size : TStore_Address ;
                                   var Buffer ) ; virtual ; { Write data }
                               function SpaceAvail : TStore_Address ; virtual ;
                               function XSpaceAvail : TStore_Address ; virtual ;
                               function MaxSpace : TStore_Address ; virtual ;
                               procedure Get_Root( var Buf ) ; virtual ;
                               procedure Set_Root( var Buf ) ; virtual ;
{$IFNDEF MOBILE}
                               function Debugger : PSDebug_Interface ; virtual ;
{$ENDIF}
                         end ; { TSManaged_Store }
     PSManaged_Store = ^TSManaged_Store ;

type TSManaged_Store64 = object( TSBase_Store64 )
{$IFDEF CC_Debug}
  {$IFDEF INT64}
                           private { Instance data... }
                              Allocations : TInteger64_List ;
                              _Sizes : TInteger64_List ;
                              IDs : TInteger_List ;
                              Allocation_Count : integer ;
  {$ENDIF}
{$ENDIF}

                           public { API... }
                               Statistics : TManaged_Store_Statistics ;
{$IFDEF CC_Debug}
                               function Check_Access( PTR, Count : TStore_Address64 ) : boolean ;
                               procedure Log_Allocation( PTR, Size : TStore_Address64 ) ;
                               procedure Log_Deallocation( PTR, Size : TStore_Address64 ) ;
{$ENDIF}

                           public { Constructors and destructors... }
                               constructor Init ;
                               destructor Done ; virtual ;

                           public { Overrides... }
                               function Max_Storage : TStore_Size64 ; virtual ;

                               function Read_Data( var Data ;
                                   Address, _Size : TStore_Address64 ;
                                   var UEC : PSUnified_Exception ) : TStore_Address64 ;
                                   virtual ;

                               function Write_Data( var Data ;
                                   Address, _Size : TStore_Address64 ;
                                   var UEC : PSUnified_Exception ) : TStore_Address64 ;
                                   virtual ;

                           public { API... }
                               function Allocate( Size : TStore_Address64 ) : TStore_Address64 ;
                                   virtual ; { Allocate space}
                               function Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;
                                   virtual ; { Allocate space at specific location }
                               procedure Copy( Source, Destination : TStore_Address64 ;
                                   Count : integer ) ; virtual ; { Copy data }
                               procedure Deallocate( PTR, Size : TStore_Address64 ) ;
                                   virtual ;
                               { Deallocate space }
                               procedure Fill( PTR : TStore_Address64 ; Count : integer ;
                                   Value : byte ) ; virtual ;
                               function Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
                                   virtual ;
                               procedure Read( PTR, Size : TStore_Address64 ;
                                   var Buffer ) ; virtual ; { Read data }
                               procedure Write( PTR, Size : TStore_Address64 ;
                                   var Buffer ) ; virtual ; { Write data }
                               function SpaceAvail : TStore_Address64 ;
                                   virtual ;
                               function XSpaceAvail : TStore_Address64 ;
                                   virtual ;
                               function MaxSpace : TStore_Address64 ; virtual ;
                               procedure Get_Root( var Buf ) ; virtual ;
                               procedure Set_Root( var Buf ) ; virtual ;
{$IFNDEF MOBILE}
                               function Debugger : PSDebug_Interface ; virtual ;
{$ENDIF}
                         end ; { TSManaged_Store64 }
     PSManaged_Store64 = ^TSManaged_Store64 ;

type TManaged_Store = class( TBase_Store )
                          public { Instance data... }
                              Statistics : TManaged_Store_Statistics ;

                          public { Overrides... }
                              function Max_Storage : TStore_Size ; override ;

                              function Read_Data( var Data ;
                                  Address, _Size : TStore_Address ;
                                  var UEC : TUnified_Exception ) : TStore_Address ; override ;

                              function Write_Data( var Data ;
                                  Address, _Size : TStore_Address ;
                                  var UEC : TUnified_Exception ) : TStore_Address ; override ;

{$IFNDEF MOBILE}
                              function Debugger : TCOM_Debug_Interface ;
                                  override ;
{$ENDIF}

                          public { API... }
                              function Allocate( Size : TStore_Address ) : TStore_Address ;
                                  virtual ; { Allocate space}
                              function Allocate_At( Offset, Size : TStore_Address ) : boolean ;
                                  virtual ; { Allocate space at specific location }
                              procedure Copy( Source, Destination : TStore_Address ;
                                  Count : integer ) ; virtual ; { Copy data }
                              procedure Deallocate( PTR, Size : TStore_Address ) ;
                                  virtual ;
                              { Deallocate space }
                              procedure Fill( PTR : TStore_Address ; Count : integer ;
                                  Value : byte ) ; virtual ;
                              function Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
                                  virtual ;
                              function Read( PTR, Size : TStore_Address ;
                                  var Buffer ) : TStore_Address ; virtual ; { Read data }
                              procedure Write( PTR, Size : TStore_Address ;
                                  var Buffer ) ; virtual ; { Write data }
                              function SpaceAvail : TStore_Address ; virtual ;
                              function XSpaceAvail : TStore_Address ; virtual ;
                              function MaxSpace : TStore_Address ; virtual ;
                              procedure Get_Root( var Buf ) ; virtual ;
                              procedure Set_Root( var Buf ) ; virtual ;
                        end ; { TManaged_Store }

type TManaged_Store64 = class( TBase_Store64 )
                            public { Instance data... }
                                Statistics : TManaged_Store_Statistics ;

                            public { Overrides... }
                                function Max_Storage : TStore_Size64 ;
                                    override ;

                                function Read_Data( var Data ;
                                    Address, _Size : TStore_Address64 ;
                                    var UEC : TUnified_Exception ) : TStore_Address64 ;
                                    override ;

                                function Write_Data( var Data ;
                                    Address, _Size : TStore_Address64 ;
                                    var UEC : TUnified_Exception ) : TStore_Address64 ;
                                    override ;

{$IFNDEF MOBILE}
                                function Debugger : TCOM_Debug_Interface ;
                                    override ;
{$ENDIF}

                            public { API... }
                                function Allocate( Size : TStore_Address64 ) : TStore_Address64 ;
                                    virtual ; { Allocate space}
                                function Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;
                                    virtual ; { Allocate space at specific location }
                                procedure Copy( Source, Destination : TStore_Address64 ;
                                    Count : integer ) ; virtual ; { Copy data }
                                procedure Deallocate( PTR, Size : TStore_Address64 ) ;
                                    virtual ;
                                { Deallocate space }
                                procedure Fill( PTR : TStore_Address64 ; Count : integer ;
                                    Value : byte ) ; virtual ;
                                function Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
                                    virtual ;
                                procedure Read( PTR, Size : TStore_Address64 ;
                                    var Buffer ) ; virtual ; { Read data }
                                procedure Write( PTR, Size : TStore_Address64 ;
                                    var Buffer ) ; virtual ; { Write data }
                                function SpaceAvail : TStore_Address64 ;
                                    virtual ;
                                function XSpaceAvail : TStore_Address64 ;
                                    virtual ;
                                function MaxSpace : TStore_Address64 ; virtual ;
                                procedure Get_Root( var Buf ) ; virtual ;
                                procedure Set_Root( var Buf ) ; virtual ;
                          end ; { TManaged_Store64 }

type TCOM_Managed_Store = class( TStandard_COM_Store )
                              public { Instance data... }
                                  Statistics : TManaged_Store_Statistics ;

                              public { Overrides... }
                                  function Is_Class( Name : PAnsiChar ) : boolean ;
                                      override ;

                                  function Max_Storage : TStore_Size ;
                                      override ;

                                  function Read_Data( var Data ;
                                      Address, _Size : TStore_Address ;
                                      var UEC : TUnified_Exception ) : TStore_Address ;
                                          override ;

                                  function Write_Data( var Data ;
                                      Address, _Size : TStore_Address ;
                                      var UEC : TUnified_Exception ) : TStore_Address ;
                                      override ;

{$IFNDEF MOBILE}
                                  function Debugger : TCOM_Debug_Interface ;
                                      override ;
{$ENDIF}

                              public { API... }
                                  function Allocate( Size : TStore_Address ) : TStore_Address ;
                                      virtual ; stdcall ; { Allocate space}
                                  function Allocate_At( Offset, Size : TStore_Address ) : boolean ;
                                      virtual ; stdcall ; { Allocate space at specific location }
                                  procedure Copy( Source, Destination : TStore_Address ;
                                      Count : integer ) ; virtual ; stdcall ; { Copy data }
                                  procedure Deallocate( PTR, Size : TStore_Address ) ;
                                      virtual ; stdcall ;
                                  { Deallocate space }
                                  procedure Fill( PTR : TStore_Address ; Count : integer ;
                                      Value : byte ) ; virtual ; stdcall ;
                                  function Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
                                      virtual ; stdcall ;
                                  procedure Read( PTR, Size : TStore_Address ;
                                      var Buffer ) ; virtual ; stdcall ; { Read data }
                                  procedure Write( PTR, Size : TStore_Address ;
                                      var Buffer ) ; virtual ; stdcall ; { Write data }
                                  function SpaceAvail : TStore_Address ;
                                      virtual ; stdcall ;
                                  function XSpaceAvail : TStore_Address ;
                                      virtual ; stdcall ;
                                  function MaxSpace : TStore_Address ;
                                      virtual ; stdcall ;
                                  procedure Get_Root( var Buf ) ;
                                      virtual ; stdcall ;
                                  procedure Set_Root( var Buf ) ;
                                      virtual ; stdcall ;
                          end ; { TCOM_Managed_Store }

type TCOM_Managed_Store64 = class( TStandard_COM_Store64 )
                                public { Instance data... }
                                    Statistics : TManaged_Store_Statistics64 ;

                                public { Overrides... }
                                    function Is_Class( Name : PAnsiChar ) : boolean ;
                                        override ;

                                    function Max_Storage : TStore_Size64 ;
                                        override ;

                                    function Read_Data( var Data ;
                                        Address, _Size : TStore_Address64 ;
                                        var UEC : TUnified_Exception ) : TStore_Address64 ;
                                        override ;

                                    function Write_Data( var Data ;
                                        Address, _Size : TStore_Address64 ;
                                        var UEC : TUnified_Exception ) : TStore_Address64 ;
                                        override ;

{$IFNDEF MOBILE}
                                    function Debugger : TCOM_Debug_Interface ;
                                        override ;
{$ENDIF}

                                public { API... }
                                    function Allocate( Size : TStore_Address64 ) : TStore_Address64 ;
                                        virtual ; stdcall ; { Allocate space}
                                    function Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;
                                        virtual ; stdcall ; { Allocate space at specific location }
                                    procedure Copy( Source, Destination : TStore_Address64 ;
                                        Count : int64 ) ; virtual ; stdcall ; { Copy data }
                                    procedure Deallocate( PTR, Size : TStore_Address64 ) ;
                                        virtual ; stdcall ; { Deallocate space }
                                    procedure Fill( PTR : TStore_Address64 ; Count : integer ;
                                        Value : byte ) ; virtual ; stdcall ;
                                    function Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
                                        virtual ; stdcall ;
                                    function Read( PTR, Size : TStore_Address64 ;
                                        var Buffer ) : TStore_Address64 ;
                                        virtual ; stdcall ; { Read data }
                                    function Write( PTR, Size : TStore_Address64 ;
                                        var Buffer ) : TStore_Address64 ;
                                        virtual ; stdcall ; { Write data }
                                    function SpaceAvail : TStore_Address64 ;
                                        virtual ; stdcall ;
                                    function XSpaceAvail : TStore_Address64 ;
                                        virtual ; stdcall ;
                                    function MaxSpace : TStore_Address64 ;
                                        virtual ; stdcall ;
                                    procedure Get_Root( var Buf ) ;
                                        virtual ; stdcall ;
                                    procedure Set_Root( var Buf ) ;
                                        virtual ; stdcall ;
                                    function Allocation_Table_Size : longint ;
                                        virtual ; stdcall ;
                                    procedure Get_Allocation_Table_Data( Buff : pointer ;
                                        _Size : longint ) ;
                                        virtual ; stdcall ;
                                    procedure Set_Allocation_Table_Data( Buff : pointer ;
                                        _Size : longint ) ;
                                        virtual ; stdcall ;
                                    procedure Set_Min_Storage( _Size : longint ) ;
                                        virtual ; stdcall ;
                                    procedure Set_Allocation_Table_Offset( Value : TStore_Address64 ) ;
                                        virtual ; stdcall ;
                                    function Get_Allocation_Table_Offset : TStore_Address64 ;
                                        virtual ; stdcall ;
                                    procedure Flush_Allocation_Table ;
                                        virtual ; stdcall ;
                                    function Get_Store : TCOM_Store64 ;
                                        virtual ; stdcall ;
                                    procedure Set_Store( Value : TCOM_Store64 ) ;
                                        virtual ; stdcall ;

                                    property Store : TCOM_Store64
                                        read Get_Store
                                        write Set_Store ;
                            end ; { TCOM_Managed_Store64 }

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}


implementation

uses { Subroutine Library - Layer 0... }
     _Common { Abstract_Error }
{$IFDEF FPC_Or_DELPHI}
     ,SysUtils // Lowercase
{$ENDIF}

     { Subroutine Library - Layer 1... }
{$IFDEF CC_Debug}
     ,_Applic { Application_Manager }
{$ENDIF}
     ;

{ TSManaged_Store methods... }

{ Constructors and destructors... }

constructor TSManaged_Store.Init ;

begin
    inherited Init ;
{$IFDEF CC_Debug}
    Allocations := TList.Create ;
    _Sizes := TInteger_List.Create ;
    IDs := TInteger_List.Create ;
    Allocation_Count := 0 ;
{$ENDIF}
end ;


destructor TSManaged_Store.Done ;

{$IFDEF CC_Debug}
  {$IFDEF Log_Abandoned_Heap}
var F : textfile ;
    Loop : integer ;
  {$ENDIF}
{$ENDIF}

begin
    inherited Done ;
{$IFDEF CC_Debug}
  {$IFDEF Log_Abandoned_Heap}
    assignfile( F, 'c:\heap.txt' ) ;
    {$I-}
    append( F ) ;
    {$I+}
    if( IOResult <> 0 ) then
    begin
        rewrite( F ) ;
    end ;
    for Loop := 0 to Allocations.Count - 1 do
    begin
        writeln( F, IDs[ Loop ], ': ', _Sizes[ Loop ], ' @', integer( Allocations[ Loop ] ) ) ;
    end ;
    closefile( F ) ;
  {$ENDIF}

    Allocations.Free ;
    Allocations := nil ;
    _Sizes.Free ;
    _Sizes := nil ;
    IDs.Free ;
    IDs := nil ;
{$ENDIF}
end ;


{ Overrides... }

function TSManaged_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := MaxSpace ;
end ;


{$IFDEF CC_Debug}
{ Debug support... }

function TSManaged_Store.Check_Access( PTR, Count : TStore_Address ) : boolean ;

var Loop : integer ;
    Address, Size : integer ;

begin
    for Loop := 0 to Allocations.Count - 1 do
    begin
        Address := integer( Allocations[ Loop ] ) ;
        Size := integer( _Sizes[ Loop ] ) ;
        if( ( PTR >= Address ) and ( PTR + Count <= Address + Size ) ) then
        begin
            Check_Access := True ;
            exit ;
        end ;
    end ;
    Check_Access := False ;
end ;


procedure TSManaged_Store.Log_Allocation( PTR, Size : TStore_Address ) ;

begin
    Allocations.Add( pointer( Ptr ) ) ;
    _Sizes.Add( Size ) ;
    IDs.Add( Allocation_Count ) ;
    inc( Allocation_Count ) ;
end ;


procedure TSManaged_Store.Log_Deallocation( PTR, Size : TStore_Address ) ;

var Index : integer ;

begin
    Index := Allocations.IndexOf( pointer( PTR ) ) ;
    if( Index = -1 ) then { Attempt to deallocate unallocated memory }
    begin
        if( Application_Manager = nil ) then
        begin
            halt ; { UNTESTED }
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Deallocation of unallocated store' ) ;
            exit ;
        end ;
    end ;
    Allocations.Delete( Index ) ;
    _Sizes.Delete( Index ) ;
    IDs.Delete( Index ) ;
end ;
{$ENDIF}


{ API... }

function TSManaged_Store.Allocate( Size : TStore_Address ) : TStore_Address ;

begin
    Allocate := 0 ;
end ;


function TSManaged_Store.Allocate_At( Offset, Size : TStore_Address ) : boolean ;

begin
    Allocate_At := False ;
end ;


procedure TSManaged_Store.Copy( Source, Destination : TStore_Address ; Count : integer ) ; {UNTESTED}
{ Copy data }

var Buffer : array[ 0..32767 ] of char ;

begin
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Read( Source, sizeof( Buffer ), Buffer ) ;
            Write( Destination, sizeof( Buffer ), Buffer ) ;
            Source := Source + sizeof( Buffer ) ;
            Destination := Destination + Sizeof( Buffer ) ;
        end else
        begin
            Read( Source, Count, Buffer ) ;
            Write( Destination, Count, Buffer ) ;
            break ; { All done }
        end ;
        dec( Count, sizeof( Buffer ) ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TSManaged_Store.Deallocate( PTR, Size : TStore_Address ) ; {UNTESTED}

begin
    Abstract_Error ;
end ;


procedure TSManaged_Store.Fill( PTR : TStore_Address ; Count : integer ;
    Value : byte ) ;

var Buffer : array[ 0..32767 ] of char ;

begin
    if( Count > sizeof( Buffer ) ) then
    begin
        fillchar( Buffer, sizeof( Buffer ), Value ) ;
    end else
    begin
        fillchar( Buffer, Count, Value ) ;
    end ;
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Write( PTR, sizeof( Buffer ), Buffer ) ;
        end else
        begin
            Write( PTR, Count, Buffer ) ;
        end ;
        Count := Count - sizeof( Buffer ) ;
        PTR := PTR + sizeof( Buffer ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


function TSManaged_Store.Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
{ This function does a reallocation operation by allocating new memory, copying
  it and then deallocating the old.  Descendants should override this default
  method if they have a more efficient reallocation method. }

var Dummy : longint ;

begin
    Dummy := Allocate( New ) ;
    Reallocate := Dummy ;
    if( Dummy <> 0 ) then { Something was allocated } {UNTESTED}
    begin
        { Copy old data to new location... }
        if( New > Old ) then
        begin
            New := Old ; { Amount to move }
        end ;
        Copy( PTR, Dummy, New ) ;

        { All done with old data }
        Deallocate( PTR, Old ) ;
    end ;

    { Update statistics }
    inc( Statistics.Reallocations ) ;
end ;


function TSManaged_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : PSUnified_Exception ) : TStore_Address ; {UNTESTED}

begin
    UEC := nil ;
    Read( Address, _Size, Data ) ;
    Read_Data := _Size ;
end ;


function TSManaged_Store.Write_Data( var Data ;
    Address, _Size : TStore_Address ; var UEC : PSUnified_Exception ) : TStore_Address ;

begin
    UEC := nil ;
    Write( Address, _Size, Data ) ;
    Write_Data := _Size ;
end ;


procedure TSManaged_Store.Read( PTR, Size : TStore_Address ; var Buffer ) ; {UNTESTED}

begin
    Abstract_Error ;
end ;


procedure TSManaged_Store.Write( PTR, Size : TStore_Address ; var Buffer ) ;

begin
end ;


function TSManaged_Store.SpaceAvail : TStore_Address ;

begin
    SpaceAvail := 0 ;
end ;


function TSManaged_Store.XSpaceAvail : TStore_Address ;

begin
    XSpaceAvail := 0 ;
end ;


function TSManaged_Store.MaxSpace : TStore_Address ;

begin
    MaxSpace := 0 ;
end ;


procedure TSManaged_Store.Get_Root( var Buf ) ;

begin
end ;


procedure TSManaged_Store.Set_Root( var Buf ) ;

begin
end ;


{$IFNDEF MOBILE}
function TSManaged_Store.Debugger : PSDebug_Interface ; {UNTESTED}

begin
    Debugger := nil ;
end ;
{$ENDIF}



{ TSManaged_Store64 methods... }

{ Constructors and destructors... }

constructor TSManaged_Store64.Init ;

begin
    inherited Init ;
  {$IFDEF CC_Debug}
    Allocations := TInteger64_List.Create ;
    _Sizes := TInteger64_List.Create ;
    IDs := TInteger_List.Create ;
    Allocation_Count := 0 ;
  {$ENDIF}
end ;


destructor TSManaged_Store64.Done ;

begin
    inherited Done ;
  {$IFDEF CC_Debug}
    Allocations.Free ;
    Allocations := nil ;
    _Sizes.Free ;
    _Sizes := nil ;
    IDs.Free ;
    IDs := nil ;
  {$ENDIF}
end ;


{ Overrides... }

function TSManaged_Store64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := MaxSpace ;
end ;


{$IFDEF CC_Debug}
{ Debug support... }

function TSManaged_Store64.Check_Access( PTR, Count : TStore_Address64 ) : boolean ;

var Loop : integer ;
    Address, Size : int64 ;

begin
    for Loop := 0 to Allocations.Count - 1 do
    begin
        Address := Allocations[ Loop ] ;
        Size := integer( _Sizes[ Loop ] ) ;
        if( ( PTR >= Address ) and ( PTR + Count <= Address + Size ) ) then
        begin
            Check_Access := True ;
            exit ;
        end ;
    end ;
    Check_Access := False ;
end ;


procedure TSManaged_Store64.Log_Allocation( PTR, Size : TStore_Address64 ) ;

begin
    Allocations.Add( Ptr ) ;
    _Sizes.Add( Size ) ;
    IDs.Add( Allocation_Count ) ;
    inc( Allocation_Count ) ;
end ;


procedure TSManaged_Store64.Log_Deallocation( PTR, Size : TStore_Address64 ) ;

var Index : integer ;

begin
    Index := Allocations.IndexOf( PTR ) ;
    if( Index = -1 ) then
    begin
        if( Application_Manager = nil ) then
        begin
            halt ; { UNTESTED }
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Deallocation of unallocated store' ) ;
            exit ;
        end ;
    end ;
    if( Size < _Sizes[ Index ] ) then // Partial deallocation
    begin
        Allocations[ Index ] := Allocations[ Index ] + Size ;
        _Sizes[ Index ] := _Sizes[ Index ] - Size ;
    end else
    begin
        Allocations.Delete( Index ) ;
        _Sizes.Delete( Index ) ;
        IDs.Delete( Index ) ;
    end ;
end ;
{$ENDIF}


{ API... }

function TSManaged_Store64.Allocate( Size : TStore_Address64 ) : TStore_Address64 ;

begin
    Allocate := 0 ;
end ;


function TSManaged_Store64.Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;

begin
    Allocate_At := False ;
end ;


procedure TSManaged_Store64.Copy( Source, Destination : TStore_Address64 ;
    Count : integer ) ; {UNTESTED}
{ Copy data }

var Buffer : array[ 0..32767 ] of char ;

begin
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Read( Source, sizeof( Buffer ), Buffer ) ;
            Write( Destination, sizeof( Buffer ), Buffer ) ;
            Source := Source + sizeof( Buffer ) ;
            Destination := Destination + Sizeof( Buffer ) ;
        end else
        begin
            Read( Source, Count, Buffer ) ;
            Write( Destination, Count, Buffer ) ;
            break ; { All done }
        end ;
        dec( Count, sizeof( Buffer ) ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TSManaged_Store64.Deallocate( PTR, Size : TStore_Address64 ) ; {UNTESTED}

begin
    Abstract_Error ;
end ;


procedure TSManaged_Store64.Fill( PTR : TStore_Address64 ; Count : integer ;
    Value : byte ) ;

var Buffer : array[ 0..32767 ] of char ;

begin
    if( Count > sizeof( Buffer ) ) then
    begin
        fillchar( Buffer, sizeof( Buffer ), Value ) ;
    end else
    begin
        fillchar( Buffer, Count, Value ) ;
    end ;
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Write( PTR, sizeof( Buffer ), Buffer ) ;
        end else
        begin
            Write( PTR, Count, Buffer ) ;
        end ;
        Count := Count - sizeof( Buffer ) ;
        PTR := PTR + sizeof( Buffer ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


function TSManaged_Store64.Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
{ This function does a reallocation operation by allocating new memory, copying
  it and then deallocating the old.  Descendants should override this default
  method if they have a more efficient reallocation method. }

var Dummy : TStore_Address64 ;

begin
    Dummy := Allocate( New ) ;
    Reallocate := Dummy ;
    if( Dummy <> 0 ) then { Something was allocated } {UNTESTED}
    begin
        { Copy old data to new location... }
        if( New > Old ) then
        begin
            New := Old ; { Amount to move }
        end ;
{$IFDEF INT64}
        Copy( PTR, Dummy, New ) ;
{$ELSE}
        Copy( PTR, Dummy, trunc( New ) ) ;
{$ENDIF}

        { All done with old data }
        Deallocate( PTR, Old ) ;
    end ;

    { Update statistics }
    inc( Statistics.Reallocations ) ;
end ;


function TSManaged_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : PSUnified_Exception ) : TStore_Address64 ; {UNTESTED}

begin
    UEC := nil ;
    Read( Address, _Size, Data ) ;
    Read_Data := _Size ;
end ;


function TSManaged_Store64.Write_Data( var Data ;
    Address, _Size : TStore_Address64 ; var UEC : PSUnified_Exception ) : TStore_Address64 ;

begin
    UEC := nil ;
    Write( Address, _Size, Data ) ;
    Write_Data := _Size ;
end ;


procedure TSManaged_Store64.Read( PTR, Size : TStore_Address64 ; var Buffer ) ; {UNTESTED}

begin
    Abstract_Error ;
end ;


procedure TSManaged_Store64.Write( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
end ;


function TSManaged_Store64.SpaceAvail : TStore_Address64 ;

begin
    SpaceAvail := 0 ;
end ;


function TSManaged_Store64.XSpaceAvail : TStore_Address64 ;

begin
    XSpaceAvail := 0 ;
end ;


function TSManaged_Store64.MaxSpace : TStore_Address64 ;

begin
    MaxSpace := 0 ;
end ;


procedure TSManaged_Store64.Get_Root( var Buf ) ;

begin
end ;


procedure TSManaged_Store64.Set_Root( var Buf ) ;

begin
end ;


{$IFNDEF MOBILE}
function TSManaged_Store64.Debugger : PSDebug_Interface ; {UNTESTED}

begin
    Debugger := nil ;
end ;
{$ENDIF}



{ TManaged_Store methods... }

{ Overrides... }

function TManaged_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := MaxSpace ;
end ;


{ API... }

function TManaged_Store.Allocate( Size : TStore_Address ) : TStore_Address ;

begin
    Allocate := 0 ;
end ;


function TManaged_Store.Allocate_At( Offset, Size : TStore_Address ) : boolean ;

begin
    Allocate_At := False ;
end ;


procedure TManaged_Store.Copy( Source, Destination : TStore_Address ;
    Count : integer ) ; {UNTESTED}
{ Copy data }

var Buffer : array[ 0..32767 ] of char ;

begin
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Read( Source, sizeof( Buffer ), Buffer ) ;
            Write( Destination, sizeof( Buffer ), Buffer ) ;
            Source := Source + sizeof( Buffer ) ;
            Destination := Destination + Sizeof( Buffer ) ;
        end else
        begin
            Read( Source, Count, Buffer ) ;
            Write( Destination, Count, Buffer ) ;
            break ; { All done }
        end ;
        dec( Count, sizeof( Buffer ) ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TManaged_Store.Deallocate( PTR, Size : TStore_Address ) ; {UNTESTED}

begin
    Abstract_Error ;
end ;


procedure TManaged_Store.Fill( PTR : TStore_Address ; Count : integer ; Value : byte ) ;

var Buffer : array[ 0..32767 ] of char ;

begin
    if( Count > sizeof( Buffer ) ) then
    begin
        fillchar( Buffer, sizeof( Buffer ), Value ) ;
    end else
    begin
        fillchar( Buffer, Count, Value ) ;
    end ;
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Write( PTR, sizeof( Buffer ), Buffer ) ;
        end else
        begin
            Write( PTR, Count, Buffer ) ;
        end ;
        Count := Count - sizeof( Buffer ) ;
        PTR := PTR + sizeof( Buffer ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


function TManaged_Store.Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
{ This function does a reallocation operation by allocating new memory, copying
  it and then deallocating the old.  Descendants should override this default
  method if they have a more efficient reallocation method. }

var Dummy : longint ;

begin
    Dummy := Allocate( New ) ;
    Reallocate := Dummy ;
    if( Dummy <> 0 ) then { Something was allocated } {UNTESTED}
    begin
        { Copy old data to new location... }
        if( New > Old ) then
        begin
            New := Old ; { Amount to move }
        end ;
        Copy( PTR, Dummy, New ) ;

        { All done with old data }
        Deallocate( PTR, Old ) ;
    end ;

    { Update statistics }
    inc( Statistics.Reallocations ) ;
end ;


function TManaged_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUnified_Exception ) : TStore_Address ; {UNTESTED}

begin
    UEC := nil ;
    Read( Address, _Size, Data ) ;
    Read_Data := _Size ;
end ;


function TManaged_Store.Write_Data( var Data ;
    Address, _Size : TStore_Address ; var UEC : TUnified_Exception ) : TStore_Address ; 

begin
    UEC := nil ;
    Write( Address, _Size, Data ) ;
    Write_Data := _Size ;
end ;


function TManaged_Store.Read( PTR, Size : TStore_Address ; var Buffer ) : TStore_Address ; {UNTESTED}

begin
    Abstract_Error ;
    Result := 0 ;
end ;


procedure TManaged_Store.Write( PTR, Size : TStore_Address ; var Buffer ) ;

begin
end ;


function TManaged_Store.SpaceAvail : TStore_Address ;

begin
    SpaceAvail := 0 ;
end ;


function TManaged_Store.XSpaceAvail : TStore_Address ;

begin
    XSpaceAvail := 0 ;
end ;


function TManaged_Store.MaxSpace : TStore_Address ;

begin
    MaxSpace := 0 ;
end ;


procedure TManaged_Store.Get_Root( var Buf ) ;

begin
end ;


procedure TManaged_Store.Set_Root( var Buf ) ;

begin
end ;


{$IFNDEF MOBILE}
function TManaged_Store.Debugger : TCOM_Debug_Interface ; {UNTESTED}

begin
    Debugger := nil ;
end ;
{$ENDIF}



{ TManaged_Store64 methods... }

{ Overrides... }

function TManaged_Store64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := MaxSpace ;
end ;


{ API... }

function TManaged_Store64.Allocate( Size : TStore_Address64 ) : TStore_Address64 ;

begin
    Allocate := 0 ;
end ;


function TManaged_Store64.Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;

begin
    Allocate_At := False ;
end ;


procedure TManaged_Store64.Copy( Source, Destination : TStore_Address64 ;
    Count : integer ) ; { Copy data } {UNTESTED}

var Buffer : array[ 0..32767 ] of char ;

begin
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Read( Source, sizeof( Buffer ), Buffer ) ;
            Write( Destination, sizeof( Buffer ), Buffer ) ;
            Source := Source + sizeof( Buffer ) ;
            Destination := Destination + Sizeof( Buffer ) ;
        end else
        begin
            Read( Source, Count, Buffer ) ;
            Write( Destination, Count, Buffer ) ;
            break ; { All done }
        end ;
        dec( Count, sizeof( Buffer ) ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TManaged_Store64.Deallocate( PTR, Size : TStore_Address64 ) ; {UNTESTED}

begin
    Abstract_Error ;
end ;


procedure TManaged_Store64.Fill( PTR : TStore_Address64 ; Count : integer ;
    Value : byte ) ;

var Buffer : array[ 0..32767 ] of char ;

begin
    if( Count > sizeof( Buffer ) ) then
    begin
        fillchar( Buffer, sizeof( Buffer ), Value ) ;
    end else
    begin
        fillchar( Buffer, Count, Value ) ;
    end ;
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Write( PTR, sizeof( Buffer ), Buffer ) ;
        end else
        begin
            Write( PTR, Count, Buffer ) ;
        end ;
        Count := Count - sizeof( Buffer ) ;
        PTR := PTR + sizeof( Buffer ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


function TManaged_Store64.Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
{ This function does a reallocation operation by allocating new memory, copying
  it and then deallocating the old.  Descendants should override this default
  method if they have a more efficient reallocation method. }

var Dummy : TStore_Address64 ;

begin
    Dummy := Allocate( New ) ;
    Reallocate := Dummy ;
    if( Dummy <> 0 ) then { Something was allocated } {UNTESTED}
    begin
        { Copy old data to new location... }
        if( New > Old ) then
        begin
            New := Old ; { Amount to move }
        end ;
{$IFDEF INT64}
        Copy( PTR, Dummy, New ) ;
{$ELSE}
        Copy( PTR, Dummy, trunc( New ) ) ;
{$ENDIF}

        { All done with old data }
        Deallocate( PTR, Old ) ;
    end ;

    { Update statistics }
    inc( Statistics.Reallocations ) ;
end ;


function TManaged_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ; {UNTESTED}

begin
    UEC := nil ;
    Read( Address, _Size, Data ) ;
    Read_Data := _Size ;
end ;


function TManaged_Store64.Write_Data( var Data ;
    Address, _Size : TStore_Address64 ; var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    UEC := nil ;
    Write( Address, _Size, Data ) ;
    Write_Data := _Size ;
end ;


procedure TManaged_Store64.Read( PTR, Size : TStore_Address64 ; var Buffer ) ; {UNTESTED}

begin
    Abstract_Error ;
end ;


procedure TManaged_Store64.Write( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
end ;


function TManaged_Store64.SpaceAvail : TStore_Address64 ;

begin
    SpaceAvail := 0 ;
end ;


function TManaged_Store64.XSpaceAvail : TStore_Address64 ;

begin
    XSpaceAvail := 0 ;
end ;


function TManaged_Store64.MaxSpace : TStore_Address64 ;

begin
    MaxSpace := 0 ;
end ;


procedure TManaged_Store64.Get_Root( var Buf ) ;

begin
end ;


procedure TManaged_Store64.Set_Root( var Buf ) ;

begin
end ;


{$IFNDEF MOBILE}
function TManaged_Store64.Debugger : TCOM_Debug_Interface ; {UNTESTED}

begin
    Debugger := nil ;
end ;
{$ENDIF}



{ TCOM_Managed_Store methods... }

{ Overrides... }

function TCOM_Managed_Store.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := lowercase( string( Name ) ) ;
    Result := S = 'tcom_managed_store' ;
end ;


function TCOM_Managed_Store.Max_Storage : TStore_Size ; stdcall ;

begin
    Max_Storage := MaxSpace ;
end ;


{ API... }

function TCOM_Managed_Store.Allocate( Size : TStore_Address ) : TStore_Address ; stdcall ;

begin
    Allocate := 0 ;
end ;


function TCOM_Managed_Store.Allocate_At( Offset, Size : TStore_Address ) : boolean ; stdcall ;

begin
    Allocate_At := False ;
end ;


procedure TCOM_Managed_Store.Copy( Source, Destination : TStore_Address ;
    Count : integer ) ; { Copy data } {UNTESTED} stdcall ;

var Buffer : array[ 0..32767 ] of char ;

begin
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Read( Source, sizeof( Buffer ), Buffer ) ;
            Write( Destination, sizeof( Buffer ), Buffer ) ;
            Source := Source + sizeof( Buffer ) ;
            Destination := Destination + Sizeof( Buffer ) ;
        end else
        begin
            Read( Source, Count, Buffer ) ;
            Write( Destination, Count, Buffer ) ;
            break ; { All done }
        end ;
        dec( Count, sizeof( Buffer ) ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TCOM_Managed_Store.Deallocate( PTR, Size : TStore_Address ) ; {UNTESTED} stdcall ;

begin
    Abstract_Error ;
end ;


procedure TCOM_Managed_Store.Fill( PTR : TStore_Address ; Count : integer ; Value : byte ) ; stdcall ;

var Buffer : array[ 0..32767 ] of char ;

begin
    if( Count > sizeof( Buffer ) ) then
    begin
        fillchar( Buffer, sizeof( Buffer ), Value ) ;
    end else
    begin
        fillchar( Buffer, Count, Value ) ;
    end ;
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Write( PTR, sizeof( Buffer ), Buffer ) ;
        end else
        begin
            Write( PTR, Count, Buffer ) ;
        end ;
        Count := Count - sizeof( Buffer ) ;
        PTR := PTR + sizeof( Buffer ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


function TCOM_Managed_Store.Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ; stdcall ;
{ This function does a reallocation operation by allocating new memory, copying
  it and then deallocating the old.  Descendants should override this default
  method if they have a more efficient reallocation method. }

var Dummy : longint ;

begin
    Dummy := Allocate( New ) ;
    Reallocate := Dummy ;
    if( Dummy <> 0 ) then { Something was allocated } {UNTESTED}
    begin
        { Copy old data to new location... }
        if( New > Old ) then
        begin
            New := Old ; { Amount to move }
        end ;
        Copy( PTR, Dummy, New ) ;

        { All done with old data }
        Deallocate( PTR, Old ) ;
    end ;

    { Update statistics }
    inc( Statistics.Reallocations ) ;
end ;


function TCOM_Managed_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUnified_Exception ) : TStore_Address ; {UNTESTED} stdcall ;

begin
    UEC := nil ;
    Read( Address, _Size, Data ) ;
    Read_Data := _Size ;
end ;


function TCOM_Managed_Store.Write_Data( var Data ;
    Address, _Size : TStore_Address ; var UEC : TUnified_Exception ) : TStore_Address ; stdcall ;

begin
    UEC := nil ;
    Write( Address, _Size, Data ) ;
    Write_Data := _Size ;
end ;


procedure TCOM_Managed_Store.Read( PTR, Size : TStore_Address ; var Buffer ) ; {UNTESTED} stdcall ;

begin
    Abstract_Error ;
end ;


procedure TCOM_Managed_Store.Write( PTR, Size : TStore_Address ; var Buffer ) ; stdcall ;

begin
end ;


function TCOM_Managed_Store.SpaceAvail : TStore_Address ; stdcall ;

begin
    SpaceAvail := 0 ;
end ;


function TCOM_Managed_Store.XSpaceAvail : TStore_Address ; stdcall ;

begin
    XSpaceAvail := 0 ;
end ;


function TCOM_Managed_Store.MaxSpace : TStore_Address ; stdcall ;

begin
    MaxSpace := 0 ;
end ;


procedure TCOM_Managed_Store.Get_Root( var Buf ) ; stdcall ;

begin
end ;


procedure TCOM_Managed_Store.Set_Root( var Buf ) ; stdcall ;

begin
end ;


{$IFNDEF MOBILE}
function TCOM_Managed_Store.Debugger : TCOM_Debug_Interface ; {UNTESTED} stdcall ;

begin
    Debugger := nil ;
end ;
{$ENDIF}



{ TCOM_Managed_Store64 methods... }

{ Overrides... }

function TCOM_Managed_Store64.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := lowercase( string( Name ) ) ;
    Result := S = 'tcom_managed_store64' ;
end ;


function TCOM_Managed_Store64.Max_Storage : TStore_Size64 ; stdcall ;

begin
    Max_Storage := MaxSpace ;
end ;


{ API... }

function TCOM_Managed_Store64.Allocate( Size : TStore_Address64 ) : TStore_Address64 ; stdcall ;

begin
    Allocate := 0 ;
end ;


function TCOM_Managed_Store64.Allocate_At( Offset, Size : TStore_Address64 ) : boolean ; stdcall ;

begin
    Allocate_At := False ;
end ;


procedure TCOM_Managed_Store64.Copy( Source, Destination : TStore_Address64 ;
    Count : int64 ) ; { Copy data } {UNTESTED} stdcall ;

var Buffer : array[ 0..32767 ] of char ;

begin
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Read( Source, sizeof( Buffer ), Buffer ) ;
            Write( Destination, sizeof( Buffer ), Buffer ) ;
            Source := Source + sizeof( Buffer ) ;
            Destination := Destination + Sizeof( Buffer ) ;
        end else
        begin
            Read( Source, Count, Buffer ) ;
            Write( Destination, Count, Buffer ) ;
            break ; { All done }
        end ;
        dec( Count, sizeof( Buffer ) ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TCOM_Managed_Store64.Deallocate( PTR, Size : TStore_Address64 ) ; {UNTESTED} stdcall ;

begin
    Abstract_Error ;
end ;


procedure TCOM_Managed_Store64.Fill( PTR : TStore_Address64 ; Count : integer ;
    Value : byte ) ; stdcall ;

var Buffer : array[ 0..32767 ] of char ;

begin
    if( Count > sizeof( Buffer ) ) then
    begin
        fillchar( Buffer, sizeof( Buffer ), Value ) ;
    end else
    begin
        fillchar( Buffer, Count, Value ) ;
    end ;
    while( Count > 0 ) do
    begin
        if( Count > sizeof( Buffer ) ) then
        begin
            Write( PTR, sizeof( Buffer ), Buffer ) ;
        end else
        begin
            Write( PTR, Count, Buffer ) ;
        end ;
        Count := Count - sizeof( Buffer ) ;
        PTR := PTR + sizeof( Buffer ) ;
    end ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


function TCOM_Managed_Store64.Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ; stdcall ;
{ This function does a reallocation operation by allocating new memory, copying
  it and then deallocating the old.  Descendants should override this default
  method if they have a more efficient reallocation method. }

var Dummy : TStore_Address64 ;

begin
    Dummy := Allocate( New ) ;
    Reallocate := Dummy ;
    if( Dummy <> 0 ) then { Something was allocated } {UNTESTED}
    begin
        { Copy old data to new location... }
        if( New > Old ) then
        begin
            New := Old ; { Amount to move }
        end ;
{$IFDEF INT64}
        Copy( PTR, Dummy, New ) ;
{$ELSE}
        Copy( PTR, Dummy, trunc( New ) ) ;
{$ENDIF}

        { All done with old data }
        Deallocate( PTR, Old ) ;
    end ;

    { Update statistics }
    inc( Statistics.Reallocations ) ;
end ;


function TCOM_Managed_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ; {UNTESTED} stdcall ;

begin
    UEC := nil ;
    Read( Address, _Size, Data ) ;
    Read_Data := _Size ;
end ;


function TCOM_Managed_Store64.Write_Data( var Data ;
    Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ; stdcall ;

begin
    UEC := nil ;
    Write( Address, _Size, Data ) ;
    Write_Data := _Size ;
end ;


function TCOM_Managed_Store64.Read( PTR, Size : TStore_Address64 ;
    var Buffer ) : TStore_Address64 ; {UNTESTED} stdcall ;

var UEC : TUnified_Exception ;

begin
    Result := Read_Data( Buffer, PTR, Size, UEC ) ;
end ;


function TCOM_Managed_Store64.Write( PTR, Size : TStore_Address64 ;
    var Buffer ) : TStore_Address64 ; stdcall ;

var UEC : TUnified_Exception ;

begin
    Result := Write_Data( Buffer, PTR, Size, UEC ) ;
end ;


function TCOM_Managed_Store64.SpaceAvail : TStore_Address64 ; stdcall ;

begin
    SpaceAvail := 0 ;
end ;


function TCOM_Managed_Store64.XSpaceAvail : TStore_Address64 ; stdcall ;

begin
    XSpaceAvail := 0 ;
end ;


function TCOM_Managed_Store64.MaxSpace : TStore_Address64 ; stdcall ;

begin
    MaxSpace := 0 ;
end ;


procedure TCOM_Managed_Store64.Get_Root( var Buf ) ; stdcall ;

begin
end ;


procedure TCOM_Managed_Store64.Set_Root( var Buf ) ; stdcall ;

begin
end ;


{$IFNDEF MOBILE}
function TCOM_Managed_Store64.Debugger : TCOM_Debug_Interface ; {UNTESTED} stdcall ;

begin
    Debugger := nil ;
end ;
{$ENDIF}


function TCOM_Managed_Store64.Allocation_Table_Size : longint ;

begin
    Result := 0 ;
end ;


procedure TCOM_Managed_Store64.Get_Allocation_Table_Data( Buff : pointer ;
    _Size : longint ) ;

begin
end ;


procedure TCOM_Managed_Store64.Set_Allocation_Table_Data( Buff : pointer ;
    _Size : longint ) ;

begin
end ;


procedure TCOM_Managed_Store64.Set_Min_Storage( _Size : longint ) ;

begin
end ;


procedure TCOM_Managed_Store64.Set_Allocation_Table_Offset( Value : TStore_Address64 ) ;

begin
end ;


function TCOM_Managed_Store64.Get_Allocation_Table_Offset : TStore_Address64 ;

begin
    Result := 0 ;
end ;


procedure TCOM_Managed_Store64.Flush_Allocation_Table ;

begin
end ;


function TCOM_Managed_Store64.Get_Store : TCOM_Store64 ;

begin
    Result := nil ;
end ;


procedure TCOM_Managed_Store64.Set_Store( Value : TCOM_Store64 ) ;

begin
end ;


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


  {$IFDEF CC_Debug}
procedure Test_Logging ;

var OK : boolean ;
    Test : TSManaged_Store ;
    Test64 : TSManaged_Store64 ;

begin
    Test.Init ;
    if( Test.Check_Access( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Test.Log_Allocation( 0, 10 ) ;
    if( not Test.Check_Access( 0, 10 ) ) then
    begin
        Error ;
    end ;
    if( Test.Check_Access( 0, 16 ) ) then
    begin
        Error ;
    end ;
    Test.Log_Deallocation( 0, 10 ) ;
  {$IFDEF FPC_OR_DELPHI}
    try
        Test.Log_Deallocation( 0, 10 ) ;
        OK := False ;
    except
        OK := True ;
    end ;
    if( not OK ) then
    begin
        Error ;
    end ;
  {$ENDIF}
    Test.Done ;

    Test64.Init ;
    if( Test64.Check_Access( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Test64.Log_Allocation( 0, 10 ) ;
    if( not Test64.Check_Access( 0, 10 ) ) then
    begin
        Error ;
    end ;
    if( Test64.Check_Access( 0, 16 ) ) then
    begin
        Error ;
    end ;
    Test64.Log_Deallocation( 0, 10 ) ;
  {$IFDEF FPC_OR_DELPHI}
    try
        Test64.Log_Deallocation( 0, 10 ) ;
        OK := False ;
    except
        OK := True ;
    end ;
    if( not OK ) then
    begin
        Error ;
    end ;
  {$ENDIF}
    Test64.Done ;
end ;
{$ENDIF}


procedure Test_Unit ;

var Buffer : integer ;
    Test : TSManaged_Store ;
    UEC : PSUnified_Exception ;
    U_E : TUnified_Exception ;
{$IFDEF FPC_OR_DELPHI}
    Test64 : TSManaged_Store64 ;
    Test1 : TManaged_Store ;
    Test164 : TManaged_Store64 ;
    Test_Com : TCOM_Managed_Store ;
    Test_Com64 : TCOM_Managed_Store64 ;
{$ENDIF}

begin
  {$IFDEF CC_Debug}
    Test_Logging ;
  {$ENDIF}

    { Test TSManaged_Store... }
    Test.Init ;

    if( Test.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate( 10 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate_At( 10, 10 ) ) then
    begin
        Error ;
    end ;
    Test.Reallocate( 10, 10, 16 ) ;

    Test.Fill( 10, 10, 0 ) ;
    Test.Fill( 10, 10, 0 ) ;
    Test.Fill( 10, 65535, 0 ) ;

    if( Test.SpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.XSpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace <> 0 ) then
    begin
        Error ;
    end ;

    Test.Get_Root( Buffer ) ;
    Test.Set_Root( Buffer ) ;

    if( Test.Write_Data( Buffer, 10, 5, UEC ) <> 5 ) then
    begin
        Error ;
    end ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;

    Test.Done ;

{$IFDEF FPC_OR_DELPHI}
    { Test TSManaged_Store64... }
    Test64.Init ;

    if( Test64.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.Allocate( 10 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.Allocate_At( 10, 10 ) ) then
    begin
        Error ;
    end ;
    Test64.Reallocate( 10, 10, 16 ) ;

    Test64.Fill( 10, 10, 0 ) ;
    Test64.Fill( 10, 10, 0 ) ;
    Test64.Fill( 10, 65535, 0 ) ;

    if( Test64.SpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.XSpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.MaxSpace <> 0 ) then
    begin
        Error ;
    end ;

    Test64.Get_Root( Buffer ) ;
    Test64.Set_Root( Buffer ) ;

    if( Test64.Write_Data( Buffer, 10, 5, UEC ) <> 5 ) then
    begin
        Error ;
    end ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;

    Test64.Done ;

    { Test TManaged_Store... }
    Test1 := TManaged_Store.Create ;

    if( Test1.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.Allocate( 10 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.Allocate_At( 10, 10 ) ) then
    begin
        Error ;
    end ;
    Test1.Reallocate( 10, 10, 16 ) ;

    Test1.Fill( 10, 10, 0 ) ;
    Test1.Fill( 10, 10, 0 ) ;
    Test1.Fill( 10, 65535, 0 ) ;

    if( Test1.SpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.XSpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.MaxSpace <> 0 ) then
    begin
        Error ;
    end ;

    Test1.Get_Root( Buffer ) ;
    Test1.Set_Root( Buffer ) ;

    if( Test1.Write_Data( Buffer, 10, 5, U_E ) <> 5 ) then
    begin
        Error ;
    end ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;

    Test1.Free ;

    { Test TManaged_Store64... }
    Test164 := TManaged_Store64.Create ;

    if( Test164.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.Allocate( 10 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.Allocate_At( 10, 10 ) ) then
    begin
        Error ;
    end ;
    Test164.Reallocate( 10, 10, 16 ) ;

    Test164.Fill( 10, 10, 0 ) ;
    Test164.Fill( 10, 10, 0 ) ;
    Test164.Fill( 10, 65535, 0 ) ;

    if( Test164.SpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.XSpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.MaxSpace <> 0 ) then
    begin
        Error ;
    end ;

    Test164.Get_Root( Buffer ) ;
    Test164.Set_Root( Buffer ) ;

    if( Test164.Write_Data( Buffer, 10, 5, U_E ) <> 5 ) then
    begin
        Error ;
    end ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;

    Test164.Free ;

    { Test TCOM_Managed_Store... }
    Test_COM := TCOM_Managed_Store.Create ;

    if( Test_COM.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Allocate( 10 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Allocate_At( 10, 10 ) ) then
    begin
        Error ;
    end ;
    Test_COM.Reallocate( 10, 10, 16 ) ;

    Test_COM.Fill( 10, 10, 0 ) ;
    Test_COM.Fill( 10, 10, 0 ) ;
    Test_COM.Fill( 10, 65535, 0 ) ;

    if( Test_COM.SpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.XSpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.MaxSpace <> 0 ) then
    begin
        Error ;
    end ;

    Test_COM.Get_Root( Buffer ) ;
    Test_COM.Set_Root( Buffer ) ;

    if( Test_COM.Write_Data( Buffer, 10, 5, U_E ) <> 5 ) then
    begin
        Error ;
    end ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;

    Test_COM.Free ;

    { Test TCOM_Managed_Store64... }
    Test_COM64 := TCOM_Managed_Store64.Create ;

    if( Test_COM64.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM64.Allocate( 10 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM64.Allocate_At( 10, 10 ) ) then
    begin
        Error ;
    end ;
    Test_COM64.Reallocate( 10, 10, 16 ) ;

    Test_COM64.Fill( 10, 10, 0 ) ;
    Test_COM64.Fill( 10, 10, 0 ) ;
    Test_COM64.Fill( 10, 65535, 0 ) ;

    if( Test_COM64.SpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM64.XSpaceAvail <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM64.MaxSpace <> 0 ) then
    begin
        Error ;
    end ;

    Test_COM64.Get_Root( Buffer ) ;
    Test_COM64.Set_Root( Buffer ) ;

    if( Test_COM64.Write_Data( Buffer, 10, 5, U_E ) <> 5 ) then
    begin
        Error ;
    end ;
    if( UEC <> nil ) then
    begin
        Error ;
    end ;

    Test_COM64.Free ;
  {$ENDIF}
end ;
{$ENDIF}



end.
