{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
{
        Program Name : NameVals
        Package Name : Subroutine Library - Layer 1
        Purpose      : Standard named values class
        Institution  : Conroy & Conroy Co.
        Date Written : 14-April-2008
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2008 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TStandard_Named_Values is the default implementation of the named
        values class.
}

unit NameVals ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF FPC_OR_DELPHI}
     Classes, // TStringList
{$ENDIF}

     { Subroutine Library - Layer 0... }
     _Common, { TObject }
     _NameVal ; { TNamed_Values }

{$IFDEF DYNAMIC_OBJECTS}
type TStandard_Named_Values = class( TNamed_Values )
                                  public // Constructors and destructors...
                                      constructor Create ;
                                      destructor Destroy ; override ;

                                  private // Instance data...
                                      Names, Values : TStringList ;
                                      References : integer ;

                                  public { API... }
                                      procedure Attach ; override ;
                                      procedure Detach ; override ;
                                      function Is_Class( const Name : Ansistring ) : boolean ;
                                          override ;
                                      function Index_Of( Name : Ansistring ) : integer ;
                                          override ;
                                      function Index_Of_Value( Value : Ansistring ) : integer ;
                                          override ;
                                      function Get( Index : integer ) : Ansistring ;
                                          override ;
                                      function Get_Name( Index : integer ) : Ansistring ;
                                          override ;
                                      function Put( Index : integer ; Value : Ansistring ) : integer ;
                                          override ;
                                      function Put_by_Name( Name, Value : Ansistring ) : integer ;
                                          override ;
                                      procedure Delete( Index : integer ) ;
                                          override ;
                              end ;

     TStandard_COM_Named_Values = class( TCOM_Named_Values )
                                      public // Constructors and destructors...
                                          constructor Create ;
                                          destructor Destroy ; override ;

                                      private // Instance data...
                                          Names, Values : TStringList ;
                                          Temp : Ansistring ;
                                          References : integer ;

                                      public // API...
                                          function Is_Class( _N : PAnsiChar ) : boolean ;
                                              override ;
                                          procedure Attach ; override ;
                                          procedure Detach ; override ;
                                          function Index_Of( Name : PAnsiChar ) : integer ;
                                              override ;
                                          function Index_Of_Value( Value : PAnsiChar ) : integer ;
                                              override ;
                                          function Get( Index : integer ) : PAnsiChar ;
                                              override ;
                                          function Get_Name( Index : integer ) : PAnsiChar ;
                                              override ;
                                          function Put( Index : integer ;
                                              Value : PAnsiChar ) : integer ;
                                              override ;
                                          function Put_by_Name( Name, Value : PAnsiChar ) : integer ;
                                              override ;
                                          procedure Delete( Index : integer ) ;
                                              override ;
                                  end ;
{$ENDIF}

type PSStandard_Named_Values = ^TSStandard_Named_Values ;

     TSStandard_Named_Values = object( TSNamed_Values )
                                   public { API... }
                                      function Index_Of( Name : Ansistring ) : integer ;
                                          virtual ;
                                      function Index_Of_Value( Value : Ansistring ) : integer ;
                                          virtual ;
                                      function Get( Index : integer ) : Ansistring ;
                                          virtual ;
                                      function Get_Name( Index : integer ) : Ansistring ;
                                          virtual ;
                                      function Put( Index : integer ; Value : Ansistring ) : integer ;
                                          virtual ;
                                      function Put_by_Name( Name, Value : Ansistring ) : integer ;
                                          virtual ;
                                      procedure Delete( Index : integer ) ;
                                          virtual ;
                               end ;

implementation

{$IFDEF DYNAMIC_OBJECTS}
uses Sysutils ; { lowercase }
{$ENDIF}

{ TSNamed_Values methods... }

function TSStandard_Named_Values.Index_Of( Name : Ansistring ) : integer ;

begin
    Index_Of := -1 ;
    Abstract_Error ;
end ;


function TSStandard_Named_Values.Index_Of_Value( Value : Ansistring ) : integer ;

begin
    Index_Of_Value := -1 ;
    Abstract_Error ;
end ;


function TSStandard_Named_Values.Get( Index : integer ) : Ansistring ;

begin
    Get := '' ;
    Abstract_Error ;
end ;


function TSStandard_Named_Values.Get_Name( Index : integer ) : Ansistring ;

begin
    Get_Name := '' ;
    Abstract_Error ;
end ;


function TSStandard_Named_Values.Put( Index : integer ; Value : Ansistring ) : integer ;

begin
    Put := -1 ;
    Abstract_Error ;
end ;


function TSStandard_Named_Values.Put_by_Name( Name, Value : Ansistring ) : integer ;

begin
    Put_by_Name := -1 ;
    Abstract_Error ;
end ;


procedure TSStandard_Named_Values.Delete( Index : integer ) ;

begin
    Abstract_Error ;
end ;


{$IFDEF DYNAMIC_OBJECTS}
{ TStandard_Named_Values methods... }

constructor TStandard_Named_Values.Create ;

begin
    inherited Create ;

    Names := TStringList.Create ;
    Values := TStringList.Create ;
end ;


destructor TStandard_Named_Values.Destroy ;

begin
    Names.Free ;
    Values.Free ;
    Names := nil ;
    Values := nil ;

    inherited Destroy ;
end ;


{ API... }

procedure TStandard_Named_Values.Attach ;

begin
    inc( References ) ;
end ;


procedure TStandard_Named_Values.Detach ;

begin
    dec( References ) ;
    if( References <= 0 ) then
    begin
        Free ;
    end ;
end ;


function TStandard_Named_Values.Is_Class( const Name : Ansistring ) : boolean ;

begin
    if( Name = 'TStandard_Named_Values' ) then
    begin
        Result := True ;
    end else
    begin
        Result := False ; // No call to inherited since all ancestors have this method as abstract
    end ;
end ;


function TStandard_Named_Values.Index_Of( Name : Ansistring ) : integer ;

begin
    Result := Names.IndexOf( Name ) ;
end ;


function TStandard_Named_Values.Index_Of_Value( Value : Ansistring ) : integer ;

begin
    Result := Values.IndexOf( Value ) ;
end ;


function TStandard_Named_Values.Get( Index : integer ) : Ansistring ;

begin
    if( ( Index < 0 ) or ( Index >= Values.Count ) ) then
    begin
        Result := '' ;
        exit ;
    end ;
    Result := Values[ Index ] ;
end ;


function TStandard_Named_Values.Get_Name( Index : integer ) : Ansistring ;

begin
    if( ( Index < 0 ) or ( Index >= Values.Count ) ) then
    begin
        Result := '' ;
        exit ;
    end ;
    Result := Names[ Index ] ;
end ;


function TStandard_Named_Values.Put( Index : integer ; Value : Ansistring ) : integer ;

begin
    if( ( Index < 0 ) or ( Index >= Values.Count ) ) then
    begin
        Result := -1 ;
        exit ;
    end ;

    Result := Index ;
    Values[ Index ] := Value ;
end ;


function TStandard_Named_Values.Put_by_Name( Name, Value : Ansistring ) : integer ;

begin
    Result := Index_Of( Name ) ;
    if( Result = -1 ) then
    begin
        Result := Values.Count ;
        Names.Add( string( Name ) ) ;
        Values.Add( string( Value ) ) ;
        exit ;
    end ;
    Values[ Result ] := Value ;
end ;


procedure TStandard_Named_Values.Delete( Index : integer ) ;

begin
    Names.Delete( Index ) ;
    Values.Delete( Index ) ;
end ;



{ TStandard_COM_Named_Values methods... }

constructor TStandard_COM_Named_Values.Create ;

begin
    inherited Create ;

    Names := TStringList.Create ;
    Values := TStringList.Create ;
end ;


destructor TStandard_COM_Named_Values.Destroy ;

begin
    Names.Free ;
    Values.Free ;
    Names := nil ;
    Values := nil ;

    inherited Destroy ;
end ;


{ API... }

function TStandard_COM_Named_Values.Is_Class( _N : PAnsiChar ) : boolean ;

var N : Ansistring ;

begin
    N := lowercase( string( _N ) ) ;
    Result := ( N = 'tstandard_com_named_values' ) ;
end ;


procedure TStandard_COM_Named_Values.Attach ;

begin
    inc( References ) ;
end ;


procedure TStandard_COM_Named_Values.Detach ;

begin
    dec( References ) ;
    if( References <= 0 ) then
    begin
        Free ;
    end ;
end ;


function TStandard_COM_Named_Values.Index_Of( Name : PAnsiChar ) : integer ;

begin
    Result := Names.IndexOf( string( Name ) ) ;
end ;


function TStandard_COM_Named_Values.Index_Of_Value( Value : PAnsiChar ) : integer ;

begin
    Result := Values.IndexOf( string( Value ) ) ;
end ;


function TStandard_COM_Named_Values.Get( Index : integer ) : PAnsiChar ;

begin
    if( ( Index < 0 ) or ( Index >= Values.Count ) ) then
    begin
        Result := nil ;
        exit ;
    end ;
    Temp := Values[ Index ] ;
    Result := PAnsiChar( Temp ) ;
end ;


function TStandard_COM_Named_Values.Get_Name( Index : integer ) : PAnsiChar ;

begin
    if( ( Index < 0 ) or ( Index >= Values.Count ) ) then
    begin
        Result := nil ;
        exit ;
    end ;
    Temp := Names[ Index ] ;
    Result := PAnsiChar( Temp ) ;
end ;


function TStandard_COM_Named_Values.Put( Index : integer ; Value : PAnsiChar ) : integer ;

begin
    if( ( Index < 0 ) or ( Index >= Values.Count ) ) then
    begin
        Result := -1 ;
        exit ;
    end ;

    Result := Index ;
    Values[ Index ] := string( Value ) ;
end ;


function TStandard_COM_Named_Values.Put_by_Name( Name, Value : PAnsiChar ) : integer ;

begin
    Result := Index_Of( Name ) ;
    if( Result = -1 ) then
    begin
        Result := Values.Count ;
        Names.Add( string( Name ) ) ;
    end ;
    Values[ Result ] := string( Value ) ;
end ;


procedure TStandard_COM_Named_Values.Delete( Index : integer ) ;

begin
    Names.Delete( Index ) ;
    Values.Delete( Index ) ;
end ;
{$ENDIF}



end.
