{
        Program Name : O_S
        Package Name : Subroutine Library - Layer 3
        Purpose      : Operating system interface
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Nov-2002
        Written By   : Alan Conroy
        Version      : 1.1

        Copyright (C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

          Mar-07-2006   RBJ         Cleaned up memory leaks.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit causes the proper TOperating_System object to be included
        in the application.
}

unit O_S ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

uses { Subroutine Library - Layer 1... }
     _OS ; { POperating_System }

{$I EDEFINES.INC}

procedure Terminate ;
function OS : PSOperating_System ;
{$IFDEF FPC_OR_DELPHI}
function COM_OS : TCOM_Operating_System ;
{$ENDIF}

implementation

uses { Pascal... }
{$IFDEF FPC_OR_DELPHI}
     SysUtils, { lowercase }
{$ENDIF}
     { Subroutine Library - Layer 0... }
{$IFNDEF MOBILE}
     _DebugIn, { TCOM_Debug_Interface }
{$ENDIF}
    _Types, { Ansistring }
     _UE, { TUnified_Exception }

     { Subroutine Library - Layer 1... }
     _DevMan, { TSDevice_Manager }
{$IFDEF FPC_OR_DELPHI}
     DevMan, { COM_Device_Manager }
{$ENDIF}
     _FiP, { TCOM_File_Processor }
     _TPM, { TSText_Presentation_Manager }
     UE, // Create_UE_Wrapper
     //Win_UE, // PSOS_Windows

     { Subroutine Library - Layer 3... }
{$IFDEF MAC}
  {$DEFINE LINUX}
{$ENDIF}
{$IFDEF OSX}
  {$DEFINE LINUX}
{$ENDIF}
{$IFDEF UOS}
{$ELSE}
  {$IFDEF ANDROID}
     OS_Android, { TOS_Android }
  {$ELSE}
    {$IFDEF MOBILE}
     OS_IOS, { TOS_IOS }
    {$ELSE}
      {$IFDEF LINUX}
     OS_Linux, { TOS_Linux }
      {$ELSE}
        {$IFDEF FPC_OR_DELPHI}
     OS_Windows, { TOS_Windows }
        {$ELSE}
     OS_DOS, { TOS_DOS }
        {$ENDIF}
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
{$ENDIF}
     FiP ; // COM_File_Processor

{ Unit API... }

function OS : PSOperating_System ;

begin
    if( _OS.OS = nil ) then
    begin
{$IFDEF UOS}
{$ELSE}
  {$IFDEF ANDROID}
        _OS.OS := new( PSOS_Android, Init ) ; { We don't ever deallocate since this is a singleton needed everywhere }
  {$ELSE}
    {$IFDEF MOBILE}
        _OS.OS := new( PSOS_IOS, Init ) ; { We don't ever deallocate since this is a singleton needed everywhere }
    {$ELSE}
      {$IFDEF LINUX}
          _OS.OS := new( PSOS_Linux, Init ) ; { We don't ever deallocate since this is a singleton needed everywhere }
      {$ELSE}
        {$IFDEF FPC_OR_DELPHI}
          _OS.OS := new( PSOS_Windows, Init ) ; { We don't ever deallocate since this is a singleton needed everywhere }
        {$ELSE}
          _OS.OS := new( POS_DOS, Init ) ; { We don't ever deallocate since this is a singleton needed everywhere }
        {$ENDIF}
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
{$ENDIF}
    end ;
    OS := _OS.OS ;
end ;


{ Unit API... }

{$IFDEF FPC_OR_DELPHI}
type TCOM_OS_Wrapper = class( TCOM_Operating_System )
                           public // Constructors and destructors...
                               constructor Create( OS : PSOperating_System ) ;
                               destructor Destroy ; override ;

                           private // Instance data...
                               _DM : TCOM_Device_Manager ;
                               __OS : PSOperating_System ;
                               Temp : Ansistring ;

                           public // Overrides...
                               function Is_Class( Name : PAnsiChar ) : boolean ;
                                   override ;

                               function Name : PAnsiChar ; override ;
                               function Version : PAnsiChar ; override ;

                               procedure Get_Date( var Year, Month, Day : word ) ;
                                   override ;
                               procedure Get_Time( var Hour, Minute, Second, Hundreths, NS : word ) ;
                                   override ;
                               function Timestamp : int64 ; override ;

                               function PID : longint ; override ;
                               procedure Reschedule ; override ;

                               function Spawn( Path, Parameters, Default_Dir : PAnsiChar ;
                                   Flags : integer ) : boolean ; override ;
                               function Thread_ID : longint ; override ;
                               function Logical_CPUs : integer ; override ;

                               function User_Name : PAnsiChar ; override ;

                               function Device_Manager : TCOM_Device_Manager ;
                                   override ;

                               procedure Beep( Typ : integer ) ; override ;

                               procedure Cancel_Typeahead( Handle : integer ) ;
                                   override ;

                               function Get_Char( Wait : boolean ;
                                   var Valid : boolean ;
                                   Handle : integer ) : Ansichar ;
                                   override ;

                               function Pending_Input( Handle : integer ) : boolean ;
                                   override ;

                               function Current_Directory : PAnsiChar ; override ;

                               procedure Print( T : PAnsiChar ) ; override ;

                               function Last_Error : TUnified_Exception ;
                                   override ;

                               function Error_Text( Code : longint ;
                                   Language : PAnsiChar ) : PAnsiChar ; override ;

                               function Application_Path : PAnsiChar ; override ;

                               function Command_Line : PAnsiChar ; override ;

{$IFNDEF MOBILE}
                               function Debugger : TCOM_Debug_Interface ;
                                   override ;
{$ENDIF}

                               function Get_Symbol( X : PAnsiChar ) : PAnsiChar ;
                                   override ;

                               function Switch_Delimiter_Count : integer ;
                                   override ;

                               function Switch_Delimiters( Index : integer ) : PAnsiChar ;
                                   override ;

                               function File_List_Delimiter : PAnsiChar ; override ;

                               function Get_Temp_Path : PAnsiChar ; override ;

                               function System_Path : PAnsiChar ; override ;

                               function Text_Presentation_Manager : TCOM_Text_Presentation_Manager ;
                                   override ;

                               function Swap_File_List : PAnsiChar ; override ;

                               function Get_Clipboard( Path : PAnsiChar ) : PAnsiChar ;
                                   override ;

                               procedure Set_Clipboard( Path, Value : PAnsiChar ) ;
                                   override ;

                               function Max_File_Size : int64 ; override ;

                               function Object_Version : longint ; override ;

                               function File_Processor : TCOM_File_Processor ;
                                   override ;

                               function Load_Library( Nam : PAnsiChar ) : longint ;
                                   override ;

                               function Get_Symbol_Address( Lib : longint ;
                                   Nam : PAnsiChar ) : int64 ; override ;

                               function Show_Message( Handle : longint ;
                                   Msg, Title : PAnsiChar ; Options : integer ) : integer ;
                                   override ;

                               function Get_Country : PAnsiChar ; override ;

                               function Max_Memory_Map : int64 ; override ;

                               function Memory_Map_File( _Name : PAnsiChar ;
                                   Writeable : boolean ;
                                   Start, Size : int64 ;
                                   Flag : boolean ) : integer ; override ;
                               { Map a file into memory }

                               procedure Unmap_File_From_Memory( Handle : integer ) ;
                                   override ;

                               function Memory_Mapped_File_Base( Handle : integer ) : pointer ;
                                   override ;

                               function Memory_Mapped_File_Size( Handle : integer ) : int64 ;
                                   override ;

                               { Memory... }
                               function Total_Physical_Memory : int64 ;
                                   override ;

                               function Available_Physical_Memory : int64 ;
                                   override ;

                               function Total_Virtual_Memory : int64 ;
                                   override ;

                               function Available_Virtual_Memory : int64 ;
                                   override ;

                               function Application_Data_Path( Default_Root, Path : PAnsiChar ;
                                   Flags : integer ) : PAnsiChar ; override ;

                               function Create_Path( Path : PAnsiChar ) : integer ;
                                   override ;

                               function Shared_Library_Extension : PAnsiChar ;
                                   override ;

                               function Path_Separator : PAnsiChar ;
                                   override ;

                               function Mouse_Position( var X, Y : single ) : boolean ;
                                   override ;

                               function Hardware_Name : PAnsiChar ; override ;
                       end ; // TCOM_OS_Wrapper

// Constructors and destructors...

constructor TCOM_OS_Wrapper.Create( OS : PSOperating_System ) ;

begin
    inherited Create ;

    __OS := OS ;
end ;


destructor TCOM_OS_Wrapper.Destroy ;

begin
    if( _DM <> nil ) then
    begin
        _DM.Free ;
        _DM := nil ;
    end ;
    __OS := nil ;

    inherited Destroy ;
end ;


// Overrides...

function TCOM_OS_Wrapper.Is_Class( Name : PAnsiChar ) : boolean ;

var S : string ;

begin
    S := Name ;
    Result := lowercase( S ) = 'tcom_os_wrapper' ;
end ;


function TCOM_OS_Wrapper.Name : PAnsiChar ;

begin
    Temp := __OS^.Name ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Version : PAnsiChar ;

begin
    Temp := __OS^.Version ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TCOM_OS_Wrapper.Get_Date( var Year, Month, Day : word ) ;

begin
    __OS^.Get_Date( Year, Month, Day ) ;
end ;


procedure TCOM_OS_Wrapper.Get_Time( var Hour, Minute, Second, Hundreths, NS : word ) ;

begin
    __OS^.Get_Time( Hour, Minute, Second, Hundreths, NS ) ;
end ;


function TCOM_OS_Wrapper.Timestamp : int64 ;

begin
    Result := __OS^.Timestamp ;
end ;


function TCOM_OS_Wrapper.PID : longint ;

begin
    Result := __OS^.PID ;
end ;


procedure TCOM_OS_Wrapper.Reschedule ;

begin
    __OS^.Reschedule ;
end ;


function TCOM_OS_Wrapper.Spawn( Path, Parameters, Default_Dir : PAnsiChar ;
    Flags : integer ) : boolean ;

begin
    Spawn := __OS^.Spawn( Ansistring( Path ), Ansistring( Parameters ), Ansistring( Default_Dir ), Flags ) ;
end ;


function TCOM_OS_Wrapper.Thread_ID : longint ;

begin
    Result := __OS^.Thread_ID ;
end ;


function TCOM_OS_Wrapper.Logical_CPUs : integer ;

begin
    Result := __OS^.Logical_CPUs ;
end ;


function TCOM_OS_Wrapper.User_Name : PAnsiChar ;

begin
    Temp := __OS^.User_Name ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Device_Manager : TCOM_Device_Manager ;

begin
    if( _DM = nil ) then
    begin
        _DM := COM_Device_Manager( __OS^.Device_Manager ) ;
    end ;
    Result := _DM ;
end ;


procedure TCOM_OS_Wrapper.Beep( Typ : integer ) ;

begin
    __OS^.Beep( Typ ) ;
end ;


procedure TCOM_OS_Wrapper.Cancel_Typeahead( Handle : integer ) ;

begin
    __OS^.Cancel_Typeahead( Handle ) ;
end ;


function TCOM_OS_Wrapper.Get_Char( Wait : boolean ;
   var Valid : boolean ;
   Handle : integer ) : Ansichar ;

begin
    Result := __OS^.Get_Char( Wait, Valid, Handle ) ;
end ;


function TCOM_OS_Wrapper.Pending_Input( Handle : integer ) : boolean ;

begin
    Result := __OS^.Pending_Input( Handle ) ;
end ;


function TCOM_OS_Wrapper.Current_Directory : PAnsiChar ;

begin
    Temp := __OS^.Current_Directory ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TCOM_OS_Wrapper.Print( T : PAnsiChar ) ;

begin
    __OS^.Print( Ansistring( T ) ) ;
end ;


function TCOM_OS_Wrapper.Last_Error : TUnified_Exception ;

begin
    Result := Create_UE_Wrapper( __OS^.Last_Error ) ;
end ;


function TCOM_OS_Wrapper.Error_Text( Code : longint ;
   Language : PAnsiChar ) : PAnsiChar ;

begin
    Temp := __OS^.Error_Text( Code, string( Language ) ) ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Application_Path : PAnsiChar ;

begin
    Temp := __OS^.Application_Path ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Command_Line : PAnsiChar ;

begin
    Temp := __OS^.Command_Line ;
    Result := PAnsiChar( Temp ) ;
end ;


{$IFNDEF MOBILE}
function TCOM_OS_Wrapper.Debugger : TCOM_Debug_Interface ;

begin
    Result := nil ; // TODO: Implement
end ;
{$ENDIF}


function TCOM_OS_Wrapper.Get_Symbol( X : PAnsiChar ) : PAnsiChar ;

begin
    Temp := __OS^.Get_Symbol( Ansistring( X ) ) ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Switch_Delimiter_Count : integer ;

begin
    Result := __OS^.Switch_Delimiter_Count ;
end ;


function TCOM_OS_Wrapper.Switch_Delimiters( Index : integer ) : PAnsiChar ;

begin
    Temp := __OS^.Switch_Delimiters( Index ) ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.File_List_Delimiter : PAnsiChar ;

begin
    Temp := __OS^.File_List_Delimiter ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Get_Temp_Path : PAnsiChar ;

begin
    Temp := __OS^.Get_Temp_Path ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.System_Path : PAnsiChar ;

begin
    Temp := __OS^.System_Path ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Text_Presentation_Manager : TCOM_Text_Presentation_Manager ;

begin
    Result := nil ; // TODO: Implement a wrapper for Text_Presentation_Manager
end ;


function TCOM_OS_Wrapper.Swap_File_List : PAnsiChar ;

begin
    Temp := __OS^.Swap_File_List ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Get_Clipboard( Path : PAnsiChar ) : PAnsiChar ;

begin
    Temp := __OS^.Get_Clipboard( Ansistring( Path ) ) ;
    Result := PAnsichar( Temp ) ;
end ;


procedure TCOM_OS_Wrapper.Set_Clipboard( Path, Value : PAnsiChar ) ;

begin
    __OS^.Set_Clipboard( Ansistring( Path ), Ansistring( Value ) ) ;
end ;


function TCOM_OS_Wrapper.Max_File_Size : int64 ;

begin
    Result := __OS^.Max_File_Size ;
end ;


function TCOM_OS_Wrapper.Object_Version : longint ;

begin
    Result := __OS^.Object_Version ;
    if( Result > OS_Object_Version ) then
    begin
        Result := OS_Object_Version ;
    end ;
end ;


function TCOM_OS_Wrapper.File_Processor : TCOM_File_Processor ;

begin
    Result := COM_File_Processor ;
end ;


function TCOM_OS_Wrapper.Load_Library( Nam : PAnsiChar ) : longint ;

begin
    Result := __OS^.Load_Library( Ansistring( Nam ) ) ;
end ;


function TCOM_OS_Wrapper.Get_Symbol_Address( Lib : longint ;
    Nam : PAnsiChar ) : int64 ;

begin
    Get_Symbol_Address := __OS^.Get_Symbol_Address( Lib, Ansistring( Nam ) ) ;
end ;


function TCOM_OS_Wrapper.Show_Message( Handle : longint ;
   Msg, Title : PAnsiChar ; Options : integer ) : integer ;

begin
    Result := __OS^.Show_Message( Handle, Ansistring( Msg ), Ansistring( Title ), Options ) ;
end ;


function TCOM_OS_Wrapper.Get_Country : PAnsiChar ;

begin
    Temp := __OS^.Get_Country ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Max_Memory_Map : int64 ;

begin
    Result := __OS^.Max_Memory_Map ;
end ;


function TCOM_OS_Wrapper.Memory_Map_File( _Name : PAnsiChar ; Writeable : boolean ;
    Start, Size : int64 ; Flag : boolean ) : integer ;

begin
    Result := __OS^.Memory_Map_File( Ansistring( _Name ), Writeable, Start, Size, Flag ) ;
end ;


procedure TCOM_OS_Wrapper.Unmap_File_From_Memory( Handle : integer ) ;

begin
    __OS^.Unmap_File_From_Memory( Handle ) ;
end ;


function TCOM_OS_Wrapper.Memory_Mapped_File_Base( Handle : integer ) : pointer ;

begin
    Result := __OS^.Memory_Mapped_File_Base( Handle ) ;
end ;


function TCOM_OS_Wrapper.Memory_Mapped_File_Size( Handle : integer ) : int64 ;

begin
    Result := __OS^.Memory_Mapped_File_Size( Handle ) ;
end ;


function TCOM_OS_Wrapper.Total_Physical_Memory : int64 ;

begin
    Result := __OS^.Total_Physical_Memory ;
end ;


function TCOM_OS_Wrapper.Available_Physical_Memory : int64 ;

begin
    Result := __OS^.Available_Physical_Memory ;
end ;


function TCOM_OS_Wrapper.Total_Virtual_Memory : int64 ;

begin
    Result := __OS^.Total_Virtual_Memory ;
end ;


function TCOM_OS_Wrapper.Available_Virtual_Memory : int64 ;

begin
    Result := __OS^.Available_Virtual_Memory ;
end ;


function TCOM_OS_Wrapper.Application_Data_Path( Default_Root, Path : PAnsiChar ;
    Flags : integer ) : PAnsiChar ;

begin
    Temp := __OS^.Application_Data_Path( Ansistring( Default_Root ), Ansistring( Path ), Flags ) ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Create_Path( Path : PAnsiChar ) : integer ;

begin
    Result := __OS^.Create_Path( Ansistring( Path ) ) ;
end ;


function TCOM_OS_Wrapper.Shared_Library_Extension : PAnsiChar ;

begin
    Temp := __OS^.Shared_Library_Extension ;
    Result := PAnsichar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Path_Separator : PAnsiChar ;

begin
    Temp := __OS^.Path_Separator ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_OS_Wrapper.Mouse_Position( var X, Y : single ) : boolean ;

begin
    Result := __OS^.Mouse_Position( X, Y ) ;
end ;


function TCOM_OS_Wrapper.Hardware_Name : PAnsiChar ;

begin
    Result := nil ;
end ;


var _COM_OS : TCOM_Operating_System = nil ;

function COM_OS : TCOM_Operating_System ;

begin
    if( _COM_OS = nil ) then
    begin
        _COM_OS := TCOM_OS_Wrapper.Create( OS ) ;
    end ;
    Result := _COM_OS ;
end ;
{$ENDIF}


procedure Terminate ;

begin
    dispose( _OS.OS, Done ) ;
    _OS.OS := nil ;
{$IFDEF FPC_OR_DELPHI}
    if( _COM_OS <> nil ) then
    begin
        _COM_OS.Free ;
        _COM_OS := nil ;
    end ;
{$ENDIF}
end ;


{$IFNDEF UNIT_INIT_AND_FINAL}
begin
    __OS := nil ;

{$ELSE}
initialization
{$ENDIF}
    { We don't ever deallocate since this is a singleton needed everywhere }
{$IFDEF UOS}
{$ELSE}
  {$IFNDEF VARIABLE_DECLARATION_ASSIGN}
  _OS.OS := nil ;
  {$ENDIF}
{$ENDIF}

{$IFDEF UNIT_INIT_AND_FINAL}
finalization
  {$IFDEF CC_Auto_Free_Operating_System}
      Terminate ;
  {$ENDIF}
{$ENDIF}

end.
