{
	Program Name : SMUBIOS
	Package Name : Subroutine Library - Layer 3
	Purpose      : Screen Management Utility I/O routines
	Institution  : Conroy & Conroy
	Date Written : 21-Jul-98
	Written By   : Alan Conroy
	Version      : 2.0A

        Copyright (C) 1998 by Alan Conroy.  Released to the public domain.

	TO THE GLORY OF GOD THROUGH JESUS CHRIST

	*********************************************************
	*                                                       *
	*        M O D I F I C A T I O N   H I S T O R Y        *
	*                                                       *
	*********************************************************

	DATE           BY             REASON
        ----           --             ------
        28-Jan-2008    EAC            Rewrite using driver classes.

	*********************************************************
	*                                                       *
	*           P R O G R A M   P U R P O S E               *
	*                                                       *
	*********************************************************

	  These routines provide SMU Basic Input/Output Services.
}

unit SMUBIOS ;

interface

{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF FPC_OR_DELPHI}
     Windows,
     WinAPI, { GetDC }
{$ELSE}
  {$IFDEF CONSOLE}
     Windows,
  {$ENDIF}
{$ENDIF}

     { C&C Subroutine Library - Layer 0... }
     _Common, { TObject }
     TypeDefs, { max_Memory }

     { C&C Subroutine Library - Layer 1... }
     _RGB ; { TRGB }

{ Video modes... }
const VM_PC_Graphics = 5 ;
      VM_Text = 7 ;

{ Hardware models... }
const { 0 is a special value meaning uninitialized }
      Model_Unknown = 1 ; { Use ANSI }
      Model_Rainbow = 2 ; { DEC Rainbow }
      Model_PC_Monochrome = 3 ;
      Model_PC_Color_Text = 4 ; {CGA/EGA/VGA/SVGA in text mode}
      Model_PC_Graphics = 5 ; {CGA/EGA/VGA/SVGA in graphics mode}
      Model_Hercules = 6 ;

type Window_Colors = array[ 0..( Max_Memory div sizeof( TRGB ) ) - 1 ] of TRGB ;
     Window_Colors_Pointer = ^Window_Colors ;

type TSVideo_Driver = object( TObject ) { "Abstract" video driver - unknown video type }
                          public
                              { Hide cursor }
                              procedure Clear_Cursor( VR, VC, Background, Foreground, Attribute : byte ) ;
                                  virtual ;

                              { Return character on screen at row, col. }
                              function Get_Screen( Row, Col : Integer ;
                                  Default : byte ) : Byte ; virtual ;

                              { Position cursor relative to screen }
                              procedure Position( X, Y : integer ) ; virtual ;

                              { Set cursor position, relative to window, and colors }
                              procedure Set_Cursor( X, Y : byte ;
                                  Foreground, Background : byte ) ; virtual ;

                              { Write data to screen }
                              procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                  const Screen_String, Attribute_String : string ;
                                  Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                  Background_Offset, Foreground_Offset : longint ) ;
                                  virtual ;
                      end ;
     PSVideo_Driver = ^TSVideo_Driver ;

function Default_Driver : PSVideo_Driver ; { Return video driver }

procedure Set_Default_Driver( D : PSVideo_Driver ) ; { Set video driver }

procedure IBM_Pos( X, Y : Integer ) ; { Position cursor relative to screen }

procedure _Write_Screen( Driver : PSVideo_Driver ;
    ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

function _Get_Screen( Row, Col : Integer ; Default : byte ) : Byte ;

{ Set cursor position }
procedure _Set_Cursor( X, Y : byte ; Foreground, Background : byte ) ;

{ Set the hardware model }
procedure Set_Model( X : integer ) ;

{ Set the video mode to the specified value. }
procedure Set_Video_Mode( X : integer ) ;


{$IFDEF FPC_OR_DELPHI}
type TSWindows_Driver = object( TSVideo_Driver )
                            public
                                procedure Position( X, Y : integer ) ;
                                    virtual ;
                                procedure Set_Cursor( X, Y : byte ;
                                    Foreground, Background : byte ) ; virtual ;
                                procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                    const Screen_String, Attribute_String : string ;
                                    Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                    Background_Offset, Foreground_Offset : longint ) ;
                                    virtual ;
                        end ;
     PSWindows_Driver = ^TSWindows_Driver ;
{$ENDIF}

implementation

{$IFDEF FPC_OR_DELPHI}
uses Graphics ;
{$ELSE}
uses WinDOS
    {$IFDEF WINDOWS}
    , WinTypes, WinProcs
    {$ELSE}
    , DOS, Graph
    {$ENDIF}
;
{$ENDIF}

{$IFNDEF FPC_OR_DELPHI}
  {$IFNDEF WINDOWS}
type TRegisters = Registers ;
  {$ENDIF}
{$ENDIF}

const Max_Row = 24 ;
      Max_Column = 80 ;

{$IFNDEF FPC_OR_DELPHI}
type Font_Array_Big = Array[ 0..2559 ] of Byte ; {Font array}
     Font_Array_Small = Array[ 0..2047 ] of Byte ; {Ditto}

{$I FONT.INC}

var Monochrome : Array[ 0..1999, 0..1 ] of Byte Absolute $B000 : 0 ;
     Rainbow_Data : Array[ 0..4095 ] of Byte Absolute $Ee00 : 0 ;
     Rainbow_Attributes : Array[ 0..4095 ] of Byte Absolute $Ef00 : 0 ;
{$ENDIF}

var _Driver : PSVideo_Driver
{$IFDEF VARIABLE_DECLARATION_ASSIGN}
      = nil
{$ENDIF}
    ;
    Video_Mode : integer ; { VM_* }
    Model : integer ; { Model_* }
    Row_Start : Array[ 1..Max_Row ] of Integer ; {Start of rows in memory}

// VIDEO DRIVERS ---------------------------------------------------------------

// TSVideo_Driver methods...

procedure TSVideo_Driver.Clear_Cursor( VR, VC, Background, Foreground, Attribute : byte ) ;

begin
    { This routine intentionally left blank }
end ;


function TSVideo_Driver.Get_Screen( Row, Col : Integer ;
    Default : byte ) : Byte ;

begin
    Get_Screen := Default ;
end ;


procedure TSVideo_Driver.Position( X, Y : integer ) ;

begin
    Write( '[', X, ';', Y, 'H' ) ; { ANSI positioning }
end ;


procedure TSVideo_Driver.Set_Cursor( X, Y : byte ;
    Foreground, Background : byte ) ;

begin
    Write( '[', X, ';', Y - 1, 'H' ) ;
end ;


procedure TSVideo_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

begin
    Position( L, C ) ;
    Write( Screen_String ) ;
end ;


{$IFDEF DIRECT_RAM_ACCESS}
type TSRainbow_Driver = object( TSVideo_Driver ) { DEC Rainbow video driver }
                            public
                                procedure Clear_Cursor( VR, VC, Background, Foreground, Attribute : byte ) ; virtual ;
                                function Get_Screen( Row, Col : Integer ;
                                    Default : byte ) : Byte ; virtual ;
                                procedure Set_Cursor( X, Y : byte ;
                                    Foreground, Background : byte ) ; virtual ;
                                procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                    const Screen_String, Attribute_String : string ;
                                    Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                    Background_Offset, Foreground_Offset : longint ) ;
                                    virtual ;
                        end ;
     PSRainbow_Driver = ^TSRainbow_Driver ;

procedure TSRainbow_Driver.Clear_Cursor( VR, VC, Background, Foreground, Attribute : byte ) ;

{$IFNDEF FPC_OR_DELPHI}
var Z : integer ;
    Reg : TRegisters ;
{$ENDIF}

begin
    If( Video_Mode <> VM_Text ) then { Not text mode }
    begin
        Set_Cursor( VR, VC, Foreground, Background ) ;
    end Else
    begin
        Mem[ $Ef00 : Row_Start[ Vr ] + Vc ] := Attribute ; { Set attribute }
    end ;
end ;


function TSRainbow_Driver.Get_Screen( Row, Col : Integer ;
    Default : byte ) : Byte ;

begin
    Get_Screen := Rainbow_Data[ Row_Start[ Row ] + Col - 1 ] ;
end ;


procedure TSRainbow_Driver.Set_Cursor( X, Y : byte ;
    Foreground, Background : byte ) ;

var Offset : integer ;

begin
    Offset := Row_Start[ X ] + Y ; { Offset to cursor position }
    Mem[ $Ef00 : Offset ] := ( Mem[ $Ef00 : Offset ] or 1 ) and ( not 4 ) ;
    { Reverse-video with blink }
end ;


procedure TSRainbow_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

var Temp : integer ;

begin
    Temp := Row_Start[ L ] + C + 1 ; { Destination offset }
    Move( Screen_String[ 1 ], Rainbow_Data[ Temp ], Len ) ; { Write characters }
    Move( Attribute_String[ 1 ], Rainbow_Attributes[ Temp ], Len ) ;
    { Write attributes }
end ;



type TSRainbow_Graphics_Driver = object( TSRainbow_Driver ) { Rainbow graphics mode video driver }
                                     private { Internal urility routines... }
                                         { Wait for GDC to be idle - awaiting commands }
                                         procedure Wait_Idle ;

                                         { Write character C to location X,Y.  1,1 is home position. }
                                         procedure Write_Text( X, Y : Byte ;
                                             C : Char ) ;

                                     public { API... }
                                         procedure Position( X, Y : integer ) ; virtual ;
                                         procedure Set_Cursor( X, Y : byte ;
                                             Foreground, Background : byte ) ; virtual ;
                                         procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                             const Screen_String, Attribute_String : string ;
                                             Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                             Background_Offset, Foreground_Offset : longint ) ;
                                             virtual ;
                                 end ;
     PSRainbow_Graphics_Driver = ^TSRainbow_Graphics_Driver ;

{ Internal urility routines... }

procedure TSRainbow_Graphics_Driver.Wait_Idle ; { Wait for GDC to be idle - awaiting commands }

begin
    inline(
	$E4 / $56 /    {    IN   AL,56H ; FIFO Full?}
	$A8 / $02 /    {    TEST AL,2}
	$74 / $0B /    {    JZ   B2     ; Nope}
	$B9 / $8000 /  {    MOV  CX,8000H ; Maximum wait}
	$E4 / $56 /    {B0: IN   AL,56H ; FIFO full?}
	$A8 / $02 /    {    TEST AL,2}
	$74 / $02 /    {    JZ   B2     ; Nope}
	$E2 / $F8 /    {    LOOP B0     ; Check again}
	$B0 / $0D /    {B2: MOV  AL,0DH}
	$E6 / $57 /    {    OUT  57H,AL ; Turn screen on}
	$E4 / $56 /    {    IN   AL,56H ; Last command fill buffer?}
	$A8 / $02 /    {    TEST AL,2}
	$74 / $0B /    {    JZ   B4     ; Nope}
	$B9 / $8000 /  {    MOV  CX,8000H ; Maximum wait}
	$E4 / $56 /    {B3: IN   AL,56H ; FIFO full?}
	$A8 / $02 /    {    TEST AL,2}
	$75 / $02 /    {    JNZ  B4     ; Nope}
	$E2 / $F8 /    {    LOOP B3     ; Check again}
	$B8 / $040D /  {B4: MOV  AX,40DH}
	$E6 / $57 /    {    OUT  57H,AL ; Another screen-on command}
	$B9 / $8000 /  {    MOV  CX,8000H ; Maximum wait}
	$E4 / $56 /    {B5: IN   AL,56H ; FIFO full?}
	$84 / $E0 /    {    TEST AH,AL}
	$75 / $02 /    {    JNZ  B6     ; Nope}
	$E2 / $F8 /    {    LOOP B5     ; Check again}
	$90 ) ;        {B6: NOP         ; Done}
end ;


procedure TSRainbow_Graphics_Driver.Write_Text( X, Y : Byte ; C : Char ) ;

var Line_Count, Lines_So_Far, Offset, Z, Rows : Integer ;

begin
    X := X - 1 ;
    Y := Y - 1 ; {Normalize row and column}
    If( Video_Mode < 4 ) then {Text mode}
    begin
	Rows := 10 ; {Lines per character}
	Offset := Ord( C ) * Rows {Get offset of character in font table}
    end else
    begin
	Rows := 8 ; {Lines per character}
	Offset := Ord( C ) Shl 3 {Get offset of character in font table}
    end ;
    Z := X * Rows ;
    If( Video_Mode = 6 ) then
        Z := Z Shl 6
    else
        Z := Z Shl 5 ;
    Z := Z + ( Y Shr 1 ) ; {Z:=X*Rows*(either 64 or 32)+Y/2 - Get word address}
    Wait_Idle ;
    If Odd( Y ) then
    begin
	Port[ $54 ] := 00 ;
	Port[ $55 ] := $Ff
    end else
    begin
	Port[ $54 ] := $Ff ;
	Port[ $55 ] := 00
    end ;
    {If Y is odd, then write in right half of word, else write in left half}

    Port[ $53 ] := $Bf ; {Select mode register}
    If( Video_Mode = 6 ) then
        Port[ $51 ] := $B1
    else
        Port[ $51 ] := $B0 ;
    {Word mode}
    Port[ $57 ] := $49 ; {Cursor command}
    Port[ $56 ] := Lo( Z ) ;
    Port[ $56 ] := Hi( Z ) ;
    Port[ $57 ] := $4A ; {Mask command}
    Port[ $56 ] := $Ff ;
    Port[ $56 ] := $Ff ; {Enable everything}
    Line_Count := Rows ; {Lines to display}
    while Line_Count > 0 Do {while lines to write}
    begin
	Lines_So_Far := 0 ;
	Port[ $53 ] := $Fe ; {Select write buffer}
	Port[ $51 ] := 0 ; {Clear counter}
	while ( Lines_So_Far < 8 ) and ( Line_Count > 0 ) Do
	begin
	    If Rows = 10 then
	    begin
		Port[ $52 ] := Font0[ Offset ] ;
		Port[ $52 ] := Font0[ Offset ]
	    end else
	    begin
		Port[ $52 ] := Font1[ Offset ] ;
		Port[ $52 ] := Font1[ Offset ]
	    end ;
	    Offset := Offset + 1 ;
	    Line_Count := Line_Count - 1 ;
	    Lines_So_Far := Lines_So_Far + 1
	    {Increment count of lines in write buffer}
	end ;
	Port[ $53 ] := $Fe ; {Select write buffer} ;
	Port[ $51 ] := 0 ; {Reset counter}
	Port[ $57 ] := $4C ; {FIGs command}
	Port[ $56 ] := 0 ; {Write down}
	Port[ $56 ] := Lines_So_Far - 1 ; {Number of words to write}
	Port[ $56 ] := 0 ;
	Port[ $57 ] := $22 ; {WDAT command}
	Port[ $56 ] := $Ff ;
	Port[ $56 ] := $Ff ;
	Port[ $57 ] := $22 ; {Issue dummy WDAT}
	while ( Port[ $56 ] and 8 ) = 8 do ; {Wait 'till done}
    end ;

    Port[ $53 ] := $Bf ; {Select mode register}
    If( Video_Mode = 6 ) then
        Port[ $51 ] := $B3
    else
        Port[ $51 ] := $B2 ;
    {Restore mode}
    Port[ $54 ] := 0 ;
    Port[ $55 ] := 0
end ;


{ API... }

procedure TSRainbow_Graphics_Driver.Position( X, Y : integer ) ;

begin
    { This routine does nothing in this mode }
end ;


procedure TSRainbow_Graphics_Driver.Set_Cursor( X, Y : byte ;
    Foreground, Background : byte ) ;

var Line_Count, Lines_So_Far : Integer ;
    Offset, Z, Rows : integer ;
    Reg : Tregisters ;

begin
    X := X - 1 ;
    Y := Y - 1 ; {Normalize row and column}
    If( Video_Mode < 4 ) then {Text mode}
    begin
	Rows := 10 ; {Lines per character}
    end Else
    begin
	Rows := 8 ; {Lines per character}
    end ;
    Z := X * Rows ;
    If( Video_Mode = 6 ) then
        Z := Z shl 6
    Else
        Z := Z shl 5 ;
    Z := Z + ( Y shr 1 ) ; {Z:=X*Rows*(either 64 or 32)+Y/2 - Get word address}
    Wait_Idle ;
    If Odd( Y ) Then
    Begin
	Port[ $54 ] := 00 ;
	Port[ $55 ] := $FF ;
    End Else
    Begin
	Port[ $54 ] := $FF ;
	Port[ $55 ] := 00 ;
    End ;
    {If Y is odd, then write in right half of word, else write in left half}

    Wait_Idle ;
    Port[ $53 ] := $EF ; {Select ALU/PS}
    Port[ $51 ] := $10 ; {Complement mode}
    Port[ $53 ] := $F7 ; {Select fgbg}
    Port[ $51 ] := ( Background shl 4 ) + Foreground ;
    {Invert colors}
    Port[ $53 ] := $Bf ; {Select mode register}
    If( Video_Mode = 6 ) Then
        Port[ $51 ] := $B1
    Else
        Port[ $51 ] := $B0 ;
    {Word mode}
    Port[ $57 ] := $49 ; {Cursor command}
    Port[ $56 ] := Lo( Z ) ;
    Port[ $56 ] := Hi( Z ) ;
    Port[ $57 ] := $4A ; {Mask command}
    Port[ $56 ] := $Ff ;
    Port[ $56 ] := $Ff ; {Enable everything}
    Line_Count := Rows ;
    while( Line_Count > 0 ) do {While lines to write}
    begin
	Lines_So_Far := 0 ;
	Port[ $53 ] := $Fe ; {Select write buffer}
	Port[ $51 ] := 0 ; {Clear counter}
	while( ( Lines_So_Far < 8 ) and ( Line_Count > 0 ) ) do
	begin
	    Port[ $52 ] := 0 ;
	    Port[ $52 ] := 0 ;
	    Line_Count := Line_Count - 1 ;
	    Lines_So_Far := Lines_So_Far + 1
	    {Increment count of lines in write buffer}
	end ;
	Port[ $53 ] := $Fe ; {Select write buffer} ;
	Port[ $51 ] := 0 ; {Reset counter}
	Port[ $57 ] := $4C ; {FIGs command}
	Port[ $56 ] := 0 ; {Write down}
	Port[ $56 ] := Lines_So_Far - 1 ; {Number of words to write}
	Port[ $56 ] := 0 ;
	Port[ $57 ] := $22 ; {WDAT command}
	Port[ $56 ] := $Ff ;
	Port[ $56 ] := $Ff ;
	Port[ $57 ] := $22 ; {Issue dummy WDAT}
	while( ( Port[ $56 ] and 8 ) = 8 ) do ; {Wait 'till done}
    end ;

    Port[ $53 ] := $Bf ; {Select mode register}
    If( Video_Mode = 6 ) then
        Port[ $51 ] := $B3
    Else
        Port[ $51 ] := $B2 ;
    {Restore mode}
    Port[ $54 ] := 0 ;
    Port[ $55 ] := 0 ;
    Wait_Idle ;
    Port[ $53 ] := $Ef ; {Select ALU/PS}
    Port[ $51 ] := 0 ; {Replace mode}
    Port[ $53 ] := $F7 ; {Select fgbg}
    Port[ $51 ] := ( Foreground shl 4 ) + Background ;
    {Reinvert colors}
    Wait_Idle ;
end ;


procedure TSRainbow_Graphics_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

var X : integer ;

begin
    X := 1 ;
    while( X <= Len ) do {For each character}
    begin
        Write_Text( L, C, Screen_String[ X ] ) ;
        Inc( C ) ;
        Inc( X ) ;
    end ;
end ;



type TSDOS_Driver = object( TSVideo_Driver ) { PC (DOS) monochrome text mode video driver }
                        public
                            procedure Clear_Cursor( VR, VC, Background, Foreground, Attribute : byte ) ; virtual ;
                            function Get_Screen( Row, Col : Integer ;
                                Default : byte ) : Byte ; virtual ;
                            procedure Position( X, Y : integer ) ; virtual ;
                            procedure Set_Cursor( X, Y : byte ;
                                Foreground, Background : byte ) ; virtual ;
                            procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                const Screen_String, Attribute_String : string ;
                                Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                Background_Offset, Foreground_Offset : longint ) ;
                                virtual ;
                    end ;
     PSDOS_Driver = ^TSDOS_Driver ;

procedure TDOS_Driver.Clear_Cursor( VR, VC, Background, Foreground, Attribute : byte ) ;

begin
    If( Video_Mode <> VM_Text ) then {Non-text}
    Begin
        Reg.Ax := $0800 ; {Read character at current cursor position}
        Intr( $10, Reg ) ; {Get info}
        Z := Reg.Ax ; {Get character}
        Reg.Ax := $0900 or Lo( Z ) ; {Character to write}
        Reg.Cx := 1 ; {Multiplier}
        Reg.Bx := ( Background Shl 4 ) + Foreground ;
        Intr( $10, Reg ) ; {Reverse it now}
    End else {Text mode}
    Begin
        Reg.Ax := $0200 ;
        Reg.Bx := 0 ;
        Reg.Dx := 80 ;
        Intr( $10, Reg ) ; {Make cursor vanish}
    End ;
end ;


function TSDOS_Driver.Get_Screen( Row, Col : Integer ; Default : byte ) : Byte ;

begin
    Get_Screen := Monochrome[ Row_Start[ Row ] + Col - 1, 0 ] ;
end ;


procedure TSDOS_Driver.Position( X, Y : integer ) ; virtual ;

var R : TRegisters ;

begin
    fillchar( R, sizeof( R ), 0 ) ;
    R.AX := $0200 ; { Set cursor position }
    R.BX := 0 ; {Page}
    R.DH := X ;
    R.DL := Y ; {Position}
    Intr( $10, R ) ; {Call BIOS}
end ;


procedure TSDOS_Driver.Set_Cursor( X, Y : byte ;
    Foreground, Background : byte ) ;

begin
end ;


procedure TSDOS_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                  const Screen_String, Attribute_String : string ;
                                  Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                  Background_Offset, Foreground_Offset : longint ) ;

var Row, Char, Char1, Temp : integer ;
    Reg : TRegisters ;

begin
  {$IFDEF MSDOS}
    Temp := ( Row_Start[ L ] + C - 1 ) Shl 1 ; {Destination offset}
    inline(
        $1E /                    {PUSH DS}
        $55 /                    {PUSH BP}
        $B9 / $B000 /            {MOV  CX,0B00H}
        $8E / $C1 /              {MOV  ES,CX}
        $8B / $Be / Temp /       {MOV  DI,[BP].Temp}
        $8B / $8E / Len /        {MOV  CX,[BP].Len}
        $Bb / Attribute_String / {MOV  BX,OFFSET Attribute_String}
        $Be / Screen_String /    {MOV  SI,OFFSET Screen_String}
        $46 /                    {INC  SI}
        $43 /                    {INC  BX}
                                 {10:}
        $8A / $24 /              {MOV  AH,[SI]}
        $26 / $88 / $25 /        {MOV ES:[DI],AH}
        $47 /                    {INC  DI}
        $8A / $27 /              {MOV AH,[BX]}
        $26 / $88 / $25 /        {MOV  ES:[DI],AH}
        $47 /                    {INC  DI}
        $46 /                    {INC  SI}
        $43 /                    {INC  BX}
        $E0 / $F0 /              {LOOPNZ 10}
        $5D /                    {POP  BP}
        $1F                      {POP  DS}
    ) ;
  {$ELSE}
      { Use BIOS if protected mode }
      Dec( L ) ;
      Dec( C ) ;
      for Temp := 1 to Len do
      begin
          Position( L, C ) ; {Position cursor}
          Inc( C ) ;
          Reg.Ah := $09 ;
          Reg.Al := Ord( Screen_String[ Temp ] ) ;
          Reg.Bh := 0 ;
          Reg.Bl := Ord( Attribute_String[ Temp ] ) and 127 ;
          Reg.Cx := 1 ;
          Intr( $10, Reg ) ; {Write character}
      end ;
  {$ENDIF}
end ;


type TSDOS_Color_Driver = object( TSVideo_Driver ) { PC (DOS) color text mode video driver }
                              public
                                  function Get_Screen( Row, Col : Integer ;
                                      Default : byte ) : Byte ; virtual ;
                                  procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                      const Screen_String, Attribute_String : string ;
                                      Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                      Background_Offset, Foreground_Offset : longint ) ;
                                      virtual ;
                          end ;
     PSDOS_Color_Driver = ^TSDOS_Color_Driver ;

function TSDOS_Color_Driver.Get_Screen( Row, Col : Integer ;
    Default : byte ) : Byte ;

var Reg : TRegisters ;

begin
    Reg.Ax := $0800 ;
    Reg.Bx := 0 ;
    Intr( $10, Reg ) ; {Read character}
    _Get_Screen := Reg.Al ;
end ;


procedure TSDOS_Color_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

var CC : String[ 3 ] ;
    Temp : integer ;

begin
    Dec( L ) ;
    Dec( C ) ;
    for Temp := 1 to Len do
    begin
        Position( L, C ) ; {Position cursor}
        Inc( C ) ;
        Reg.Ah := $09 ;
        Reg.Al := Ord( Screen_String[ Temp ] ) ;
        Reg.Bh := 0 ;
        Reg.Bl := Ord( Attribute_String[ Temp ] ) and 127 ;
        Reg.Cx := 1 ;
        Intr( $10, Reg ) ; {Write character}
    end ;
end ;


type TSDOS_Graphics_Driver = object( TSDOS_Driver ) { PC (DOS) graphics mode video driver }
                        public
                            procedure Position( X, Y : integer ) ; virtual ;
                            procedure Set_Cursor( X, Y : byte ;
                                Foreground, Background : byte ) ; virtual ;
                            procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                const Screen_String, Attribute_String : string ;
                                Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                Background_Offset, Foreground_Offset : longint ) ;
                                virtual ;
                    end ;
     PSDOS_Graphics_Driver = ^TSDOS_Graphics_Driver ;

procedure TSDOS_Graphics_Driver.Position( X, Y : integer ) ;

begin
    { Nothing to do in this case }
end ;


procedure TSDOS_Graphics_Driver.Set_Cursor( X, Y : byte ;
    Foreground, Background : byte ) ;

var Reg : Tregisters ;

begin
    Reg.Ax := $0900 ; {Read character}
    Intr( $10, Reg ) ; {Get info}
    Z := Reg.Ax ; {Get character}
    Reg.Ax := $0900 or Lo( Z ) ; {Character to write}
    Reg.Cx := 1 ; {Multiplier}
    Reg.Bx := Foreground * 16 + Background ;
    Intr( $10, Reg ) ; {Reverse it now}
end ;


procedure TSDOS_Graphics_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

begin
    Position( L - 1, C - 1 ) ; { Position cursor }
    C := ( C - 1 ) * Text_Width ;
    L := ( L - 1 ) * Text_Height ;
    Outtextxy( C + Border, L + Border, Screen_String ) ; { Output text }
end ;
{$ENDIF}


{$IFNDEF DIRECT_RAM_ACCESS}
type TSWindows_Console_Driver = object( TSVideo_Driver )
                                    public
                                        procedure Position( X, Y : integer ) ;
                                            virtual ;
                                        procedure Set_Cursor( X, Y : byte ;
                                            Foreground, Background : byte ) ; virtual ;
                                        procedure Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
                                            const Screen_String, Attribute_String : string ;
                                            Backgrounds, Foregrounds : Window_Colors_Pointer ;
                                            Background_Offset, Foreground_Offset : longint ) ;
                                            virtual ;
                                end ;
     PSWindows_Console_Driver = ^TSWindows_Console_Driver ;

procedure TSWindows_Console_Driver.Position( X, Y : integer ) ;

var Point : COORD ;

begin
    Point.Y := X ;
    Point.X := Y ;
    SetConsoleCursorPosition( GetStdHandle( STD_OUTPUT_HANDLE ), Point ) ;
end ;


procedure TSWindows_Console_Driver.Set_Cursor( X, Y : byte ;
    Foreground, Background : byte ) ;

var Info : CONSOLE_CURSOR_INFO ;

begin
    GetConsoleCursorInfo( GetStdHandle( STD_OUTPUT_HANDLE ), Info ) ;
    Info.bVisible := False ;
    SetConsoleCursorInfo( GetStdHandle( STD_OUTPUT_HANDLE ), Info ) ;
    Position( X - 1, Y - 1 ) ;
end ;


procedure TSWindows_Console_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

var Point : COORD ;
    S : string ;
    X : DWord ;

begin
    if( ID = 0 ) then
    begin
        ID := GetStdHandle( STD_OUTPUT_HANDLE ) ;
    end ;
    Point.X := C - 1 ;
    Point.Y := L - 1 ;
    WriteConsoleOutputCharacter( ID,
        PChar( Screen_String ), length( Screen_String ), Point, X ) ;
    if( length( Attribute_String ) > 0 ) then
    begin
        S := '' ;
        for X := 1 to length( Attribute_String ) do
        begin
           S := S + Attribute_String[ X ] + #0 ;
        end ;
        WriteConsoleOutputAttribute( ID,
            PChar( S ), length( Attribute_String ), Point, X ) ;
     end ;
end ;
{$ENDIF}



{$IFDEF FPC_OR_DELPHI}
procedure TSWindows_Driver.Position( X, Y : integer ) ;

begin
    { Do nothing in this mode }
end ;


procedure TSWindows_Driver.Set_Cursor( X, Y : byte ;
    Foreground, Background : byte ) ;

begin
    { Do nothing in this mode }
end ;


procedure TSWindows_Driver.Write_Screen( ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

var Background, Foreground : TRGB ;
    DC : HDC ;
    Column, Len, X : integer ;

begin
    Len := Length( Screen_String ) ;
    DC := GetDC( ID ) ;
    try
        SelectObject( DC, GetStockObject( ANSI_FIXED_FONT ) ) ;
        X := 1 ;
        Column := C ;
        while( Column <= C + Len - 1 ) do
        begin
            Background := Backgrounds[ Background_Offset + X - 1 ] ;
            Foreground := Foregrounds[ Foreground_Offset + X - 1 ] ;
            if( ( ord( Attribute_String[ X ] ) and 1 ) = 0 ) then { Normal video }
            begin
                SetTextColor( DC, RGB( Background.R, Background.G, Background.B ) ) ;
                SetBKColor( DC, RGB( Foreground.R, Foreground.G, Foreground.B ) ) ;
            end else
            begin
                SetTextColor( DC, RGB( Foreground.R, Foreground.G, Foreground.B ) ) ;
                SetBKColor( DC, RGB( Background.R, Background.G, Background.B ) ) ;
            end ;
            SetBKMode( DC, OPAQUE ) ;
            TextOut( DC,
                ( Column - 1 ) * Text_Width + Border,
                ( L - 1 ) * Text_Height + Border,
                @Screen_String[ X ], 1 ) ;
            X := X + 1 ;
            inc( Column ) ;
        end ;
    finally
        ReleaseDC( ID, DC ) ;
    end ;
end ;
{$ENDIF}


// UNIT API --------------------------------------------------------------------

procedure _Clear_Cursor( VR, VC, Background, Foreground, Attribute : byte ) ;

begin
    Default_Driver^.Clear_Cursor( VR, VC, Background, Foreground, Attribute ) ;
end ;


function _Get_Screen( Row, Col : Integer ; Default : byte ) : Byte ;

begin
    _Get_Screen := Default_Driver^.Get_Screen( Row, Col, Default ) ;
end ;


procedure IBM_Pos( X, Y : Integer ) ; { Position cursor relative to screen }

begin
    Default_Driver^.Position( X, Y ) ;
end ;


procedure _Set_Cursor( X, Y : byte ; Foreground, Background : byte ) ;

begin
    Default_Driver^.Set_Cursor( X, Y, Foreground, Background ) ;
end ;


{ Set the hardware model }
procedure Set_Model( X : integer ) ;

var R : integer ;

begin
    if( X <> Video_Mode ) then
    begin
        Model := X ;
        case Model of
             Model_Rainbow : {Rainbow}
                 For R := 1 To 24 Do Row_Start[ R ] := ( R - 1 ) * 135 + 16 ;
             Model_PC_Monochrome..Model_PC_Color_Text : {IBM PC in text mode}
                 For R := 1 To 24 Do Row_Start[ R ] := ( R - 1 ) * 80 ;
        end ;
        if( _Driver <> nil ) then
        begin
            dispose( _Driver, Done ) ;
            _Driver := nil ;
        end ;
    end ;
end ;


{ Set the video mode to the specified value. }
procedure Set_Video_Mode( X : integer ) ;

begin
    if( X <> Video_Mode ) then
    begin
        Video_Mode := X ;
        if( _Driver <> nil ) then
        begin
            dispose( _Driver, Done ) ;
            _Driver := nil ;
        end ;
    end ;
end ;


procedure _Write_Screen( Driver : PSVideo_Driver ;
    ID, L, C, Text_Height, Text_Width, Border : integer ;
    const Screen_String, Attribute_String : string ;
    Backgrounds, Foregrounds : Window_Colors_Pointer ;
    Background_Offset, Foreground_Offset : longint ) ;

begin
    if( Driver = nil ) then
    begin
        Driver := Default_Driver ;
    end ;
    Driver^.Write_Screen( ID, L, C, Text_Height, Text_Width, Border,
        Screen_String, Attribute_String, Backgrounds, Foregrounds,
        Background_Offset, Foreground_Offset ) ;
end ;


function Default_Driver : PSVideo_Driver ; { Return video driver }

begin
    if( _Driver = nil ) then { Need to create a driver }
    begin
{$IFDEF CONSOLE}
        _Driver := new( PSWindows_Console_Driver, Init ) ;
{$ELSE}
  {$IFNDEF DIRECT_RAM_ACCESS}
        _Driver := new( PSWindows_Driver, Init ) ;
  {$ELSE}
       case Model of
           Model_Unknown : _Driver := new( PSVideo_Driver, Init ) ;
           Model_Rainbow :
               if( Video_Mode <> VM_Text ) then { Non-text }
               begin
                   _Driver := new( PSRainbow_Graphics_Driver, Init )
               end else
               begin
                   _Driver := new( PSRainbow_Driver, Init )
               end ;
           Model_PC_Monochrome : _Driver := new( PSDOS_Driver, Init ) ;
           Model_PC_Color_Text : _Driver := new( PSDOS_Color_Driver, Init ) ;
           else _Driver := new( PSDOS_Graphics_Driver, Init ) ;
       end ;
  {$ENDIF}
{$ENDIF}
    end ;
    Default_Driver := _Driver ;
end ;


procedure Set_Default_Driver( D : PSVideo_Driver ) ; { Set video driver }

begin
    _Driver := D ;
end ;


begin
    _Driver := nil ;
end.
