{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
unit Stores ;
{
        Program Name : Stores
        Package Name : Subroutine Library - Layer 1
        Purpose      : Base Storage objects
        Institution  : Conroy & Conroy Co.
        Date Written : 1999
        Written By   : Alan Conroy
        Version      : 1.1B

        Copyright (C) 1999-2008 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

          1-Nov-2002    EAC         Use TObject from Common.
         13-Nov-2002    EAC         Rename Tallocator to
                                    TManaged_Store.
         16-Nov-2005    EAC         Changed const parameters to var
                                    to compile with BP.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TBase_Store is the base class for all storage-related classes.  It
        implements the generic store functionality upon which descendant stores
        are built.

          LEVEL 2 CERTIFICATION: 6-Sep-2008 by EAC
}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

{$I EDEFINES.INC}

uses { C&C Subroutine Library - Layer 0 }
     _Types, { Ansistring }

     { Layer 1... }
     _Cache, { PSCache }
{$IFNDEF MOBILE}
     _DebugIn, { PSDebug_Interface }
{$ENDIF}
     _Stores, { TSStore }
     _UE ; { TUnified_Exception }

const StoreErr_Facility = 48 ;

      StoreErr_Success = 0 ;
      StoreErr_Invalid_Operation = 1 ; { The operation is not valid }
      StoreErr_Abstract = 2 ;

type TStore_Statistics = record
                             Bytes_Read : longint ; { Data bytes read }
                             Bytes_Written : longint ; { Data bytes written }
                             Reads : longint ; { Data read count }
                             Writes : longint ; { Data write count }
                             Error_Count : longint ; { Error count }
                         end ;

type TSBase_Store = object( TSStore )
                        public { Instance data... } {protected}
                            _Cache : PSCache ;
                            _Read_Only : boolean ;
                            _Write_Only : boolean ;
                            Store_Statistics : TStore_Statistics ;

                        public { API... }
                            function Read_Data( var Data ;
                                Address, _Size : TStore_Address ;
                                var UEC : PSUnified_Exception ) : TStore_Address ;
                                virtual ;
                            { Read data from specified Address in store to Data.
                              Read _Size bytes.  Returns success of operation in UEC.
                              Function returns the number of bytes actually read. }

                            function Write_Data( var Data ; Address,
                                _Size : TStore_Address ;
                                var UEC : PSUnified_Exception ) : TStore_Address ;
                                virtual ;
                            { Write data from Data to specified Address in store.
                              Write _Size bytes.  Returns success of operation in UEC.
                              Function returns the number of bytes actually written. }

                            function Max_Storage : TStore_Size ; virtual ;
                            { Returns the maximum (current) size of the store, in
                              bytes.  In other words, this is the highest valid
                              address (plus 1). }

                            function Min_Storage : TStore_Address ; virtual ;
                            { Returns the minimum atomic size of the store.  For
                              instance, this would be the sector size for a disk. }

                            function Extend( Amount : TStore_Address ) : TStore_Address ;
                                virtual ;
                            { Requests extension of store by the amount specified.
                              Returns the amount actually added to store.  If Amount
                              is 0, function returns 0 if store is extendable, and -1
                              if it is not. }

                            function Get_Read_Only : boolean ; virtual ;
                            { Indicates that the store can only be read from. }

                            function Get_Write_Only : boolean ; virtual ;
                            { Indicates that the store can only be written to. }

                            procedure Initialize ; virtual ;
                            { Initializes the TStore object. }

                            procedure Terminate ; virtual ;
                            { Terminates the TStore object. }

                            procedure Format ; virtual ;
                            { Does a low-level format of the store. }

                            function Get_Name : Ansistring ; virtual ;
                            { Returns name of store. }

{$IFNDEF MOBILE}
                            function Debugger : PSDebug_Interface ; virtual ;
                            { Returns a debugger object. }
{$ENDIF}

                            function Get_Cache : PSCache ; virtual ;
                            { Returns current cache object. }

                            procedure Set_Cache( Value : PSCache ) ; virtual ;
                            { Sets the current cache object. }

                            function Contiguous_Store : boolean ; virtual ;
                            { Returns True if the store's addresses are contiguous
                              (non-sparse). }

                            procedure Set_Max_Storage( Value : TStore_Address ;
                                var Res : PSUnified_Exception ) ; virtual ;
                            { Sets the maximum (current) size of the store, in bytes.
                              In other words, this is the new highest valid address
                              (plus 1).  If the store is not resizable, the call
                              fails and an error code is returned in Res.  This should
                              only be used to shrink the size of the store, not to
                              extend it.  To extend, use the Extend method. }

                            function Extended_Size : TStore_Address ; virtual ;
                            { Returns the maximum theoretical size that the store can
                              be extended to.  If the store is not extendable, this
                              returns the same value as Max_Storage. }

                            procedure Set_Read_Only( Value : boolean ) ; virtual ;
                            { Sets the store read-only value. }

                            procedure Set_Write_Only( Value : boolean ) ; virtual ;
                            { Sets the store write-only value. }

                            function Get_Bytes_Read : longint ;
                                virtual ; { Data bytes read }
                            function Get_Bytes_Written : longint ;
                                virtual ; { Data bytes written }
                            function Get_Reads : longint ;
                                virtual ; { Data read count }
                            function Get_Writes : longint ;
                                virtual ; { Data write count }
                            function Get_Error_Count : longint ;
                                virtual ; { Error count }
                            procedure Set_Bytes_Read( Value : longint ) ;
                                virtual ; { Data bytes read }
                            procedure Set_Bytes_Written( Value : longint ) ;
                                virtual ; { Data bytes written }
                            procedure Set_Reads( Value : longint ) ;
                                virtual ; { Data read count }
                            procedure Set_Writes( Value : longint ) ;
                                 virtual ; { Data write count }
                            procedure Set_Error_Count( Value : longint ) ;
                                virtual ; { Error count }
                    end ; { TSBase_Store }
     PSBase_Store = ^TSBase_Store ;

{$IFDEF INT64}
type TSBase_Store64 = object( TSStore64 )
                          protected { Instance data... }
                              _Cache : PSCache64 ;
                              _Read_Only : boolean ;
                              _Write_Only : boolean ;
                              Store_Statistics : TStore_Statistics ;

                          public { API... }
                              function Read_Data( var Data ;
                                  Address, _Size : TStore_Address64 ;
                                  var UEC : PSUnified_Exception ) : TStore_Address64 ; virtual ;
                              { Read data from specified Address in store to Data.
                                Read _Size bytes.  Returns success of operation in UEC.
                                Function returns the number of bytes actually read. }

                              function Write_Data( var Data ; Address,
                                  _Size : TStore_Address64 ;
                                  var UEC : PSUnified_Exception ) : TStore_Address64 ; virtual ;
                              { Write data from Data to specified Address in store.
                                Write _Size bytes.  Returns success of operation in UEC.
                                Function returns the number of bytes actually written. }

                              function Max_Storage : TStore_Size64 ; virtual ;
                              { Returns the maximum (current) size of the store, in
                                bytes.  In other words, this is the highest valid
                                address (plus 1). }

                              function Min_Storage : TStore_Address64 ; virtual ;
                              { Returns the minimum atomic size of the store.  For
                                instance, this would be the sector size for a disk. }

                              function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                  virtual ;
                              { Requests extension of store by the amount specified.
                                Returns the amount actually added to store.  If Amount
                                is 0, function returns 0 if store is extendable, and -1
                                if it is not. }

                              function Get_Read_Only : boolean ; virtual ;
                              { Indicates that the store can only be read from. }

                              function Get_Write_Only : boolean ; virtual ;
                              { Indicates that the store can only be written to. }

                              procedure Initialize ; virtual ;
                              { Initializes the TStore object. }

                              procedure Terminate ; virtual ;
                              { Terminates the TStore object. }

                              procedure Format ; virtual ;
                              { Does a low-level format of the store. }

                              function Get_Name : Ansistring ; virtual ;
                              { Returns name of store. }

{$IFNDEF MOBILE}
                              function Debugger : PSDebug_Interface ; virtual ;
                              { Returns a debugger object. }
{$ENDIF}

                              function Get_Cache : PSCache64 ; virtual ;
                              { Returns current cache object. }

                              procedure Set_Cache( Value : PSCache64 ) ; virtual ;
                              { Sets the current cache object. }

                              function Contiguous_Store : boolean ; virtual ;
                              { Returns True if the store's addresses are contiguous
                                (non-sparse). }

                              procedure Set_Max_Storage( Value : TStore_Address64 ;
                                  var Res : PSUnified_Exception ) ; virtual ;
                              { Sets the maximum (current) size of the store, in bytes.
                                In other words, this is the new highest valid address
                                (plus 1).  If the store is not resizable, the call
                                fails and an error code is returned in Res.  This should
                                only be used to shrink the size of the store, not to
                                extend it.  To extend, use the Extend method. }

                              function Extended_Size : TStore_Address64 ; virtual ;
                              { Returns the maximum theoretical size that the store can
                                be extended to.  If the store is not extendable, this
                                returns the same value as Max_Storage. }

                              procedure Set_Read_Only( Value : boolean ) ; virtual ;
                              { Sets the store read-only value. }

                              procedure Set_Write_Only( Value : boolean ) ; virtual ;
                              { Sets the store write-only value. }

                              function Get_Bytes_Read : longint ;
                                  virtual ; { Data bytes read }
                              function Get_Bytes_Written : longint ;
                                  virtual ; { Data bytes written }
                              function Get_Reads : longint ;
                                  virtual ; { Data read count }
                              function Get_Writes : longint ;
                                  virtual ; { Data write count }
                              function Get_Error_Count : longint ;
                                  virtual ; { Error count }
                              procedure Set_Bytes_Read( Value : longint ) ;
                                  virtual ; { Data bytes read }
                              procedure Set_Bytes_Written( Value : longint ) ;
                                  virtual ; { Data bytes written }
                              procedure Set_Reads( Value : longint ) ;
                                  virtual ; { Data read count }
                              procedure Set_Writes( Value : longint ) ;
                                  virtual ; { Data write count }
                              procedure Set_Error_Count( Value : longint ) ;
                                  virtual ; { Error count }
                      end ; { TSBase_Store64 }
     PSBase_Store64 = ^TSBase_Store64 ;
{$ENDIF}

{$IFDEF DYNAMIC_OBJECTS}
type TBase_Store = class( TStore )
                      protected { Instance data... }
                          Store_Statistics : TStore_Statistics ;
                          _Cache : TCache ;
                          _Read_Only : boolean ;
                          _Write_Only : boolean ;

                      public { API... }
                          function Read_Data( var Data ; Address, _Size : TStore_Address ;
                              var UEC : TUnified_Exception ) : TStore_Address ;
                              override ;
                          { Read data from specified Address in store to Data.
                            Read _Size bytes.  Returns success of operation in UEC.
                            Function returns the number of bytes actually read. }

                          function Write_Data( var Data ; Address, _Size : TStore_Address ;
                              var UEC : TUnified_Exception ) : TStore_Address ;
                              override ;
                          { Write data from Data to specified Address in store.
                            Write _Size bytes.  Returns success of operation in UEC.
                            Function returns the number of bytes actually written. }

                          function Max_Storage : TStore_Size ; override ;
                          { Returns the maximum (current) size of the store, in
                            bytes.  In other words, this is the highest valid
                            address (plus 1). }

                          function Min_Storage : TStore_Address ;
                              override ;
                          { Returns the minimum atomic size of the store.  For
                            instance, this would be the sector size for a disk. }

                          function Extend( Amount : TStore_Address ) : TStore_Address ;
                              override ;
                          { Requests extension of store by the amount specified.
                            Returns the amount actually added to store.  If Amount
                            is 0, function returns 0 if store is extendable, and -1
                            if it is not. }

                          function Get_Read_Only : boolean ; override ;
                          { Indicates that the store can only be read from. }

                          function Get_Write_Only : boolean ; override ;
                          { Indicates that the store can only be written to. }

                          procedure Initialize ; override ;
                          { Initializes the TStore object. }

                          procedure Terminate ; override ;
                          { Terminates the TStore object. }

                          procedure Format ; override ;
                          { Does a low-level format of the store. }

                          function Get_Name : Ansistring ; override ;
                          { Returns name of store. }

                          function Get_Cache : TCache ; override ;
                          { Returns current cache object. }

                          procedure Set_Cache( Value : TCache ) ;
                              override ;
                          { Sets the current cache object. }

                          function Contiguous_Store : boolean ;
                              override ;
                          { Returns True if the store's addresses are contiguous
                            (non-sparse). }

                          procedure Set_Max_Storage( Value : TStore_Address ;
                              var Res : TUnified_Exception ) ; override ;
                          { Sets the maximum (current) size of the store, in bytes.
                            In other words, this is the new highest valid address
                            (plus 1).  If the store is not resizable, the call
                            fails and an error code is returned in Res.  This should
                            only be used to shrink the size of the store, not to
                            extend it.  To extend, use the Extend method. }

                          function Extended_Size : TStore_Address ;
                              override ;
                          { Returns the maximum theoretical size that the store can
                            be extended to.  If the store is not extendable, this
                            returns the same value as Max_Storage. }

                          procedure Set_Read_Only( Value : boolean ) ; override ;
                          { Sets the store read-only value. }

                          procedure Set_Write_Only( Value : boolean ) ; override ;
                          { Sets the store write-only value. }

                          function Get_Bytes_Read : longint ;
                              override ; { Data bytes read }
                          function Get_Bytes_Written : longint ;
                              override ; { Data bytes written }
                          function Get_Reads : longint ;
                              override ; { Data read count }
                          function Get_Writes : longint ;
                              override ; { Data write count }
                          function Get_Error_Count : longint ;
                              override ; { Error count }
                          procedure Set_Bytes_Read( Value : longint ) ;
                              override ; { Data bytes read }
                          procedure Set_Bytes_Written( Value : longint ) ;
                              override ; { Data bytes written }
                          procedure Set_Reads( Value : longint ) ;
                              override ; { Data read count }
                          procedure Set_Writes( Value : longint ) ;
                              override ; { Data write count }
                          procedure Set_Error_Count( Value : longint ) ;
                              override ; { Error count }
                   end ;

type TBase_Store64 = class( TStore64 )
                         protected { Instance data... }
                             Store_Statistics : TStore_Statistics ;
                             _Cache : TCache64 ;
                             _Read_Only : boolean ;
                             _Write_Only : boolean ;

                         public { API... }
                             function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                                 var UEC : TUnified_Exception ) : TStore_Address64 ;
                                 override ;
                             { Read data from specified Address in store to Data.
                               Read _Size bytes.  Returns success of operation in UEC.
                               Function returns the number of bytes actually read. }

                             function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                                 var UEC : TUnified_Exception ) : TStore_Address64 ;
                                 override ;
                             { Write data from Data to specified Address in store.
                               Write _Size bytes.  Returns success of operation in UEC.
                               Function returns the number of bytes actually written. }

                             function Max_Storage : TStore_Size64 ;
                                 override ;
                             { Returns the maximum (current) size of the store, in
                               bytes.  In other words, this is the highest valid
                               address (plus 1). }

                             function Min_Storage : TStore_Address64 ;
                                 override ;
                             { Returns the minimum atomic size of the store.  For
                               instance, this would be the sector size for a disk. }

                             function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                 override ;
                             { Requests extension of store by the amount specified.
                               Returns the amount actually added to store.  If Amount
                               is 0, function returns 0 if store is extendable, and -1
                               if it is not. }

                             function Get_Read_Only : boolean ; override ;
                             { Indicates that the store can only be read from. }

                             function Get_Write_Only : boolean ; override ;
                             { Indicates that the store can only be written to. }

                             procedure Initialize ; override ;
                             { Initializes the TStore object. }

                             procedure Terminate ; override ;
                             { Terminates the TStore object. }

                             procedure Format ; override ;
                             { Does a low-level format of the store. }

                             function Get_Name : Ansistring ; override ;
                             { Returns name of store. }

                             function Get_Cache : TCache64 ; override ;
                             { Returns current cache object. }

                             procedure Set_Cache( Value : TCache64 ) ;
                                 override ;
                             { Sets the current cache object. }

                             function Contiguous_Store : boolean ;
                                 override ;
                             { Returns True if the store's addresses are contiguous
                               (non-sparse). }

                             procedure Set_Max_Storage( Value : TStore_Address64 ;
                                 var Res : TUnified_Exception ) ; override ;
                             { Sets the maximum (current) size of the store, in bytes.
                               In other words, this is the new highest valid address
                               (plus 1).  If the store is not resizable, the call
                               fails and an error code is returned in Res.  This should
                               only be used to shrink the size of the store, not to
                               extend it.  To extend, use the Extend method. }

                             function Extended_Size : TStore_Address64 ;
                                 override ;
                             { Returns the maximum theoretical size that the store can
                               be extended to.  If the store is not extendable, this
                               returns the same value as Max_Storage. }

                            procedure Set_Read_Only( Value : boolean ) ;
                                override ;
                            { Sets the store read-only value. }

                            procedure Set_Write_Only( Value : boolean ) ;
                                override ;
                            { Sets the store write-only value. }

                            function Get_Bytes_Read : longint ;
                                override ; { Data bytes read }
                            function Get_Bytes_Written : longint ;
                                override ; { Data bytes written }
                            function Get_Reads : longint ;
                                override ; { Data read count }
                            function Get_Writes : longint ;
                                override ; { Data write count }
                            function Get_Error_Count : longint ;
                                override ; { Error count }
                            procedure Set_Bytes_Read( Value : longint ) ;
                                override ; { Data bytes read }
                            procedure Set_Bytes_Written( Value : longint ) ;
                                override ; { Data bytes written }
                            procedure Set_Reads( Value : longint ) ;
                                override ; { Data read count }
                            procedure Set_Writes( Value : longint ) ;
                                override ; { Data write count }
                            procedure Set_Error_Count( Value : longint ) ;
                                override ; { Error count }
                     end ; { TBase_Store64 }

type TStandard_COM_Store = class( TCOM_Store )
                               private { Instance data... }
                                   Store_Statistics : TStore_Statistics ;
                                   _Cache : TCOM_Cache ;
                                   _Read_Only : boolean ;
                                   _Write_Only : boolean ;

                               public { API... }
                                   function Is_Class( Name : PAnsiChar ) : boolean ;
                                       override ;
                                       
                                   function Read_Data( var Data ;
                                       Address, _Size : TStore_Address ;
                                       var UEC : TUnified_Exception ) : TStore_Address ;
                                       override ;
                                   { Read data from specified Address in store to Data.
                                     Read _Size bytes.  Returns success of operation in UEC.
                                     Function returns the number of bytes actually read. }

                                   function Write_Data( var Data ;
                                       Address, _Size : TStore_Address ;
                                       var UEC : TUnified_Exception ) : TStore_Address ;
                                           override ;
                                   { Write data from Data to specified Address in store.
                                     Write _Size bytes.  Returns success of operation in UEC.
                                     Function returns the number of bytes actually written. }

                                   function Max_Storage : TStore_Size ;
                                       override ;
                                   { Returns the maximum (current) size of the store, in
                                     bytes.  In other words, this is the highest valid
                                     address (plus 1). }

                                   function Min_Storage : TStore_Address ;
                                       override ;
                                   { Returns the minimum atomic size of the store.  For
                                     instance, this would be the sector size for a disk. }

                                   function Extend( Amount : TStore_Address ) : TStore_Address ;
                                      override ;
                                   { Requests extension of store by the amount specified.
                                     Returns the amount actually added to store.  If Amount
                                     is 0, function returns 0 if store is extendable, and -1
                                     if it is not. }

                                   function Get_Read_Only : boolean ;
                                       override ;
                                   { Indicates that the store can only be read from. }

                                   function Get_Write_Only : boolean ;
                                       override ;
                                   { Indicates that the store can only be written to. }

                                   procedure Format ; override ;
                                   { Does a low-level format of the store. }

                                   function Get_Name : PAnsiChar ; override ;
                                   { Returns name of store. }

                                   function Get_Cache : TCOM_Cache ;
                                       override ;
                                   { Returns current cache object. }

                                   procedure Set_Cache( Value : TCOM_Cache ) ;
                                       override ;
                                   { Sets the current cache object. }

                                   function Contiguous_Store : boolean ;
                                       override ;
                                   { Returns True if the store's addresses are contiguous
                                     (non-sparse). }

                                   procedure Set_Max_Storage( Value : TStore_Address ;
                                       var Res : TUnified_Exception ) ; override ;
                                   { Sets the maximum (current) size of the store, in bytes.
                                     In other words, this is the new highest valid address
                                     (plus 1).  If the store is not resizable, the call
                                     fails and an error code is returned in Res.  This should
                                     only be used to shrink the size of the store, not to
                                     extend it.  To extend, use the Extend method. }

                                   function Extended_Size : TStore_Address ;
                                       override ;
                                   { Returns the maximum theoretical size that the store can
                                     be extended to.  If the store is not extendable, this
                                     returns the same value as Max_Storage. }

                                   function Get_Bytes_Read : longint ;
                                       override ; { Data bytes read }
                                   function Get_Bytes_Written : longint ;
                                       override ; { Data bytes written }
                                   function Get_Reads : longint ;
                                       override ; { Data read count }
                                   function Get_Writes : longint ;
                                       override ; { Data write count }
                                   function Get_Error_Count : longint ;
                                       override ; { Error count }
                                   procedure Set_Bytes_Read( Value : longint ) ;
                                       override ; { Data bytes read }
                                   procedure Set_Bytes_Written( Value : longint ) ;
                                       override ; { Data bytes written }
                                   procedure Set_Reads( Value : longint ) ;
                                       override ; { Data read count }
                                   procedure Set_Writes( Value : longint ) ;
                                       override ; { Data write count }
                                   procedure Set_Error_Count( Value : longint ) ;
                                       override ; { Error count }
                                   procedure Set_Read_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store read-only value. }
                                   procedure Set_Write_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store write-only value. }
                           end ;

type TStandard_COM_Store64 = class( TCOM_Store64 )
                                 private { Instance data... }
                                     Store_Statistics : TStore_Statistics ;
                                     _Cache : TCOM_Cache64 ;
                                     _Read_Only : boolean ;
                                     _Write_Only : boolean ;
                                     _Debugger : TCOM_Managed_Store64_Debugger ;

                                 public { API... }
                                     function Is_Class( Name : PAnsiChar ) : boolean ;
                                         override ;
                                         
                                     function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                                         var UEC : TUnified_Exception ) : TStore_Address64 ;
                                         override ;
                                     { Read data from specified Address in store to Data.
                                       Read _Size bytes.  Returns success of operation in UEC.
                                       Function returns the number of bytes actually read. }

                                     function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                                         var UEC : TUnified_Exception ) : TStore_Address64 ;
                                         override ;
                                     { Write data from Data to specified Address in store.
                                       Write _Size bytes.  Returns success of operation in UEC.
                                       Function returns the number of bytes actually written. }

                                     function Max_Storage : TStore_Size64 ;
                                         override ;
                                     { Returns the maximum (current) size of the store, in
                                       bytes.  In other words, this is the highest valid
                                       address (plus 1). }

                                     function Min_Storage : TStore_Address64 ;
                                         override ;
                                     { Returns the minimum atomic size of the store.  For
                                       instance, this would be the sector size for a disk. }

                                     function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                         override ;
                                     { Requests extension of store by the amount specified.
                                       Returns the amount actually added to store.  If Amount
                                       is 0, function returns 0 if store is extendable, and -1
                                       if it is not. }

                                     function Get_Read_Only : boolean ;
                                         override ;
                                     { Indicates that the store can only be read from. }

                                     function Get_Write_Only : boolean ;
                                         override ;
                                     { Indicates that the store can only be written to. }

                                     procedure Format ; override ;
                                     { Does a low-level format of the store. }

                                     function Get_Name : PAnsiChar ; override ;
                                     { Returns name of store. }

                                     function Get_Cache : TCOM_Cache64 ;
                                         override ;
                                     { Returns current cache object. }

                                     procedure Set_Cache( Value : TCOM_Cache64 ) ;
                                         override ;
                                     { Sets the current cache object. }

                                     function Contiguous_Store : boolean ;
                                         override ;
                                     { Returns True if the store's addresses are contiguous
                                       (non-sparse). }

                                     procedure Set_Max_Storage( Value : TStore_Address64 ;
                                         var Res : TUnified_Exception ) ;
                                         override ;
                                     { Sets the maximum (current) size of the store, in bytes.
                                       In other words, this is the new highest valid address
                                       (plus 1).  If the store is not resizable, the call
                                       fails and an error code is returned in Res.  This should
                                       only be used to shrink the size of the store, not to
                                       extend it.  To extend, use the Extend method. }

                                     function Extended_Size : TStore_Address64 ;
                                         override ;
                                     { Returns the maximum theoretical size that the store can
                                       be extended to.  If the store is not extendable, this
                                       returns the same value as Max_Storage. }

                                   function Get_Bytes_Read : longint ;
                                       override ; { Data bytes read }
                                   function Get_Bytes_Written : longint ;
                                       override ; { Data bytes written }
                                   function Get_Reads : longint ;
                                       override ; { Data read count }
                                   function Get_Writes : longint ;
                                       override ; { Data write count }
                                   function Get_Error_Count : longint ;
                                       override ; { Error count }
                                   procedure Set_Bytes_Read( Value : longint ) ;
                                       override ; { Data bytes read }
                                   procedure Set_Bytes_Written( Value : longint ) ;
                                       override ; { Data bytes written }
                                   procedure Set_Reads( Value : longint ) ;
                                       override ; { Data read count }
                                   procedure Set_Writes( Value : longint ) ;
                                       override ; { Data write count }
                                   procedure Set_Error_Count( Value : longint ) ;
                                       override ; { Error count }
                                   procedure Set_Read_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store read-only value. }
                                   procedure Set_Write_Only( Value : boolean ) ;
                                       override ;
                                   { Sets the store write-only value. }
                                   function Get_Verifier : TCOM_Managed_Store64_Debugger ;
                                       override ;
                                   procedure Set_Verifier( Value : TCOM_Managed_Store64_Debugger ) ;
                                       override ;
                             end ; { TStandard_COM_Store64 }
{$ENDIF}

{$IFDEF CC_Test}
procedure Test_Unit ;
{$ENDIF}


implementation

uses { Pascal }
{$IFDEF FPC_OR_DELPHI}
     SysUtils, { lowercase }
{$ENDIF}

     { C&C Subroutine Library - Layer 1 }
     CommonUt, { Boolean_To_String }
{$IFNDEF MOBILE}
     DebugInt, { TSText_Debugger }
{$ENDIF}
     Num1s, { Num1 }
     UE ; { Get_Abstract_Error_Exception }


type TSBase_Store_Exception = object( TSUnified_Exception )
                                  private
                                      _Code : longint ;

                                  public
                                      function Get_Facility : longint ; virtual ;
                                      function Get_Facility_Version : longint ;
                                          virtual ;
                                      function Version : longint ; virtual ;
                                      function Severity : longint ; virtual ;
                                      function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                          virtual ;
                                      function Get_Error : longint ; virtual ;
                                      function Get_Previous : PsUnified_Exception ;
                                          virtual ;
                                      procedure Terminate ; virtual ;
                              end ;
     PSBase_Store_Exception = ^TSBase_Store_Exception ;

function TSBase_Store_Exception.Get_Facility : longint ;

begin
    Get_Facility := StoreErr_Facility ;
end ;


function TSBase_Store_Exception.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 10 ;
end ;


function TSBase_Store_Exception.Version : longint ;

begin
    Version := 10 ;
end ;


function TSBase_Store_Exception.Severity : longint ;

begin
    case _Code of
        StoreErr_Success: Severity := UE_Success ;
        else Severity := UE_Error ;
    end ;
end ;


function TSBase_Store_Exception.Error_Text( var Size, Typ : longint ) : Ansistring ;

var S : Ansistring ;

begin
    case _Code of
        StoreErr_Success: S := 'Success' ;
        StoreErr_Invalid_Operation: S := 'Invalid operation' ;
        StoreErr_Abstract: S := 'Abstract error' ;
    end ;
    Size := length( S ) ;
    Typ := 0 ;
    Error_Text := S ;
end ;


function TSBase_Store_Exception.Get_Error : longint ;

begin
    Get_Error := _Code ;
end ;


function TSBase_Store_Exception.Get_Previous : PsUnified_Exception ;

begin
    Get_Previous := nil ;
end ;


procedure TSBase_Store_Exception.Terminate ;

begin
    Free ;
end ;



function Create_SBase_Store_Exception( Code : longint ) : PSUnified_Exception ;

var E : PSBase_Store_Exception ;

begin
    E := new( PSBase_Store_Exception, Init ) ;
    PSBase_Store_Exception( E )^._Code := Code ;
    Create_SBase_Store_Exception := E ;
end ;


{$IFDEF DYNAMIC_OBJECTS}
type TBase_Store_Exception = class( TStandard_Error_Interface )
                                 private
                                     _Code : longint ;
                                     Temp : Ansistring ;

                                 public
                                     function Get_Facility : longint ; override ;
                                     function Get_Facility_Version : longint ;
                                         override ;
                                     function Version : longint ; override ;
                                     function Severity : longint ; override ;
                                     function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                         override ;
                                     function Get_Error : longint ; override ;
                                     function Get_Previous : TUnified_Exception ;
                                         override ;
                                     procedure Terminate ; override ;
                             end ;

function TBase_Store_Exception.Get_Facility : longint ;

begin
    Get_Facility := StoreErr_Facility ;
end ;


function TBase_Store_Exception.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 10 ;
end ;


function TBase_Store_Exception.Version : longint ;

begin
    Version := 10 ;
end ;


function TBase_Store_Exception.Severity : longint ;

begin
    case _Code of
        StoreErr_Success: Severity := UE_Success ;
        else Severity := UE_Error ;
    end ;
end ;


function TBase_Store_Exception.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

begin
    case _Code of
        StoreErr_Success: Temp := 'Success' ;
        StoreErr_Invalid_Operation: Temp := 'Invalid operation' ;
        StoreErr_Abstract: Temp := 'Abstract error' ;
    end ;
    Size := length( Temp ) ;
    Typ := 0 ;
    Error_Text := PAnsiChar( Temp ) ;
end ;


function TBase_Store_Exception.Get_Error : longint ;

begin
    Get_Error := _Code ;
end ;


function TBase_Store_Exception.Get_Previous : TUnified_Exception ;

begin
    Get_Previous := nil ;
end ;


procedure TBase_Store_Exception.Terminate ;

begin
    Free ;
end ;



function Create_Base_Store_Exception( Code : longint ) : TUnified_Exception ;

begin
    Result := TBase_Store_Exception.Create ;
    TBase_Store_Exception( Result )._Code := Code ;
end ;
{$ENDIF}


{$IFDEF CC_INTERNALS}
type PSBase_Store_Debugger = ^TSBase_Store_Debugger ;
     TSBase_Store_Debugger = object( TSText_Debugger )
                                 private
                                     _Instance : PSBase_Store ;
                                     S : Small_String ;

                                 public
                                     constructor Init( I : PSBase_Store ) ;

                                     function Count : longint ; virtual ;

                                     function Child( Ordinal : longint ) : PSDebug_Interface ;
                                         virtual ;
                             end ;


constructor TSBase_Store_Debugger.Init( I : PSBase_Store ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TSBase_Store_Debugger.Count : longint ;

begin
    Count := 8 ;
end ;


function TSBase_Store_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
    new( R, Init ) ;
    case Ordinal of
        0 : begin
                if( _Instance^._Cache = nil ) then
                begin
                    S := '_Cache = nil' ;
                end else
                begin
                    dispose( R, done ) ;
                    R := PSText_Debugger( _Instance^._Cache^.Debugger ) ;
                    if( R = nil ) then
                    begin
                        new( R, Init ) ;
                    end ;
                    S := '_Cache' ;
                end ;
            end ;
        1 : S := '_Read_Only = ' + Boolean_To_String( _Instance^._Read_Only ) ;
        2 : S := '_Write_Only = ' + Boolean_To_String( _Instance^._Write_Only ) ;
        3 : S := 'Store_Statistics.Bytes_Read = ' + Num1( _Instance^.Store_Statistics.Bytes_Read ) ;
        4 : S := 'Store_Statistics.Bytes_Written = ' + Num1( _Instance^.Store_Statistics.Bytes_Written ) ;
        5 : S := 'Store_Statistics.Reads = ' + Num1( _Instance^.Store_Statistics.Reads ) ;
        6 : S := 'Store_Statistics.Writes = ' + Num1( _Instance^.Store_Statistics.Writes ) ;
        7 : S := 'Store_Statistics.Error_Count = ' + Num1( _Instance^.Store_Statistics.Error_Count ) ;
    end ;
    S := S + #0 ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ;


  {$IFDEF INT64}
type PSBase_Store64_Debugger = ^TSBase_Store64_Debugger ;
     TSBase_Store64_Debugger = object( TSText_Debugger )
                                 private
                                     _Instance : PSBase_Store64 ;
                                     S : Small_String ;

                                 public
                                     constructor Init( I : PSBase_Store64 ) ;

                                     function Count : longint ; virtual ;

                                     function Child( Ordinal : longint ) : PSDebug_Interface ;
                                         virtual ;
                             end ;


constructor TSBase_Store64_Debugger.Init( I : PSBase_Store64 ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TSBase_Store64_Debugger.Count : longint ;

begin
    Count := 8 ;
end ;


function TSBase_Store64_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
    new( R, Init ) ;
    case Ordinal of
        0 : begin
                if( _Instance^._Cache = nil ) then
                begin
                    S := '_Cache = nil' ;
                end else
                begin
                    dispose( R, done ) ;
                    R := PSText_Debugger( _Instance^._Cache^.Debugger ) ;
                    if( R = nil ) then
                    begin
                        new( R, Init ) ;
                    end ;
                    S := '_Cache' ;
                end ;
            end ;
        1 : S := '_Read_Only = ' + Boolean_To_String( _Instance^._Read_Only ) ;
        2 : S := '_Write_Only = ' + Boolean_To_String( _Instance^._Write_Only ) ;
        3 : S := 'Store_Statistics.Bytes_Read = ' + Num1( _Instance^.Store_Statistics.Bytes_Read ) ;
        4 : S := 'Store_Statistics.Bytes_Written = ' + Num1( _Instance^.Store_Statistics.Bytes_Written ) ;
        5 : S := 'Store_Statistics.Reads = ' + Num1( _Instance^.Store_Statistics.Reads ) ;
        6 : S := 'Store_Statistics.Writes = ' + Num1( _Instance^.Store_Statistics.Writes ) ;
        7 : S := 'Store_Statistics.Error_Count = ' + Num1( _Instance^.Store_Statistics.Error_Count ) ;
    end ;
    S := S + #0 ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ;
  {$ENDIF}
{$ENDIF}



{ TSBase_Store methods... }

{ API... }

function TSBase_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : PSUnified_Exception ) : TStore_Address ;

begin
    Read_Data := 0 ;
    UEC := Get_Abstract_Error_Exception ;
end ;


function TSBase_Store.Write_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : PSUnified_Exception ) : TStore_Address ;

begin
    Write_Data := 0 ;
    UEC := Get_Abstract_Error_Exception ;
end ;


function TSBase_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := 0 ;
end ;


function TSBase_Store.Min_Storage : TStore_Address ;

begin
    Min_Storage := 1 ;
end ;


function TSBase_Store.Extend( Amount : TStore_Address ) : TStore_Address ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TSBase_Store.Get_Read_Only : boolean ;

begin
    Get_Read_Only := _Read_Only ;
end ;


function TSBase_Store.Get_Write_Only : boolean ;

begin
    Get_Write_Only := _Write_Only ;
end ;


procedure TSBase_Store.Initialize ;

begin
    _Cache := nil ;
    _Read_Only := False ;
    _Write_Only := False ;
    fillchar( Store_Statistics, sizeof( Store_Statistics ), 0 ) ;
end ;


procedure TSBase_Store.Terminate ;

begin
    { This routine left intentionally blank. }
end ;


procedure TSBase_Store.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TSBase_Store.Get_Name : Ansistring ;

begin
    Get_Name := '' ;
end ;


{$IFNDEF MOBILE}
function TSBase_Store.Debugger : PSDebug_Interface ; {UNTESTED}

begin
  {$IFDEF CC_INTERNALS}
    Debugger := new( PSBase_Store_Debugger, Init( @self ) ) ;
  {$ELSE}
    Debugger := nil ;
  {$ENDIF}
end ;
{$ENDIF}


function TSBase_Store.Get_Cache : PSCache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TSBase_Store.Set_Cache( Value : PSCache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value^.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache^.Terminate ;
            _Cache^.Initialize ;
            _Cache^.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TSBase_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TSBase_Store.Set_Max_Storage( Value : TStore_Address ; var Res : PSUnified_Exception ) ;

begin
    Res := Create_SBase_Store_Exception( StoreErr_Invalid_Operation ) ;
end ;


function TSBase_Store.Extended_Size : TStore_Address ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TSBase_Store.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TSBase_Store.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;


function TSBase_Store.Get_Bytes_Read : longint ;

begin
    Get_Bytes_Read := Store_Statistics.Bytes_Read ;
end ;


function TSBase_Store.Get_Bytes_Written : longint ;

begin
    Get_Bytes_Written := Store_Statistics.Bytes_Written ;
end ;


function TSBase_Store.Get_Reads : longint ;

begin
    Get_Reads := Store_Statistics.Reads ;
end ;


function TSBase_Store.Get_Writes : longint ;

begin
    Get_Writes := Store_Statistics.Writes ;
end ;


function TSBase_Store.Get_Error_Count : longint ;

begin
    Get_Error_Count := Store_Statistics.Error_Count ;
end ;


procedure TSBase_Store.Set_Bytes_Read( Value : longint ) ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TSBase_Store.Set_Bytes_Written( Value : longint ) ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TSBase_Store.Set_Reads( Value : longint ) ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TSBase_Store.Set_Writes( Value : longint ) ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TSBase_Store.Set_Error_Count( Value : longint ) ;

begin
    Store_Statistics.Error_Count := Value ;
end ;



{$IFDEF INT64}
{ TSBase_Store64 methods... }

{ API... }

function TSBase_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : PSUnified_Exception ) : TStore_Address64 ;

begin
    Read_Data := 0 ;
    UEC := Get_Abstract_Error_Exception ;
end ;


function TSBase_Store64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : PSUnified_Exception ) : TStore_Address64 ;

begin
    Write_Data := 0 ;
    UEC := Get_Abstract_Error_Exception
end ;


function TSBase_Store64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := 0 ;
end ;


function TSBase_Store64.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := 1 ;
end ;


function TSBase_Store64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TSBase_Store64.Get_Read_Only : boolean ;

begin
    Get_Read_Only := _Read_Only ;
end ;


function TSBase_Store64.Get_Write_Only : boolean ;

begin
    Get_Write_Only := _Write_Only ;
end ;


procedure TSBase_Store64.Initialize ;

begin
    _Cache := nil ;
    _Read_Only := False ;
    _Write_Only := False ;
    fillchar( Store_Statistics, sizeof( Store_Statistics ), 0 ) ;
end ;


procedure TSBase_Store64.Terminate ;

begin
    { This routine left intentionally blank. }
end ;


procedure TSBase_Store64.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TSBase_Store64.Get_Name : Ansistring ;

begin
    Get_Name := '' ;
end ;


{$IFNDEF MOBILE}
function TSBase_Store64.Debugger : PSDebug_Interface ; {UNTESTED}

begin
  {$IFDEF CC_INTERNALS}
    Debugger := new( PSBase_Store64_Debugger, Init( @self ) ) ;
  {$ELSE}
    Debugger := nil ;
  {$ENDIF}
end ;
{$ENDIF}


function TSBase_Store64.Get_Cache : PSCache64 ;

begin
    Get_Cache := _Cache ;
end ;


procedure TSBase_Store64.Set_Cache( Value : PSCache64 ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value^.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache^.Terminate ;
            _Cache^.Initialize ;
            _Cache^.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TSBase_Store64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TSBase_Store64.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : PSUnified_Exception ) ;

begin
    Res := Get_Abstract_Error_Exception ;
end ;


function TSBase_Store64.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TSBase_Store64.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TSBase_Store64.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;


function TSBase_Store64.Get_Bytes_Read : longint ;

begin
    Get_Bytes_Read := Store_Statistics.Bytes_Read ;
end ;


function TSBase_Store64.Get_Bytes_Written : longint ;

begin
    Get_Bytes_Written := Store_Statistics.Bytes_Written ;
end ;


function TSBase_Store64.Get_Reads : longint ;

begin
    Get_Reads := Store_Statistics.Reads ;
end ;


function TSBase_Store64.Get_Writes : longint ;

begin
    Get_Writes := Store_Statistics.Writes ;
end ;


function TSBase_Store64.Get_Error_Count : longint ;

begin
    Get_Error_Count := Store_Statistics.Error_Count ;
end ;


procedure TSBase_Store64.Set_Bytes_Read( Value : longint ) ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TSBase_Store64.Set_Bytes_Written( Value : longint ) ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TSBase_Store64.Set_Reads( Value : longint ) ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TSBase_Store64.Set_Writes( Value : longint ) ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TSBase_Store64.Set_Error_Count( Value : longint ) ;

begin
    Store_Statistics.Error_Count := Value ;
end ;
{$ENDIF}


{$IFDEF DYNAMIC_OBJECTS}
{ TBase_Store methods... }

{ API... }

function TBase_Store.Read_Data( var Data ; Address, _Size : TStore_Address ;
  var UEC : TUnified_Exception ) : TStore_Address ;

begin
    Read_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TBase_Store.Write_Data( var Data ; Address, _Size : TStore_Address ;
  var UEC : TUnified_Exception ) : TStore_Address ;

begin
    Write_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TBase_Store.Max_Storage : TStore_Size ;

begin
    Max_Storage := 0 ;
end ;


function TBase_Store.Min_Storage : TStore_Address ;

begin
    Min_Storage := 1 ;
end ;


function TBase_Store.Extend( Amount : TStore_Address ) : TStore_Address ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TBase_Store.Get_Read_Only : boolean ;

begin
    Get_Read_Only := _Read_Only ;
end ;


function TBase_Store.Get_Write_Only : boolean ;

begin
    Get_Write_Only := _Write_Only ;
end ;


procedure TBase_Store.Initialize ;

begin
    { This routine left intentionally blank. }
end ;


procedure TBase_Store.Terminate ;

begin
    Free ;
end ;


procedure TBase_Store.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TBase_Store.Get_Name : Ansistring ;

begin
    Get_Name := '' ;
end ;


function TBase_Store.Get_Cache : TCache ;

begin
    Get_Cache := _Cache ;
end ;


procedure TBase_Store.Set_Cache( Value : TCache ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TBase_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TBase_Store.Set_Max_Storage( Value : TStore_Address ;
    var Res : TUnified_Exception ) ;

begin
    Res := Create_Base_Store_Exception( StoreErr_Invalid_Operation ) ;
end ;


function TBase_Store.Extended_Size : TStore_Address ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TBase_Store.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TBase_Store.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;


function TBase_Store.Get_Bytes_Read : longint ;

begin
    Get_Bytes_Read := Store_Statistics.Bytes_Read ;
end ;


function TBase_Store.Get_Bytes_Written : longint ;

begin
    Get_Bytes_Written := Store_Statistics.Bytes_Written ;
end ;


function TBase_Store.Get_Reads : longint ;

begin
    Get_Reads := Store_Statistics.Reads ;
end ;


function TBase_Store.Get_Writes : longint ;

begin
    Get_Writes := Store_Statistics.Writes ;
end ;


function TBase_Store.Get_Error_Count : longint ;

begin
    Get_Error_Count := Store_Statistics.Error_Count ;
end ;


procedure TBase_Store.Set_Bytes_Read( Value : longint ) ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TBase_Store.Set_Bytes_Written( Value : longint ) ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TBase_Store.Set_Reads( Value : longint ) ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TBase_Store.Set_Writes( Value : longint ) ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TBase_Store.Set_Error_Count( Value : longint ) ;

begin
    Store_Statistics.Error_Count := Value ;
end ;



  {$IFDEF INT64}
{ TBase_Store64 methods... }

{ API... }

function TBase_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
   var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Read_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TBase_Store64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
   var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Write_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TBase_Store64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := 0 ;
end ;


function TBase_Store64.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := 1 ;
end ;


function TBase_Store64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TBase_Store64.Get_Read_Only : boolean ;

begin
    Get_Read_Only := _Read_Only ;
end ;


function TBase_Store64.Get_Write_Only : boolean ;

begin
    Get_Write_Only := _Write_Only ;
end ;


procedure TBase_Store64.Initialize ;

begin
    { This routine left intentionally blank. }
end ;


procedure TBase_Store64.Terminate ;

begin
    Free ;
end ;


procedure TBase_Store64.Format ;

begin
    { This routine left intentionally blank. }
end ;


function TBase_Store64.Get_Name : Ansistring ;

begin
    Get_Name := '' ;
end ;


function TBase_Store64.Get_Cache : TCache64 ;

begin
    Get_Cache := _Cache ;
end ;


procedure TBase_Store64.Set_Cache( Value : TCache64 ) ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TBase_Store64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TBase_Store64.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : TUnified_Exception ) ;

begin
    Res := Create_Base_Store_Exception( StoreErr_Invalid_Operation ) ;
end ;


function TBase_Store64.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Max_Storage ;
end ;


procedure TBase_Store64.Set_Read_Only( Value : boolean ) ;

begin
    _Read_Only := Value ;
end ;


procedure TBase_Store64.Set_Write_Only( Value : boolean ) ;

begin
    _Write_Only := Value ;
end ;


function TBase_Store64.Get_Bytes_Read : longint ;

begin
    Get_Bytes_Read := Store_Statistics.Bytes_Read ;
end ;


function TBase_Store64.Get_Bytes_Written : longint ;

begin
    Get_Bytes_Written := Store_Statistics.Bytes_Written ;
end ;


function TBase_Store64.Get_Reads : longint ;

begin
    Get_Reads := Store_Statistics.Reads ;
end ;


function TBase_Store64.Get_Writes : longint ;

begin
    Get_Writes := Store_Statistics.Writes ;
end ;


function TBase_Store64.Get_Error_Count : longint ;

begin
    Get_Error_Count := Store_Statistics.Error_Count ;
end ;


procedure TBase_Store64.Set_Bytes_Read( Value : longint ) ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TBase_Store64.Set_Bytes_Written( Value : longint ) ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TBase_Store64.Set_Reads( Value : longint ) ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TBase_Store64.Set_Writes( Value : longint ) ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TBase_Store64.Set_Error_Count( Value : longint ) ;

begin
    Store_Statistics.Error_Count := Value ;
end ;
  {$ENDIF}



{ TStandard_COM_Store methods... }

{ API... }

function TStandard_COM_Store.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := Name ;
    Result := lowercase( S ) = 'tstandard_com_store' ;
end ;


function TStandard_COM_Store.Read_Data( var Data ;
    Address, _Size : TStore_Address ; var UEC : TUnified_Exception ) : TStore_Address ; stdcall ;

begin
    Read_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TStandard_COM_Store.Write_Data( var Data ;
    Address, _Size : TStore_Address ; var UEC : TUnified_Exception ) : TStore_Address ; stdcall ;

begin
    Write_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TStandard_COM_Store.Max_Storage : TStore_Size ; stdcall ;

begin
    Max_Storage := 0 ;
end ;


function TStandard_COM_Store.Min_Storage : TStore_Address ; stdcall ;

begin
    Min_Storage := 1 ;
end ;


function TStandard_COM_Store.Extend( Amount : TStore_Address ) : TStore_Address ; stdcall ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TStandard_COM_Store.Get_Read_Only : boolean ; stdcall ;

begin
    Get_Read_Only := _Read_Only ;
end ;


function TStandard_COM_Store.Get_Write_Only : boolean ; stdcall ;

begin
    Get_Write_Only := _Write_Only ;
end ;


procedure TStandard_COM_Store.Format ; stdcall ;

begin
    { This routine left intentionally blank. }
end ;


function TStandard_COM_Store.Get_Name : PAnsiChar ; stdcall ;

begin
    Get_Name := nil ;
end ;


function TStandard_COM_Store.Get_Cache : TCOM_Cache ; stdcall ;

begin
    Get_Cache := _Cache ;
end ;


procedure TStandard_COM_Store.Set_Cache( Value : TCOM_Cache ) ; stdcall ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TStandard_COM_Store.Contiguous_Store : boolean ; stdcall ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TStandard_COM_Store.Set_Max_Storage( Value : TStore_Address ;
    var Res : TUnified_Exception ) ;  stdcall ;

begin
    Res := Create_Base_Store_Exception( StoreErr_Invalid_Operation ) ;
end ;


function TStandard_COM_Store.Extended_Size : TStore_Address ; stdcall ;

begin
    Extended_Size := Max_Storage ;
end ;


function TStandard_COM_Store.Get_Bytes_Read : longint ; stdcall ;

begin
    Result := Store_Statistics.Bytes_Read ;
end ;


function TStandard_COM_Store.Get_Bytes_Written : longint ; stdcall ;

begin
    Result := Store_Statistics.Bytes_Written ;
end ;


function TStandard_COM_Store.Get_Reads : longint ; stdcall ;

begin
    Result := Store_Statistics.Reads ;
end ;


function TStandard_COM_Store.Get_Writes : longint ; stdcall ;

begin
    Result := Store_Statistics.Writes ;
end ;


function TStandard_COM_Store.Get_Error_Count : longint ; stdcall ;

begin
    Result := Store_Statistics.Error_Count ;
end ;


procedure TStandard_COM_Store.Set_Bytes_Read( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TStandard_COM_Store.Set_Bytes_Written( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TStandard_COM_Store.Set_Reads( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TStandard_COM_Store.Set_Writes( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TStandard_COM_Store.Set_Error_Count( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Error_Count := Value ;
end ;


procedure TStandard_COM_Store.Set_Read_Only( Value : boolean ) ; stdcall ;

begin
    _Read_Only := Value ;
end ;


procedure TStandard_COM_Store.Set_Write_Only( Value : boolean ) ; stdcall ;

begin
    _Write_Only := Value ;
end ;


  {$IFDEF INT64}
{ TStandard_COM_Store64 methods... }

{ API... }

function TStandard_COM_Store64.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := Name ;
    Result := lowercase( S ) = 'tstandard_com_store64' ;
end ;


function TStandard_COM_Store64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ; stdcall ;

begin
    Read_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TStandard_COM_Store64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ; stdcall ;

begin
    Write_Data := 0 ;
    UEC := Create_Abstract_Error_Exception ;
end ;


function TStandard_COM_Store64.Max_Storage : TStore_Size64 ; stdcall ;

begin
    Max_Storage := 0 ;
end ;


function TStandard_COM_Store64.Min_Storage : TStore_Address64 ; stdcall ;

begin
    Min_Storage := 1 ;
end ;


function TStandard_COM_Store64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ; stdcall ;

begin
    if( Amount = 0 ) then
    begin
        Extend := -1 ; { Cannot extend }
    end else
    begin
        Extend := 0 ; { Nothing added because we cannot extend }
    end ;
end ;


function TStandard_COM_Store64.Get_Read_Only : boolean ; stdcall ;

begin
    Get_Read_Only := _Read_Only ;
end ;


function TStandard_COM_Store64.Get_Write_Only : boolean ; stdcall ;

begin
    Get_Write_Only := _Write_Only ;
end ;


procedure TStandard_COM_Store64.Format ; stdcall ;

begin
    { This routine left intentionally blank. }
end ;


function TStandard_COM_Store64.Get_Name : PAnsiChar ; stdcall ;

begin
    Get_Name := nil ;
end ;


function TStandard_COM_Store64.Get_Cache : TCOM_Cache64 ; stdcall ;

begin
    Get_Cache := _Cache ;
end ;


procedure TStandard_COM_Store64.Set_Cache( Value : TCOM_Cache64 ) ; stdcall ;

begin
    if( Value <> _Cache ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Cache <> nil ) then
        begin
            _Cache.Detach ;
        end ;
        _Cache := Value ;
    end ;
end ;


function TStandard_COM_Store64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ; { Assume not }
end ;


procedure TStandard_COM_Store64.Set_Max_Storage( Value : TStore_Address64 ;
     var Res : TUnified_Exception ) ;

begin
    Res := Create_Base_Store_Exception( StoreErr_Invalid_Operation ) ;
end ;


function TStandard_COM_Store64.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Max_Storage ;
end ;


function TStandard_COM_Store64.Get_Bytes_Read : longint ;

begin
    Result := Store_Statistics.Bytes_Read ;
end ;


function TStandard_COM_Store64.Get_Bytes_Written : longint ;

begin
    Result := Store_Statistics.Bytes_Written ;
end ;


function TStandard_COM_Store64.Get_Reads : longint ; stdcall ;

begin
    Result := Store_Statistics.Reads ;
end ;


function TStandard_COM_Store64.Get_Writes : longint ; stdcall ;

begin
    Result := Store_Statistics.Writes ;
end ;


function TStandard_COM_Store64.Get_Error_Count : longint ; stdcall ;

begin
    Result := Store_Statistics.Error_Count ;
end ;


procedure TStandard_COM_Store64.Set_Bytes_Read( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Bytes_Read := Value ;
end ;


procedure TStandard_COM_Store64.Set_Bytes_Written( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Bytes_Written := Value ;
end ;


procedure TStandard_COM_Store64.Set_Reads( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Reads := Value ;
end ;


procedure TStandard_COM_Store64.Set_Writes( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Writes := Value ;
end ;


procedure TStandard_COM_Store64.Set_Error_Count( Value : longint ) ; stdcall ;

begin
    Store_Statistics.Error_Count := Value ;
end ;


procedure TStandard_COM_Store64.Set_Read_Only( Value : boolean ) ; stdcall ;

begin
    _Read_Only := Value ;
end ;


procedure TStandard_COM_Store64.Set_Write_Only( Value : boolean ) ; stdcall ;

begin
    _Write_Only := Value ;
end ;


function TStandard_COM_Store64.Get_Verifier : TCOM_Managed_Store64_Debugger ;

begin
    Get_Verifier := _Debugger ;
end ;


procedure TStandard_COM_Store64.Set_Verifier( Value : TCOM_Managed_Store64_Debugger ) ;

begin
    _Debugger := Value ;
end ;
  {$ENDIF}
{$ENDIF}


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


type TTest_Cache = object( TSCache )
                       procedure Initialize ; virtual ;
                       procedure Terminate ; virtual ;
                       procedure Attach ; virtual ;
                       procedure Detach ; virtual ;
                   end ;

procedure TTest_Cache.Initialize ;

begin
end ;


procedure TTest_Cache.Terminate ;

begin
end ;


procedure TTest_Cache.Attach ;

begin
end ;


procedure TTest_Cache.Detach ;

begin
end ;


  {$IFDEF DYNAMIC_OBJECTS}
type TTest_Cache1 = class( TCache )
                        procedure Initialize ; override ;
                        procedure Terminate ; override ;
                        procedure Attach ; override ;
                        procedure Detach ; override ;

                        function Read_Cache( var Data ;
                            Address, Size : longint ) : longint ; override ;
                        procedure Write_Cache( var Data ;
                            Address, Size : longint ) ; override ;
                        procedure Clear ; override ;
                        procedure Set_Max_Size( Size : longint ) ; override ;
                        function Get_Max_Size : longint ; override ;
                        function Get_Size : longint ; override ;
                        function Lock( Address : longint ) : boolean ; override ;
                        function Unlock( Address : longint ) : boolean ;
                            override ;
                        function Get_Hits : longint ; override ;
                        function Get_Misses : longint ; override ;
                        function Get_Writes : longint ; override ;
                        function Get_Turns : longint ; override ;
                        function Get_Replacements : longint ; override ;
                        procedure Set_Hits( Value : longint ) ; override ;
                        procedure Set_Misses( Value : longint ) ; override ;
                        procedure Set_Writes( Value : longint ) ; override ;
                        procedure Set_Turns( Value : longint ) ; override ;
                        procedure Set_Replacements( Value : longint ) ; override ;
                    end ;

procedure TTest_Cache1.Initialize ;

begin
end ;


procedure TTest_Cache1.Terminate ;

begin
end ;


procedure TTest_Cache1.Attach ;

begin
end ;


procedure TTest_Cache1.Detach ;

begin
end ;


function TTest_Cache1.Read_Cache( var Data ; Address, Size : longint ) : longint ;

begin
    Read_Cache := 0 ;
end ;


procedure TTest_Cache1.Write_Cache( var Data ; Address, Size : longint ) ;

begin
end ;


procedure TTest_Cache1.Clear ;

begin
end ;


procedure TTest_Cache1.Set_Max_Size( Size : longint ) ;

begin
end ;


function TTest_Cache1.Get_Max_Size : longint ;

begin
    Get_Max_Size := 0 ;
end ;


function TTest_Cache1.Get_Size : longint ;

begin
    Get_Size := 0 ;
end ;


function TTest_Cache1.Lock( Address : longint ) : boolean ;

begin
    Lock := False ;
end ;


function TTest_Cache1.Unlock( Address : longint ) : boolean ;

begin
    Unlock := False ;
end ;


function TTest_Cache1.Get_Hits : longint ;

begin
    Get_Hits := 0 ;
end ;


function TTest_Cache1.Get_Misses : longint ;

begin
    Get_Misses := 0 ;
end ;


function TTest_Cache1.Get_Writes : longint ;

begin
    Get_Writes := 0 ;
end ;


function TTest_Cache1.Get_Turns : longint ;

begin
    Get_Turns := 0 ;
end ;


function TTest_Cache1.Get_Replacements : longint ;

begin
    Get_Replacements := 0 ;
end ;


procedure TTest_Cache1.Set_Hits( Value : longint ) ;

begin
end ;


procedure TTest_Cache1.Set_Misses( Value : longint ) ;

begin
end ;


procedure TTest_Cache1.Set_Writes( Value : longint ) ;

begin
end ;


procedure TTest_Cache1.Set_Turns( Value : longint ) ;

begin
end ;


procedure TTest_Cache1.Set_Replacements( Value : longint ) ;

begin
end ;



type TTest_Cache164 = class( TCache64 )
                          procedure Initialize ; override ;
                          procedure Terminate ; override ;
                          procedure Attach ; override ;
                          procedure Detach ; override ;

                          function Read_Cache( var Data ;
                              Address, Size : int64 ) : longint ; override ;
                          procedure Write_Cache( var Data ;
                              Address, Size : int64 ) ; override ;
                          procedure Clear ; override ;
                          procedure Set_Max_Size( Size : longint ) ; override ;
                          function Get_Max_Size : longint ; override ;
                          function Get_Size : longint ; override ;
                          function Lock( Address : int64 ) : boolean ; override ;
                          function Unlock( Address : int64 ) : boolean ;
                              override ;
                          function Get_Hits : longint ; override ;
                          function Get_Misses : longint ; override ;
                          function Get_Writes : longint ; override ;
                          function Get_Turns : longint ; override ;
                          function Get_Replacements : longint ; override ;
                          procedure Set_Hits( Value : longint ) ; override ;
                          procedure Set_Misses( Value : longint ) ; override ;
                          procedure Set_Writes( Value : longint ) ; override ;
                          procedure Set_Turns( Value : longint ) ; override ;
                          procedure Set_Replacements( Value : longint ) ; override ;
                      end ;

procedure TTest_Cache164.Initialize ;

begin
end ;


procedure TTest_Cache164.Terminate ;

begin
end ;


procedure TTest_Cache164.Attach ;

begin
end ;


procedure TTest_Cache164.Detach ;

begin
end ;


function TTest_Cache164.Read_Cache( var Data ; Address, Size : int64 ) : longint ;

begin
    Read_Cache := 0 ;
end ;


procedure TTest_Cache164.Write_Cache( var Data ; Address, Size : int64 ) ;

begin
end ;


procedure TTest_Cache164.Clear ;

begin
end ;


procedure TTest_Cache164.Set_Max_Size( Size : longint ) ;

begin
end ;


function TTest_Cache164.Get_Max_Size : longint ;

begin
    Get_Max_Size := 0 ;
end ;


function TTest_Cache164.Get_Size : longint ;

begin
    Get_Size := 0 ;
end ;


function TTest_Cache164.Lock( Address : int64 ) : boolean ;

begin
    Lock := False ;
end ;


function TTest_Cache164.Unlock( Address : int64 ) : boolean ;

begin
    Unlock := False ;
end ;


function TTest_Cache164.Get_Hits : longint ;

begin
    Get_Hits := 0 ;
end ;


function TTest_Cache164.Get_Misses : longint ;

begin
    Get_Misses := 0 ;
end ;


function TTest_Cache164.Get_Writes : longint ;

begin
    Get_Writes := 0 ;
end ;


function TTest_Cache164.Get_Turns : longint ;

begin
    Get_Turns := 0 ;
end ;


function TTest_Cache164.Get_Replacements : longint ;

begin
    Get_Replacements := 0 ;
end ;


procedure TTest_Cache164.Set_Hits( Value : longint ) ;

begin
end ;


procedure TTest_Cache164.Set_Misses( Value : longint ) ;

begin
end ;


procedure TTest_Cache164.Set_Writes( Value : longint ) ;

begin
end ;


procedure TTest_Cache164.Set_Turns( Value : longint ) ;

begin
end ;


procedure TTest_Cache164.Set_Replacements( Value : longint ) ;

begin
end ;



type TTest_COM_Cache = class( TCOM_Cache )
                           function Initialize : TUnified_Exception ; override ;
                           function Terminate : TUnified_Exception ; override ;
                           procedure Attach ; override ;
                           procedure Detach ; override ;

                           function Is_Class( Name : PAnsiChar ) : boolean ;
                               override ;

                           function Read_Cache( var Data ;
                               Address, Size : longint ) : longint ; override ;
                           procedure Write_Cache( var Data ;
                               Address, Size : longint ) ; override ;
                           procedure Clear ; override ;
                           procedure Set_Max_Size( Size : longint ) ; override ;
                           function Get_Max_Size : longint ; override ;
                           function Get_Size : longint ; override ;
                           function Lock( Address : longint ) : boolean ; override ;
                           function Unlock( Address : longint ) : boolean ;
                               override ;
                           function Get_Hits : longint ; override ;
                           function Get_Misses : longint ; override ;
                           function Get_Writes : longint ; override ;
                           function Get_Turns : longint ; override ;
                           function Get_Replacements : longint ; override ;
                           procedure Set_Hits( Value : longint ) ; override ;
                           procedure Set_Misses( Value : longint ) ; override ;
                           procedure Set_Writes( Value : longint ) ; override ;
                           procedure Set_Turns( Value : longint ) ; override ;
                           procedure Set_Replacements( Value : longint ) ; override ;
                       end ;

function TTest_COM_Cache.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := Name ;
    Result := lowercase( S ) = 'ttest_com_cache' ;
end ;


function TTest_COM_Cache.Initialize : TUnified_Exception ;

begin
    Initialize := nil ;
end ;


function TTest_COM_Cache.Terminate : TUnified_Exception ;

begin
    Terminate := nil ;
end ;


procedure TTest_COM_Cache.Attach ;

begin
end ;


procedure TTest_COM_Cache.Detach ;

begin
end ;


function TTest_Com_Cache.Read_Cache( var Data ; Address, Size : longint ) : longint ;

begin
    Read_Cache := 0 ;
end ;


procedure TTest_Com_Cache.Write_Cache( var Data ; Address, Size : longint ) ;

begin
end ;


procedure TTest_Com_Cache.Clear ;

begin
end ;


procedure TTest_Com_Cache.Set_Max_Size( Size : longint ) ;

begin
end ;


function TTest_Com_Cache.Get_Max_Size : longint ;

begin
    Get_Max_Size := 0 ;
end ;


function TTest_Com_Cache.Get_Size : longint ;

begin
    Get_Size := 0 ;
end ;


function TTest_Com_Cache.Lock( Address : longint ) : boolean ;

begin
    Lock := False ;
end ;


function TTest_Com_Cache.Unlock( Address : longint ) : boolean ;

begin
    Unlock := False ;
end ;


function TTest_Com_Cache.Get_Hits : longint ;

begin
    Get_Hits := 0 ;
end ;


function TTest_Com_Cache.Get_Misses : longint ;

begin
    Get_Misses := 0 ;
end ;


function TTest_Com_Cache.Get_Writes : longint ;

begin
    Get_Writes := 0 ;
end ;


function TTest_Com_Cache.Get_Turns : longint ;

begin
    Get_Turns := 0 ;
end ;


function TTest_Com_Cache.Get_Replacements : longint ;

begin
    Get_Replacements := 0 ;
end ;


procedure TTest_Com_Cache.Set_Hits( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache.Set_Misses( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache.Set_Writes( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache.Set_Turns( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache.Set_Replacements( Value : longint ) ;

begin
end ;



type TTest_COM_Cache64 = class( TCOM_Cache64 )
                             function Initialize : TUnified_Exception ; override ;
                             function Terminate : TUnified_Exception ; override ;
                             procedure Attach ; override ;
                             procedure Detach ; override ;

                             function Is_Class( Name : PAnsiChar ) : boolean ;
                                 override ;

                             function Read_Cache( var Data ;
                                 Address, Size : int64 ) : longint ; override ;
                             procedure Write_Cache( var Data ;
                                 Address, Size : int64 ) ; override ;
                             procedure Clear ; override ;
                             procedure Set_Max_Size( Size : longint ) ; override ;
                             function Get_Max_Size : longint ; override ;
                             function Get_Size : longint ; override ;
                             function Lock( Address : int64 ) : boolean ; override ;
                             function Unlock( Address : int64 ) : boolean ;
                                 override ;
                             function Get_Hits : longint ; override ;
                             function Get_Misses : longint ; override ;
                             function Get_Writes : longint ; override ;
                             function Get_Turns : longint ; override ;
                             function Get_Replacements : longint ; override ;
                             procedure Set_Hits( Value : longint ) ; override ;
                             procedure Set_Misses( Value : longint ) ; override ;
                             procedure Set_Writes( Value : longint ) ; override ;
                             procedure Set_Turns( Value : longint ) ; override ;
                             procedure Set_Replacements( Value : longint ) ; override ;
                         end ;

function TTest_COM_Cache64.Initialize : TUnified_Exception ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TTest_COM_Cache64.Terminate : TUnified_Exception ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


procedure TTest_COM_Cache64.Attach ;

begin
end ;


procedure TTest_COM_Cache64.Detach ;

begin
end ;


function TTest_Com_Cache64.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := Name ;
    Result := lowercase( S ) = 'ttest_com_cache64' ;
end ;


function TTest_Com_Cache64.Read_Cache( var Data ; Address, Size : int64 ) : longint ;

begin
    Read_Cache := 0 ;
end ;


procedure TTest_Com_Cache64.Write_Cache( var Data ; Address, Size : int64 ) ;

begin
end ;


procedure TTest_Com_Cache64.Clear ;

begin
end ;


procedure TTest_Com_Cache64.Set_Max_Size( Size : longint ) ;

begin
end ;


function TTest_Com_Cache64.Get_Max_Size : longint ;

begin
    Get_Max_Size := 0 ;
end ;


function TTest_Com_Cache64.Get_Size : longint ;

begin
    Get_Size := 0 ;
end ;


function TTest_Com_Cache64.Lock( Address : int64 ) : boolean ;

begin
    Lock := False ;
end ;


function TTest_Com_Cache64.Unlock( Address : int64 ) : boolean ;

begin
    Unlock := False ;
end ;


function TTest_Com_Cache64.Get_Hits : longint ;

begin
    Get_Hits := 0 ;
end ;


function TTest_Com_Cache64.Get_Misses : longint ;

begin
    Get_Misses := 0 ;
end ;


function TTest_Com_Cache64.Get_Writes : longint ;

begin
    Get_Writes := 0 ;
end ;


function TTest_Com_Cache64.Get_Turns : longint ;

begin
    Get_Turns := 0 ;
end ;


function TTest_Com_Cache64.Get_Replacements : longint ;

begin
    Get_Replacements := 0 ;
end ;


procedure TTest_Com_Cache64.Set_Hits( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache64.Set_Misses( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache64.Set_Writes( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache64.Set_Turns( Value : longint ) ;

begin
end ;


procedure TTest_Com_Cache64.Set_Replacements( Value : longint ) ;

begin
end ;
  {$ENDIF}


procedure Test_Unit ;

var Buffer : integer ;
    Cache : TTest_Cache ;
    Test : TSBase_Store ;
{$IFDEF INT64}
    Test64 : TSBase_Store64 ;
{$ENDIF}
{$IFDEF DYNAMIC_OBJECTS}
  {$IFDEF INT64}
    Test164 : TBase_Store64 ;
    Cache164 : TTest_Cache164 ;
    Test_Com64 : TStandard_COM_Store64 ;
    Cache_Com64 : TTest_COM_Cache64 ;
  {$ENDIF}
    Test1 : TBase_Store ;
    Cache1 : TTest_Cache1 ;
    Test_Com : TStandard_COM_Store ;
    Cache_Com : TTest_COM_Cache ;
{$ENDIF}
    UEC : PSUnified_Exception ;
    U_E : TUnified_Exception ;

begin
    // Test TSBase_Store...
    Test.Init ;
    Test.Initialize ;
    Test.Format ;

    if( Test.Read_Data( Buffer, 0, 0, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    if( Test.Write_Data( Buffer, 0, 0, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 1 ) then
    begin
        Error ;
    end ;

    if( Test.Extend( 0 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test.Extend( 1 ) <> 0 ) then
    begin
        Error ;
    end ;

    if( Test.Get_Name <> '' ) then
    begin
        Error ;
    end ;

    if( Test.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache.Init ;
    Test.Set_Cache( @Cache ) ;
    if( Test.Get_Cache <> @Cache ) then
    begin
        Error ;
    end ;
    Test.Set_Cache( @Cache ) ;
    if( Test.Get_Cache <> @Cache ) then
    begin
        Error ;
    end ;
    Test.Set_Cache( nil ) ;
    if( Test.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache.Done ;

    if( Test.Contiguous_Store ) then
    begin
        Error ;
    end ;

    Test.Set_Max_Storage( 0, UEC ) ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test.Extended_Size <> 0 ) then
    begin
        Error ;
    end ;

    if( Test.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test.Get_Write_Only ) then
    begin
        Error ;
    end ;
    Test.Set_Read_Only( True ) ;
    Test.Set_Write_Only( True ) ;
    if( not Test.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( not Test.Get_Write_Only ) then
    begin
        Error ;
    end ;

    if( Test.Get_Bytes_Read <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Bytes_Written <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Reads <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Writes <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Error_Count <> 0 ) then
    begin
        Error ;
    end ;
    Test.Set_Bytes_Read( 1 ) ;
    Test.Set_Bytes_Written( 2 ) ;
    Test.Set_Reads( 3 ) ;
    Test.Set_Writes( 4 ) ;
    Test.Set_Error_Count( 5 ) ;
    if( Test.Get_Bytes_Read <> 1 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Bytes_Written <> 2 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Reads <> 3 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Writes <> 4 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Error_Count <> 5 ) then
    begin
        Error ;
    end ;

    Test.Terminate ;
    Test.Done ;

  {$IFDEF INT64}
    // Test TSBase_Store64...
    Test64.Init ;
    Test64.Initialize ;
    Test64.Format ;

    if( Test64.Read_Data( Buffer, 0, 0, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    if( Test64.Write_Data( Buffer, 0, 0, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test64.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.Min_Storage <> 1 ) then
    begin
        Error ;
    end ;

    if( Test64.Extend( 0 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test64.Extend( 1 ) <> 0 ) then
    begin
        Error ;
    end ;

    if( Test64.Get_Name <> '' ) then
    begin
        Error ;
    end ;

    if( Test64.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache.Init ;
    Test64.Set_Cache( @Cache ) ;
    if( Test64.Get_Cache <> @Cache ) then
    begin
        Error ;
    end ;
    Test64.Set_Cache( @Cache ) ;
    if( Test64.Get_Cache <> @Cache ) then
    begin
        Error ;
    end ;
    Test64.Set_Cache( nil ) ;
    if( Test64.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache.Done ;

    if( Test64.Contiguous_Store ) then
    begin
        Error ;
    end ;

    Test64.Set_Max_Storage( 0, UEC ) ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test64.Extended_Size <> 0 ) then
    begin
        Error ;
    end ;

    if( Test64.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Write_Only ) then
    begin
        Error ;
    end ;
    Test64.Set_Read_Only( True ) ;
    Test64.Set_Write_Only( True ) ;
    if( not Test64.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( not Test64.Get_Write_Only ) then
    begin
        Error ;
    end ;

    if( Test64.Get_Bytes_Read <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Bytes_Written <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Reads <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Writes <> 0 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Error_Count <> 0 ) then
    begin
        Error ;
    end ;
    Test64.Set_Bytes_Read( 1 ) ;
    Test64.Set_Bytes_Written( 2 ) ;
    Test64.Set_Reads( 3 ) ;
    Test64.Set_Writes( 4 ) ;
    Test64.Set_Error_Count( 5 ) ;
    if( Test64.Get_Bytes_Read <> 1 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Bytes_Written <> 2 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Reads <> 3 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Writes <> 4 ) then
    begin
        Error ;
    end ;
    if( Test64.Get_Error_Count <> 5 ) then
    begin
        Error ;
    end ;

    Test64.Terminate ;
    Test64.Done ;

    // Test TBase_Store...
    Test1 := TBase_Store.Create ;
    Test1.Initialize ;
    Test1.Format ;

    if( Test1.Read_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    if( Test1.Write_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test1.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.Min_Storage <> 1 ) then
    begin
        Error ;
    end ;

    if( Test1.Extend( 0 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test1.Extend( 1 ) <> 0 ) then
    begin
        Error ;
    end ;

    if( Test1.Get_Name <> '' ) then
    begin
        Error ;
    end ;

    if( Test1.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache1 := TTest_Cache1.Create ;
    Test1.Set_Cache( Cache1 ) ;
    if( Test1.Get_Cache <> Cache1 ) then
    begin
        Error ;
    end ;
    Test1.Set_Cache( Cache1 ) ;
    if( Test1.Get_Cache <> Cache1 ) then
    begin
        Error ;
    end ;
    Test1.Set_Cache( nil ) ;
    if( Test1.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache1.Free ;

    if( Test1.Contiguous_Store ) then
    begin
        Error ;
    end ;

    Test1.Set_Max_Storage( 0, U_E ) ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test1.Extended_Size <> 0 ) then
    begin
        Error ;
    end ;

    if( Test1.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Write_Only ) then
    begin
        Error ;
    end ;
    Test1.Set_Read_Only( True ) ;
    Test1.Set_Write_Only( True ) ;
    if( not Test1.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( not Test1.Get_Write_Only ) then
    begin
        Error ;
    end ;

    if( Test1.Get_Bytes_Read <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Bytes_Written <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Reads <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Writes <> 0 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Error_Count <> 0 ) then
    begin
        Error ;
    end ;
    Test1.Set_Bytes_Read( 1 ) ;
    Test1.Set_Bytes_Written( 2 ) ;
    Test1.Set_Reads( 3 ) ;
    Test1.Set_Writes( 4 ) ;
    Test1.Set_Error_Count( 5 ) ;
    if( Test1.Get_Bytes_Read <> 1 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Bytes_Written <> 2 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Reads <> 3 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Writes <> 4 ) then
    begin
        Error ;
    end ;
    if( Test1.Get_Error_Count <> 5 ) then
    begin
        Error ;
    end ;

    Test1.Terminate ;

    // Test TBase_Store...
    Test164 := TBase_Store64.Create ;
    Test164.Initialize ;
    Test164.Format ;

    if( Test164.Read_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    if( Test164.Write_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test164.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.Min_Storage <> 1 ) then
    begin
        Error ;
    end ;

    if( Test164.Extend( 0 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test164.Extend( 1 ) <> 0 ) then
    begin
        Error ;
    end ;

    if( Test164.Get_Name <> '' ) then
    begin
        Error ;
    end ;

    if( Test164.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache164 := TTest_Cache164.Create ;
    Test164.Set_Cache( Cache164 ) ;
    if( Test164.Get_Cache <> Cache164 ) then
    begin
        Error ;
    end ;
    Test164.Set_Cache( Cache164 ) ;
    if( Test164.Get_Cache <> Cache164 ) then
    begin
        Error ;
    end ;
    Test164.Set_Cache( nil ) ;
    if( Test164.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache164.Free ;

    if( Test164.Contiguous_Store ) then
    begin
        Error ;
    end ;

    Test164.Set_Max_Storage( 0, U_E ) ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test164.Extended_Size <> 0 ) then
    begin
        Error ;
    end ;

    if( Test164.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Write_Only ) then
    begin
        Error ;
    end ;
    Test164.Set_Read_Only( True ) ;
    Test164.Set_Write_Only( True ) ;
    if( not Test164.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( not Test164.Get_Write_Only ) then
    begin
        Error ;
    end ;

    if( Test164.Get_Bytes_Read <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Bytes_Written <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Reads <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Writes <> 0 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Error_Count <> 0 ) then
    begin
        Error ;
    end ;
    Test164.Set_Bytes_Read( 1 ) ;
    Test164.Set_Bytes_Written( 2 ) ;
    Test164.Set_Reads( 3 ) ;
    Test164.Set_Writes( 4 ) ;
    Test164.Set_Error_Count( 5 ) ;
    if( Test164.Get_Bytes_Read <> 1 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Bytes_Written <> 2 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Reads <> 3 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Writes <> 4 ) then
    begin
        Error ;
    end ;
    if( Test164.Get_Error_Count <> 5 ) then
    begin
        Error ;
    end ;

    Test164.Terminate ;

    // Test TStandard_COM_Store...
    Test_COM := TStandard_COM_Store.Create ;
    Test_COM.Initialize ;
    Test_COM.Format ;

    if( Test_COM.Read_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    if( Test_COM.Write_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test_COM.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Min_Storage <> 1 ) then
    begin
        Error ;
    end ;

    if( Test_COM.Extend( 0 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Extend( 1 ) <> 0 ) then
    begin
        Error ;
    end ;

    if( Test_COM.Get_Name <> '' ) then
    begin
        Error ;
    end ;

    if( Test_COM.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache_COM := TTest_COM_Cache.Create ;
    Test_COM.Set_Cache( Cache_COM ) ;
    if( Test_COM.Get_Cache <> Cache_COM ) then
    begin
        Error ;
    end ;
    Test_COM.Set_Cache( Cache_COM ) ;
    if( Test_COM.Get_Cache <> Cache_COM ) then
    begin
        Error ;
    end ;
    Test_COM.Set_Cache( nil ) ;
    if( Test_COM.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache_COM.Free ;

    if( Test_COM.Contiguous_Store ) then
    begin
        Error ;
    end ;

    Test_COM.Set_Max_Storage( 0, U_E ) ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test_COM.Extended_Size <> 0 ) then
    begin
        Error ;
    end ;

    if( Test_COM.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Write_Only ) then
    begin
        Error ;
    end ;
    Test_COM.Set_Read_Only( True ) ;
    Test_COM.Set_Write_Only( True ) ;
    if( not Test_COM.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( not Test_COM.Get_Write_Only ) then
    begin
        Error ;
    end ;

    if( Test_COM.Get_Bytes_Read <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Bytes_Written <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Reads <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Writes <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Error_Count <> 0 ) then
    begin
        Error ;
    end ;
    Test_COM.Set_Bytes_Read( 1 ) ;
    Test_COM.Set_Bytes_Written( 2 ) ;
    Test_COM.Set_Reads( 3 ) ;
    Test_COM.Set_Writes( 4 ) ;
    Test_COM.Set_Error_Count( 5 ) ;
    if( Test_COM.Get_Bytes_Read <> 1 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Bytes_Written <> 2 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Reads <> 3 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Writes <> 4 ) then
    begin
        Error ;
    end ;
    if( Test_COM.Get_Error_Count <> 5 ) then
    begin
        Error ;
    end ;

    Test_COM.Terminate ;

    // Test TStandard_COM_Store64...
    Test_Com64 := TStandard_Com_Store64.Create ;
    Test_Com64.Initialize ;
    Test_Com64.Format ;

    if( Test_Com64.Read_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Write_Data( Buffer, 0, 0, U_E ) <> 0 ) then
    begin
        Error ;
    end ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test_Com64.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Min_Storage <> 1 ) then
    begin
        Error ;
    end ;

    if( Test_Com64.Extend( 0 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Extend( 1 ) <> 0 ) then
    begin
        Error ;
    end ;

    if( Test_Com64.Get_Name <> '' ) then
    begin
        Error ;
    end ;

    if( Test_Com64.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache_Com64 := TTest_Com_Cache64.Create ;
    Test_Com64.Set_Cache( Cache_Com64 ) ;
    if( Test_Com64.Get_Cache <> Cache_Com64 ) then
    begin
        Error ;
    end ;
    Test_Com64.Set_Cache( Cache_Com64 ) ;
    if( Test_Com64.Get_Cache <> Cache_Com64 ) then
    begin
        Error ;
    end ;
    Test_Com64.Set_Cache( nil ) ;
    if( Test_Com64.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Cache_Com64.Free ;

    if( Test_Com64.Contiguous_Store ) then
    begin
        Error ;
    end ;

    Test_Com64.Set_Max_Storage( 0, U_E ) ;
    if( UEC = nil ) then
    begin
        Error ;
    end ;

    if( Test_Com64.Extended_Size <> 0 ) then
    begin
        Error ;
    end ;

    if( Test_Com64.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Write_Only ) then
    begin
        Error ;
    end ;
    Test_Com64.Set_Read_Only( True ) ;
    Test_Com64.Set_Write_Only( True ) ;
    if( not Test_Com64.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( not Test_Com64.Get_Write_Only ) then
    begin
        Error ;
    end ;

    if( Test_Com64.Get_Bytes_Read <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Bytes_Written <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Reads <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Writes <> 0 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Error_Count <> 0 ) then
    begin
        Error ;
    end ;
    Test_Com64.Set_Bytes_Read( 1 ) ;
    Test_Com64.Set_Bytes_Written( 2 ) ;
    Test_Com64.Set_Reads( 3 ) ;
    Test_Com64.Set_Writes( 4 ) ;
    Test_Com64.Set_Error_Count( 5 ) ;
    if( Test_Com64.Get_Bytes_Read <> 1 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Bytes_Written <> 2 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Reads <> 3 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Writes <> 4 ) then
    begin
        Error ;
    end ;
    if( Test_Com64.Get_Error_Count <> 5 ) then
    begin
        Error ;
    end ;

    Test_Com64.Terminate ;
  {$ENDIF}
end ;
{$ENDIF}



end.
