{
        Program Name : TestMan
        Package Name : Subroutine Library - Layer 2
        Purpose      : Default Test Manager
        Institution  : Conroy & Conroy Co.
        Date Written : 15-August-2008
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2008 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains the default Test Manager implementation.
}

unit TestMan ;

interface

uses { Subroutine Library - Layer 0... }
     Collect, { TCollection }
     _Streams, { TStream }

     { Subroutine Library - Layer 2... }
     _TestMa, { TSTest_Manager }

     { Layer 1... }
     _Applic ; { Application_Manager }

{$I EDEFINES.INC}

type TSDefault_Test_Manager = object( TSTest_Manager )
                                 public { Constructors and destructors... }
                                     constructor Init ;
                                     destructor Done ; virtual ;

                                 private { Instance data... }
                                     Collection : TCollection ;

                                 public { Unit Tester support... }
                                     function Get_Unit_Tester( Index : longint ) : PSUnit_Tester ;
                                         virtual ;

                                     { Register a unit Tester }
                                     procedure Register_Unit_Tester( D : PSUnit_Tester ) ;
                                         virtual ;

                                     { Remove a registered unit Tester }
                                     procedure Remove_Unit_Tester( D : PSUnit_Tester ) ;
                                         virtual ;

                                     { Returns count of registered unit Testers }
                                     function Unit_Tester_Count : longint ;
                                         virtual ;

                                     { Runs test }
                                     function Run_Test( Flags : longint ;
                                       Stream : PSStream ) : boolean ;
                                       virtual ;
                             end ; { TSDefault_Test_Manager }
     PSDefault_Test_Manager = ^TSDefault_Test_Manager ;

implementation

{ Constructors and destructors... }

constructor TSDefault_Test_Manager.Init ;

begin
    inherited Init ;

    Collection.Init( 0, 4 ) ;
end ;


destructor TSDefault_Test_Manager.Done ;

begin
    Collection.Done ;

    inherited Done ;
end ;


{ Unit Tester support... }

function TSDefault_Test_Manager.Get_Unit_Tester( Index : longint ) : PSUnit_Tester ;

begin
    Get_Unit_Tester := PSUnit_Tester( Collection.At( Index ) ) ;
end ;


procedure TSDefault_Test_Manager.Register_Unit_Tester( D : PSUnit_Tester ) ;

begin
    Collection.Insert( D ) ;
end ;


procedure TSDefault_Test_Manager.Remove_Unit_Tester( D : PSUnit_Tester ) ;

begin
    Collection.Delete( D ) ;
end ;


function TSDefault_Test_Manager.Unit_Tester_Count : longint ;

begin
    Unit_Tester_Count := Collection.Count ;
end ;


function TSDefault_Test_Manager.Run_Test( Flags : longint ;
    Stream : PSStream ) : boolean ;

var Loop : integer ;
    T : boolean ;

begin
    Run_Test := True ;
    for Loop := 0 to Unit_Tester_Count - 1 do
    begin
        T := Get_Unit_Tester( Loop )^.Run_Test( Flags, Stream ) ;
        if( not T ) then
        begin
            Run_Test := False ;
        end ;
    end ;
end ;



var _Test_Manager : TSDefault_Test_Manager ;
{$IFDEF FPC_OR_DELPHI}
    _TM_Initialized : boolean = False ;
{$ENDIF}

function Test_Manager : PSTest_Manager ;

begin
    Test_Manager := @_Test_Manager ;
end ;


procedure Initialize_Unit ;

begin
{$IFDEF FPC_OR_DELPHI}
    if( _TM_Initialized ) then
    begin
        exit ;
    end ;
{$ENDIF}
    if(
        ( Application_Manager <> nil )
        and
        ( Application_Manager^.Test_Manager = nil )
      ) then
    begin
        _Test_Manager.Init ;
        Application_Manager^.Set_Test_Manager( @_Test_Manager ) ;
{$IFDEF FPC_OR_DELPHI}
        _TM_Initialized := True ;
{$ENDIF}
    end ;
end ;


begin
    Initialize_Unit ;
end.
