{
        Program Name : Text_Editor_Interface
        Package Name : LB
        Purpose      : Visual interface to a Text_Editor
        Institution  : Conroy & Conroy Co.
        Date Written :
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2012 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit TE_Interface ;

interface

uses // Borland...
     Classes, // TShiftStates
     Extctrls, // TPanel
     StdCtrls, // TScrollBar
     Graphics, // TColor

     // C&C...
     _TE_Interface, // TText_Editor_Interface
     Paint_Panel, // TPaint_Panel
     TextEditor ; // TSimple_Text_Editor

type TStandard_Text_Editor_Interface = class( TText_Editor_Interface )
                                           public // Constructors and destructors...
                                               constructor Create ;
                                               destructor Destroy ; override ;

                                           private // Data....
                                               Editor : TSimple_Text_Editor ;
                                               Panel : TPaint_Panel ;
                                               Gutter, Text_Panel : TPaint_Panel ;
                                               Temp_Get_Line : string ;
                                               Horizontal_Scrollbar : TScrollBar ;
                                               Vertical_ScrollBar : TScrollBar ;
                                               Cursor_X, Cursor_Y : integer ;
                                               Have_Caret : boolean ;
                                               Notifier : TText_Editor_Interface_Notifier ;
                                               Max_C, Max_R : integer ;
                                               H, W : integer ;
                                               Selecting : boolean ; // True if selecting text
                                               Selection_Anchor : TText_Editor_Position ;

                                           private // Internal utility routines...
                                               procedure CB_Paint( Sender : TObject ) ;
                                               procedure CB_Horizontal_Scroll( Sender : TObject ; ScrollCode : TScrollCode ; var ScrollPos : Integer ) ;
                                               procedure CB_Vertical_Scroll( Sender : TObject ; ScrollCode : TScrollCode ; var ScrollPos : Integer ) ;
                                               procedure CB_SetFocus( Sender : TObject ) ;
                                               procedure CB_KillFocus( Sender : TObject ) ;
                                               procedure CB_Click( Sender : TObject ) ;
                                               procedure CB_KeyDown( Sender : TObject ; var Key : Word ; Shift : TShiftState ) ;
                                               procedure CB_Focus( Sender : TObject ) ;

                                               procedure Set_Caret ;
                                               procedure Calculate_Metrics ;
                                               procedure Adjust_Scrollbars ;
                                               procedure Invalidate_Region( Start_R, End_R, Start_C, End_C : integer ) ;

                                           public // API...
                                               procedure Set_Parent( Handle : integer ) ;
                                                   override ;
                                               function Get_Parent : integer ; override ;
                                               procedure Set_Size( Height, Width : integer ) ;
                                                   override ;
                                               function Column_Count : integer ;
                                                   override ;
                                               function Line_Count : integer ;
                                                   override ;

                                               procedure Set_Background( Color : TColor ) ;
                                                   override ;
                                               procedure Set_Foreground( Color : TColor ) ;
                                                   override ;
                                               procedure Set_Font( Name : PChar ) ;
                                                   override ;
                                               procedure Set_Font_Size( Siz : integer ) ;
                                                   override ;

                                               procedure Load_File( Nam : PChar ) ;
                                                   override ;
                                               procedure Save_File( Nam : PChar ) ;
                                                   override ;

                                               function Get_Line( Row : integer ) : PChar ;
                                                   override ;
                                               procedure Locate( S : string ;
                                                  var R, C : integer ;
                                                  CS : boolean ) ;
                                                  override ;
                                               procedure Position_To( R, C : integer ) ;
                                                   override ;
                                               procedure Get_Position( var R, C : integer ) ;
                                                   override  ;
                                               procedure Set_Notifier( V : TText_Editor_Interface_Notifier ) ;
                                                   override  ;
                                               procedure Key_Down( var Key : Word ;
                                                   Shift : TShiftState ) ;
                                                   override  ;
                                               procedure Bring_Cursor_Into_View ;
                                                   override  ;
                                               procedure SetFocus ; override ;
                                               procedure Set_Position( P : TText_Editor_Position ) ;
                                                   override ;
                                               function Get_Selected_Text : string ;
                                                   override ;
                                       end ; { TStandard_Text_Editor_Interface }

implementation

uses // Borland...
     Windows, // SetCaretPos
     Controls, // alBottom
     Forms, // TApplication
     Sysutils ; // inttostr

procedure Normalize_Range( var Start_Row, Start_Column, End_Row, End_Column : integer ) ;

var Temp : integer ;

begin
    if( Start_Row > End_Row ) then
    begin
        Temp := End_Row ;
        End_Row := Start_Row ;
        Start_Row := Temp ;
        Temp := End_Column ;
        End_Column := Start_Column ;
        Start_Column := Temp ;
        exit ;
    end ;
    if( Start_Row = End_Row ) then
    begin
        if( Start_Column > End_Column ) then
        begin
            Temp := End_Column ;
            End_Column := Start_Column ;
            Start_Column := Temp ;
        end ;
    end ;
end ;


type TStandard_Text_Editor_Observer = class( TText_Editor_Observer )
                                          private
                                              Notifier : TText_Editor_Interface_Notifier ;

                                          public
                                              procedure Line_Read( Count : integer ) ;
                                                  override ;
                                      end ;


procedure TStandard_Text_Editor_Observer.Line_Read( Count : integer ) ;

begin
    Notifier.Line_Read( Count ) ;
end ;



// TStandard_Text_Editor_Interface methods...

// Constructors and destructors...

constructor TStandard_Text_Editor_Interface.Create ;

begin
    inherited Create ;

    Editor := TSimple_Text_Editor.Create ;
    Panel := TPaint_Panel.Create( Application ) ;
    Panel.DoubleBuffered := True ;
    Panel.OnSetFocus := CB_Focus ;

    Horizontal_Scrollbar := TScrollBar.Create( Panel ) ;
    Horizontal_Scrollbar.Parent := Panel ;
    Horizontal_Scrollbar.Align := alBottom ;
    Horizontal_Scrollbar.TabStop := False ;
    Horizontal_Scrollbar.OnScroll := CB_Horizontal_Scroll ;
    Vertical_ScrollBar := TScrollBar.Create( Panel ) ;
    Vertical_ScrollBar.Parent := Panel ;
    Vertical_Scrollbar.Kind := sbVertical ;
    Vertical_ScrollBar.Align := alRight ;
    Vertical_ScrollBar.TabStop := False ;
    Vertical_ScrollBar.OnScroll := CB_Vertical_Scroll ;
    Gutter := TPaint_Panel.Create( Application ) ;
    Gutter.Parent := Panel ;
    Gutter.Align := alLeft ;
    Gutter.Color := clBtnFace ;
    Gutter.Font.Color := clWindowText ;
    Gutter.Width := 25 ;
    Gutter.On_Paint := CB_Paint ;
    Gutter.DoubleBuffered := True ;

    Text_Panel := TPaint_Panel.Create( Application ) ;
    Text_Panel.Parent := Panel ;
    Text_Panel.Align := alClient ;
    Text_Panel.Color := clWindow ;
    Text_Panel.Font.Color := clWindowText ;
    Text_Panel.Tabstop := True ;
    Text_Panel.On_Paint := CB_Paint ;
    Text_Panel.Font.Name := 'Courier New' ;
    Text_Panel.DoubleBuffered := True ;
    Text_Panel.OnSetFocus := CB_SetFocus ;
    Text_Panel.OnKillFocus := CB_KillFocus ;
    Text_Panel.OnClick := CB_Click ;
    Text_Panel.OnKeyDown := CB_KeyDown ;
end ;


destructor TStandard_Text_Editor_Interface.Destroy ;

begin
    Editor.Free ;
    Editor := nil ;

    inherited Destroy ;
end ;


// Internal utility routines...

procedure TStandard_Text_Editor_Interface.CB_Paint( Sender : TObject ) ;

    function After( Start_R, Start_C, R, C : integer ) : boolean ;

    begin
        Result := ( R > Start_R ) or ( ( R = Start_R ) and ( C >= Start_C ) ) ;
    end ;


    function Before( End_R, End_C, R, C : integer ) : boolean ;

    begin
        Result := ( R < End_R ) or ( ( R = End_R ) and ( C < End_C ) ) ;
    end ;

var C, R, I : integer ;
    Ch : char ;
    S, S1 : string ;
    This_Column, This_Row : integer ;
    Start_R, Start_C, End_R, End_C : integer ;
    Rect : TRect ;

begin
    // Setup...
    if( Selecting ) then
    begin
        Start_R := Selection_Anchor.Row ;
        Start_C := Selection_Anchor.Column ;
        End_R := Cursor_Y ;
        End_C := Cursor_X ;
        Normalize_Range( Start_R, Start_C, End_R, End_C ) ;
    end ;
    if( Have_Caret ) then
    begin
        HideCaret( Text_Panel.Handle ) ;
    end ;

    // Erase background...
    Text_Panel.Brush.Color := Panel.Color ;
    Text_Panel.Canvas.FillRect( Panel.ClientRect ) ;
    Text_Panel.Canvas.Font := Text_Panel.Font ;
    Gutter.Canvas.Brush.Color := Gutter.Color ;
    Gutter.Canvas.FillRect( Gutter.ClientRect ) ;
    Gutter.Canvas.Font := Gutter.Font ;

    // Draw foreground...
    S1 := inttostr( Line_Count ) ;
    for R := 0 to Max_R do
    begin
        This_Row := R + Vertical_Scrollbar.Position ;
        if( This_Row >= Line_Count ) then
        begin
            break ; // Past end of last line
        end ;

        // Write number to gutter...
        S := inttostr( This_Row + 1 ) ;
        I := length( S1 ) - length( S ) ;
        I := I * Gutter.Canvas.TextWidth( '0' ) ;
        Gutter.Canvas.TextOut( I + 2, R * H, S ) ;

        // Write row...
        S := Get_Line( This_Row ) ;
        for C := 0 to Max_C do
        begin
            This_Column := C + Horizontal_Scrollbar.Position + 1 ;
            if( This_Column >= length( S ) ) then
            begin
                Ch := ' ' ;
            end else
            begin
                Ch := S[ This_Column ] ;
            end ;
            if( Selecting ) then
            begin
                if(
                    After( Start_R, Start_C, This_Row, This_Column )
                    and
                    Before( End_R, End_C, This_Row, This_Column )
                  ) then
                begin
                    Text_Panel.Canvas.Font.Color := clWhite ;
                    Text_Panel.Canvas.Brush.Color := clNavy ;
                    Rect.Top := R * H ;
                    Rect.Bottom := ( R + 1 ) * H + 1 ;
                    Rect.Left := C * W ;
                    Rect.Right := ( C + 1 ) * W + 1 ;
                    Text_Panel.Canvas.FillRect( Rect ) ;
                    SetbkMode(Text_Panel.Canvas.Handle, TRANSPARENT) ;
                    Text_Panel.Canvas.TextOut( C * W, R * H, Ch ) ;
                    Text_Panel.Canvas.Font.Color := Text_Panel.Font.Color ;
                    Text_Panel.Canvas.Brush.Color := Panel.Color ;
                    continue ;
                end ;
            end ; // if( Selecting )
            Text_Panel.Canvas.TextOut( C * W, R * H, Ch ) ;
        end ; // for C
    end ; // for R

    Set_Caret ;
    if( Have_Caret ) then
    begin
        ShowCaret( Text_Panel.Handle ) ;
    end ;
end ; // TStandard_Text_Editor_Interface.CB_Paint


procedure TStandard_Text_Editor_Interface.CB_Horizontal_Scroll( Sender : TObject ;
    ScrollCode : TScrollCode ; var ScrollPos : Integer ) ;

begin
    Text_Panel.Invalidate ;
end ;


procedure TStandard_Text_Editor_Interface.CB_Vertical_Scroll( Sender : TObject ;
    ScrollCode : TScrollCode ; var ScrollPos : Integer ) ;

begin
    Panel.Invalidate ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Caret ;

begin
    if( Have_Caret ) then
    begin
        SetCaretPos( ( Cursor_X - Horizontal_ScrollBar.Position ) * W, ( Cursor_Y - Vertical_ScrollBar.Position ) * H ) ;
    end ;
    if( Notifier <> nil ) then
    begin
        Notifier.Cursor_Change( Cursor_Y, Cursor_X ) ;
    end ;
end ;


procedure TStandard_Text_Editor_Interface.Calculate_Metrics ;

begin
    H := Text_Panel.Canvas.TextHeight( 'W' ) ;
    W := Text_Panel.Canvas.TextWidth( 'W' ) ;
    Max_C := Text_Panel.ClientWidth div W + 1 ;
    Max_R := Text_Panel.ClientHeight div H + 1 ;
end ;


procedure TStandard_Text_Editor_Interface.Adjust_Scrollbars ;

begin
    // Adjust scrollbars...
    if( Column_Count = 0 ) then
    begin
        Horizontal_Scrollbar.Max := 0 ;
    end else
    begin
        Horizontal_Scrollbar.Max := Column_Count - 1 ;
    end ;
    Horizontal_Scrollbar.PageSize := Max_C ;
    Horizontal_Scrollbar.LargeChange := Max_C ;
    if( Line_Count = 0 ) then
    begin
        Vertical_Scrollbar.Max := 0 ;
    end else
    begin
        Vertical_Scrollbar.Max := Line_Count - 1 ;
    end ;
    Vertical_Scrollbar.PageSize := Max_R ;
    Vertical_Scrollbar.LargeChange := Max_R ;
end ;


// API...

procedure TStandard_Text_Editor_Interface.Set_Parent( Handle : integer ) ;

begin
    Panel.ParentWindow := Handle ;
end ;


function TStandard_Text_Editor_Interface.Get_Parent : integer ;

begin
    Result := Panel.ParentWindow ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Size( Height, Width : integer ) ;

begin
    Panel.Height := Height ;
    Panel.Width := Width ;
    Panel.Cursor := crIBeam ;

    Calculate_Metrics ;
    Adjust_Scrollbars ;
end ;


function TStandard_Text_Editor_Interface.Column_Count : integer ;

begin
    Result := Editor.Column_Count ;
end ;


function TStandard_Text_Editor_Interface.Line_Count : integer ;

begin
    Result := Editor.Row_Count ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Background( Color : TColor ) ;

begin
    Text_Panel.Color := Color ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Foreground( Color : TColor ) ;

begin
    Text_Panel.Font.Color := Color ;
    Gutter.Font.Color := Color ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Font( Name : PChar ) ;

begin
    Text_Panel.Font.Name := Name ;
    Text_Panel.Canvas.Font := Panel.Font ;
    Calculate_Metrics ;
    Text_Panel.Invalidate ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Font_Size( Siz : integer ) ;

begin
    Text_Panel.Font.Size := Siz ;
    Text_Panel.Canvas.Font := Panel.Font ;
    Gutter.Font.Size := Siz ;
    Calculate_Metrics ;
    Panel.Invalidate ;
end ;


procedure TStandard_Text_Editor_Interface.Load_File( Nam : PChar ) ;

var Observer : TStandard_Text_Editor_Observer ;

begin
    Text_Panel.Cursor := crHourGlass ;
    Observer := TStandard_Text_Editor_Observer.Create ;
    try
        Observer.Notifier := Notifier ;
        Editor.Open_File( Nam, Observer ) ;
    finally
        Observer.Free ;
    end ;
    Gutter.Width := Gutter.Canvas.TextWidth( inttostr( Editor.Row_Count )+ '*' ) + 4 ;
    Adjust_Scrollbars ;
    Text_Panel.Repaint ;
    Text_Panel.Cursor := crIBeam ;
end ;


procedure TStandard_Text_Editor_Interface.Save_File( Nam : PChar ) ;

begin
    Editor.SaveToFile( Nam ) ;
end ;


function TStandard_Text_Editor_Interface.Get_Line( Row : integer ) : PChar ;

begin
    Temp_Get_Line := Editor.Get_Row( Row ) ;
    Result := PChar( Temp_Get_Line ) ;
end ;


procedure TStandard_Text_Editor_Interface.Locate( S : string ;
    var R, C : integer ; CS : boolean ) ;

var L : string ;
    P : integer ;

begin
    Text_Panel.Cursor := crHourGlass ;
    if( CS ) then
    begin
        S := lowercase( S ) ;
    end ;
    try
        if( R < 0 ) then
        begin
            R := Cursor_Y ;
        end ;
        if( C < 0 ) then
        begin
            C := Cursor_X ;
        end ;
        while( R < Line_Count ) do
        begin
            L := Get_Line( R ) ;
            if( CS ) then
            begin
                L := lowercase( L ) ;
            end ;
            P := pos( S, copy( L, C, length( L ) ) ) ;
            if( P > 0 ) then
            begin
                C := C + P - 1 ;
                exit ;
            end ;
            R := R + 1 ;
            C := 0 ;
        end ;
        R := -1 ; { Not found }
    finally
        Text_Panel.Cursor := crIBeam ;
    end ;
end ; { TStandard_Text_Editor_Interface.Locate }


procedure TStandard_Text_Editor_Interface.Position_To( R, C : integer ) ;

var Dummy : integer ;

begin
    // Boundary checks...
    if( R < 0 ) then
    begin
        R := 0 ;
    end ;
    if( R >= Editor.Row_Count ) then
    begin
        R := Editor.Row_Count - 1 ;
    end ;
    if( C < 0 ) then
    begin
        C := 0 ;
    end ;

    // Position...
    if(
        ( Vertical_ScrollBar.Position > R )
        or
        ( Vertical_ScrollBar.Position + Max_R < R )
      ) then
    begin
        Dummy := R - Max_R div 2 ;
        if( Dummy < 0 ) then
        begin
            Dummy := 0 ;
        end ;
        Vertical_ScrollBar.Position := Dummy ;
    end ;
    if(
        ( Horizontal_ScrollBar.Position > C )
        or
        ( Horizontal_ScrollBar.Position + Max_C < C )
      ) then
    begin
        Dummy := C - Max_C div 2 ;
        if( Dummy < 0 ) then
        begin
            Dummy := 0 ;
        end ;
        if( Dummy > C ) then
        begin
            Dummy := C ;
        end ;
        Horizontal_ScrollBar.Position := Dummy ;
    end ;
    Cursor_X := C ;
    Cursor_Y := R ;
    Bring_Cursor_Into_View ;
    Set_Caret ;

    Panel.Invalidate ;
end ; // TStandard_Text_Editor_Interface.Position_To


procedure TStandard_Text_Editor_Interface.CB_SetFocus( Sender : TObject ) ;

begin
    CreateCaret( Text_Panel.Handle, 0, GetSystemMetrics( SM_CXBORDER ), H ) ;
    ShowCaret( Text_Panel.Handle ) ;
    Have_Caret := True ;
end ;


procedure TStandard_Text_Editor_Interface.CB_KillFocus( Sender : TObject ) ;

begin
    HideCaret( Text_Panel.Handle ) ;
    DestroyCaret ;
    Have_Caret := False ;
end ;


procedure TStandard_Text_Editor_Interface.CB_Click( Sender : TObject ) ;

var P : TPoint ;

begin
    GetCursorPos( P ) ;
    P := Text_Panel.ScreenToClient( P ) ;
    if(
        ( P.Y < 0 )
        or
        ( P.X < 0 )
        or
        ( P.Y > Text_Panel.ClientHeight )
        or
        ( P.X > Text_Panel.ClientWidth )
      ) then
    begin
        exit ; // Out of range
    end ;
    P.X := ( P.X + H div 2 ) div W ;
    P.Y := P.Y div H ;
    Cursor_X := P.X + Horizontal_Scrollbar.Position ;
    Cursor_Y := P.Y + Vertical_Scrollbar.Position ;
    Set_Caret ;
    SetFocus ;
end ;


procedure TStandard_Text_Editor_Interface.Get_Position( var R, C : integer ) ;

begin
    R := Cursor_Y ;
    C := Cursor_X ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Notifier( V : TText_Editor_Interface_Notifier ) ;

begin
    Notifier := V ;
end ;


procedure TStandard_Text_Editor_Interface.CB_KeyDown( Sender : TObject ;
    var Key : Word ; Shift : TShiftState ) ;

begin
    Key_Down( Key, Shift ) ;
end ;


procedure TStandard_Text_Editor_Interface.Invalidate_Region( Start_R, End_R, Start_C, End_C : integer ) ;

var Rect : TRect ;
    Temp : integer ;

begin
    if( ( Start_R = End_R ) and ( Start_C = End_C ) ) then
    begin
        Text_Panel.Invalidate ; // Repaint the whole thing
        exit ;
    end ;
    if( Start_R < End_R ) then
    begin
        Temp := End_R ;
        End_R := Start_R ;
        Start_R := Temp ;
    end else
    if( Start_R = End_R ) then
    begin
       if( Start_C > End_C ) then
       begin
           Temp := End_C ;
           End_C := Start_C ;
           Start_C := Temp ;
       end ;
    end ;
    Rect.Top := ( Start_R - Vertical_Scrollbar.Position ) * H ;
    Rect.Bottom := ( ( End_R + 1 ) - Vertical_Scrollbar.Position ) * H ;
    if( Start_R < End_R ) then // Different lines - select whole lines
    begin
        Rect.Left := 0 ;
        Rect.Right := Text_Panel.ClientWidth ;
    end else
    begin
        Rect.Left := ( Start_C - Horizontal_Scrollbar.Position ) * W ;
        Rect.Right := ( ( End_C + 1 ) - Horizontal_Scrollbar.Position ) * W ;
    end ;
rect.right := text_panel.clientwidth ;
    InvalidateRect( Text_Panel.Handle, @Rect, False ) ;
end ;


procedure TStandard_Text_Editor_Interface.Key_Down( var Key : Word ;
    Shift : TShiftState ) ;

    procedure Process_Selection( R, C : integer ) ;

    begin
        if( ssShift in Shift ) then
        begin
            if( not Selecting ) then
            begin
                Selecting := True ;
                Selection_Anchor.Row := Cursor_Y ;
                Selection_Anchor.Column := Cursor_X ;
            end ;
            Invalidate_Region( Cursor_Y, Cursor_Y + R, Cursor_X, Cursor_X + C ) ;
        end else
        begin
            if( Selecting ) then
            begin
                Selecting := False ;
                Invalidate_Region( Cursor_Y, Cursor_Y + R, Cursor_X, Cursor_X + C ) ;
            end ;
        end ;
    end ;

begin
    case Key of
        VK_UP :
            begin
                Key := 0 ;
                if( Cursor_Y > 0 ) then
                begin
                    Process_Selection( -1, 0 ) ;
                    Bring_Cursor_Into_View ;
                    Cursor_Y := Cursor_Y - 1 ;
                    if( Cursor_Y = Vertical_Scrollbar.Position - 1  ) then
                    begin
                        Vertical_Scrollbar.Position := Vertical_Scrollbar.Position - 1 ;
                        Panel.Invalidate ;
                    end ;
                    Set_Caret ;
                end ;
            end ;
        VK_DOWN :
            begin
                Key := 0 ;
                if( Cursor_Y < Line_Count - 1 ) then
                begin
                    Process_Selection( 1, 0 ) ;
                    Bring_Cursor_Into_View ;
                    Cursor_Y := Cursor_Y + 1 ;
                    if( Cursor_Y = Vertical_Scrollbar.Position + Max_R ) then
                    begin
                        Vertical_Scrollbar.Position := Vertical_Scrollbar.Position + 1 ;
                        Panel.Invalidate ;
                    end ;
                    Set_Caret ;
                end ;
            end ;
        VK_LEFT :
            begin
                Key := 0 ;
                if( Cursor_X > 0 ) then
                begin
                    Process_Selection( 0, -1 ) ;
                    Bring_Cursor_Into_View ;
                    Cursor_X := Cursor_X - 1 ;
                    if( Cursor_X = Horizontal_Scrollbar.Position - 1 ) then
                    begin
                        Horizontal_Scrollbar.Position := Horizontal_Scrollbar.Position - 1 ;
                        Panel.Invalidate ;
                    end ;
                    Set_Caret ;
                end ;
            end ;
        VK_RIGHT :
            begin
                Key := 0 ;
                if( Cursor_X < Editor.Column_Count - 1 ) then
                begin
                    Process_Selection( 0, 1 ) ;
                    Bring_Cursor_Into_View ;
                    Cursor_X := Cursor_X + 1 ;
                    if( Cursor_X = Horizontal_Scrollbar.Position + Max_C ) then
                    begin
                        Horizontal_Scrollbar.Position := Horizontal_Scrollbar.Position + 1 ;
                        Panel.Invalidate ;
                    end ;
                    Set_Caret ;
                end ;
            end ;
        VK_PRIOR :
            begin
                Key := 0 ;
                Process_Selection( 0, 0 ) ;
                Cursor_Y := Cursor_Y - Max_R ;
                if( Cursor_Y < 0 ) then
                begin
                    Cursor_Y := 0 ;
                end ;
                Bring_Cursor_Into_View ;
                Set_Caret ;
            end ;
        VK_NEXT :
            begin
                Key := 0 ;
                Process_Selection( 0, 0 ) ;
                Cursor_Y := Cursor_Y + Max_R ;
                if( Cursor_Y >= Line_Count ) then
                begin
                    Cursor_Y := Line_Count - 1 ;
                end ;
                Bring_Cursor_Into_View ;
                Set_Caret ;
            end ;
        VK_HOME :
            begin
                Key := 0 ;
                Process_Selection( 0, 0 ) ;
                Cursor_X := 0 ;
                Bring_Cursor_Into_View ;
                Set_Caret ;
            end ;
        VK_END :
            begin
                Key := 0 ;
                Process_Selection( 0, 0 ) ;
                Cursor_X := length( Get_Line( Cursor_Y ) ) ;
                Bring_Cursor_Into_View ;
                Set_Caret ;
            end ;
    end ; // case Key
end ; // TStandard_Text_Editor_Interface.Key_Down


procedure TStandard_Text_Editor_Interface.Bring_Cursor_Into_View ;

var Dummy : integer ;

begin
    if( ( Cursor_Y < Vertical_Scrollbar.Position ) or ( Cursor_Y > Vertical_Scrollbar.Position + Max_R ) ) then
    begin
        Dummy := Cursor_Y - Max_R div 2 ;
        if( Dummy < 0 ) then
        begin
            Dummy := 0 ;
        end ;
        Vertical_Scrollbar.Position := Dummy ;
        Panel.Invalidate ;
    end ;
    if( ( Cursor_X < Horizontal_Scrollbar.Position ) or ( Cursor_X > Horizontal_Scrollbar.Position + Max_C ) ) then
    begin
        Dummy := Cursor_X - Max_C div 2 ;
        if( Dummy < 0 ) then
        begin
            Dummy := 0 ;
        end ;
        Horizontal_Scrollbar.Position := Dummy ;
        Panel.Invalidate ;
    end ;
end ;


procedure TStandard_Text_Editor_Interface.CB_Focus( Sender : TObject ) ;

begin
    SetFocus ;
end ;


procedure TStandard_Text_Editor_Interface.SetFocus ;

begin
    Windows.SetFocus( Text_Panel.Handle ) ;
end ;


procedure TStandard_Text_Editor_Interface.Set_Position( P : TText_Editor_Position ) ;

begin
    if( P.Row <> -1 ) then
    begin
        Cursor_Y := P.Row ;
    end ;
    if( P.Column <> -1 ) then
    begin
        Cursor_X := P.Column ;
    end ;
    Bring_Cursor_Into_View ;
end ;


function TStandard_Text_Editor_Interface.Get_Selected_Text : string ;

begin
    Result := '' ;
    if( not Selecting ) then
    begin
        exit ;
    end ;
    Result := Editor.Get_Text_Range( Selection_Anchor.Row, Selection_Anchor.Column, Cursor_Y, Cursor_X ) ;
end ;


end.

