{
        Program Name : UOS_File_System
        Package Name : Subroutine Library Layer 2
        Purpose      : UOS File class and File Header structure
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Nov-2015
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Manages the native UOS File System.
}

unit UOS_File_System ;

interface

uses // Pascal
     Classes, // TList
     
     // C&C Subroutine Library - Layer 0...
     _Stores, // TStore_Address64
     TypeDefs, // Max_Memory
     _UE, // TUnified_Exception

     // Subroutine Library - Layer 1...
     Heaps, // TCOM_Heap64
     MStores, // TCOM_Managed_Store64

     // C&C Subroutine Library - Layer 2...
     UOS_File, // TUOS_Native_File
     _UOS_File_System, // TUOS_File_System
     UOS_String_Table ; // TUOS_String_Table

{$I EDEFINES.INC}

type TUOS_Native_File_System = class ;

     TFS_File = class( TUOS_Native_File )
                    public // Constructors and destructors...
                        constructor Create( _FS : TUOS_Native_File_System ) ;
                        destructor Destroy ; override ;

                    protected // Overrides...
                        procedure Update_Header ; override ;

                    private // Instance data...
                        FS : TUOS_Native_File_System ;
{$IFDEF CC_Track_Resources}
                        Serial_Number : integer ;
{$ENDIF}
                end ;

     TUOS_Native_File_System = class( TUOS_File_System )
                                   public // Constructors and destructors...
                                       constructor Create ;
                                       destructor Destroy ; override ;

                                   private // Instance data...
                                       _Buffer : PChar ;
                                       _Header : FS_Header ;
                                       _Store : TCOM_Managed_Store64 ;
                                       _Root : TFS_File ;
                                       _Store_Folder : TFS_File ;
                                       _String_Table : TUOS_String_Table ;
                                       Mounted : boolean ;
                                       Open_Files : TList ;
                                       Last_Code : longint ;

                                       Temp : string ;
{$IFDEF CC_Track_Resources}
                                       Serial_Number : integer ;
{$ENDIF}

                                   protected // Internal utility methods...
                                       function Buffer : PChar ;
                                       procedure Get_Header ;
                                       function Root : TFS_File ;
                                       function Store_Folder : TFS_File ;
                                       function Create_File_Instance( Parent : TUOS_Native_File ;
                                           Index : cardinal ) : TFS_File ;
                                       function Get_Store_File_Index( Index : integer ) : TFS_File ;
                                       function Find_File( const FileSpec : string ;
                                           Directory : TUOS_Native_File ;
                                           var Free_Index : longint ) : longint ;
                                       function Lookup_Directory( Spec : string ) : TFS_File ;
                                       function Parse_Filespec( Filespec : string ;
                                           var Directory : TFS_File ) : string ;
                                       procedure Destroy_File( Sender : TObject ) ;
                                       function Create_Exception( C : longint ;
                                           E : TUnified_Exception ) : tUnified_Exception ;

                                   public
                                       function Get_String( Index : cardinal ) : string ;

                                   public // Property handlers...
                                       function Get_Flags : int64 ; override ;
                                       function Is_Class( Name : PChar ) : boolean ;
                                           override ;
                                       function Get_Store : TCOM_Managed_Store64 ;
                                           override ;
                                       procedure Set_Store( Value : TCOM_Managed_Store64 ) ;
                                           override ;
                                       function Get_Heap : TCOM_Heap64 ;
                                           override ;
                                       procedure Set_Heap( Value : TCOM_Heap64 ) ;
                                           override ;
                                      function Get_File( Name : PChar ) : TUOS_File ;
                                           override ;

                                   public // API...
                                       procedure Init( FSLabel, Password : PChar ;
                                           Patterns, Clustersize, Folder_Clustersize, Flags : longint ;
                                           var AT_Offset : TStore_Address64 ;
                                           Notifier : TScan_Notify ) ;
                                           override ;
                                       function Rebuild : boolean ; override ;
                                       function Mount( _Label, Password : PChar ;
                                           Flags : longint ) : TUnified_Exception ;
                                           override ;
                                       function Dismount : TUnified_Exception ;
                                           override ;
                                       function Valid_File_System( var P : TStore_Address64 ) : boolean ;
                                           override ;

                                       function Get_File_Info( Name : Pchar ;
                                           Stream : longint ) : TUOS_File_Info ;
                                           override ;
                                       function Set_File_Info( Name : Pchar ;
                                           Stream : longint ;
                                           Info, Flags : TUOS_File_Info ) : TUnified_Exception ;
                                           override ;
                                       function Lookup( Spec : PChar ;
                                           var Context : pointer ) : PChar ;
                                           override ;
                                       procedure Clear_Context( var Context : pointer ) ;
                                           override ;
                                       function Create_File( Filespec : PChar ;
                                           Info : TUOS_File_Info ) : TUnified_Exception ;
                                           override ;
                                       function Delete_File( Filespec : PChar ) : TUnified_Exception ;
                                           override ;
                                       function Rename( Filespec, New_Name : PChar ) : TUnified_Exception ;
                                           override ;
                                       function Move_File( Old_Filespec, New_Filespec : PChar ) : TUnified_Exception ;
                                           override ;
                                       function Translate_Error : longint ;
                                           override ;
                                       function Normalize( Name : PAnsiChar ) : PAnsiChar ;
                                           override ;
                               end ; // TUOS_Native_File_System

const UOS_File_System_Facility_ID = 135 ;

const UOS_File_System_Error_Success = 0 ;
      UOS_File_System_Error_Bad_Boot_Sector = 1 ;
      UOS_File_System_Write_Error = 2 ;
      UOS_File_System_Out_Of_Space = 3 ;
      UOS_File_System_Already_Mounted = 4 ;
      UOS_File_System_Not_Mounted = 5 ;
      UOS_File_System_Invalid_File_System_Structure = 6 ;
      UOS_File_System_Read_Error = 7 ;
      UOS_File_System_Path_Not_Found = 8 ;
      UOS_File_System_Invalid_Filename = 9 ;
      UOS_File_System_File_Exists = 10 ;
      UOS_File_System_File_Creation_Error = 11 ;
      UOS_File_System_File_Not_Found = 12 ;
      UOS_File_System_Illegal_Operation = 13 ;
      UOS_File_System_Is_Dirty = 14 ;
      UOS_File_System_Source_Same_As_Destination = 15 ;
      UOS_File_System_Clustersize_Too_Small = 16 ;

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}


implementation

uses // Pascal...
     sysutils,
{$IFDEF CC_Regression_Test}
     Forms,
     stdctrls,
     controls,
     dialogs,
     commonut,
     RAMDisk,
     Dates, // Format_Sirius_Date
{$ENDIF}
{$IFDEF CC_Track_Resources}
     ResMon,
{$ENDIF}

     // C&C Subroutine Library - Layer 1...
     UE, // TStandard_Error_Interface
     Unicode, // WC_Match

     // C&C Subroutine Library - Layer 2...
     UOS_MStore, // TUOS_Managed_Store

     // UOS...
     UOS_API ; // TUOS_API

{$IFDEF CC_Track_Resources}
var Next_Serial_Number : integer = 0 ;
{$ENDIF}


// TFS_File methods...

// Constructors and destructors...

constructor TFS_File.Create( _FS : TUOS_Native_File_System ) ;

begin
    inherited Create ;

    FS := _FS ;
    if( FS <> nil ) then
    begin
        FS.Open_Files.Add( self ) ;
    end ;
end ;


destructor TFS_File.Destroy ;

var Dummy : integer ;

begin
    if( ( FS <> nil ) and ( FS.Open_Files <> nil ) ) then
    begin
        Dummy := FS.Open_Files.Indexof( self ) ;
        if( Dummy <> -1 ) then
        begin
            FS.Open_Files.Remove( self ) ;
        end ;
    end ;

    inherited Destroy ;
end ;


procedure TFS_File.Update_Header ;

var Dummy : integer ;
    F : TFS_File ;

begin
    inherited Update_Header ;

    if(
        ( FS <> nil )
        and
        ( FS.Open_Files <> nil )
        and
        ( ( Header.Flags and FAF_DIRECTORY ) <> 0 )
      ) then
    begin
        for Dummy := 0 to FS.Open_Files.Count - 1 do
        begin
            F := TFS_File( FS.Open_Files[ Dummy ] ) ;
            if( F.Directory = self ) then
            begin
                F.Directory_Cluster := Header.Clusters[ 0 ] ;
            end ;
        end ;
    end ;
end ;



function ERT( C : longint ) : string ;

begin
    case C of
        UOS_File_System_Error_Success : Result := 'Success' ;
        UOS_File_System_Error_Bad_Boot_Sector : Result := 'Bad boot sector' ;
        UOS_File_System_Write_Error : Result := 'Write error' ;
        UOS_File_System_Out_Of_Space : Result := 'Out of space' ;
        UOS_File_System_Already_Mounted : Result := 'Already mounted' ;
        UOS_File_System_Not_Mounted : Result := 'Not mounted' ;
        UOS_File_System_Invalid_File_System_Structure : Result := 'Invalid file system structure' ;
        UOS_File_System_Read_Error : Result := 'Read error' ;
        UOS_File_System_Path_Not_Found : Result := 'Path not found' ;
        UOS_File_System_Invalid_Filename : Result := 'Invalid filename' ;
        UOS_File_System_File_Exists : Result := 'File exists' ;
        UOS_File_System_File_Creation_Error : Result := 'File creation error' ;
        UOS_File_System_File_Not_Found : Result := 'File not found' ;
        UOS_File_System_Illegal_Operation : Result := 'Illegal operation' ;
        UOS_File_System_Is_Dirty : Result := 'File system is dirty' ;
        UOS_File_System_Source_Same_As_Destination : Result := 'Source and destination are the same' ;
        UOS_File_System_Clustersize_Too_Small : Result := 'Clustersize is too small' ;
        else Result := 'Unknown error' ;
    end ;
end ;


var _Lookup_Contexts : TList = nil ;

function Lookup_Contexts : TList ;

begin
    if( _Lookup_Contexts = nil ) then
    begin
        _Lookup_Contexts := TList.Create ;
    end ;
    Result := _Lookup_Contexts ;
end ;


type TLookup_Context = class
                           public // Constructors and destructors...
                               constructor Create ;
                               destructor Destroy ; override ;

                           private // Instance data...
                               _Directory : TUOS_Native_File ;
                               _Spec, _Spec_Extension : TUnicode_String ;

                           protected // Property handlers...
                               procedure Set_Directory( Value : TUOS_Native_File ) ;

                           public // Instance data...
                               Dir, Spec : string ;
                               Index : longint ;

                           public // API...
                               property Directory : TUOS_Native_File
                                   read _Directory
                                   write Set_Directory ;
                       end ; // TLookup_Context


// Constructors and destructors...

constructor TLookup_Context.Create ;

begin
    inherited Create ;

    Lookup_Contexts.Add( self ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Object, integer( self ) ) ;
{$ENDIF}
end ;


destructor TLookup_Context.Destroy ;

begin
    _Spec.Free ;
    _Spec := nil ;
    _Spec_Extension.Free ;
    _Spec_Extension := nil ;
    Lookup_Contexts.Remove( self ) ;
    Directory := nil ;
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Object, integer( self ) ) ;
{$ENDIF}

    inherited Destroy ;
end ;


// Property handlers...

procedure TLookup_Context.Set_Directory( Value : TUOS_Native_File ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Directory <> nil ) then
    begin
        _Directory.Detach ;
    end ;
    _Directory := Value ;
end ;



var _Heap : TCOM_Heap64 = nil ;

function NewGetMem( Size : Integer ) : Pointer ;

begin
    Result := pointer( _Heap.Allocate( Size ) ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Memory, integer( Result ) ) ;
{$ENDIF}
end ;


function NewFreeMem( P : Pointer ) : Integer ;

begin
    Result := 0 ;
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Memory, integer( P ) ) ;
{$ENDIF}
    _Heap.Deallocate( integer( P ), 0 ) ;
end ;


function NewReallocMem( P : Pointer ; Size : Integer ) : Pointer ;

begin
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Memory, integer( P ) ) ;
{$ENDIF}
    Result := pointer( _Heap.Reallocate( integer( P ), 0, Size ) ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Memory, integer( P ) ) ;
{$ENDIF}
end ;


{$WARNINGS OFF}
const NewMemMgr : TMemoryManager = (
                                      GetMem : NewGetMem ;
                                      FreeMem : NewFreeMem ;
                                      ReallocMem : NewReallocMem ;
                                   ) ;
{$WARNINGS ON}



type tUOS_File_Unified_Exception = class( TStandard_Error_Interface )
                                       private
                                           _Code : longint ;
                                           _Text : string ;
                                           _Previous : tUnified_Exception ;

                                       public
                                           function Get_Facility : longint ;
                                               override ;
                                           function Get_Facility_Version : longint ;
                                               override ;
                                           function Version : longint ;
                                               override ;
                                           function Severity : longint ;
                                               override ;
                                           function Error_Text( var Size, Typ : longint ) : PChar ;
                                               override ;
                                           function Get_Error : longint ;
                                               override ;
                                           function Get_Previous : tUnified_Exception ;
                                               override ;
                                           procedure Terminate ; override ;
                                   end ;


function tUOS_File_Unified_Exception.Get_Facility : longint ;

begin
    Result := UOS_File_System_Facility_ID ;
end ;


function tUOS_File_Unified_Exception.Get_Facility_Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_File_Unified_Exception.Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_File_Unified_Exception.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function tUOS_File_Unified_Exception.Error_Text( var Size, Typ : longint ) : PChar ;

begin
    _Text := ERT( _Code ) ;
    Size := length( _Text ) ;
    Typ := 0 ;
    Result := PChar( _Text ) ;
end ;


function tUOS_File_Unified_Exception.Get_Error : longint ;

begin
    Result := _Code ;
end ;


function tUOS_File_Unified_Exception.Get_Previous : tUnified_Exception ;

begin
    Result := _Previous ;
end ;


procedure tUOS_File_Unified_Exception.Terminate ;

begin
    Free ;
end ;



// TUOS_Native_File_System methods...

// Constructors and destructors...

constructor TUOS_Native_File_System.Create ;

begin
    inherited Create ;

    Open_Files := TList.Create ;
end ;


destructor TUOS_Native_File_System.Destroy ;

begin
    Dismount ;
    if( _String_Table <> nil ) then
    begin
        _String_Table.Free ;
        _String_Table := nil ;
    end ;
    if( Open_Files <> nil ) then
    begin
        Open_Files.Free ;
        Open_Files := nil ;
    end ;
    if( _Root <> nil ) then
    begin
        _Root.Free ;
        _Root := nil ;
    end ;
    if( _Store_Folder <> nil ) then
    begin
        _Store_Folder.Free ;
        _Store_Folder := nil ;
    end ;
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    freemem( _Buffer ) ;
    _Buffer := nil ;

    inherited Destroy ;
end ;


// Internal utility methods...

function TUOS_Native_File_System.Create_Exception( C : longint ;
    E : TUnified_Exception ) : tUnified_Exception ;

begin
    Result := tUOS_File_Unified_Exception.Create ;
    tUOS_File_Unified_Exception( Result )._Code := C ;
    tUOS_File_Unified_Exception( Result )._Previous := E ;
    case C of
        //UOS_File_System_Write_Error : ;
        //UOS_File_System_Out_Of_Space : ;
        //UOS_File_System_File_Exists : ;
        //UOS_File_System_File_Creation_Error : ;
        //UOS_File_System_Source_Same_As_Destination : ;
        //UOS_File_System_Clustersize_Too_Small : ;
        UOS_File_System_Read_Error : Last_Code := 13 ;
        UOS_File_System_Already_Mounted : Last_Code := 6 ;
        UOS_File_System_Not_Mounted : Last_Code := 8 ;
        UOS_File_System_Path_Not_Found : Last_Code := 11 ;
        UOS_File_System_File_Not_Found : Last_Code := 10 ;
        UOS_File_System_Invalid_Filename : Last_Code := 12 ;

        // Any other errors have no UOS equivalent, so return generic error
        else Last_Code := 7 ; // Invalid operation
    end ;
end ;


function TUOS_Native_File_System.Buffer : PChar ;

begin
    if( _Buffer = nil ) then
    begin
        _Buffer := allocmem( Store.Min_Storage ) ;
{$IFDEF CC_Track_Resources}
        Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    end ;
    Result := _Buffer ;
end ;


procedure TUOS_Native_File_System.Get_Header ;

var P : TStore_Address64 ;

begin
    if( _Header.ID = 0 ) then
    begin
        Store.Get_Root( Buffer^ ) ;
        move( Buffer[ 16 ], P, sizeof( P ) ) ;
        Store.Read( P, Store.Min_Storage, Buffer^ ) ;
        move( Buffer[ 0 ], _Header, sizeof( _Header ) ) ;
{$IFDEF CC_Track_Resources}
        Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        reallocmem( _Buffer, _Header.Folder_Clustersize ) ;
{$IFDEF CC_Track_Resources}
        Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    end ; // if( _Header.ID = 0 )
end ;


function TUOS_Native_File_System.Root : TFS_File ;

begin
    if( _Root = nil ) then
    begin
        _Root := TFS_File.Create( self ) ;
        _Root.Store := _Store ;
        Get_Header ;
        _Store.Read( _Header.Root, sizeof( _Root.Header ), Buffer^ ) ;
        move( Buffer^, _Root.Header, sizeof( _Root.Header ) ) ;
        _Root.Directory_Cluster := _Header.Root ;
        _Root.Attach ;
    end ; // if( _Root = nil )
    Result := _Root ;
end ;


function TUOS_Native_File_System.Store_Folder : TFS_File ;

begin
    if( _Store_Folder = nil ) then
    begin
        _Store_Folder := Create_File_Instance( Root, 0 ) ;
        _Store_Folder.Attach ;
        Root.Read( 0, 0, sizeof( _Store_Folder.Header ), Buffer^ ) ;
        move( Buffer^, _Store_Folder.Header, sizeof( _Store_Folder.Header ) ) ;
    end ;
    Result := _Store_Folder ;
end ;


function TUOS_Native_File_System.Create_File_Instance( Parent : TUOS_Native_File ;
    Index : cardinal ) : TFS_File ;

var Dummy : integer ;

begin
    for Dummy := 0 to Open_Files.Count - 1 do
    begin
        Result := TFS_File( Open_Files[ Dummy ] ) ;
        if(
            ( Result.Directory_Cluster = Parent.Header.Clusters[ 0 ] )
            and
            ( Result.Directory_Index = Index )
          ) then
        begin
            Result.Attach ; // EAC, 3-Oct-2016
            exit ;
        end ;
    end ;
    Result := TFS_File.Create( self ) ;
    Result.Store := _Store ;
    Result.Directory_Index := Index ;
    Result.Directory_Cluster := Parent.Header.Clusters[ 0 ] ;
    Result.Directory := Parent ;
    Result.Attach ;
end ;


function TUOS_Native_File_System.Get_Store_File_Index( Index : integer ) : TFS_File ;

begin
    Result := Create_File_Instance( Store_Folder, Index ) ;
    Store_Folder.Read( 0, Index * sizeof( Result.Header ), sizeof( Result.Header ), Buffer^ ) ;
    move( Buffer^, Result.Header, sizeof( Result.Header ) ) ;
end ;


function TUOS_Native_File_System.Get_String( Index : cardinal ) : string ;

begin
    // Setup...
    Result := '' ;
    if( ( Index = 0 ) or odd( Index ) ) then
    begin
        exit ;
    end ;

    if( _String_Table = nil ) then
    begin
        _String_Table := TUOS_String_Table.Create ;
        _String_Table.Set_AT_File( Get_Store_File_Index( 0 ) ) ;
        _String_Table.Set_Index_File( Get_Store_File_Index( 1 ) ) ;
        _String_Table.Set_Data_File( Get_Store_File_Index( 2 ) ) ;
    end ;

    Result := _String_Table.Get_String( Index ) ;
end ;


{ Lookup file in directory, returning index in directory (or -1 if not found).
  Free_Index will be set to the first free header in the index (or -1 if none
  before file). }
function TUOS_Native_File_System.Find_File( const FileSpec : string ;
    Directory : TUOS_Native_File ; var Free_Index : longint ) : longint ;

var Header : TUOS_File_Header ;
    S : string ;

begin
    Free_Index := -1 ;
    Result := 0 ;
    while( true ) do
    begin
        if( Directory.Read( 0, ( Result ) * sizeof( Header ), sizeof( Header ),
          Header ) = 0 ) then
        begin
            // No more entries in this directory...
            Set_Last_Error( Create_Exception( UOS_File_System_File_Not_Found,
                nil ) ) ;
            Result := -1 ;
            exit ;
        end ;
        if( ( Header.Name <> 0 ) and ( ( Header.Flags and FAF_DELETED ) = 0 ) ) then
        begin // Not a null entry or deleted file
            S := Get_String( Header.Name ) ;
            if( WC_Match( FileSpec, S, 3 ) ) then
            begin
                exit ;
            end ;
        end else
        if( ( Header.Name = 0 ) and ( Free_Index < 0 ) ) then
        begin
            Free_Index := Result ;
        end ; // if( H.Name <> 0 )
        inc( Result ) ;
    end ;  // while( true )
    Result := -1 ;
end ;


function TUOS_Native_File_System.Lookup_Directory( Spec : string ) : TFS_File ;

var D : string ;
    Dummy : integer ;
    F : TFS_File ;
    Found : boolean ;
    H : TUOS_File_Header ;
    P : cardinal ;

begin
    Result := nil ;

    // Get root directory...
    if( copy( Spec, 1, 1 ) = '\' ) then
    begin
        Spec := copy( Spec, 2, length( Spec ) ) ;
    end ;
    if( copy( Spec, length( Spec ), 1 ) <> '\' ) then
    begin
        Spec := Spec + '\' ;
    end ;
    if( Spec = '\' ) then // Asked for root
    begin
        Result := Root ;
        exit ;
    end ;
    F := Root ;
    F.Attach ;

    // Find specified directory...
    while( Spec <> '' ) do
    begin
        // Parse next directory in the path...
        Dummy := pos( '\', Spec + '\' ) ;
        D := copy( Spec, 1, Dummy - 1 ) ;
        Spec := copy( Spec, Dummy + 1, length( Spec ) ) ;

        // Look through all file headers in the directory...
        P := 0 ;
        Found := False ;
        while( ( P < F.Header.EOF ) and ( F.Read( 0, P, sizeof( H ), H ) > 0 ) ) do
        begin
            if( ( H.Name <> 0 ) and ( ( H.Flags and FAF_DELETED ) = 0 ) ) then
            begin // Not a null entry or deleted file
                if( WC_Match( D, Get_String( H.Name ), 3 ) ) then
                begin
                    F := Create_File_Instance( F, P div sizeof( H ) ) ;
                    //F.Attach ; // Removed, 3-Oct-2016
                    F.Header := H ;
                    if( Spec = '' ) then
                    begin
                        dec( F.Reference_Count ) ;
                        Result := F ;
                        exit ;
                    end ;
                    if( ( H.Flags and FAF_DIRECTORY ) = 0 ) then
                    begin // Path contained a non-folder
                        F.Detach ;
                        exit ;
                    end ;
//                    P := 0 ;
                    Found := True ;
                    break ;
                end ; // if
            end ; // if
            P := P + sizeof( H ) ;
        end ; // while
        if( not Found ) then // We didn't find a match in the directory
        begin
            F.Detach ;
            exit ;
        end ;
    end ; // while( Spec <> '' )
    dec( F.Reference_Count ) ;
    Result := F ;
end ; // TUOS_Native_File_System.Lookup_Directory


function TUOS_Native_File_System.Parse_Filespec( Filespec : string ;
    var Directory : TFS_File ) : string ;

var Dummy : integer ;

begin
    Directory := nil ;
    Result := '' ;
    if( copy( Filespec, 1, 1 ) = '\' ) then
    begin
        Filespec := copy( Filespec, 2, length( Filespec ) ) ;
    end ;
    if( Filespec = '' ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Invalid_Filename, nil ) ) ;
        exit ;
    end ;
    Dummy := length( Filespec ) ;
    while( ( Dummy > 0 ) and ( Filespec[ Dummy ] <> '\' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    Directory := Lookup_Directory( copy( Filespec, 1, Dummy ) ) ;
    Result := copy( Filespec, Dummy + 1, length( Filespec ) + 1 ) ;
    if( Directory = nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Path_Not_Found, nil ) ) ;
    end ;
end ; // TUOS_Native_File_System.Parse_Filespec


procedure TUOS_Native_File_System.Destroy_File( Sender : TObject ) ;

var F : TFS_File ;

begin
    F := TFS_File( Sender ) ;
    if( F.Directory <> nil ) then
    begin
        F.Directory.Detach ;
    end ;
    if( F = _Root ) then
    begin
        _Root := nil ;
    end else
    if( F = _Store_Folder ) then
    begin
        _Store_Folder := nil ;
    end ;
end ;


function TUOS_Native_File_System.Valid_File_System( var P : TStore_Address64 ) : boolean ;

var UEC : TUnified_Exception ;

begin
    // Validate that the file system structure is valid...
    Result := False ;

    Store.Get_Root( Buffer^ ) ;
    move( Buffer[ 16 ], P, 8 ) ; // Get store header address
    if( odd( P ) or ( P <= Store.Min_Storage ) ) then
    begin
        exit ;
    end ;
    Store.Read_Data( Buffer[ 0 ], P, sizeof( _Header ), UEC ) ;
    if( UEC <> nil ) then
    begin
        Set_Last_Error( UEC ) ;
        exit ;
    end ;
    move( Buffer[ 0 ], _Header, sizeof( _Header ) ) ;
    if( ( _Header.ID <> -1 ) or ( _Header.ID1 <> 135 ) or ( _Header.Version > 10 ) ) then
    begin
        exit ;
    end ;

    Result := True ;
end ;


// API...

function TUOS_Native_File_System.Is_Class( Name : PChar ) : boolean ;

begin
    Result := lowercase( Name ) = 'tuos_native_file_system' ;
    if( not Result ) then
    begin
        Result := inherited Is_Class( Name ) ;
    end ;
end ;


function TUOS_Native_File_System.Get_Flags : int64 ;

var P : TStore_Address64 ;

begin
    Result := 0 ;
    if( _Store = nil ) then
    begin
        exit ;
    end ;
    if( not Valid_File_System( P ) ) then
    begin
        exit ;
    end ;
    Result := _Header.Flags ;
end ;


function TUOS_Native_File_System.Get_Store ;

begin
    Result := _Store ;
end ;


procedure TUOS_Native_File_System.Set_Store( Value : TCOM_Managed_Store64 ) ;

begin
    Set_Last_Error( nil ) ;
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Store <> nil ) then
    begin
        _Store.Detach ;
    end ;
    _Store := Value ;
end ;


function TUOS_Native_File_System.Get_Heap : TCOM_Heap64 ;

begin
    Result := _Heap ;
end ;


procedure TUOS_Native_File_System.Set_Heap( Value : TCOM_Heap64 ) ;

begin
    Set_Last_Error( nil ) ;
    if( Value <> nil ) then
    begin
        Value.Attach ;
{$WARNINGS OFF}
  {$IFNDEF No_DLL}
        SetMemoryManager( NewMemMgr ) ;
  {$ENDIF}
{$WARNINGS ON}
    end ;
    if( _Heap <> nil ) then
    begin
        _Heap.Detach ;
    end ;
    _Heap := Value ;
end ;


const Data_Stream_Index = 0 ;

procedure TUOS_Native_File_System.Init( FSLabel, Password : PChar ;
    Patterns, Clustersize, Folder_Clustersize, Flags : longint ;
    var AT_Offset : TStore_Address64 ; Notifier : TScan_Notify ) ;

var BadBlocks : Pint64_Array ;
    BadBlock_Count : cardinal ;

    procedure Add_Bad_Block( P : TStore_Address64 ) ;

    begin
{$IFDEF CC_Track_Resources}
        Remove_Resource( RT_Memory, integer( BadBlocks ) ) ;
{$ENDIF}
        reallocmem( Badblocks, ( Badblock_Count + 1 ) * 8 ) ;
{$IFDEF CC_Track_Resources}
        Add_Resource( RT_Memory, integer( BadBlocks ) ) ;
{$ENDIF}
        BadBlocks[ Badblock_Count ] := P ;
        inc( Badblock_Count ) ;
    end ;

var Abort : boolean ;
    BootRec : FS_Boot_Record ;
    Buffer, Buffer1 : pByte_Array ;
    Buff_Size : longint ;
    C : cardinal ;
    Cluster : TStore_Address64 ;
    CN, Current : int64 ;
    FAT : TUOS_Native_File ;
    FBadBlocks : TUOS_Native_File ;
    FHeader : TUOS_File_Header ;
    FIndex : TUOS_Native_File ;
    FRoot : TUOS_Native_File ;
    FStrings : TUOS_Native_File ;
    FStore : TUOS_Native_File ;
    Header : FS_Header ;
    I : integer ;
    Max : int64 ;
    P : TStore_Address64 ;
    Percent : cardinal ;
    Position : longint ;
    Second_Attempt : boolean ;
    ST : string ;
    UEC : TUnified_Exception ;
    V : byte ;

begin // TUOS_Native_File_System.Init
    // Setup...
    if( _Store = nil ) then
    begin
        exit ;
    end ;
    if( _Store.Min_Storage < sizeof( FAT.Header ) ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Clustersize_Too_Small, nil ) ) ;
        exit ; // Cannot initialize a file system on a store with a min clustersize less than 256
    end ;
    if( Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Already_Mounted, nil ) ) ;
        exit ; // Cannot initialize a mounted file system
    end ;
    Percent := 0 ;

    // Default the values...
    if( Clustersize < _Store.Min_Storage ) then
    begin
        Clustersize := _Store.Min_Storage ;
    end ;
    if( Folder_Clustersize < Clustersize ) then
    begin
        Folder_Clustersize := Clustersize ;
    end ;
    Clustersize := Clustersize and ( not ( _Store.Min_Storage - 1 ) ) ;
    Folder_Clustersize := Folder_Clustersize and ( not ( _Store.Min_Storage - 1 ) ) ;
    _Store.Set_Min_Storage( Clustersize ) ;
    C := _Store.Max_Storage div Clustersize ; // Total clusters
    if( AT_Offset = 0 ) then // Default to center of store
    begin
        AT_Offset :=  _Store.Max_Storage div 2 ;
    end ;
    AT_Offset := ( AT_Offset + Clustersize - 1 ) and ( not ( Clustersize - 1 ) ) ;
    if( AT_Offset < Clustersize * 2 ) then
    begin
        AT_Offset := Clustersize * 2 ;
    end ;

    // Allocate buffers...
    Buff_Size := Folder_Clustersize ;
    Buffer := Allocmem( Buff_Size ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Memory, integer( Buffer ) ) ;
{$ENDIF}
    Buffer1 := allocmem( Buff_Size ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Memory, integer( Buffer1 ) ) ;
{$ENDIF}

    try
        // Run patterns...
        BadBlocks := nil ;
        BadBlock_Count := 0 ;
        Max := C * Patterns div 100 ; // Number of operations that comprise 1% of the scan
        Current := Max ;
        if( Patterns > 0 ) then
        begin
            _Store.Format ;
        end ;
        while( Patterns > 0 ) do
        begin
            dec( Patterns ) ;
            case( Patterns ) of
                0 : V := 0 ;
                1 : V := 255 ;
                2 : V := 170 ; // 10101010
                3 : V := 85 ; // 01010101
                4 : V := 205 ; // 11001100
                5 : V := 51 ; // 00110011
                else V := Patterns ;
            end ;
            fillchar( Buffer^, Clustersize, V ) ;
            _Store.Set_Root( Buffer^ ) ;
            if( _Store.Last_Error <> nil ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Error_Bad_Boot_Sector, _Store.Last_Error ) ) ;
                exit ;
            end ;
            _Store.Get_Root( Buffer1^ ) ;
            if( _Store.Last_Error <> nil ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Error_Bad_Boot_Sector, _Store.Last_Error ) ) ;
                exit ;
            end ;
            for I := 0 to Clustersize - 1 do
            begin
                if( Buffer^[ I ] <> Buffer1^[ I ] ) then
                begin
                    Set_Last_Error( Create_Exception( UOS_File_System_Error_Bad_Boot_Sector, nil ) ) ;
                    exit ;
                end ;
            end ;
            P := Clustersize ;
            while( P < _Store.Max_Storage ) do
            begin
                dec( Current ) ;
                if( Current <= 0 ) then
                begin
                    Current := Max ;
                    inc( Percent ) ;
                    if( assigned( Notifier ) ) then
                    begin
                        Abort := False ;
                        Notifier( self, Percent, BadBlock_Count, Abort ) ;
                        if( Abort ) then
                        begin
                            exit ;
                        end ;
                    end ;
                end ;
                _Store.Write_Data( Buffer^, P, Clustersize, UEC ) ;
                if( UEC <> nil ) then
                begin
                    Add_Bad_Block( P ) ;
                    continue ;
                end ;
                _Store.Read_Data( Buffer1^, P, Clustersize, UEC ) ;
                if( UEC <> nil ) then
                begin
                    Add_Bad_Block( P ) ;
                    continue ;
                end ;
                for I := 0 to Clustersize - 1 do
                begin
                    if( Buffer[ I ] <> Buffer1[ I ] ) then
                    begin
                        Add_Bad_Block( P ) ;
                        continue ;
                    end ;
                end ;
                P := P + Clustersize ;
            end ; // while( P < _Store.Max_Storage )
        end ; // while( Patterns > 0 )

        // Allocate bad blocks and allocation table...
        for I := 0 to Badblock_Count - 1 do // Mark bad blocks as allocated
        begin
            Store.Allocate_At( Badblocks^[ I ], Clustersize ) ;
        end ;
        Second_Attempt := False ;
        while( not Store.Allocate_At( AT_Offset, C div 8 ) ) do
        begin
            AT_Offset := At_Offset + Clustersize ; // Try next cluster
            if( AT_Offset >= Store.Max_Storage ) then
            begin
                if( Second_Attempt ) then
                begin
                    Set_Last_Error( Create_Exception( UOS_File_System_Out_Of_Space, UEC ) ) ;
                    exit ;
                end ;
                AT_Offset := Clustersize ; // Start over at beginning of store
                Second_Attempt := True ;
            end ;
        end ; // while

        // Write root folder header...
        P := Store.Allocate( Clustersize ) ;
        FRoot := TUOS_Native_File.Create ;
        try
            FRoot.Store := Store ;
            FRoot.Header.Clustersize := Folder_Clustersize ;
            FRoot.Header.Creation := Get_Timestamp ;
            FRoot.Header.Flags := FAF_PERMANENT or FAF_SYSTEM or FAF_DIRECTORY ;
            FRoot.Header.Last_Access := FRoot.Header.Creation ;
            FRoot.Header.Last_Modified := FRoot.Header.Creation ;

            // Create String table...
            FStrings := TUOS_Native_File.Create ;
            try
                FStrings.Store := Store ;
                FStrings.Header.Creation := Get_Timestamp ;
                FStrings.Header.Flags := FAF_PERMANENT or FAF_SYSTEM ;
                FStrings.Header.Last_Access := FStrings.Header.Creation ;
                FStrings.Header.Last_Modified := FStrings.Header.Creation ;
                FStrings.Header.Name := 48 ;
                ST := // Note: Each string entry is a multiple of 8 bytes
                    #0#0#0#0 + #0#0#0#0 + // Initial zero entry (0)
                    #1#0#0#0 + #5'Sto' + 're'#0#0 + #0#0#0#0 + // "Store" (8)
                    #1#0#0#0 + #13'Bad' + 'Bloc' + 'ks.s' + 'ys'#0#0 + #0#0#0#0 + // "BadBlocks.sys" (24)
                    #1#0#0#0 + #11'Str' + 'ings' + '.sys' + // "Strings.sys" (48)
                    #1#0#0#0 + #9'Ind' + 'ex.s' + 'ys'#0#0 + // "Index.sys" (64)
                    #1#0#0#0 + #6'AT.' + 'sys'#0 + #0#0#0#0 ; // "AT.sys" (80)
                FStrings.Write( 0, 0, length( ST ), PChar( ST )[ 0 ] ) ; // 96 bytes total

                // String index...
                FIndex := TUOS_Native_File.Create ;
                try
                    FIndex.Store := Store ;
                    FIndex.Header.Creation := Get_Timestamp ;
                    FIndex.Header.Flags := FAF_PERMANENT or FAF_SYSTEM ;
                    FIndex.Header.Last_Access := FIndex.Header.Creation ;
                    FIndex.Header.Last_Modified := FIndex.Header.Creation ;
                    FIndex.Header.Name := 64 ;
                    ST := #80#0#0#0 + #24#0#0#0 + #64#0#0#0 + #8#0#0#0 + #48#0#0#0 ;
                    FIndex.Write( 0, 0, length( ST ), PChar( ST )[ 0 ] ) ;

                    // String table allocation table...
                    FAT := TUOS_Native_File.Create ;
                    try
                        FAT.Store := Store ;
                        FAT.Header.Creation := Get_Timestamp ;
                        FAT.Header.Flags := FAF_PERMANENT or FAF_SYSTEM ;
                        FAT.Header.Last_Access := FAT.Header.Creation ;
                        FAT.Header.Last_Modified := FAT.Header.Creation ;
                        FAT.Header.Name := 80 ;
                        C := $FFF ; // First 12 bits of AT = 96 bytes
                        FAT.Write( 0, 0, 2, C ) ;

                        // BadBlocks...
                        FBadBlocks := TUOS_Native_File.Create ;
                        try
                            FBadBlocks.Store := Store ;
                            FBadBlocks.Header.Creation := Get_Timestamp ;
                            FBadBlocks.Header.Flags := FAF_PLACED or FAF_PERMANENT or FAF_SYSTEM ;
                            FBadBlocks.Header.Last_Access := FBadBlocks.Header.Creation ;
                            FBadBlocks.Header.Last_Modified := FBadBlocks.Header.Creation ;
                            FBadBlocks.Header.Name := 24 ;
                            FBadBlocks.Header.Size := Badblock_Count * Store.Min_Storage ;
                            for I := 0 to BadBlock_Count - 1 do
                            begin
                                move( BadBlocks[ I ], Cluster, sizeof( Cluster ) ) ;
                                FBadBlocks.Set_Allocation_Cluster( 0, I, Cluster ) ;
                            end ;

                            // Store folder...
                            FStore := TUOS_Native_File.Create ;
                            try
                                FStore.Store := Store ;
                                FStore.Header.Creation := Get_Timestamp ;
                                FStore.Header.Clustersize := Folder_Clustersize ;
                                FStore.Header.Flags := FAF_PERMANENT or FAF_SYSTEM or FAF_DIRECTORY ;
                                FStore.Header.Last_Access := FStore.Header.Creation ;
                                FStore.Header.Last_Modified := FStore.Header.Creation ;
                                FStore.Header.Name := 8 ;

                                // Write files to store folder...
                                Position := 0 ;
                                fillchar( Buffer^, Buff_Size, 0 ) ;

                                move( FAT.Header, Buffer[ 0 ], sizeof( FAT.Header ) ) ;
                                FStore.Write( Data_Stream_Index, Position, sizeof( FAT.Header ), Buffer^ ) ;
                                Position := Position + sizeof( FAT.Header ) ;

                                move( Findex.Header, Buffer[ 0 ], sizeof( Findex.Header ) ) ;
                                FStore.Write( Data_Stream_Index, Position, sizeof( Findex.Header ), Buffer^ ) ;
                                Position := Position + sizeof( FAT.Header ) ;

                                move( FStrings.Header, Buffer[ 0 ], sizeof( FStrings.Header ) ) ;
                                FStore.Write( Data_Stream_Index, Position, sizeof( FStrings.Header ), Buffer^ ) ;
                                Position := Position + sizeof( FAT.Header ) ;

                                move( FBadBlocks.Header, Buffer[ 0 ], sizeof( FBadBlocks.Header ) ) ;
                                FStore.Write( Data_Stream_Index, Position, sizeof( FBadBlocks.Header ), Buffer^ ) ;
                                Position := Position + sizeof( FAT.Header ) ;
                            finally
                                FStore.Detach ;
                            end ;
                        finally
                            FBadBlocks.Detach ;
                        end ;
                    finally
                        FAT.Detach ;
                    end ;
                finally
                    FIndex.Detach ;
                end ;
            finally
                FStrings.Detach ;
            end ;

            // Update store folder header to root folder...
            move( FStore.Header, Buffer[ 0 ], sizeof( FStore.Header ) ) ;
            FRoot.Write( 0, 0, Clustersize, Buffer^ ) ;

            // Write root directory file header...
            move( FRoot.Header, Buffer^, sizeof( FRoot.Header ) ) ;
            _Store.Write_Data( Buffer^, P, sizeof( FRoot.Header ), UEC ) ;
            if( UEC <> nil ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Write_Error, UEC ) ) ;
                exit ;
            end ;
        finally
            FRoot.Detach ;
        end ;


        // Build store header...
        fillchar( Header, sizeof( Header ), 0 ) ;
        Header.ID := -1 ;
        Header.ID1 := 135 ;
        Header.Version := 10 ; // V1.0
        Header.AT_Offset := AT_Offset ;
        CN := _Store.Max_Storage div _Store.Min_Storage ;
        Header.AT_Size := CN div 8 ;
        Header.Flags := Flags ;
        Header.Clustersize := Clustersize ;
        Header.Folder_Clustersize := Folder_Clustersize ;
        Header.Root := P ;
        Header.Volume_Label := FSLabel ;
        Header.Password := Password ;

        // Write store header...
        move( Header, Buffer^, sizeof( Header ) ) ;
        P := Store.Allocate( Clustersize ) ;
        _Store.Write_Data( Buffer^, P, Clustersize, UEC ) ;
        if( UEC <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Write_Error, UEC ) ) ;
            exit ;
        end ;

        Store.Set_Allocation_Table_Offset( AT_Offset ) ;
        Store.Flush_Allocation_Table ;

        // Write Boot cluster...
        fillchar( Buffer^, Clustersize, 0 ) ;
        BootRec.Header := P ; // File system header
        move( BootRec, Buffer^, sizeof( BootRec ) ) ;
        _Store.Set_Root( Buffer^ ) ;
        UEC := _Store.Last_Error ;
        if( UEC <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Error_Bad_Boot_Sector, UEC ) ) ;
            exit ;
        end ;
    finally
        freemem( BadBlocks ) ;
        freemem( Buffer ) ;
{$IFDEF CC_Track_Resources}
        Remove_Resource( RT_Memory, integer( Buffer ) ) ;
{$ENDIF}
        freemem( Buffer1 ) ;
{$IFDEF CC_Track_Resources}
        Remove_Resource( RT_Memory, integer( Buffer1 ) ) ;
{$ENDIF}
    end ;
end ; // TUOS_Native_File_System.Init


function TUOS_Native_File_System.Rebuild : boolean ;

    function Valid_Pointer( P : int64 ) : boolean ;

    begin
        if( P = 0 ) then // Nil is always valid
        begin
            Result := True ;
            exit ;
        end ;
        Result := False ;
        if( P < Store.Min_Storage ) then // Too small
        begin
            exit ;
        end ;
        if( P > Store.Max_Storage ) then // Too big
        begin
            exit ;
        end ;
        if( ( P mod Store.Min_Storage ) <> 0 ) then // Not an exact multiple of store clustersize
        begin
            exit ;
        end ;
        Result := True ;
    end ;


    procedure Process( Folder : TUOS_Native_File ) ;

    var F : TUOS_Native_File ;
        Header : TUOS_File_Header ;
        Index, Loop : integer ;
        Need_Update : boolean ;

    begin
        Index := 0 ;
        while( Folder.Read( 0, Index * sizeof( Header ), sizeof( Header ), Header ) > 0 ) do
        begin
            if( Header.Name <> 0 ) then // A valid file header
            begin
                // Check for, and correct, obviously bad data...
                Need_Update := False ;
                if(
                    ( Header.Clustersize < Store.Min_Storage ) // Too small
                    or
                    ( Header.Clustersize >= Store.Max_Storage ) // Too big
                    or
                    ( ( Header.Clustersize mod Store.Min_Storage ) <> 0 ) // Not an exact multiple of store clustersize
                  ) then
                begin
                    Need_Update := True ;
                    Header.Clustersize := Store.Min_Storage ;
                end ;
                if( not Valid_Pointer( Header.ACL ) ) then
                begin
                    Header.ACL := 0 ;
                    Need_Update := True ;
                end ;
                if( not Valid_Pointer( Header.Extension ) ) then
                begin
                    Header.Extension := 0 ;
                    Need_Update := True ;
                end ;
                if( not Valid_Pointer( Header.Data_Stream ) ) then
                begin
                    Header.Data_Stream := 0 ;
                    Need_Update := True ;
                end ;
                for Loop := 0 to Max_Header_Cluster do
                begin
                    if( not Valid_Pointer( Header.Clusters[ Loop ] ) ) then
                    begin
                        Header.Clusters[ Loop ] := 0 ;
                        Need_Update := True ;
                        if( Header.Size > Loop * Header.Clustersize ) then
                        begin
                            Header.Size := Loop * Header.Clustersize ;
                        end ;
                    end ;
                end ;
                for Loop := 0 to Max_Header_Data_Stream do
                begin
                    if( not Valid_Pointer( Header.Streams[ Loop ].Pointer ) ) then
                    begin
                        Header.Streams[ Loop ].Pointer := 0 ;
                        Header.Streams[ Loop ].Name := 0 ;
                        Need_Update := True ;
                    end ;
                end ; // for Loop := 0 to Max_Header_Data_Stream
                if( Header.EOF > Header.Size ) then
                begin
                    Header.EOF := Header.Size ;
                    Need_Update := True ;
                end ;
                if( Need_Update ) then
                begin
                    Folder.Write( 0, Index * sizeof( Header ), sizeof( Header ), Header ) ;
                end ;

                // Allocate clusters for this file...
                F := TFS_File.Create( self ) ;
                try
                    F.Store := Store ;
                    F.Header := Header ;
                    if( F.Rebuild ) then
                    begin
                        Result := True ;
                    end ;
                    if( ( Header.Flags and FAF_DIRECTORY ) <> 0 ) then // This file is a directory
                    begin
                        Process( F ) ;
                    end ;
                finally
                    F.Detach ;
                end ;
            end ; // if( Header.Name <> 0 )
            inc( Index ) ;
        end ; // while
    end ; // TUOS_Native_File_System.Rebuild.Process


var Buffer : PChar ;
    FSHeader : FS_Header ;
    P : TStore_Address64 ;
    FRoot : TUOS_Native_File ;
    UEC : TUnified_Exception ;
    
begin // TUOS_Native_File_System.Rebuild
    Result := False ;
    if( _Store = nil ) then
    begin
        exit ;
    end ;
    if( Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Already_Mounted, nil ) ) ;
        exit ; // Cannot rebuild a mounted file system
    end ;

    Set_Last_Error( nil ) ;
    if( not Valid_File_System( P ) ) then
    begin
        exit ;
    end ;
    Buffer := allocmem( Store.Min_Storage ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Memory, integer( Buffer ) ) ;
{$ENDIF}
    try
        // Ensure the store is marked as dirty...
        Store.Read_Data( Buffer[ 0 ], P, sizeof( FSHeader ), UEC ) ;
        move( Buffer[ 0 ], FSHeader, sizeof( FSHeader ) ) ;
        FSHeader.Flags := FSHeader.Flags or UOSFSF_Dirty ;
        move( FSHeader, Buffer[ 0 ], sizeof( FSHeader ) ) ;
        Store.Write_Data( Buffer[ 0 ], P, sizeof( FSHeader ), UEC ) ;
        if( UEC <> nil ) then
        begin
            exit ;
        end ;

        // Get root directory...
        FRoot := TUOS_Native_File.Create ;
        try
            Store.Read_Data( Buffer[ 0 ], FSHeader.Root, sizeof( FRoot.Header ), UEC ) ;
            if( UEC <> nil ) then
            begin
                exit ;
            end ;
            move( Buffer^, FRoot.Header, sizeof( FRoot.Header ) ) ;
            FRoot.Store := Store ;

            // Clear the allocation table...
            Store.Set_Min_Storage( FSHeader.Clustersize ) ;
            Store.Deallocate( FSHeader.Clustersize, Store.Max_Storage ) ;

            // Reallocate allocation rable, store header, and root directory header...
    //        C := _Store.Max_Storage div FSHeader.Clustersize ; // Total clusters
            Store.Allocate_At( FSHeader.AT_Offset, FSHeader.AT_Size ) ; // AT
            Store.Allocate_At( FSHeader.Root, FSHeader.Clustersize ) ; // Root folder
            Store.Allocate_At( P, FSHeader.Clustersize ) ; // FS Header

            // Handle allocation chain of the root folder...
            Result := FRoot.Rebuild ;

            // Process the file structure...
            Process( FRoot ) ;

        finally
            // Cleanup...
            FRoot.Detach ;
        end ;

        // Persist the updated allocation table...
        Store.Set_Allocation_Table_Offset( FSHeader.AT_Offset ) ;
        Store.Flush_Allocation_Table ;

        // Mark store as clean...
        FSHeader.Flags := FSHeader.Flags and ( not UOSFSF_Dirty ) ;
        move( FSHeader, Buffer[ 0 ], sizeof( FSHeader ) ) ;
        Store.Write_Data( Buffer[ 0 ], P, sizeof( FSHeader ), UEC ) ;
        if( UEC <> nil ) then
        begin
            exit ;
        end ;
    finally
        freemem( Buffer ) ;
{$IFDEF CC_Track_Resources}
        Add_Resource( RT_Memory, integer( Buffer ) ) ;
{$ENDIF}
    end ;
end ; // TUOS_Native_File_System.Rebuild


function Matched_Strings( S : TUnicode_String ; const N : string ) : boolean ;

var _N : TUnicode_String ;

begin
    _N := TUnicode_String.Create ;
    try
        _N.Assign_From_String( N, 3 ) ;
        _N.Lowercase ;
        Result := S.Equal( _N ) ;
    finally
        _N.Free ;
    end ;
end ;


function TUOS_Native_File_System.Get_File_Info( Name : Pchar ;
    Stream : longint ) : TUOS_File_Info ;

var Directory : TFS_File ;
    F : TFS_File ;
    H : TUOS_File_Header ;
    Index : integer ;
    S : string ;
    _S : TUnicode_String ;

begin
    // Setup...
    fillchar( Result, sizeof( Result ), 0 ) ;
    Set_Last_Error( nil ) ;

    // Parse specification...
    S := Name ;
    if( S = '' ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Invalid_Filename, nil ) ) ;
        exit ; // No context and no file spec
    end ;
    S := Parse_Filespec( S, Directory ) ;
    if( Directory = nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Path_Not_Found, nil ) ) ;
        exit ; // Directory not found
    end ;

    _S := TUnicode_String.Create ;
    try
        _S.Assign_From_String( S, 3 ) ;
        _S.Lowercase ;
        Index := 0 ;
        while( true ) do
        begin
            inc( Index ) ;
            if( Directory.Read( 0, ( Index - 1 ) * sizeof( H ), sizeof( H ), H ) = 0 ) then
            begin
                // No more entries in this directory...
                Set_Last_Error( Create_Exception( UOS_File_System_File_Not_Found, nil ) ) ;
                exit ;
            end ;
            if( ( H.Name <> 0 ) and ( ( H.Flags and FAF_DELETED ) = 0 ) ) then
            begin // Not a null entry or deleted file
                if( Matched_Strings( _S, Get_String( H.Name ) ) ) then
                begin
                    Result.Size := H.Size ;
                    Result.EOF := H.EOF ;
                    Result.Uncompressed_Size := H.Uncompressed_Size ;
                    Result.Clustersize := H.Clustersize ;
                    Result.Record_Size := H.Record_Size ;
                    Result.Creation := H.Creation ;
                    Result.Last_Modified := H.Last_Modified ;
                    Result.Last_Backup := H.Last_Backup ;
                    Result.Last_Access := H.Last_Access ;
                    Result.Expiration := H.Expiration ;
                    Result.Creator := H.Creator ;
                    Result.Owner := H.Owner ;
                    Result.Flags := H.Flags ;
                    Result.Version_Limit := H.Version_Limit ;
                    Result.Location := H.Clusters[ 0 ] ; // Note: Location is meaningless if the file is not placed and/or contiguous
                    if( Stream > 0 ) then
                    begin
                        F := Create_File_Instance( Directory, Index - 1 ) ;
                        try
                            F.Attach ;
                            F.Header := H ;
                            F.Store := _Store ;
                            if( F.Stream_Name( Stream ) = 0 ) then
                            begin // Stream not defined
                                Set_Last_Error( Create_Exception( UOS_File_System_Illegal_Operation,
                                    nil ) ) ;
                                exit ;
                            end ;
                            Result.Size := F.Get_Stream_Size( Stream ) ;
                            Result.EOF := Result.Size ;
                        finally
                            F.Detach ;
                        end ;
                    end ;
                    exit ;
                end ;
            end ; // if( H.Name <> 0 )
        end ;  // while( true )
    finally
        _S.Free ;
    end ;
end ; // TUOS_Native_File_System.Get_File_Info


function TUOS_Native_File_System.Set_File_Info( Name : Pchar ;
    Stream : longint ; Info, Flags : TUOS_File_Info ) : TUnified_Exception ;

var Directory : TFS_File ;
    Dummy, Dummy1 : integer ;
    F : TFS_File ;
    FileSpec, S : string ;
    H : TUOS_File_Header ;
    Index : integer ;
    _S : TUnicode_String ;

begin
    // Setup...
    if( not Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Not_Mounted, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    Result := nil ;
    Set_Last_Error( nil ) ;

    // Parse specification...
    S := Name ;
    if( S = '' ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Invalid_Filename, nil ) ) ;
        Result := Last_Error ;
        exit ; // No context and no file spec
    end ;
    S := Parse_Filespec( S, Directory ) ;
    FileSpec := S ;
    if( Directory = nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Path_Not_Found, nil ) ) ;
        Result := Last_Error ;
        exit ; // Directory not found
    end ;

    S := lowercase( S ) ;
    _S := TUnicode_String.Create ;
    try
        _S.Assign_From_String( S, 3 ) ;
        _S.Lowercase ;
        Index := 0 ;
        while( true ) do
        begin
            inc( Index ) ;
            if( Directory.Read( 0, ( Index - 1 ) * sizeof( H ), sizeof( H ),
              H ) = 0 ) then
            begin
                // No more entries in this directory...
                Set_Last_Error( Create_Exception( UOS_File_System_File_Not_Found,
                    nil ) ) ;
                Result := Last_Error ;
                exit ;
            end ;
            if( ( H.Name <> 0 ) and ( ( H.Flags and FAF_DELETED ) = 0 ) ) then
            begin // Not a null entry or deleted file
                if( Matched_Strings( _S, Get_String( H.Name ) ) ) then
                begin
                    Dummy := Info.Clustersize ;
                    if( Dummy = 0 ) then
                    begin
                        Dummy := Store.Min_Storage ;
                    end ;
                    Dummy1 := H.Clustersize ;
                    if( Dummy1 = 0 ) then
                    begin
                        Dummy1 := Store.Min_Storage ;
                    end ;
                    if(
                        ( Flags.Clustersize <> 0 )
                        and
                        ( Dummy <> Dummy1 )
                      ) then
                    begin
                        if( H.Size <> 0 ) then
                        begin // Can't set clustersize on a file that already has data
                            Set_Last_Error( Create_Exception( UOS_File_System_Illegal_Operation,
                                nil ) ) ;
                            Result := Last_Error ;
                            exit ;
                        end ;
                        H.Clustersize := Info.Clustersize ;
                    end ;
                    if( Flags.Record_Size <> 0 ) then
                    begin
                        H.Record_Size := Info.Record_Size ;
                    end ;
                    if( Flags.Creation <> 0 ) then
                    begin
                        H.Creation := Info.Creation ;
                    end ;
                    if( Flags.Last_Modified <> 0 ) then
                    begin
                        H.Last_Modified := Info.Last_Modified ;
                    end ;
                    if( Flags.Last_Backup <> 0 ) then
                    begin
                        H.Last_Backup := Info.Last_Backup ;
                    end ;
                    if( Flags.Last_Access <> 0 ) then
                    begin
                        H.Last_Access := Info.Last_Access ;
                    end ;
                    if( Flags.Expiration <> 0 ) then
                    begin
                        H.Expiration := Info.Expiration ;
                    end ;
                    if( Flags.Creator <> 0 ) then
                    begin
                        H.Creator := Info.Creator ;
                    end ;
                    if( Flags.Owner <> 0 ) then
                    begin
                        H.Owner := Info.Owner ;
                    end ;
                    if( Flags.Flags <> 0 ) then
                    begin
                        H.Flags := Info.Flags
                                   and
                                   ( not FAF_Contiguous ) ;
                    end ;
                    if( Flags.Version_Limit <> 0 ) then
                    begin
                        H.Version_Limit := Info.Version_Limit ;
                    end ;
                    if( Flags.EOF <> 0 ) then
                    begin
                        if( Info.EOF > H.Size ) then
                        begin
                            Info.EOF := H.Size ;
                        end ;
                        H.EOF := Info.EOF ;
                    end ;
                    F := Create_File_Instance( Directory, Index - 1 ) ;
                    try
                        F.Attach ;
                        F.Header := H ;
                        F.Store := _Store ;
                        if(
                            ( Stream > 0 )
                            and
                            ( F.Stream_Name( Stream ) = 0 )
                          ) then // Stream not defined
                        begin
                            Set_Last_Error( Create_Exception( UOS_File_System_Illegal_Operation,
                                nil ) ) ;
                            exit ;
                        end ;
                        if( Flags.Flags <> 0 ) then
                        begin
                            F.Set_Contiguous( ( ( Info.Flags and FAF_Contiguous ) <> 0 ) ) ;
                        end ;
                        if( F.Last_Error <> nil ) then
                        begin
                            Set_Last_Error( Create_Exception( UOS_File_System_Illegal_Operation,
                                F.Last_Error ) ) ;
                            Result := Last_Error ;
                            exit ;
                        end ;
                        if( Flags.Size <> 0 ) then
                        begin
                            F.Set_Stream_Size( Stream, Info.Size ) ;
                        end ;
                        if( Directory.Write( 0, ( Index - 1 ) * sizeof( H ),
                          sizeof( H ), F.Header ) = 0 ) then
                        begin
                            Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
                                Directory.Last_Error ) ) ;
                            Result := Last_Error ;
                            exit ;
                        end ;
                        if( Directory.Last_Error <> nil ) then
                        begin
                            Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
                                Directory.Last_Error ) ) ;
                            Result := Last_Error ;
                            exit ;
                        end ;
                    finally
                        F.Detach ;
                    end ;
                    exit ;
                end ; // if( Matched_Strings( _S, Get_String( H.Name ) ) )
            end ; // if( H.Name <> 0 )
        end ; // while( true )
    finally
        _S.Free ;
    end ;
end ; //  TUOS_Native_File_System.Set_File_Info


function TUOS_Native_File_System.Lookup( Spec : PChar ; var Context : pointer ) : PChar ;

var _Context : TLookup_Context ;
    Dir : string ;
    Dummy : integer ;
    Directory : TFS_File ;
    Filespec : string ;
    H : TUOS_File_Header ;
    S : string ;

begin
    Result := nil ;
    if( Context = nil ) then // Initial lookup
    begin
        // Parse specification...
        S := Spec ;
        if( S = '' ) then
        begin
            exit ; // No context and no file spec
        end ;
        Dummy := length( S ) ;
        while( ( Dummy > 0 ) and ( S[ Dummy ] <> '\' ) ) do
        begin
            dec( Dummy ) ;
        end ;
        Dir := copy( S, 1, Dummy ) ;
        S := copy( S, Dummy + 1, length( S ) + 1 ) ;
        FileSpec := S ;
        Directory := Lookup_Directory( Dir ) ;
        if( Directory = nil ) then
        begin
            exit ; // Directory not found
        end ;

        // Create context...
        _Context := TLookup_Context.Create ;
        _Context.Spec := S ;
        _Context.Dir := Dir ;
        _Context.Index := 0 ; // Next file to read is index 0
        _Context.Directory := Directory ;
        _Context._Spec := Create_WC_Context( S, 3, _Context._Spec_Extension ) ;
        Context := _Context ;
    end ; // if( Context = nil )

    _Context := TLookup_Context( Context ) ;
    if( Lookup_Contexts.IndexOf( _Context ) = -1 ) then // Not a valid context
    begin
        exit ;
    end ;
    while( true ) do
    begin
        inc( _Context.Index ) ;
        if( _Context.Directory.Read( 0, ( _Context.Index - 1 ) * sizeof( H ),
          sizeof( H ), H ) = 0 ) then
        begin
            // No more entries in this directory...
            _Context.Free ;
            Context := nil ;
            exit ;
        end ;
        if( ( H.Name <> 0 ) and ( ( H.Flags and FAF_DELETED ) = 0 ) ) then
        begin // Not a null entry or deleted file
            S := Get_String( H.Name ) ;
            if( Context_WC_Match( _Context._Spec, _Context._Spec_Extension, S, 3 ) ) then
            begin
                Temp := S ;
                Result := PChar( Temp ) ;
                exit ;
            end ;
        end ; // if( H.Name <> 0 )
    end ;  // while( true )
end ; // TUOS_Native_File_System.Lookup


procedure TUOS_Native_File_System.Clear_Context( var Context : pointer ) ;

begin
    if( Context <> nil ) then
    begin
        TLookup_Context( Context ).Free ;
        Context := nil ;
    end ;
end ;


function TUOS_Native_File_System.Create_File( Filespec : PChar ;
    Info : TUOS_File_Info ) : TUnified_Exception ;

var Directory : TFS_File ;
    S : string ;
    F : TFS_File ;
    Free_Index : longint ;
    Header : TUOS_File_Header ;
    Index : longint ;

begin
    if( not Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Not_Mounted, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    Result := nil ;

    // Parse specification...
    S := Filespec ;
    S := Parse_Filespec( S, Directory ) ;
    if( ( S = '' ) or ( Directory = nil ) ) then
    begin
        Result := Last_Error ;
        exit ;
    end ;

    Directory.Attach ;
    try
        // Find file...
        Index := Find_File( S, Directory, Free_Index ) ;
        if( _Last_Error <> nil ) then
        begin
            if(
                ( _Last_Error.Get_Facility = UOS_File_System_Facility_ID )
                and
                ( _Last_Error.Get_Error = UOS_File_System_File_Not_Found )
              ) then
            begin
                Set_Last_Error( nil ) ;
                Index := -1 ;
            end else
            begin
                exit ;
            end ;
        end ;
        if( Index >= 0 ) then // File already exists
        begin
            // Mark existing file as deleted (Delete_File does physical deletion)...
            if( Directory.Read( 0, ( Index ) * sizeof( Header ), sizeof( Header ),
              Header ) = 0 ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Read_Error,
                    Directory.Last_Error ) ) ;
                Result := Last_Error ;
                exit ;
            end ;
            Header.Flags := Header.Flags or FAF_DELETED ;
            if( Directory.Write( 0, ( Index ) * sizeof( Header ), sizeof( Header ),
              Header ) = 0 ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
                    Directory.Last_Error ) ) ;
                Result := Last_Error ;
                exit ;
            end ;
        end ; // if( Index >= 0 )

        // Make sure clustersize if correct...
        if( Info.Clustersize < Store.Min_Storage ) then
        begin
            Info.Clustersize := Store.Min_Storage ;
        end ;
        if( ( Info.Flags and FAF_DIRECTORY ) <> 0 ) then
        begin
            if( Info.Clustersize < _Header.Folder_Clustersize ) then
            begin
                Info.Clustersize := _Header.Folder_Clustersize ;
            end ;
        end ;
        Info.Clustersize := Info.Clustersize and ( not ( Store.Min_Storage - 1 ) ) ;

        // Create and write header...
        if( Free_Index < 0 ) then // Extending directory to hold new header
        begin
//            Free_Index := Directory.Get_Stream_Size( 0 ) div sizeof( Header ) ; // This was skipping every other header index (eg went from 512 to 1024)
            Free_Index := Directory.Header.EOF div sizeof( Header ) ;
        end ;
        fillchar( Header, sizeof( Header ), 0 ) ;
        Header.Clustersize := Info.Clustersize ;
        Header.Record_Size := Info.Record_Size ;
        Header.Creation := Info.Creation ;
        Header.Last_Modified := Info.Last_Modified ; // Last write date
        Header.Last_Backup := Info.Last_Backup ;
        Header.Last_Access := Info.Last_Access ; // Last read/open date
        Header.Expiration := Info.Expiration ;
        Header.Creator := Info.Creator ;
        Header.Owner := Info.Owner ;
        Header.Flags := Info.Flags ;
        Header.Version_Limit := Info.Version_Limit ;
        Header.Name := _String_Table.Add_String( S ) ;
        F := Create_File_Instance( Directory, Free_Index ) ;
        try
            //F.Attach ; // removed, 3-Oct-2016
            F.Header := Header ;
            F.Store := Store ;
            F.Set_Stream_Size( 0, Info.Size ) ;
            if(
                ( F.Last_Error <> nil )
                or
                ( F.Get_Stream_Size( 0 ) < Info.Size )
              ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_File_Creation_Error,
                    F.Last_Error ) ) ;
                Result := Last_Error ;
                exit ;
            end ;
            if( Info.EOF > Header.Size ) then
            begin
                Info.EOF := Header.Size ;
            end ;
            F.Header.EOF := Info.EOF ;
            if( Directory.Write( 0, Free_Index * sizeof( F.Header ),
                sizeof( F.Header ), F.Header ) = 0 ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
                    Directory.Last_Error ) ) ;
                Result := Last_Error ;
                exit ;
            end ;
            F.Dirty := False ;
        finally
            F.Detach ;
        end ;
    finally
        Directory.Detach ;
    end ;
end ; // TUOS_Native_File_System.Create_File


function TUOS_Native_File_System.Delete_File( Filespec : PChar ) : TUnified_Exception ;

var Directory : TFS_File ;
    Dummy : integer ;
    S : string ;
    F : TFS_File ;
    Header : TUOS_File_Header ;
    Free_Index, Index : integer ;

begin
    if( not Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Not_Mounted, nil ) ) ;
        Result := Last_Error ;
        exit ; 
    end ;
    Result := nil ;
    Set_Last_Error( nil ) ;

    // Parse specification...
    S := Filespec ;
    S := Parse_Filespec( S, Directory ) ;
    if( ( S = '' ) or ( Directory = nil ) ) then
    begin
        Result := Last_Error ;
        exit ;
    end ;

    try
        // Find file...
        Index := Find_File( S, Directory, Free_Index ) ;
        Directory.Attach ;
        if( Index >= 0 ) then
        begin
            if( Directory.Read( 0, ( Index ) * sizeof( Header ), sizeof( Header ),
              Header ) = 0 ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Read_Error,
                    Directory.Last_Error ) ) ;
                Result := Last_Error ;
                exit ;
            end ;

            // Release space allocated for file...
            F := Create_File_Instance( Directory, Index ) ;
            try
                F.Attach ;
                F.Header := Header ;
                F.Store := Store ;
                for Dummy := 0 to F.Max_Stream do
                begin
                    F.Set_Stream_Size( Dummy, 0 ) ;
                end ;
            finally
                F.Detach ;
            end ;

            // Mark the directory entry as unused...
            fillchar( Header, sizeof( Header ), 0 ) ;
            if( Directory.Write( 0, Index * sizeof( Header ), sizeof( Header ),
              Header ) = 0 ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
                    Directory.Last_Error ) ) ;
                Result := Last_Error ;
            end ;
            exit ;
        end ; // if( Index >= 0 )

        Set_Last_Error( Create_Exception( UOS_File_System_File_Not_Found, nil ) ) ;
        Result := Last_Error ;
    finally
        Directory.Detach ;
    end ;
end ; // TUOS_Native_File_System.Delete_File


function TUOS_Native_File_System.Rename( Filespec,
    New_Name : PChar ) : TUnified_Exception ;

var Directory : TFS_File ;
    Header : TUOS_File_Header ;
    Free_Index, Index : longint ;
    N, O : cardinal ;
    S : string ;

begin
    if( not Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Not_Mounted, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    Result := nil ;
    Set_Last_Error( nil ) ;

    // Parse specification...
    if( ( length( New_Name ) = 0 ) or ( pos( '\', New_Name ) > 0 ) ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Invalid_Filename, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    S := Filespec ;
    S := Parse_Filespec( S, Directory ) ;
    if( ( S = '' ) or ( Directory = nil ) ) then
    begin
        Result := Last_Error ;
        exit ;
    end ;
    Directory.Attach ;
    try
        if( S = New_Name ) then
        begin
            exit ; // No change
        end ;

        Index := Find_File( New_Name, Directory, Free_Index ) ;
        if( Index >= 0 ) then // New file already exists
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_File_Exists,
                nil ) ) ;
            Result := Last_Error ;
            exit ;
        end ;
        if(
            ( _Last_Error.Get_Facility <> UOS_File_System_Facility_ID )
            or
            ( _Last_Error.Get_Error <> UOS_File_System_File_Not_Found )
          ) then
        begin
            exit ;
        end ;
        Set_Last_Error( nil ) ;
        Index := Find_File( S, Directory, Free_Index ) ;
        if( Index < 0 ) then // Old file isn't there
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_File_Not_Found, nil ) ) ;
            Result := Last_Error ;
            exit ;
        end ;

        if( Directory.Read( 0, ( Index ) * sizeof( Header ), sizeof( Header ),
          Header ) = 0 ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Read_Error,
                Directory.Last_Error ) ) ;
            Result := Last_Error ;
            exit ;
        end ;
        N := _String_Table.Add_String( New_Name ) ;
        if( Last_Error <> nil ) then
        begin
            Result := Last_Error ;
            exit ;
        end ;
        O := Header.Name ;
        Header.Name := N ;
        if( Directory.Write( 0, ( Index ) * sizeof( Header ), sizeof( Header ),
          Header ) = 0 ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
                Directory.Last_Error ) ) ;
            Result := Last_Error ;
            _String_Table.Delete_String( N ) ;
            exit ;
        end ;
        _String_Table.Delete_String( O ) ;
    finally
        Directory.Detach ;
    end ;
end ; // TUOS_Native_File_System.Rename


function TUOS_Native_File_System.Move_File( Old_Filespec,
    New_Filespec : PChar ) : TUnified_Exception ;

var Directory, Directory1 : TFS_File ;
    Header : TUOS_File_Header ;
    Free_Index, Index, Index1 : longint ;
    New_Filename : string ;
    S : string ;

begin
    // Setup and sanity check...
    if( not Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Not_Mounted, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    Result := nil ;
    Set_Last_Error( nil ) ;

    // Parse specifications...
    S := Old_Filespec ;
    S := Parse_Filespec( S, Directory ) ;
    if( ( S = '' ) or ( Directory = nil ) ) then
    begin
        Result := Last_Error ;
        exit ;
    end ;

    New_Filename := New_Filespec ;
    if( copy( New_Filename, length( New_Filename ), 1 ) <> '\' ) then
    begin
        New_Filename := New_Filename + '\' ;
    end ;
    New_Filename := Parse_Filespec( New_Filename + S, Directory1 ) ;
    if( ( New_Filename = '' ) or ( Directory1 = nil ) ) then
    begin
        Result := Last_Error ;
        exit ;
    end ;

    if( Directory = Directory1 ) then // Moving to the same directory
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Source_Same_As_Destination, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;

    // Lookup old file...
    Index := Find_File( S, Directory, Free_Index ) ;
    if( Index < 0 ) then // Old file isn't there
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_File_Not_Found, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;

    // Validate no existing (new) file ...
    Index1 := Find_File( S, Directory1, Free_Index ) ;
    if( Index1 >= 0 ) then // New file is already there
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_File_Exists, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    if(
        ( _Last_Error.Get_Facility = UOS_File_System_Facility_ID )
        and
        ( _Last_Error.Get_Error = UOS_File_System_File_Not_Found )
      ) then
    begin
        Set_Last_Error( nil ) ;
    end else
    begin
        exit ;
    end ;
    if( Free_Index < 0 ) then
    begin
        Free_Index := Directory1.Get_Stream_Size( 0 ) div sizeof( Header ) ;
    end ;

    // Get file header...
    if( Directory.Read( 0, Index * sizeof( Header ), sizeof( Header ),
      Header ) = 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Read_Error,
            Directory.Last_Error ) ) ;
        Result := Last_Error ;
        exit ;
    end ;

    // Update old header...
    Header.Flags := Header.Flags or FAF_COPYING ;
    if( Directory.Write( 0, Index * sizeof( Header ), sizeof( Header ),
      Header ) = 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
            Directory.Last_Error ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    Header.Flags := Header.Flags and ( not FAF_COPYING ) ;

    // Write to new directory...
    if( Directory1.Write( 0, Free_Index * sizeof( Header ), sizeof( Header ),
      Header ) = 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
            Directory1.Last_Error ) ) ;
        Result := Last_Error ;
        exit ;
    end ;

    // Erase from old directory...
    fillchar( Header, sizeof( Header ), 0 ) ;
    if( Directory.Write( 0, Index * sizeof( Header ), sizeof( Header ),
      Header ) = 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Write_Error,
            Directory.Last_Error ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
end ; // TUOS_Native_File_System.Move_File


function TUOS_Native_File_System.Translate_Error : longint ;

begin
    if( Last_Error = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    Result := Last_Code ;
end ;


function TUOS_Native_File_System.Normalize( Name : PAnsiChar ) : PAnsiChar ;

var I, P : integer ;
    S : TUnicode_String ;
    Start : integer ;
    Work : string ;

begin
    // Convert to string...
    Temp := Name ;

    // Sanity check...
    if( pos( '\\', Temp ) > 0 ) then // Syntax error
    begin
        Result := nil ;
        exit ;
    end ;

    // Convert to lowercase UTF8...
    S := TUnicode_String.Create ;
    S.Assign_From_String( Temp, 3 ) ;
    S.Lowercase ;
    Temp := S.As_String ;

    // Process the specification...
    Start := 1 ;
    P := 2 ;
    while( P <= length( Temp ) + 1 ) do
    begin
        if( ( P > length( Temp ) ) or ( Temp[ P ] = '\' ) ) then
        begin
            Work := copy( Temp, Start + 1, P - Start - 1 ) ; // Get this segment
            if( Work = '.' ) then // Same directory
            begin
                Temp := copy( Temp, 1, Start ) + copy( Temp, P + 1, length( Temp ) ) ;
                P := Start + 1 ;
                continue ;
            end else
            if( Work = '..' ) then // Parent directory
            begin
                I := Start - 1 ; // Find start of previous directory
                while( ( I > 0 ) and ( Temp[ I ] <> '\' ) ) do
                begin
                    dec( I ) ;
                end ;
                if( I = 0 ) then
                begin
                    Temp := copy( Temp, Start, length( Temp ) ) ;
                    P := 2 ;
                end else
                begin
                    Temp := copy( Temp, 1, I ) + copy( Temp, P + 1, length( Temp ) ) ;
                    P := I + 1 ;
                end ;
                continue ;
            end ; // if( Work = '..' )
            Start := P ;
        end ; // if( Temp[ P ] = '\' )
        inc( P ) ;
    end ; // while( P < length( Temp ) )
    Result := PChar( Temp ) ;
end ; // TUOS_Native_File_System.Normalize


function TUOS_Native_File_System.Mount( _Label, Password : PChar ;
    Flags : longint ) : TUnified_Exception ;

var Buff : PChar ;
    I : integer ;
    P : TStore_Address64 ;
    UEC : TUnified_Exception ;

begin
    if( Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Already_Mounted, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    Result := nil ;
    if( not Valid_File_System( P ) ) then
    begin
        if( Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Invalid_File_System_Structure, nil ) ) ;
        end ;
        Result := Last_Error ;
        exit ;
    end ;

    // Mark the store as dirty and mounted...
    Store.Read_Data( Buffer[ 0 ], P, sizeof( _Header ), UEC ) ;
    if( UEC <> nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Read_Error, UEC ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    if( not Store.Read_Only ) then
    begin
        move( Buffer[ 0 ], _Header, sizeof( _Header ) ) ;
        if( ( _Header.Flags and UOSFSF_Dirty ) <> 0 ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Is_Dirty, UEC ) ) ;
            Result := Last_Error ;
            exit ;
        end ;
        _Header.Flags := _Header.Flags or UOSFSF_Dirty ;
        move( _Header, Buffer[ 0 ], sizeof( _Header ) ) ;
        Store.Write_Data( Buffer[ 0 ], P, sizeof( _Header ), UEC ) ;
        if( UEC <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Write_Error, UEC ) ) ;
            Result := Last_Error ;
            exit ;
        end ;
    end ;

    // Get allocation table...
    Store.Set_Allocation_Table_Offset( _Header.AT_Offset ) ;
    I := _Header.At_Size ;
    I := ( I + Store.Min_Storage - 1 ) and ( not ( Store.Min_Storage - 1 ) ) ;
    Buff := allocmem( I ) ;
    try
        Store.Read_Data( Buff[ 0 ], _Header.AT_Offset, I, UEC ) ;
        if( UEC <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Read_Error, UEC ) ) ;
            Result := Last_Error ;
            exit ;
        end ;
        Store.Set_Allocation_Table_Data( Buff, _Header.AT_Size ) ;
    finally
        freemem( Buff ) ;
    end ;

    Mounted := True ;
    Result := Last_Error ;
end ; // TUOS_Native_File_System.Mount


function TUOS_Native_File_System.Dismount : TUnified_Exception ;

var Dummy : integer ;
    F : TFS_File ;
    P : TStore_Address64 ;
    UEC : TUnified_Exception ;

begin
    // Close all open files...
    if( _String_Table <> nil ) then
    begin
        _String_Table.Free ;
        _String_Table := nil ;
    end ;
    for Dummy := Open_Files.Count - 1 downto 0 do
    begin
        if( Open_Files[ Dummy ] <> nil ) then
        begin
            F := TFS_File( Open_Files[ Dummy ] ) ;
            Open_Files[ Dummy ] := nil ;
            try
                if( F = _Root ) then
                begin
                    _Root := nil ;
                end else
                if( F = _Store_Folder ) then
                begin
                    _Store_Folder := nil ;
                end ;
            finally
                F.Free ;
            end ;
        end ; // if( Open_Files[ Dummy ] <> nil ) 
    end ; // for Dummy := Open_Files.Count - 1 downto 0
    Open_Files.Clear ;
    
    // Write up-to-date allocated table
    if( not Store.Read_Only ) then
    begin
        Store.Flush_Allocation_Table ;
    end ;

    // Close out our persistent file objects...
    if( _Root <> nil ) then
    begin
        _Root.Free ;
        _Root := nil ;
    end ;
    if( _Store_Folder <> nil ) then
    begin
        _Store_Folder.Free ;
        _Store_Folder := nil ;
    end ;

    // Sanity check...
    if( not Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Not_Mounted, nil ) ) ;
        Result := Last_Error ;
        exit ;
    end ;

    // Mark us as clean and dismounted...
    Mounted := False ;
    Valid_File_System( P ) ;
    Store.Read_Data( Buffer[ 0 ], P, sizeof( _Header ), UEC ) ;
    if( UEC <> nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Read_Error, UEC ) ) ;
        Result := Last_Error ;
        exit ;
    end ;
    if( not Store.Read_Only ) then
    begin
        _Header.Flags := _Header.Flags and ( not UOSFSF_Dirty ) ;
        move( _Header, Buffer[ 0 ], sizeof( _Header ) ) ;
        Store.Write_Data( Buffer[ 0 ], P, sizeof( _Header ), UEC ) ;
        if( UEC <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Write_Error, UEC ) ) ;
            Result := Last_Error ;
            exit ;
        end ;
    end ;
    
    Set_Last_Error( nil ) ;
    Result := nil ;
end ; // TUOS_Native_File_System.Dismount


function TUOS_Native_File_System.Get_File( Name : PChar ) : TUOS_File ;

var Directory : TFS_File ;
    Header : TUOS_File_Header ;
    Free_Index, Index : integer ;
    S : string ;

begin
    Result := nil ;
    if( not Mounted ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Not_Mounted, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;

    // Parse specification...
    S := Name ;
    if( length( S ) = 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_System_Invalid_Filename, nil ) ) ;
        exit ;
    end ;
    S := Parse_Filespec( S, Directory ) ;
    if( ( S = '' ) or ( Directory = nil ) ) then
    begin
        if( Directory <> nil ) then
        begin
            Directory.Detach ;
        end ;
        exit ;
    end ;
    Directory.Attach ;
    try
        Index := Find_File( S, Directory, Free_Index ) ;
        if( Index < 0 ) then // File isn't there
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_File_Not_Found, nil ) ) ;
            exit ;
        end ;

        if( Directory.Read( 0, Index * sizeof( Header ), sizeof( Header ),
          Header ) = 0 ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_System_Read_Error,
                Directory.Last_Error ) ) ;
            exit ;
        end ;
        Result := Create_File_Instance( Directory, Index ) ;
        TFS_File( Result ).Header := Header ;
    finally
        Directory.Detach ;
    end ;
end ; // TUOS_Native_File_System.Get_File


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    halt ;
end ;


var LB : TListBox = nil ;
    F : TForm = nil ;

procedure Display_Dir( FS : TUOS_Native_File_System ; S : string ) ;

var Context : pointer ;
    Dir : string ;
    Info : TUOS_File_Info ;
    S1 : string ;

begin
    if( LB = nil ) then
    begin
        F := TForm.Create( Application ) ;
        LB := TListBox.Create( F ) ;
        LB.Parent := F ;
        LB.Align := alClient ;
    end ;

    Context := nil ;
    if( copy( S, length( S ), 1 ) <> '\' ) then
    begin
        S := S + '\' ;
    end ;
    Dir := S ;
    S := FS.Lookup( PChar( S + '*.*' ), Context ) ;
    while( Context <> nil ) do
    begin
        Info := FS.Get_File_Info( PChar( Dir + S ), 0 ) ;
        while( length( S ) < 24 ) do
        begin
            S := S + ' ' ;
        end ;
        S1 := S ;
        S := '' ;
        if( ( Info.Flags and FAF_PLACED ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'Pl' ;
        end ;
        if( ( Info.Flags and FAF_CONTIGUOUS ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'C' ;
        end ;
        if( ( Info.Flags and FAF_READONLY ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'RO' ;
        end ;
        if( ( Info.Flags and FAF_SYSTEM ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'S' ;
        end ;
        if( ( Info.Flags and FAF_HIDDEN ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'H' ;
        end ;
        if( ( Info.Flags and FAF_DIRECTORY ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'Di' ;
        end ;
        if( ( Info.Flags and FAF_PERMANENT ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'Pe' ;
        end ;
        if( ( Info.Flags and FAF_DELETED ) <> 0 ) then
        begin
            if( S <> '' ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'De' ;
        end ;
        while( length( S ) < 16 ) do
        begin
            S := S + ' ' ;
        end ;
        S1 := S1 + S ;
        S := inttostr( Info.EOF ) ;
        while( length( S ) < 7 ) do
        begin
            S := ' ' + S ;
        end ;
        S := S + '/' + inttostr( Info.Size ) ;
        while( length( S ) < 12 ) do
        begin
            S := S + ' ' ;
        end ;
        S := S + '    ' + Format_Sirius_Date( '', Info.Creation ) + ' ' +
            Format_Sirius_Time( 'HH:MM', Info.Creation ) ;
        LB.Items.Add( S1 + S ) ;
        S := FS.Lookup( '', Context ) ;
    end ;

    LB.Items.Add( '------' ) ;
    F.Show ;
end ;


procedure Test_Unit ;

var E : TUnified_Exception ;
    FS : TUOS_Native_File_System ;
    Info, Flags : TUOS_File_Info ;
    MStore : TUOS_Managed_Store ;
    _Siz, _Typ : longint ;
    Siz : int64 ;
    Store : TRAM_Disk64 ;

begin
    // Initialize disk...
    Store := TRAM_Disk64.Create ;
    Store.Set_Max_Storage( 5 * 1024 * 1024, E ) ;
    MStore := TUOS_Managed_Store.Create ;
    MStore.Store := Store ;
    FS := TUOS_Native_File_System.Create ;
    FS.Store := MStore ;
    Siz := 2621440 ;
    FS.Init( 'Test', '', 0, 512, 512, 0, Siz, nil ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;

    FS.Create_File( '\test\abc.txt', Info ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Rename( '\test\abc.txt', 'abc.new' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Move_File( '\test\abc.new', '\othertest\' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Delete_File( '\othertest\abc.new' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;

    E := FS.Mount( '', '', 0 ) ;
    if( E <> nil ) then
    begin
        Error ;
    end ;
    E := FS.Mount( '', '', 0 ) ;
    if( E = nil ) then
    begin
        Error ;
    end ;
    fillchar( Info, sizeof( Info ), 0 ) ;
    fillchar( Flags, sizeof( Flags ), 0 ) ;

    FS.Create_File( '', Info ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Rename( '', 'abc.new' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Rename( 'abc.new', '' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Info.Flags := FAF_DIRECTORY ;
    FS.Create_File( '\test', Info  ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\test' ) ;
    FS.Create_File( '\othertest', Info  ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\test' ) ;
    Display_Dir( FS, '\othertest' ) ;
    Display_Dir( FS, '\' ) ;

    Info.Flags := 0 ;
    Info.Size := 1024 ;
    FS.Create_File( '\test\abc.txt', Info ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\' ) ;
    Display_Dir( FS, '\test' ) ;

    Flags.Flags := 1 ;
    Info.Flags := FAF_SYSTEM ;
    FS.Set_File_Info( '\test\abc.txt', 0, Info, Flags ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\test' ) ;

    Info.Flags := 0 ;
    Info.Size := 2048 ;
    FS.Create_File( '\test\abc.txt', Info ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\test' ) ;

    FS.Rename( '\test\abc.txt', 'abc.new' ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\test' ) ;

    FS.Move_File( '\test\abc.new', '\othertest\' ) ;
    if( FS.Last_Error <> nil ) then
    begin
        ShowMessage( FS.Last_Error.Error_Text( _Siz, _Typ ) ) ;
        Error ;
    end ;
    Display_Dir( FS, '\test' ) ;
    Display_Dir( FS, '\othertest' ) ;

    FS.Delete_File( '\test\abc.txt' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Delete_File( '\othertest\abc.new' ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\othertest' ) ;
    Info.Flags := 0 ;
    Info.Clustersize := 1024 ;
    FS.Create_File( '\test\abc.txt', Info ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, '\test' ) ;
    FS.Move_File( 'nonexistantfile.txt', 'test' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Move_File( '\test\abc.txt', '\test\' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Move_File( '', '\test\' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Move_File( '\test\abc.txt', '' ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    fillchar( Info, sizeof( Info ), 0 ) ;
    Info.Flags := FAF_DIRECTORY ;
    FS.Create_File( '\test1', Info  ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    FS.Move_File( '\test\abc.txt', 'test2' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Move_File( '\test\abc.txt', 'test1' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Info.Flags := 0 ;
    Info.Size := 2048 ;
    FS.Create_File( '\othertest\abc.txt', Info ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Info.Flags := 0 ;
    Info.Size := 2048 ;
    FS.Create_File( '\test\abc.txt', Info ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Display_Dir( FS, 'test' ) ;
    FS.Move_File( '\othertest\abc.txt', 'test' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Info.Flags := 0 ;
    Info.Size := 2048 ;
    Info.EOF := 4096 ;
    FS.Create_File( '\test\abc.new', Info ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    FS.Rename( '\test\abc.new', 'abc.txt' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Rename( '\test\abc.txt', 'abc.txt' ) ;
    if( FS.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    FS.Delete_File( '' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;
    FS.Rename( 'test\moo.txt', 'abc.txt1' ) ;
    if( FS.Last_Error = nil ) then
    begin
        Error ;
    end ;

    if( FS.Dismount <> nil ) then
    begin
        Error ;
    end ;
    if( FS.Dismount = nil ) then
    begin
        Error ;
    end ;

    FS.Free ;
    MStore.Free ;
    Store.Free ;
end ;
{$ENDIF}


begin
//    Test_Unit ;
end.
