{
        Program Name : UOS_MStore
        Package Name : Subroutine Library - Layer 1
        Purpose      : UOS Managed Store class
        Institution  : Conroy & Conroy Co.
        Date Written : 2015
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2015 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TBase_Store is the base class for all storage-related classes.  It
        implements the generic store functionality upon which descendant stores
        are built.
}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

unit UOS_MStore ;

interface

uses // Subroutine Library - Layer 0...
{$IFDEF CC_DEBUG}
     Classes, // TStringList
{$ENDIF}
{$IFDEF CC_Regression_Test}
     _TestInt, // TTest_Interface
{$ENDIF}
     _UE, // TUnified_Exception

     // Subroutine Library - Layer 1...
     _Cache, // TCOM_Cache64
     MStores, // TCOM_Managed_Store64
     _Stores, // TCOM_Store64

     // Subroutine Library - Layer 3...
     Allocatr64 ; // TAT64

type TUOS_Managed_Store = class( TCOM_Managed_Store64 )
                              public // Constructors and destructors...
                                  constructor Create ;
                                  destructor Destroy ; override ;

                              private // Instance data...
                                  _Store : TCOM_Store64 ;
                                  _Clustersize : cardinal ;
                                  _AT : TAT64 ;

                              protected // Internal utility routines...
                                  function AT : TAT64 ;
                                  function Clustersize : cardinal ;

{$IFDEF CC_DEBUG}
                              public // Debugging
                                  _Allocations : TStringList ;

                                  procedure Add_Allocation( Pos, Len : int64 ) ;
                                  procedure Remove_Allocation( Pos, Len : int64 ) ;
                                  procedure Check_Access( Pos, Len : int64 ; R : boolean ) ;
                                  procedure Do_ReAllocation( Pos, Old_Len, New_Len : int64 ) ;
{$ENDIF}

                              public { Overrides... }
                                  function Is_Class( Name : PChar ) : boolean ;
                                      override ;

                                  function Read_Data( var Data ; Address,
                                      _Size : TStore_Address64 ;
                                      var UEC : TUnified_Exception ) : TStore_Address64 ;
                                      override ;

                                  function Write_Data( var Data ; Address,
                                      _Size : TStore_Address64 ;
                                      var UEC : TUnified_Exception ) : TStore_Address64 ;
                                      override ;

                                  function Max_Storage : TStore_Size64 ;
                                      override ;

                                  function Min_Storage : TStore_Address64 ;
                                      override ;

                                  function Extend( Amount : TStore_Address64 ) :
                                      TStore_Address64 ; override ;

                                  function Get_Read_Only : boolean ;
                                      override ;

                                  function Get_Write_Only : boolean ;
                                      override ;

                                  procedure Format ; override ;

                                  function Get_Name : PChar ; override ;
                                  { Returns name of store. }

                                  function Get_Cache : TCOM_Cache64 ;
                                      override ;
                                  { Returns current cache object. }

                                  procedure Set_Cache( Value : TCOM_Cache64 ) ;
                                      override ;
                                  { Sets the current cache object. }

                                  function Contiguous_Store : boolean ;
                                      override ;

                                  procedure Set_Max_Storage( Value : TStore_Address64 ;
                                      var Res : TUnified_Exception ) ;
                                      override ;

                                  function Extended_Size : TStore_Address64 ;
                                      override ;

                                  procedure Set_Read_Only( Value : boolean ) ;
                                      override ;
                                  { Sets the store read-only value. }
                                  procedure Set_Write_Only( Value : boolean ) ;
                                      override ;
                                  { Sets the store write-only value. }

                                  function Allocate( Size : TStore_Address64 ) :
                                      TStore_Address64 ; override ;
                                      { Allocate space}
                                  function Allocate_At( Offset,
                                      Size : TStore_Address64 ) : boolean ;
                                      override ;
                                      { Allocate space at specific location }
                                  procedure Copy( Source,
                                      Destination : TStore_Address64 ;
                                      Count : int64 ) ; override ; { Copy data }
                                  procedure Deallocate( PTR,
                                      Size : TStore_Address64 ) ;
                                      override ;
                                  procedure Fill( PTR : TStore_Address64 ;
                                      Count : integer ; Value : byte ) ;
                                      override ;
                                  function Reallocate( PTR, Old,
                                      New : TStore_Address64 ) : TStore_Address64 ;
                                      override ;
                                  function SpaceAvail : TStore_Address64 ;
                                      override ;
                                  function XSpaceAvail : TStore_Address64 ;
                                      override ;
                                  function MaxSpace : TStore_Address64 ;
                                      override ;
                                  procedure Get_Root( var Buf ) ; override ;
                                  procedure Set_Root( var Buf ) ; override ;
                                  function Allocation_Table_Size : longint ;
                                      override ;
                                  procedure Get_Allocation_Table_Data( Buff : pointer ;
                                      _Size : longint ) ;
                                      override ;
                                  procedure Set_Allocation_Table_Data( Buff : pointer ;
                                      _Size : longint ) ;
                                      override ;
                                  procedure Set_Min_Storage( _Size : longint ) ;
                                      override ;
                                  procedure Set_Allocation_Table_Offset( Value : TStore_Address64 ) ;
                                      override ;
                                  function Get_Allocation_Table_Offset : TStore_Address64 ;
                                      override ;
                                  procedure Flush_Allocation_Table ; override ;
                                  function Get_Store : TCOM_Store64 ; override ;
                                  procedure Set_Store( Value : TCOM_Store64 ) ;
                                      override ;

                              public // Properties...
                                  property Allocation_Table_Offset : TStore_Address64
                                      read Get_Allocation_Table_Offset
                                      write Set_Allocation_Table_Offset ;
                                  property Cache : TCOM_Cache64
                                      read Get_Cache
                                      write Set_Cache ;
                                  property Store : TCOM_Store64
                                      read Get_Store
                                      write Set_Store ;
                          end ; { TUOS_Managed_Store }

{$IFDEF CC_Regression_Test}
procedure Test_Unit( T : TTest_Interface ) ;
{$ENDIF}


implementation

uses // Pascal...
     Sysutils, // lowercase
{$IFDEF CC_Regression_Test}
     Forms,

     // Subroutine library, layer 1...
     RAMDisk, // TRAM_Disk64
{$ENDIF}

     // Subroutine library, layer 1...
     Maths, // Round_Up
     UE ; // TStandard_Error_Interface

const UOS_MStore_Error_Success = 0 ;
      UOS_MStore_Error_Access_Violation = 1 ;
      UOS_MStore_Error_IO_Failure = 2 ;
      UOS_AT_Resize_Failure = 3 ;

function ERT( C : integer ) : string ;

begin
   case C of
       UOS_MStore_Error_Access_Violation : Result := 'Access violation' ;
       UOS_MStore_Error_IO_Failure : Result := 'I/O failure' ;
       UOS_AT_Resize_Failure : Result := 'AT resize failure' ;
       else Result := 'Unknown error' ;
   end ;
end ;


type tUOS_MStore_Unified_Exception = class( TStandard_Error_Interface )
                                         private
                                             _Code : longint ;
                                             _Text : string ;
                                             _Previous : tUnified_Exception ;

                                         public
                                             function Get_Facility : longint ; override ;
                                             function Get_Facility_Version : longint ;
                                                 override ;
                                             function Version : longint ; override ;
                                             function Severity : longint ; override ;
                                             function Error_Text( var Size, Typ : longint ) : PChar ;
                                                 override ;
                                             function Get_Error : longint ; override ;
                                             function Get_Previous : tUnified_Exception ;
                                                 override ;
                                             procedure Terminate ; override ;
                                     end ; // tUOS_MStore_Unified_Exception


function tUOS_MStore_Unified_Exception.Get_Facility : longint ;

begin
    Result := 136 ;
end ;


function tUOS_MStore_Unified_Exception.Get_Facility_Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_MStore_Unified_Exception.Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_MStore_Unified_Exception.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function tUOS_MStore_Unified_Exception.Error_Text( var Size, Typ : longint ) : PChar ;

begin
    _Text := ERT( _Code ) ;
    Size := length( _Text ) ;
    Typ := 0 ;
    Result := PChar( _Text ) ;
end ;


function tUOS_MStore_Unified_Exception.Get_Error : longint ;

begin
    Result := _Code ;
end ;


function tUOS_MStore_Unified_Exception.Get_Previous : tUnified_Exception ;

begin
    Result := _Previous ;
end ;


procedure tUOS_MStore_Unified_Exception.Terminate ;

begin
    Free ;
end ;



function Create_Exception( C : longint ; E : TUnified_Exception ) : tUOS_MStore_Unified_Exception ;

begin
    Result := tUOS_MStore_Unified_Exception.Create ;
    Result._Code := C ;
    Result._Previous := E ;
end ;



// TUOS_Managed_Store methods...

// Constructors and destructors...

constructor TUOS_Managed_Store.Create ;

begin
    inherited Create ;

{$IFDEF CC_DEBUG}
    _Allocations := TStringList.Create ;
{$ENDIF}
end ;


destructor TUOS_Managed_Store.Destroy ;

begin
    Set_Store( nil ) ;
    if( _AT <> nil ) then
    begin
        _AT.Free ;
        _AT := nil ;
    end ;
{$IFDEF CC_DEBUG}
    _Allocations.Free ;
{$ENDIF}

    inherited Destroy ;
end ;


{$IFDEF CC_DEBUG}
// Debugging...

procedure Error( const S : string ) ;

begin
    halt ;
end ;


procedure TUOS_File_Heap.Add_Allocation( Pos, Len : int64 ) ;

var I : integer ;
    S, L : string ;
    Work : string ;

begin
    S := inttostr( Pos ) ;
    while( length( S ) < 9 ) do
    begin
        S := '0' + S ;
    end ;
    L := inttostr( Pos + Len - 1 ) ;
    while( length( L ) < 9 ) do
    begin
        L := '0' + L ;
    end ;
    for I := 0 to _Allocations.Count - 1 do
    begin
        Work := _Allocations[ I ] ;
        if( ( S <= copy( Work, 10, 9 ) ) and ( L >= copy( Work, 1, 9 ) ) ) then
        begin
            Error( 'Overlapped allocation' ) ;
            exit ;
        end ;
    end ;
    _Allocations.Add( S + L ) ;
end ;


procedure TUOS_File_Heap.Remove_Allocation( Pos, Len : int64 ) ;

var I : integer ;
    S, L : string ;
    Work : string ;

begin
    S := inttostr( Pos ) ;
    while( length( S ) < 9 ) do
    begin
        S := '0' + S ;
    end ;
    L := inttostr( Pos + Len - 1 ) ;
    while( length( L ) < 9 ) do
    begin
        L := '0' + L ;
    end ;
    for I := 0 to _Allocations.Count - 1 do
    begin
        Work := _Allocations[ I ] ;
        if( S + L = Work ) then // Found
        begin
            _Allocations.Delete( I ) ;
            exit ;
        end ;
        if( ( S <= copy( Work, 10, 9 ) ) and ( L >= copy( Work, 1, 9 ) ) ) then
        begin
            Error( 'Overlapped deallocation' ) ;
            exit ;
        end ;
    end ;
    Error( 'Deallocating unallocated area' ) ;
end ;


procedure TUOS_File_Heap.Check_Access( Pos, Len : int64 ; R : boolean ) ;

var I : integer ;
    S, L : string ;
    Work : string ;

begin
    S := inttostr( Pos ) ;
    while( length( S ) < 9 ) do
    begin
        S := '0' + S ;
    end ;
    L := inttostr( Pos + Len - 1 ) ;
    while( length( L ) < 9 ) do
    begin
        L := '0' + L ;
    end ;
    for I := 0 to _Allocations.Count - 1 do
    begin
        Work := _Allocations[ I ] ;
        if( ( S <= copy( Work, 10, 9 ) ) and ( L >= copy( Work, 1, 9 ) ) ) then
        begin
            if( ( S < copy( Work, 1, 9 ) ) or ( L > copy( Work, 10, 9 ) ) ) then
            begin
                if( R ) then
                begin
                    Error( 'Read past allocated bounds' ) ;
                end else
                begin
                    Error( 'Write past allocated bounds' ) ;
                end ;
            end ;
            exit ;
        end ;
    end ;
    if( R ) then
    begin
        Error( 'Read from unallocated area' ) ;
    end else
    begin
        Error( 'Write to unallocated area' ) ;
    end ;
end ;


procedure Do_ReAllocation( Pos, Old_Len, New_Len : int64 ) ;

var I : integer ;
    S, O, N : string ;
    Work : string ;

begin
    S := inttostr( Pos ) ;
    while( length( S ) < 9 ) do
    begin
        S := '0' + S ;
    end ;
    O := inttostr( Pos + Old_Len - 1 ) ;
    while( length( O ) < 9 ) do
    begin
        O := '0' + O ;
    end ;
    N := inttostr( Pos + New_Len - 1 ) ;
    while( length( N ) < 9 ) do
    begin
        N := '0' + N ;
    end ;
    for I := 0 to _Allocations.Count - 1 do
    begin
        Work := _Allocations[ I ] ;
        if( Work = S + O ) then
        begin
            _Allocations[ I ] := S + N ;
            exit ;
        end ;
    end ;
    Error( 'Reallocation doesn''t match previous allocation' ) ;
end ;
{$ENDIF}


// Internal utility routines...

function TUOS_Managed_Store.AT : TAT64 ;

begin
    if( _AT = nil ) then
    begin
        _AT := TAT64.Create( Clustersize ) ;
        _AT.Set_Store( _Store ) ;
    end ;
    Result := _AT ;
end ;


function TUOS_Managed_Store.Clustersize : cardinal ;

begin
    if( ( _Clustersize = 0 ) and ( _Store <> nil ) ) then
    begin
        Result := _Store.Min_Storage ;
    end else
    begin
        Result := _Clustersize ;
    end ;
end ;


// Overrides...

function TUOS_Managed_Store.Is_Class( Name : PChar ) : boolean ;

begin
    Result := lowercase( Name ) = 'tuos_managed_store' ;
    if( not Result ) then
    begin
        Result := inherited Is_Class( Name ) ;
    end ;
end ;


function TUOS_Managed_Store.Read_Data( var Data ;
    Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Set_Last_Error( nil ) ;
    UEC := nil ;
    if( Address < Clustersize ) then
    begin
        // Have to use set/get root methods to read the first cluster
        UEC := Create_Exception( UOS_MStore_Error_Access_Violation, nil ) ;
        Set_Last_Error( UEC ) ;
        Result := 0 ;
        exit ;
    end ;
{$IFDEF CC_DEBUG}
    Check_Access( Address, _Size, True ) ;
{$ENDIF}
    Result := _Store.Read_Data( Data, Address, _Size, UEC ) ;
    if( UEC <> nil ) then
    begin
        UEC := Create_Exception( UOS_MStore_Error_IO_Failure, UEC ) ; { untested }
        Set_Last_Error( UEC ) ;
    end ;
end ;


function TUOS_Managed_Store.Write_Data( var Data ;
    Address, _Size : TStore_Address64 ;
    var UEC : TUnified_Exception ) : TStore_Address64 ;

begin
    Set_Last_Error( nil ) ;
    UEC := nil ;
    if( Address < Clustersize ) then
    begin
        // Have to use set/get root methods to read the first cluster
        UEC := Create_Exception( UOS_MStore_Error_Access_Violation, nil ) ;
        Set_Last_Error( UEC ) ;
        Result := 0 ;
        exit ;
    end ;
{$IFDEF CC_DEBUG}
    Check_Access( Address, _Size, False ) ;
{$ENDIF}
    Result := _Store.Write_Data( Data, Address, _Size, UEC ) ;
    if( UEC <> nil ) then
    begin
        UEC := Create_Exception( UOS_MStore_Error_IO_Failure, UEC ) ; { untested }
        Set_Last_Error( UEC ) ;
    end ;
end ;


function TUOS_Managed_Store.Max_Storage : TStore_Size64 ;

begin
    if( _Store = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := _Store.Max_Storage ;
    end ;
end ;


function TUOS_Managed_Store.Min_Storage : TStore_Address64 ;

begin
    if( _Store = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := _Store.Min_Storage ;
    end ;
end ;


function TUOS_Managed_Store.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    Set_Last_Error( nil ) ;
    if( _Store = nil ) then
    begin
        Result := -1 ; // Not extendable
    end else
    begin
        Result := _Store.Extend( Amount ) ;
        Set_Last_Error( _Store.Last_Error ) ;
        if( Amount <> 0 ) then
        begin
            inc( Statistics.Extensions ) ;
        end ;
    end ;
end ;


function TUOS_Managed_Store.Get_Read_Only : boolean ;

begin
    Set_Last_Error( nil ) ;
    if( _Store = nil ) then
    begin
        Result := True ;
    end else
    begin
        Result := _Store.Get_Read_Only ;
    end ;
end ;


function TUOS_Managed_Store.Get_Write_Only : boolean ;

begin
    if( _Store = nil ) then
    begin
        Result := False ;
    end else
    begin
        Result := _Store.Get_Write_Only ;
    end ;
end ;


procedure TUOS_Managed_Store.Format ;

begin
    Set_Last_Error( nil ) ;
    if( _Store <> nil ) then
    begin
        _Store.Format ;
        Set_Last_Error( _Store.Last_Error ) ;
    end ;
end ;


function TUOS_Managed_Store.Get_Name : PChar ;

begin
    if( _Store = nil ) then
    begin
        Result := nil ;
    end else
    begin
        Result := _Store.Get_Name ;
    end ;
end ;


function TUOS_Managed_Store.Get_Cache : TCOM_Cache64 ;

begin
    if( _Store = nil ) then
    begin
        Result := nil ;
    end else
    begin
        Result := _Store.Get_Cache ;
    end ;
end ;


procedure TUOS_Managed_Store.Set_Cache( Value : TCOM_Cache64 ) ;

begin
    if( _Store <> nil ) then
    begin
        _Store.Set_Cache( Value ) ;
    end ;
end ;


function TUOS_Managed_Store.Contiguous_Store : boolean ;

begin
    if( _Store = nil ) then
    begin
        Result := True ;
    end else
    begin
        Result := _Store.Contiguous_Store ;
    end ;
end ;


procedure TUOS_Managed_Store.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : TUnified_Exception ) ;

var C : int64 ; // Total clusters on store

begin
    Set_Last_Error( nil ) ;
    if( _Store = nil ) then
    begin
        Res := nil ;
    end else
    begin
        _Store.Set_Max_Storage( Value, Res ) ;
        if( Res <> nil ) then
        begin
            C := _Store.Max_Storage div _Store.Min_Storage ; { untested }
            AT.Set_Size( C div 8 ) ;
        end ;
        Set_Last_Error( _Store.Last_Error ) ;
    end ;
end ;


function TUOS_Managed_Store.Extended_Size : TStore_Address64 ;

begin
    Set_Last_Error( nil ) ;
    if( _Store = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := _Store.Extended_Size ;
        Set_Last_Error( _Store.Last_Error ) ;
    end ;
end ;


procedure TUOS_Managed_Store.Set_Read_Only( Value : boolean ) ;

begin
    if( _Store <> nil ) then
    begin
        _Store.Set_Read_Only( Value ) ;
    end ;
end ;


procedure TUOS_Managed_Store.Set_Write_Only( Value : boolean ) ;

begin
    if( _Store <> nil ) then
    begin
        _Store.Set_Write_Only( Value ) ;
    end ;
end ;


function TUOS_Managed_Store.Allocate( Size : TStore_Address64 ) : TStore_Address64 ;

var Calculating : boolean ;
    Original : int64 ;
    Previous_New_AT_Size : int64 ;
    Needed_Extension : int64 ; // Amount of extension needed
    New_AT_Size, Old_AT_Size : int64 ; // Number of bytes for new and old AT
    New_AT : int64 ;
    Res : TUnified_Exception ;
    Unused_Bits, Total_Bits : longint ;

begin
    Result := AT.Allocate( Size ) ;
{$IFDEF CC_DEBUG}
    if( Result <> 0 ) then
    begin
        Add_Allocation( Result, Size ) ;
    end ;
{$ENDIF}
    Set_Last_Error( AT.Last_Error ) ;
    if( ( Result = 0 ) and ( Extend( 0 ) = 0 ) ) then // Allocation failure on extendable store ////
    begin
        // Extend the store...
        New_AT := 0 ;
        Needed_Extension := ( Size + _Store.Min_Storage - 1 ) and not ( _Store.Min_Storage - 1 ) ;
        Old_AT_Size := ( _Store.Max_Storage div _Store.Min_Storage + 7 ) div 8 ;
        Old_AT_Size := ( Old_At_Size + _Store.Min_Storage - 1 ) and not ( _Store.Min_Storage - 1 ) ;
        New_AT_Size := ( ( _Store.Max_Storage + Needed_Extension ) div _Store.Min_Storage + 7 ) div 8 ;
        New_AT_Size := ( New_At_Size + _Store.Min_Storage - 1 ) and not ( _Store.Min_Storage - 1 ) ;
        if( Allocation_Table_Offset <> 0 ) then // If AT is persisted to the store
        begin
            Previous_New_AT_Size := 0 ; // So that only differential is added
            Calculating := True ; // Execute loop at least once
            while( Calculating ) do
            begin
                Calculating := False ; // Assume that one iteration is enough
                if( Old_AT_Size < New_AT_Size ) then // If new AT requires more clusters than the old
                begin
                    New_AT := AT.Reallocate( Allocation_Table_Offset, Old_AT_Size, New_AT_Size ) ;
                    // See if we can allocate in existing free space
                    if( New_AT = 0 ) then // No existing room for new AT
                    begin
                        Needed_Extension := Needed_Extension + New_AT_Size - Previous_New_AT_Size ;
                        // Room for AT also
                        Previous_New_AT_Size := New_AT_Size ;
                        New_AT_Size := ( _Store.Max_Storage + Needed_Extension )
                            div _Store.Min_Storage div 8 ;
                        New_AT_Size := ( New_At_Size + _Store.Min_Storage - 1 )
                            and not ( _Store.Min_Storage - 1 ) ; // Round up
                        Calculating := True ; // Calculate again
{$IFDEF CC_DEBUG}
                    end else
                    begin
                        Do_ReAllocation( Allocation_Table_Offset, Old_AT_Size, New_AT_Size ) ;
{$ENDIF}
                    end ;
                end ;
            end ; // while( Calculating )
        end ; // if( Allocation_Table_Offset <> 0 )
        if( ( New_AT_Size - Old_AT_Size ) * 8 * _Store.Min_Storage > Needed_Extension ) then
        begin
            Needed_Extension := ( New_AT_Size - Old_AT_Size ) * 8 * _Store.Min_Storage ;
        end ;
        Original := _Store.Max_Storage ;
        if( Original = Allocation_Table_Offset ) then // AT isn't yet physically written
        begin
            Original := Allocation_Table_Offset +
                Round_Up( AT.Size, _Store.Min_Storage ) ; // Start after AT
        end ;
        Original := Round_Up( Original, _Store.Min_Storage ) ;
        if( Extend( Needed_Extension ) >= Needed_Extension ) then // Successful
        begin
            // Adjust AT bits...
            Total_Bits := AT.Get_Size * 8 ;
            Unused_Bits := Total_Bits - ( Original div _Store.Min_Storage ) ;
            if( Unused_Bits > 0 ) then // Need to clear bits
            begin
                if( ( Needed_Extension + _Store.Min_Storage - 1 ) div _Store.Min_Storage < Unused_Bits ) then
                begin
                    Unused_Bits := ( Needed_Extension + _Store.Min_Storage - 1 )
                        div _Store.Min_Storage ;
                end ;
                AT.Deallocate( Original, Unused_Bits * _Store.Min_Storage ) ;
            end ;

            AT.Set_Size( ( ( _Store.Max_Storage div _Store.Min_Storage ) + 7 ) div 8 ) ;
            Total_Bits := AT.Get_Size * 8 ;
            Unused_Bits := Total_Bits -
                ( Round_Up( _Store.Max_Storage, _Store.Min_Storage ) div _Store.Min_Storage ) ;
            if( Unused_Bits > 0 ) then // Need to set bits mapping beyond end of store
            begin
                AT.Allocate_At( ( Total_Bits - Unused_Bits ) * _Store.Min_Storage,
                    Unused_Bits * _Store.Min_Storage ) ;
{$IFDEF CC_DEBUG}
                Add_Allocation( ( Total_Bits - Unused_Bits ) * _Store.Min_Storage,
                    Unused_Bits * _Store.Min_Storage ) ;
{$ENDIF}
            end ;

            // Flush allocation table...
            if( New_AT = 0 ) then // AT wasn't already reallocated
            begin
                New_AT := AT.Reallocate( Allocation_Table_Offset, Old_AT_Size, New_AT_Size ) ;
                if( New_AT = 0 ) then // No existing room for new AT
                begin
                    Result := 0 ;
                    Set_Last_Error( Create_Exception( UOS_AT_Resize_Failure, nil ) ) ;
                    _Store.Set_Max_Storage( Original, Res ) ;
                    exit ;
                end ;
{$IFDEF CC_DEBUG}
                Do_ReAllocation( Allocation_Table_Offset, Old_AT_Size, New_AT_Size ) ;
{$ENDIF}
            end ;
            Allocation_Table_Offset := New_AT ;
            AT.Flush ;

            // Try allocation again...
            Result := AT.Allocate( Size ) ;
{$IFDEF CC_DEBUG}
            if( Result <> 0 ) then
            begin
                Add_Allocation( Result, Size ) ;
            end ;
{$ENDIF}
            Set_Last_Error( AT.Last_Error ) ;
        end ; // if
    end ; // if
    if( Last_Error = nil ) then
    begin
        inc( Statistics.Allocations ) ;
        Statistics.Bytes_Allocated := Statistics.Bytes_Allocated + Size ;
    end ;
end ; // TUOS_Managed_Store.Allocate


function TUOS_Managed_Store.Allocate_At( Offset,
    Size : TStore_Address64 ) : boolean ;

begin
    Result := AT.Allocate_At( Offset, Size ) ;
    Set_Last_Error( AT.Last_Error ) ;
    if( Last_Error = nil ) then
    begin
{$IFDEF CC_DEBUG}
        if( Result <> 0 ) then
        begin
            Add_Allocation( Offset, Size ) ;
        end ;
{$ENDIF}
        inc( Statistics.Allocations ) ;
        Statistics.Bytes_Allocated := Statistics.Bytes_Allocated + Size ;
    end ;
end ;


procedure TUOS_Managed_Store.Copy( Source, Destination : TStore_Address64 ;
    Count : int64 ) ;

var Buffer : pointer ;
    UEC : TUnified_Exception ;

begin
    // Setup...
    Set_Last_Error( nil ) ;
    if( ( Count <= 0 ) or ( Source = Destination ) ) then
    begin
        exit ;
    end ;
    Buffer := allocmem( _Clustersize ) ;
    try
        Count := Count div _Clustersize ; // Number of clusters to copy
        Source := ( Source div Clustersize ) * Clustersize ;
        Destination := ( Destination div Clustersize ) * Clustersize ;
        if( Source < Destination ) then
        begin
            Source := Source + ( Count - 1 ) * Clustersize ;
            Destination := Destination + ( Count - 1 ) * Clustersize ;
        end ;

        // Do the copy...
        while( Count > 0 ) do
        begin
            dec( Count ) ;
{$IFDEF CC_DEBUG}
            Check_Access( Source, _Clustersize, True ) ;
{$ENDIF}
            _Store.Read_Data( Buffer^, Source, _Clustersize, UEC ) ;
            if( UEC <> nil ) then
            begin
                Set_Last_Error( UEC ) ; { untested }
                exit ;
            end ;
            {_Store.}Write_Data( Buffer^, Destination, _Clustersize, UEC ) ;
            if( UEC <> nil ) then
            begin
                Set_Last_Error( UEC ) ; { untested }
                exit ;
            end ;
            if( Source < Destination ) then
            begin
                Source := Source - Clustersize ;
                Destination := Destination - Clustersize ;
            end else
            begin
                Source := Source + Clustersize ;
                Destination := Destination + Clustersize ;
            end ;
        end ; // while( Count > 0 )

    finally
        // Clean up and update statistics...
        freemem( Buffer ) ;
    end ;
    inc( Statistics.Copies ) ;
    Statistics.Bytes_Copied := Statistics.Bytes_Copied + Count ;
end ; // TUOS_Managed_Store.Copy


procedure TUOS_Managed_Store.Deallocate( PTR, Size : TStore_Address64 ) ;

begin
    AT.Deallocate( PTR, Size ) ;
    Set_Last_Error( AT.Last_Error ) ;
    if( Last_Error = nil ) then
    begin
{$IFDEF CC_DEBUG}
        if( Result <> 0 ) then
        begin
            Remove_Allocation( PTR, Size ) ;
        end ;
{$ENDIF}
        inc( Statistics.Deallocations ) ;
        Statistics.Bytes_Deallocated := Statistics.Bytes_Deallocated + Size ;
    end ;
end ;


procedure TUOS_Managed_Store.Fill( PTR : TStore_Address64 ; Count : integer ;
    Value : byte ) ;

var Buffer : pointer ;
    Original : integer ;

begin
    Set_Last_Error( nil ) ;
    if( ( PTR < Clustersize ) or ( Count < 1 ) ) then
    begin
        exit ;
    end ;
    PTR := ( PTR div Clustersize ) * Clustersize ;
    Count := ( Count div Clustersize ) * Clustersize ;
    Original := Count ;
    Buffer := allocmem( Clustersize ) ;
    try
        fillchar( Buffer^, Clustersize, Value ) ;
        while( Count > 0 ) do
        begin
            Write( PTR, Clustersize, Buffer^ ) ;
            if( Last_Error <> nil ) then
            begin
                exit ; { untested }
            end ;
            PTR := PTR + Clustersize ;
            Count := Count - Clustersize ;
        end ;
    finally
        freemem( Buffer ) ;
    end ;
    inc( Statistics.Fills ) ;
    Statistics.Bytes_Filled := Statistics.Bytes_Filled + Original ;
end ;


function TUOS_Managed_Store.Reallocate( PTR, Old,
    New : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := 0 ;
    Set_Last_Error( nil ) ;
    if( ( PTR < Clustersize ) or ( Old < 1 ) or ( New < 1 ) ) then
    begin
        exit ;
    end ;

    PTR := ( PTR div Clustersize ) * Clustersize ;
    Old := ( Old div Clustersize ) * Clustersize ;
    New := ( New div Clustersize ) * Clustersize ;
    Result := PTR ;
    if( Old = New ) then // No change
    begin
        exit ;
    end ;
    if( New < Old ) then // Shrinking
    begin
        AT.Deallocate( PTR + New, Old - New ) ;
{$IFDEF CC_DEBUG}
        Remove_Allocation( PTR + New, Old - New ) ;
{$ENDIF}
        Set_Last_Error( AT.Last_Error ) ;
        inc( Statistics.Reallocations_In_Place ) ;
        exit ;
    end ;

    // Expand...
    if( not AT.Allocate_At( PTR + Old, New - Old ) ) then
    begin
        // Cannot expand in-place, so allocate a new area, copy old to new, and deallocate old
        Result := AT.Allocate( New ) ;
        if( Result = 0 ) then
        begin
            Result := PTR ; { untested }
            Set_Last_Error( AT.Last_Error ) ;
        end else
        begin
            Copy( PTR, Result, Old ) ;
            AT.Deallocate( PTR, Old ) ;
{$IFDEF CC_DEBUG}
            Remove_Allocation( PTR, Old ) ;
{$ENDIF}
            Set_Last_Error( AT.Last_Error ) ;
            inc( Statistics.Reallocations ) ;
        end ;
    end else
    begin
{$IFDEF CC_DEBUG}
        if( Result <> 0 ) then
        begin
            Add_Allocation( PTR + Old, New - Old ) ;
        end ;
{$ENDIF}
        inc( Statistics.Reallocations_In_Place ) ;
    end ;
end ; // TUOS_Managed_Store.Reallocate


function TUOS_Managed_Store.SpaceAvail : TStore_Address64 ;

begin
    Result := AT.SpaceAvail ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


function TUOS_Managed_Store.XSpaceAvail : TStore_Address64 ;

begin
    Result := SpaceAvail ;
end ;


function TUOS_Managed_Store.MaxSpace : TStore_Address64 ;

begin
    Result := AT.MaxSpace ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


procedure TUOS_Managed_Store.Get_Root( var Buf ) ;

var UEC : TUnified_Exception ;

begin
    _Store.Read_Data( Buf, 0, Clustersize, UEC ) ;
    Set_Last_Error( UEC ) ;
    inc( Statistics.Root_Reads ) ;
end ;


procedure TUOS_Managed_Store.Set_Root( var Buf ) ;

var UEC : TUnified_Exception ;

begin
    _Store.Write_Data( Buf, 0, Clustersize, UEC ) ;
    Set_Last_Error( UEC ) ;
    inc( Statistics.Root_Writes ) ;
end ;


function TUOS_Managed_Store.Allocation_Table_Size : longint ;

begin
    Result := AT.Get_Size ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


procedure TUOS_Managed_Store.Get_Allocation_Table_Data( Buff : pointer ;
    _Size : longint ) ;

begin
    move( AT.Get_Data^, Buff^, _Size ) ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


procedure TUOS_Managed_Store.Set_Allocation_Table_Data( Buff : pointer ;
    _Size : longint ) ;

begin
    AT.Set_Data( Buff, _Size ) ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


procedure TUOS_Managed_Store.Set_Min_Storage( _Size : longint ) ;

begin
    Set_Last_Error( nil ) ;
    _Clustersize := _Size ;
    if( _Store <> nil ) then
    begin
        if( _Clustersize < _Store.Min_Storage ) then { else untested }
        begin
            _Clustersize := _Store.Min_Storage ;
        end ;
        AT.Resolution := _Clustersize ;
    end ;
end ;


procedure TUOS_Managed_Store.Set_Allocation_Table_Offset( Value : TStore_Address64 ) ;

begin
    AT.Set_Offset( Value ) ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


function TUOS_Managed_Store.Get_Allocation_Table_Offset : TStore_Address64 ;

begin
    Result := AT.Get_Offset ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


procedure TUOS_Managed_Store.Flush_Allocation_Table ;

begin
    AT.Flush ;
    Set_Last_Error( AT.Last_Error ) ;
end ;


function TUOS_Managed_Store.Get_Store : TCOM_Store64 ;

begin
    Result := _Store ;
end ;


procedure TUOS_Managed_Store.Set_Store( Value : TCOM_Store64 ) ;

begin
    Set_Last_Error( nil ) ;
    if( _AT <> nil ) then
    begin
        _AT.Flush ;
    end ;
    if( Value <> nil ) then
    begin
        //Value.Attach ; // AT.Set_Store also attaches to Value, creating a deadly embrace
        if( _Clustersize < Value.Min_Storage ) then { else untested }
        begin
            _Clustersize := Value.Min_Storage ;
        end ;
        AT.Set_Store( Value ) ;
        AT.Set_Resolution( _Clustersize ) ;
        Set_Last_Error( AT.Last_Error ) ;
    end ;
    if( _Store <> nil ) then
    begin
        if( Value = nil ) then
        begin
            AT.Set_Store( nil ) ;
        end ;
//        _Store.Detach ; // See above comment about deadly embrace
    end ;
    _Store := Value ;
end ;


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    halt ;
end ;


var F : text ;
    Last_Op : string ;

procedure Test_Unit( T : TTest_Interface ) ;

    function Log( const S : string ) : boolean ;

    begin
        Last_Op := S ;
        append( F ) ;
        writeln( F, S ) ;
        closefile( F ) ;
        Result := T.Log_Operation( S ) ;
    end ;


var Allocations : array[ 1..1023 ] of boolean ;

    function Find_Allocated : integer ;

    begin
        Result := random( 1022 ) + 1 ;
        while( not Allocations[ Result ] ) do
        begin
            inc( Result ) ;
            if( Result > 1022 ) then
            begin
                Result := -1 ;
                exit ;
            end ;
        end ;
        Result := Result * 1024 ;
    end ;


    function Find_Unallocated : integer ;

    begin
        Result := random( 1022 ) + 1 ;
        while( Allocations[ Result ] ) do
        begin
            inc( Result ) ;
            if( Result > 1022 ) then
            begin
                Result := -1 ;
                exit ;
            end ;
        end ;
        Result := Result * 1024 ;
    end ;

var Buffer : array[ 0..511 ] of byte ;
    Disk : TRAM_Disk64 ;
    Loop : integer ;
    P, P1 : TStore_Address64 ;
    Res : TUnified_Exception ;
    Test : TUOS_Managed_Store ;

begin
    Disk := TRAM_Disk64.Create ;
    Disk.Set_Max_Storage( 10485760, Res ) ;
    Test := TUOS_Managed_Store.Create ;
    Test.Format ;
    if( not Test.Is_Class( 'tuos_managed_store' ) ) then
    begin
        Error ;
    end ;
    if( Test.Is_Class( 'moo' ) ) then
    begin
        Error ;
    end ;
    Test.Set_Min_Storage( 1 ) ;
    Test.Set_Max_Storage( 10485760, Res ) ;
    Test.Set_Read_Only( True ) ;
    Test.Set_Write_Only( True ) ;
    if( Test.Extended_Size <> 0 ) then
    begin
        Error ;
    end ;
    if( not Test.Contiguous_Store ) then
    begin
        Error ;
    end ;
    if( Test.Get_Name <> nil ) then
    begin
        Error ;
    end ;
    if( not Test.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test.Get_Write_Only ) then
    begin
        Error ;
    end ;
    if( Test.Extend( 104857600 ) <> 0 ) then
    begin
        Error ;
    end ;
    Test.Set_Cache( nil ) ;
    if( Test.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Max_Storage <> 0 ) then
    begin
        Error ;
    end ;
    if( Res <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Get_Store <> nil ) then
    begin
        Error ;
    end ;
    Test.Set_Store( Disk ) ;
    if( Test.Get_Store = nil ) then
    begin
        Error ;
    end ;
    Test.Format ;
    if( Test.Min_Storage <> 512 ) then
    begin
        Error ;
    end ;
    Test.Set_Min_Storage( 1 ) ;
    if( Test.Min_Storage <> 512 ) then
    begin
        Error ;
    end ;
    if( not Test.Contiguous_Store ) then
    begin
        Error ;
    end ;
    Test.Set_Max_Storage( 104857600, Res ) ; // 10 Mb
    if( Res <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Max_Storage <> 104857600 ) then
    begin
        Error ;
    end ;
    Test.Set_Max_Storage( 10485760, Res ) ; // 10 Mb
    if( Res <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Extended_Size <> 10485760 ) then
    begin
        Error ;
    end ;
    if( Test.Extend( 104857600 ) <> 10485760 ) then
    begin
        Error ;
    end ;
    Test.Set_Read_Only( True ) ;
    Test.Set_Write_Only( True ) ;
    if( Test.Get_Read_Only ) then
    begin
        Error ;
    end ;
    if( Test.Get_Write_Only ) then
    begin
        Error ;
    end ;
    if( Test.Read_Data( Buffer, 512, 512, Res ) <> 512 ) then
    begin
        Error ;
    end ;
    if( Res <> nil ) then
    begin
        Error ;
    end ;
    if( Test.SpaceAvail <> 10485760 - 512 ) then
    begin
        Error ;
    end ;
    if( Test.XSpaceAvail <> 10485760 - 512 ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace <> 10485760 - 512 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Name <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Allocation_Table_Size <> 2560 ) then
    begin
        Error ;
    end ;
    P := ( Test.Max_Storage + Test.Allocation_Table_Size ) div 2 ;
    P := ( P + 511 ) and ( not 511 ) ;
    if( Test.Allocate_At( P, Test.Allocation_Table_Size ) ) then
    begin
        Test.Set_Allocation_Table_Offset( P ) ;
    end else
    begin
        Error ;
    end ;
    if( Test.Get_Allocation_Table_Offset <> P ) then
    begin
        Error ;
    end ;
    Test.Set_Cache( nil ) ;
    if( Test.Get_Cache <> nil ) then
    begin
        Error ;
    end ;
    Test.Get_Allocation_Table_Data( @Buffer, sizeof( Buffer ) ) ;
    Test.Set_Allocation_Table_Data( @Buffer, sizeof( Buffer ) ) ;
    fillchar( Buffer, sizeof( Buffer ), 0 ) ;
    Test.Set_Root( Buffer ) ;
    fillchar( Buffer, sizeof( Buffer ), 1 ) ;
    Test.Get_Root( Buffer ) ;
    for Loop := 0 to sizeof( Buffer ) - 1 do
    begin
        if( Buffer[ Loop ] <> 0 ) then
        begin
            Error ;
        end ;
    end ;
    P := Test.Allocate( 512 ) ;
    if( P <> 512 ) then
    begin
        Error ;
    end ;
    Test.Fill( 512, 512, 1 ) ;
    P := Test.Allocate( 512 ) ;
    if( P <> 1024 ) then
    begin
        Error ;
    end ;
    Test.Fill( 1024, 1024, 2 ) ;
    if( Test.Reallocate( 1024, 512, 1024 ) <> 1024 ) then
    begin
        Error ;
    end ;
    if( Test.Reallocate( 512, 512, 1024 ) <> 2048 ) then
    begin
        Error ;
    end ;
    P := Test.SpaceAvail ;
    Test.Deallocate( 2048, 1024 ) ;
    if( Test.SpaceAvail <> P + 1024 ) then
    begin
        Error ;
    end ;
    P := Test.Allocate( 2048 ) ;
    fillchar( Buffer, sizeof( Buffer ), 3 ) ;
    Test.Write_Data( Buffer, P, 512, Res ) ;
    if( Res <> nil ) then
    begin
        Error ;
    end ;
    fillchar( Buffer, sizeof( Buffer ), 4 ) ;
    Test.Write_Data( Buffer, P + 512, 512, Res ) ;
    if( Res <> nil ) then
    begin
        Error ;
    end ;
    fillchar( Buffer, sizeof( Buffer ), 5 ) ;
    Test.Write_Data( Buffer, P + 1024, 512, Res ) ;
    if( Res <> nil ) then
    begin
        Error ;
    end ;
    fillchar( Buffer, sizeof( Buffer ), 6 ) ;
    Test.Write( P + 1536, 512, Buffer ) ;
    Test.Read( P, 512, Buffer ) ;
    for Loop := 0 to sizeof( Buffer ) - 1 do
    begin
        if( Buffer[ Loop ] <> 3 ) then
        begin
            Error ;
        end ;
    end ;
    Test.Read( P, 512, Buffer ) ;
    for Loop := 0 to sizeof( Buffer ) - 1 do
    begin
        if( Buffer[ Loop ] <> 3 ) then
        begin
            Error ;
        end ;
    end ;
    Test.Copy( P, P + 1024, 1024 ) ;
    Test.Read( P, 512, Buffer ) ;
    for Loop := 0 to sizeof( Buffer ) - 1 do
    begin
        if( Buffer[ Loop ] <> 3 ) then
        begin
            Error ;
        end ;
    end ;
    Test.Read( P + 512, 512, Buffer ) ;
    for Loop := 0 to sizeof( Buffer ) - 1 do
    begin
        if( Buffer[ Loop ] <> 4 ) then
        begin
            Error ;
        end ;
    end ;
    Test.Copy( P + 512, P, 1024 ) ;
    Test.Read( P, 512, Buffer ) ;
    for Loop := 0 to sizeof( Buffer ) - 1 do
    begin
        if( Buffer[ Loop ] <> 4 ) then
        begin
            Error ;
        end ;
    end ;
    Test.Read( 0, 512, Buffer ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Write( 0, 512, Buffer ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Copy( -1, 0, -1 ) ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Test.Fill( -1, -1, 0 ) ;
    if( Test.Reallocate( -1, -1, -1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Reallocate( 512, 512, 512 ) <> 512 ) then
    begin
        Error ;
    end ;
    if( Test.Reallocate( P, 2048, 1024 ) <> P ) then
    begin
        Error ;
    end ;

    Test.Flush_Allocation_Table ;
    Test.Free ;

    // Stress test...
    assignfile( F, 'k:\test.txt' ) ;
    rewrite( F ) ;
    closefile( F ) ;
    Disk := TRAM_Disk64.Create ;
    Disk.Set_Max_Storage( 1048576, Res ) ;
    fillchar( Allocations, sizeof( Allocations ), 0 ) ;
    Test := TUOS_Managed_Store.Create ;
    Test.Set_Store( Disk ) ;
    Test.Allocate( 512 ) ;
    while( true ) do
    begin
        case random( 17 ) of
            0: // Get_Root
               begin
                   if( Log( 'GR' ) ) then
                   begin
                       exit ;
                   end ;
                   Test.Get_Root( Buffer ) ;
               end ;
            1: // Set_Root
               begin
                   if( Log( 'SR' ) ) then
                   begin
                       exit ;
                   end ;
                   fillchar( Buffer, 512, 0 ) ;
                   Test.Set_Root( Buffer ) ;
                   fillchar( Buffer, 512, 1 ) ;
                   Test.Get_Root( Buffer ) ;
                   if( Buffer[ 0 ] <> 0 ) then
                   begin
                       Error ;
                   end ;
               end ;
            2: // Reallocate
               begin
                   P := Find_Allocated ;
                   if( P < 0 ) then
                   begin
                       continue ;
                   end ;
                   if( Log( 'RA' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   P1 := Test.Reallocate( P, 1024, 1536 ) ;
                   if( P1 = 0 ) then
                   begin
                       if( Test.MaxSpace >= 1536 ) then
                       begin
                           Error ;
                       end ;
                   end else
                   if( P <> P1 ) then
                   begin
                       Allocations[ P div 1024 ] := False ;
                       Allocations[ P1 div 1024 ] := True ;
                       Test.Fill( P1, 1024, P1 and 255 ) ;
                   end ;
                   P := Test.Reallocate( P1, 1536, 1024 ) ;
                   if( P1 <> P ) then
                   begin
                       Error ;
                   end ;
               end ;
            3: // Read
               begin
                   P := Find_Allocated ;
                   if( P < 0 ) then
                   begin
                       continue ;
                   end ;
                   if( Log( 'R' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   if( Test.Read( P, 512, Buffer ) <> 512 ) then
                   begin
                       Error ;
                   end ;
                   if( Buffer[ 0 ] <> P and 255 ) then
                   begin
                       Error ;
                   end ;
               end ;
            4: // Write
               begin
                   P := Find_Allocated ;
                   if( P < 0 ) then
                   begin
                       continue ;
                   end ;
                   if( Log( 'W' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   fillchar( Buffer, sizeof( Buffer ), P and 255 ) ;
                   if( Test.Write( P, 512, Buffer ) <> 512 ) then
                   begin
                       Error ;
                   end ;
               end ;
            5: // Allocate
               begin
                   P := Test.Allocate( 1024 ) ;
                   if( Log( 'A=' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   if( P = 0 ) then
                   begin
                       if( Test.MaxSpace >= 1536 ) then
                       begin
                           Error ;
                       end ;
                   end else
                   begin
                       Allocations[ P div 1024 ] := True ;
                   end ;
               end ;
            6: // Allocate_At
               begin
                   P := Find_Unallocated ;
                   if( P < 0 ) then
                   begin
                       continue ;
                   end ;
                   if( Log( 'AA' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   if( not Test.Allocate_At( P, 1024 ) ) then
                   begin
                       Test.Allocate_At( P, 1024 ) ;
                       Error ;
                   end ;
                   Allocations[ P div 1024 ] := True ;
                   Test.Fill( P, 1024, P and 255 ) ;
               end ;
            7: // Copy
               begin
                   P := Find_Allocated ;
                   if( P < 0 ) then
                   begin
                       continue ;
                   end ;
                   P1 := Find_Allocated ;
                   if( P1 < 0 ) then
                   begin
                       continue ;
                   end ;
                   if( Log( 'C' ) ) then
                   begin
                       exit ;
                   end ;
                   Test.Copy( P, P1, 1024 ) ;
                   Test.Read( P, 512, Buffer ) ;
                   if( Buffer[ 0 ] <> P and 255 ) then
                   begin
                       Error ;
                   end ;
                   Test.Fill( P1, 1024, P1 and 255 ) ;
               end ;
            8: // Deallocate
               begin
                   P := Find_Allocated ;
                   if( P < 0 ) then
                   begin
                       continue ;
                   end ;
                   if( Log( 'D' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   Test.Deallocate( P, 1024 ) ;
                   if( Test.Last_Error <> nil ) then
                   begin
                       Error ;
                   end ;
                   Allocations[ P div 1024 ] := False ;
               end ;
            9: // Fill
               begin
                   P := Find_Allocated ;
                   if( P < 0 ) then
                   begin
                       continue ;
                   end ;
                   if( Log( 'F' + inttostr( P ) ) ) then
                   begin
                       exit ;
                   end ;
                   Test.Fill( P, 1024, P and 255 ) ;
               end ;
            10: // SpaceAvail
                begin
                    if( Log( 'S' ) ) then
                    begin
                        exit ;
                    end ;
                    Test.SpaceAvail ;
                end ;
            11: // MaxSpace
                begin
                    if( Log( 'M' ) ) then
                    begin
                        exit ;
                    end ;
                    Test.MaxSpace ;
                end ;
        end ;
        if( Test.Last_Error <> nil ) then
        begin
            Error ;
        end ;
    end ;

    Test.Free ;
end ; // Test_Unit
{$ENDIF}


end.
