{
        Program Name : UOS_String_Table
        Package Name : String Table class
        Purpose      : Manage a string table
        Institution  : Conroy & Conroy Co.
        Date Written : 7-Nov-2015
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        This class manages a string table.
}

unit UOS_String_Table ;

interface

uses // C&C Subroutine Library - Layer 0...
     COMInter, // TCommon_COM_Interface
{$IFDEF CC_Regression_Test}
     _TestInt, // TTest_Interface
{$ENDIF}

     // Subroutine Library - layer 1...
     CommonUt, // TInteger_List

     // Subroutine Library - layer 2...
     UOS_File, // TUOS_Native_File

     // Subroutine Library - layer 3...
     Allocatr64 ; // TAT64

type TUOS_String_Table = class( TCommon_COM_Interface )
                             public // Constructors and destructors...
                                constructor Create ;
                                destructor Destroy ; override ;

                             private // Instance data...
                                 _AT : TAT64 ;
                                 _Buffer : PChar ; // Data when no Data_File has been provided
                                 _Buffer_Length : longint ; // Length of _Buffer
                                 _Index : TInteger_List ;
                                 _Position : longint ;
                                 _Dirty_Data : boolean ;
                                 _Dirty_Index : boolean ;
                                 _Data_File : TUOS_Native_File ;
                                 _Index_File : TUOS_Native_File ;
                                 _AT_File : TUOS_Native_File ;

                             public // Property handlers...
                                 function Get_AT_FIle : TUOS_Native_File ;
                                 procedure Set_AT_File( Value : TUOS_Native_File ) ;
                                 function Get_Data_File : TUOS_Native_File ;
                                 procedure Set_Data_File( Value : TUOS_Native_File ) ;
                                 function Get_Index_File : TUOS_Native_File ;
                                 procedure Set_Index_File( Value : TUOS_Native_File ) ;

                             private // Internal utility routines...
                                 procedure Add_Space( Value : cardinal ) ;
                                 function Indexof( const Value : string ) : longint ;
                                 procedure Read( var Data ; Size : longint ) ;
                                 function Read_Longint : longint ;
                                 procedure Seek( Value : longint ) ;
                                 procedure Update ;
                                 procedure Write( const Value : string ) ;
                                 procedure Write_Longint( Value : longint ) ;

                             public // API...
                                 function Is_Class( Name : PChar ) : boolean ;
                                     override ;

                                 function Add_String( Value : string ) : longint ;
                                 procedure Delete_String( Index : longint ) ;
                                 function Get_String( Value : longint ) : string ;

                                 property AT_File : TUOS_Native_File
                                      read Get_AT_FIle
                                      write Set_AT_File ;
                                 property Data_File : TUOS_Native_File
                                     read Get_Data_File
                                     write Set_Data_File ;
                                 property Index_File : TUOS_Native_File
                                     read Get_Index_File
                                     write Set_Index_File ;
                         end ; // TUOS_String_Table


{$IFDEF CC_Regression_Test}
procedure Test_Unit( T : TTest_Interface ) ;
{$ENDIF}

const UOS_String_Table_Error_Invalid_String_Index = 1 ;


implementation

uses // Pascal...
     Sysutils, // inttostr
{$IFDEF CC_Regression_Test}
     Classes, // TStringList
     Forms, // Application
{$ENDIF}
{$IFDEF CC_Track_Resources}
     ResMon,
{$ENDIF}

     // C&C Subroutine Library - Layer 0...
     _UE, // tUnified_Exception

     // C&C Subroutine Library - Layer 1...
     UE  ; // TStandard_Error_Interface

function ERT( C : longint ) : string ;

begin
    case C of
        UOS_String_Table_Error_Invalid_String_Index : Result := 'Invalid string index' ;
        else Result := 'Unknown error' ;
    end ;
end ;


type tUOS_String_Table_Unified_Exception = class( TStandard_Error_Interface )
                                               private
                                                   _Code : longint ;
                                                   _Text : string ;
                                                   _Previous : tUnified_Exception ;

                                               public
                                                   function Get_Facility : longint ; override ;
                                                   function Get_Facility_Version : longint ;
                                                       override ;
                                                   function Version : longint ; override ;
                                                   function Severity : longint ; override ;
                                                   function Error_Text( var Size, Typ : longint ) : PChar ;
                                                       override ;
                                                   function Get_Error : longint ; override ;
                                                   function Get_Previous : tUnified_Exception ;
                                                       override ;
                                                   procedure Terminate ; override ;
                                           end ;


function tUOS_String_Table_Unified_Exception.Get_Facility : longint ;

begin
    Result := 139 ;
end ;


function tUOS_String_Table_Unified_Exception.Get_Facility_Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_String_Table_Unified_Exception.Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_String_Table_Unified_Exception.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function tUOS_String_Table_Unified_Exception.Error_Text( var Size, Typ : longint ) : PChar ;

begin
    _Text := ERT( _Code ) ;
    Size := length( _Text ) ;
    Typ := 0 ;
    Result := PChar( _Text ) ;
end ;


function tUOS_String_Table_Unified_Exception.Get_Error : longint ;

begin
    Result := _Code ;
end ;


function tUOS_String_Table_Unified_Exception.Get_Previous : tUnified_Exception ;

begin
    Result := _Previous ;
end ;


procedure tUOS_String_Table_Unified_Exception.Terminate ;

begin
    Free ;
end ;



function Create_Exception( C : longint ; E : TUnified_Exception ) : tUOS_String_Table_Unified_Exception ;

begin
    Result := tUOS_String_Table_Unified_Exception.Create ;
    Result._Code := C ;
    Result._Previous := E ;
end ;



// TUOS_String_Table methods...

// Constructors and destructors...

constructor TUOS_String_Table.Create ;

begin
    inherited Create ;

    _Index := TInteger_List.Create ;
    _Buffer := allocmem( 64 ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    _Buffer_Length := 64 ;
    _Dirty_Data := True ;
    _AT := TAT64.Create( 8 ) ;
    _AT.Size := 1 ;
end ;


destructor TUOS_String_Table.Destroy ;

begin
    Update ;
    if( _AT <> nil ) then
    begin
        _AT.Free ;
        _AT := nil ;
    end ;
    Data_File := nil ;
    Index_File := nil ;
    AT_File := nil ;
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    freemem( _Buffer ) ;
    _Buffer := nil ;

    inherited Destroy ;
end ;


// Property handlers...

function TUOS_String_Table.Get_AT_File : TUOS_Native_File ;

begin
    Result := _AT_File ;
end ;


procedure TUOS_String_Table.Set_AT_File( Value : TUOS_Native_File ) ;

var S : longint ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
        S := Value.Header.Size ;
        if( S = 0 ) then // File is empty
        begin
            Value.Write( 0, 0, _AT.Size, _AT.Get_Data^ ) ;
        end else
        begin
            _AT.Set_Size( S ) ; // Number of bytes for allocation table
            Value.Read( 0, 0, S, _AT.Get_Data^ ) ;
            _AT.Find_Lowest_Space ;
        end ;
    end ;
    if( _AT_File <> nil ) then
    begin
        _AT_File.Detach ;
    end ;
    _AT_File := Value ;
end ;


function TUOS_String_Table.Get_Data_File : TUOS_Native_File ;

begin
    Result := _Data_File ;
end ;


procedure TUOS_String_Table.Set_Data_File( Value : TUOS_Native_File ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
        if( Value.Get_Stream_Size( 0 ) = 0 ) then // File is empty
        begin
            Value.Write( 0, 0, _Buffer_Length, _Buffer^ ) ;
        end ;
{$IFDEF CC_Track_Resources}
        Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        freemem( _Buffer ) ;
        _Buffer := nil ;
        _Dirty_Data := False ;
    end else
    begin
        _Buffer := allocmem( 64 ) ;
{$IFDEF CC_Track_Resources}
        Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        _Buffer_Length := 64 ;
        _Dirty_Data := True ;
    end ;
    if( _Data_File <> nil ) then
    begin
        _Data_File.Detach ;
    end ;
    _Data_File := Value ;
end ;


function TUOS_String_Table.Get_Index_File : TUOS_Native_File ;

begin
    Result := _Index_File ;
end ;


procedure TUOS_String_Table.Set_Index_File( Value : TUOS_Native_File ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
        if( Value.Header.EOF = 0 ) then // New file
        begin
            Value.Write( 0, 0, _Index.Count * 4, _Index.List^ ) ;
        end else
        begin
            _Index.Count := Value.Header.EOF div 4 ;
            Value.Read( 0, 0, _Index.Count * 4, _Index.List^ ) ;
        end ;
        _Dirty_Index := False ;
    end ;
    if( _Index_File <> nil ) then
    begin
        _Index_File.Detach ;
    end ;
    _Index_File := Value ;
end ;


// Internal utility routines...

function TUOS_String_Table.Indexof( const Value : string ) : longint ;

var Low, High : longint ;
    This_Value : string ;

begin
    Result := 0 ;
    if( _Index.Count < 2 ) then
    begin
        exit ;
    end ;
    Low := 0 ;
    High := _Index.Count - 1 ;
    while( True ) do
    begin
        if( High <= Low ) then
        begin
            Result := Low ;
            exit ;
        end ;
        Result := ( High - Low ) div 2 + Low ; // Start in middle
        This_Value :=  Get_String( _Index[ Result ] ) ;
        if( This_Value = Value ) then // Found our target
        begin
            // Return first instance of a value, in case there are duplicates
            while(
                   ( Result >= 0 )
                   and
                   ( Get_String( _Index[ Result - 1 ] ) = Value )
                 ) do
            begin
                dec( Result ) ;
            end ;
            exit ;
        end else
        if( This_Value < Value ) then
        begin
            Low := Result + 1 ;
        end else
        begin
            High := Result - 1 ;
        end ;
    end ; // while( True )
end ; // TUOS_String_Table.Indexof


procedure TUOS_String_Table.Read( var Data ; Size : longint ) ;

begin
    if( Data_File = nil ) then
    begin
        move( _Buffer[ _Position ], Data, Size ) ;
    end else
    begin
        Data_File.Read( 0, _Position, Size, Data ) ;
    end ;
    _Position := _Position + Size ;
end ;


function TUOS_String_Table.Read_Longint : longint ;

begin
    Read( Result, sizeof( Result ) ) ;
end ;


procedure TUOS_String_Table.Seek( Value : longint ) ;

var P : cardinal ;

begin
    _Position := Value ;
    if( Data_File = nil ) then
    begin
        if( _Position > _Buffer_Length ) then
        begin
            _Position := _Buffer_Length ;
        end ;
    end else
    begin
        P := Data_File.Get_Stream_Size( 0 ) ;
        if( _Position > P ) then
        begin
            _Position := P ;
        end ;
    end ;
end ;


procedure TUOS_String_Table.Update ;

begin
    if( _Dirty_Data ) then
    begin
        if( Data_File <> nil ) then
        begin
            _Dirty_Data := False ;
        end ;
    end ;
    if( _Dirty_Index ) then
    begin
        if( _Index_File <> nil ) then
        begin
            _Dirty_Index := False ;
            _Index_File.Write( 0, 0, _Index.Count * 4, _Index.List^ ) ;
        end ;
    end ;
    if( _AT.Modified ) then
    begin
        if( _AT_File <> nil ) then
        begin
            _AT.Modified := False ;
            _AT_File.Write( 0, 0, _AT.Size, _AT.Get_Data^ ) ;
        end ;
    end ;
end ;


procedure TUOS_String_Table.Write( const Value : string ) ;

var L : cardinal ;

begin
    if( Data_File = nil ) then
    begin
        if( _Position + length( Value ) > _Buffer_Length ) then
        begin
            L := ( _Buffer_Length + length( Value ) + 5 + 63 ) div 64 ;
            _AT.Size := L ;
            _Buffer_Length := L * 64 ;
{$IFDEF CC_Track_Resources}
            Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
            reallocmem( _Buffer, _Buffer_Length ) ;
{$IFDEF CC_Track_Resources}
            Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        end ;
        move( PChar( Value )[ 0 ], _Buffer[ _Position ], length( Value ) ) ;
    end else
    begin
        Data_File.Write( 0, _Position, length( Value ), PChar( Value )[ 0 ] ) ;
    end ;
    _Position := _Position + length( Value ) ;
    _Dirty_Data := True ;
end ;


procedure TUOS_String_Table.Write_Longint( Value : longint ) ;

var S : string ;

begin
    setlength( S, 4 ) ;
    move( Value, PChar( S )[ 0 ], 4 ) ;
    Write( S ) ;
end ;


procedure TUOS_String_Table.Add_Space( Value : cardinal ) ;

var Buff : array[ 0..63 ] of byte ;
    L : int64 ;

begin
    if( Data_File <> nil ) then
    begin
        L := ( Value + 63 ) div 64 ; // Round up to number of 64-byte chunks
        fillchar( Buff, sizeof( Buff ), 0 ) ;
        while( L > 0 ) do
        begin
            Data_File.Write( 0, Data_File.Header.Size, sizeof( Buff ), Buff ) ;
            dec( L ) ;
        end ;
        exit ;
    end ;
    L := ( _Buffer_Length + Value + 63 ) div 64 ; // New size in 64-byte chunks
    _AT.Size := L ;
    _Buffer_Length := L * 64 ;
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    reallocmem( _Buffer, _Buffer_Length ) ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
end ;


// API...

function TUOS_String_Table.Is_Class( Name : PChar ) : boolean ;

begin
    if( lowercase( Name ) = 'tuos_string_table' ) then
    begin
        Result := True ;
    end else
    begin
        Result := inherited Is_Class( Name ) ;
    end ;
end ;


function TUOS_String_Table.Add_String( Value : string ) : longint ;

var C : byte ;
    P, Ref : cardinal ;
    S : string ;

begin
    // Setup...
    Set_Last_Error( nil ) ;
    Result := 0 ;
    if( length( Value ) = 0 ) then
    begin
        exit ;
    end ;
    if( length( Value ) > 256 ) then
    begin
        Setlength( Value, 256 ) ;
    end ;
    C := length( Value ) ;

    if( _Index.Count > 0 ) then
    begin
        // Find existing string...
        Result := Indexof( Value ) ;
        S := Get_String( _Index[ Result ] ) ;
        if( Value = S ) then
        begin
            Seek( _Index[ Result ] ) ;
            Ref := Read_Longint ;
            if( Ref < $FFFFFFFF ) then // Less than 4 billion references
            begin
                // Increment the reference count...
                Result := _Index[ Result ] ;
                Seek( Result ) ;
                inc( Ref ) ;
                Write_Longint( Ref ) ;
                Update ;
                exit ;
            end ;
        end ;
    end else
    begin
        Result := -1 ;
    end ;

    // A new string...
    P := _AT.Allocate( 5 + length( Value ) ) ;
    if( P = 0 ) then // Not enough room
    begin
        Add_Space( length( Value ) + 5 ) ;
        P := _AT.Allocate( 5 + length( Value ) ) ;
    end ;
    if( ( P > $FFFFFFFE ) or ( P = 0 ) ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    Seek( P ) ;
    if( S >= Value ) then
    begin
        dec( Result ) ; // Need to insert new string before found index
    end ;
    _Index.Insert( Result + 1, P ) ;
    _Dirty_Index := True ;
    Result := P ;
    Write( #1#0#0#0 + chr( C ) + Value ) ;
    Update ;
end ; // TUOS_String_Table.Add_String


procedure TUOS_String_Table.Delete_String( Index : longint ) ;

var L, Ref : cardinal ;

begin
    if( _Index.Indexof( Index ) = -1 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_String_Table_Error_Invalid_String_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;

    // Update string reference...
    Seek( Index ) ;
    Ref := Read_Longint ;
    L := 0 ;
    Read( L, 1 ) ;
    if( Ref < 2 ) then // No more references
    begin
        _AT.Deallocate( Index, 5 + L ) ;

        // Remove from index
        Index := _Index.Indexof( Index ) ;
        _Index.Delete( Index ) ;
        _Dirty_Index := True ;

        Update ;
        exit ;
    end ; // if( Ref < 2 )

    // Decrement the reference count...
    Seek( Index ) ;
    dec( Ref ) ;
    Write_Longint( Ref ) ;
    Update ;
end ; // TUOS_String_Table.Delete_String


function TUOS_String_Table.Get_String( Value : longint ) : string ;

var C : integer ;

begin
    Set_Last_Error( nil ) ;
    Result := '' ;
    if( Value = 0 ) then
    begin
        exit ;
    end ;
    if( _Index.Indexof( Value ) = -1 ) then
    begin
        exit ;
    end ;
    Seek( Value + sizeof( longint ) ) ; // Position to length byte
    C := 0 ;
    Read( C, 1 ) ;
    if( C = 0 ) then
    begin
        C := 256 ;
    end ;
    setlength( Result, C ) ;
    Read( PChar( Result )[ 0 ], C ) ;
end ; // TUOS_String_Table.Get_String



{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    halt ;
end ;


var F : text ;

procedure Test_Unit( T : TTest_Interface ) ;

    function Log( const S : string ) : boolean ;

    begin
        append( F ) ;
        writeln( F, S ) ;
        closefile( F ) ;
        Result := T.Log_Operation( S ) ;
    end ;


var A, B, C, X, Z : cardinal ;
    S : string ;
    Test : TUOS_String_Table ;


    procedure Dump_Buffer ;

    var C, C1 : char ;
        Loop : integer ;
//        P : cardinal ;

    begin
exit;
        append( F ) ;
        for Loop := 0 to Test._Buffer_Length - 1 do
        begin
            C := Test._Buffer[ Loop ] ;
            C1 := C ;
            if( ( C < #32 ) or ( C > #126 ) ) then
            begin
                C1 := '.' ;
            end ;
            writeln( F, Loop:3, ': ', ord( C ), ' ', C1 ) ;
        end ;

(*
        Test.Seek( 0 ) ;
        P := Test.Read_Longint ;
        write( F, 'Deleted chain=' ) ;
        while( P <> 0 ) do
        begin
            write( F, P, ' ->' ) ;
            Test.Seek( P ) ;
            P := Test.Read_Longint ;
        end ;
        writeln( F, '0' ) ;
*)
        write( F, 'Indexes=' ) ;
        for Loop := 0 to Test._Index.Count - 1 do
        begin
            writeln( F, Loop, ': ', Test.Get_String( Test._Index[ Loop ] ) ) ;
        end ;
        closefile( F ) ;
    end ;


var Counts : TInteger_List ;
    Dummy : integer ;
    Strings : TStringList ;

begin
    Test := TUOS_String_Table.Create ;
    Test.Update ;
    if( Test.Get_String( 0 ) <> '' ) then
    begin
        Error ;
    end ;
    B := Test.Add_String( 'B' ) ;
    if( B <> 8 ) then
    begin
        Error ;
    end ;
    Z := Test.Add_String( 'Z' ) ;
    if( Z <> 16 ) then
    begin
        Error ;
    end ;
    A := Test.Add_String( 'A' ) ;
    if( A <> 24 ) then
    begin
        Error ;
    end ;
    if( Test.Get_String( A ) <> 'A' ) then
    begin
        Error ;
    end ;
    if( Test.Get_String( B ) <> 'B' ) then
    begin
        Error ;
    end ;
    if( Test.Get_String( Z ) <> 'Z' ) then
    begin
        Error ;
    end ;
    if( Test.Add_String( 'A' ) <> 24 ) then
    begin
        Test.Add_String( 'A' ) ;
        Error ;
    end ;
    C := Test.Add_String( 'C' ) ;
    if( C <> 32 ) then
    begin
        Error ;
    end ;
    Test.Delete_String( C ) ;
    if( Test.Get_String( A ) <> 'A' ) then
    begin
        Error ;
    end ;
    if( Test.Get_String( B ) <> 'B' ) then
    begin
        Error ;
    end ;
    if( Test.Get_String( Z ) <> 'Z' ) then
    begin
        Error ;
    end ;
    if( Test.Add_String( 'D' ) <> C ) then
    begin
        Error ;
    end ;
    Test.Delete_String( A ) ;
    setlength( S, 500 ) ;
    S[ 1 ] := 'U' ;
    X := Test.Add_String( S ) ;
    Test.Delete_String( X ) ;
    Test.Seek( A ) ;
    Test.Write_Longint( $FFFFFFFF ) ;
    X := Test.Add_String( 'A' ) ;
    if( X = A ) then
    begin
        Error ;
    end ;
    if( Test.Add_String( 'A' ) <> X ) then
    begin
        Error ;
    end ;
    if( Test.Add_String( 'Z' ) <> Z ) then
    begin
        Test.Add_String( 'Z' ) ;
        Error ;
    end ;
    Test.Add_String( 'ZZ' ) ;

    Test.Free ;

    // Stress test...
    assignfile( F, 'k:\test.txt' ) ;
    rewrite( F ) ;
    closefile( F ) ;
    Test := TUOS_String_Table.Create ;
    Strings := TStringList.Create ;
    Counts := TInteger_List.Create ;
    while( true ) do
    begin
        case random( 3 ) of
            0 : // Add
                begin
                    case( random( 2 ) ) of
                        0 : begin // Add existing
                                if( Strings.Count = 0 ) then
                                begin
                                    continue ;
                                end ;
                                Dummy := random( Strings.Count ) ;
                                S := Strings[ Dummy ] ;
                                if( Log( 'A' + S ) ) then
                                begin
                                    exit ;
                                end ;
                                Counts[ Dummy ] := Counts[ Dummy ] + 1 ;
                                X := Test.Add_String( S ) ;
                                Strings.Objects[ Dummy ] := pointer( X ) ;
                                Dump_Buffer ;
                            end ;
                        1 : begin
                                setlength( S, random( 256 ) + 1 ) ;
                                for Dummy := 1 to length( S ) do
                                begin
                                    S[ Dummy ] := chr( random( 95 ) + 32 ) ;
                                end ;
                                if( Strings.IndexOf( S ) <> -1 ) then
                                begin
                                    continue ;
                                end ;
                                if( Log( 'A' + S ) ) then
                                begin
                                    exit ;
                                end ;
                                Strings.Add( S ) ;
                                Counts.Add( 1 ) ;
                                X := Test.Add_String( S ) ;
                                Strings.Objects[ Strings.Count - 1 ] := pointer( X ) ;
                                Dump_Buffer ;
                            end ;
                    end ;
                end ;
            1 : // Delete
                begin
                    if( Strings.Count = 0 ) then
                    begin
                        continue ;
                    end ;
                    Dummy := random( Strings.Count ) ;
                    if( Counts[ Dummy ] <= 0 ) then
                    begin
                        continue ;
                    end ;
                    if( Log( 'D' + inttostr( integer( Strings.Objects[ Dummy ] ) ) ) ) then
                    begin
                        exit ;
                    end ;
                    Test.Delete_String( integer( Strings.Objects[ Dummy ] ) ) ;
                    Counts[ Dummy ] := Counts[ Dummy ] - 1 ;
                    Dump_Buffer ;
                end ;
            2 : // Get
                begin
                    if( Strings.Count = 0 ) then
                    begin
                        continue ;
                    end ;
                    Dummy := random( Strings.Count ) ;
                    if( Counts[ Dummy ] <= 0 ) then
                    begin
                        continue ;
                    end ;
                    if( Log( 'G' + inttostr( integer( Strings.Objects[ Dummy ] ) ) ) ) then
                    begin
                        exit ;
                    end ;
                    if( Test.Get_String( integer( Strings.Objects[ Dummy ] ) ) <> Strings[ Dummy ] ) then
                    begin
                        Test.Get_String( integer( Strings.Objects[ Dummy ] ) ) ;
                        Error ;
                    end ;
                end ;
        end ;
    end ;
    Strings.Free ;
    Counts.Free ;
    Test.Free ;
end ;
{$ENDIF}


end.

