{
        Program Name : VTables
        Package Name : LB
        Purpose      : Visual Tables
        Institution  : Conroy & Conroy Co.
        Date Written : 3-Jun-2017
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2017 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit VTables ;

interface

uses // Pascal...
     Windows, // TPoint
     Classes, // TStringList
     Forms, // TScrollBox
     Extctrls ; // TPanel

type TColumn_Types = ( CT_Boolean, CT_Integer, CT_String, CT_Selection ) ;

type TColumn_Info = class
                        public
                            Index : integer ;
                            Name : string ;
                            Heading : string ;
                            Data_Type : TColumn_Types ;
                            Default_Value : string ;
                            Other : string ;
                            Panel : TPanel ;
                    end ;

type TModify_Event = procedure( Row, Column : integer ; Value : string ) of object ;

type TVTable = class( TScrollBox )
                   private // Instance data...
                       _ColumnCount : integer ;
                       _RowCount : integer ;
                       _Columns : TList ;
                       _Rows : TList ; // List of stringlists
                       _Main_Panel : TPanel ;
                       _Inner_Panel : TPanel ;
                       RowHeight : integer ;
                       _Click_Column : TColumn_Info ;
                       _Click_Row : integer ;

                   public // Constructors and destructors...
                       constructor Create( AComponent : TComponent ) ; override ;
                       destructor Destroy ; override ;

                   private // Internal utility routines...
                       procedure CB_CB_Change( Sender : TObject ) ;
                       procedure CB_Column_Resize( Sender : TObject ) ;
                       procedure CB_Column_Click( Sender : TObject ) ;
                       procedure CB_Column_Right_Click( Sender : TObject ;
                           P : TPoint ; var B : boolean ) ;
                       procedure CB_Change_Selection( Sender : TObject ) ;
                       function Get_Row( Index : integer ) : TStringList ;

                   private // Private properties
                       function Get_Columns : TList ;
                       function Get_Rows : TList ;

                       property Columns : TList
                           read Get_Columns
                           write _Columns ;
                       property Rows : TList
                           read Get_Rows
                           write _Rows ;

                   public // Overrides...
                       procedure Resizing( State : TWindowState ) ; override ;

                   public // API...
                       procedure Add_Column( Name, Heading : string ;
                           Data_Type : TColumn_Types ; Default_Value, Other : string ) ;
                       procedure Add_Row( S : TStringList ) ;
                       function ColumnCount : integer ;
                       function RowCount : integer ;
                       function Get_Value( Row, Column : integer ) : string ;
                       procedure Set_Value( Row, Column : integer ; S : string ) ;

                   public // Events
                       On_Modify : TModify_Event ;
               end ; // TVTable

implementation

uses // Pascal...
     Controls, // bvNone
     Dialogs, // InputQuery
     Menus, // TPopupMenu
     Graphics, // cl*
     Stdctrls, // TLabel
     SysUtils, // trim

     // Treasure32...
     Expression_Query_Form ; // Expression_Query

// TVTable methods...

// Constructors and destructors...

constructor TVTable.Create( AComponent : TComponent ) ;

begin
    inherited Create( AComponent ) ;

    _Main_Panel := TPanel.Create( self ) ;
    _Main_Panel.Parent := self ;
    _Main_Panel.BevelInner := bvNone ;
    _Main_Panel.BevelOuter := bvNone ;
    _Main_Panel.BorderStyle := bsNone ;
    _Inner_Panel := _Main_Panel ;
    _Main_Panel.Height := 1 ;
end ;


destructor TVTable.Destroy ;

var Loop : integer ;

begin
    if( _Rows <> nil ) then
    begin
        for Loop := 0 to _Rows.Count - 1 do
        begin
            TStringList( _Rows[ Loop ] ).Free ;
            _Rows[ Loop ] := nil ;
        end ;
        _Rows.Free ;
        _Rows := nil ;
    end ;
    if( _Columns <> nil ) then
    begin
        for Loop := 0 to _Columns.Count - 1 do
        begin
            TColumn_Info( _Columns[ Loop ] ).Free ;
            _Columns[ Loop ] := nil ;
        end ;
        _Columns.Free ;
        _Columns := nil ;
    end ;

    inherited Destroy ;
end ; // TVTable.Destroy


// Internal utility routines...

procedure TVTable.CB_CB_Change( Sender : TObject ) ;

var C : TCheckBox ;
    Column : TColumn_Info ;
    S : string ;

begin
    C := TCheckBox( Sender ) ;
    Column := TColumn_Info( C.Tag ) ;
    _Click_Column := Column ;
    _Click_Row := ( C.Top + C.Height div 2 ) div RowHeight ;
    if( assigned( On_Modify ) ) then
    begin
        if( C.Checked ) then
        begin
            S := 'true' ;
        end else
        begin
            S := 'false' ;
        end ;
        On_Modify( _Click_Row, Column.Index, S ) ;
    end ;
end ;


procedure TVTable.CB_Column_Resize( Sender : TObject ) ;

var Loop : integer ;
    P : TPanel ;

begin
    P := TPanel( Sender ) ;
    for Loop := 0 to P.ControlCount - 1 do
    begin
        P.Controls[ Loop ].Width := ClientWidth ;
    end ;
end ;


procedure TVTable.CB_Column_Click( Sender : TObject ) ;

var Column : TColumn_Info ;
    Dummy, I : integer ;
    L : TLabel ;
    M : TMenuItem ;
    P : TPopupMenu ;
    S : string ;

begin
    L := TLabel( Sender ) ;
    Column := TColumn_Info( L.Tag ) ;
    _Click_Column := Column ;
    _Click_Row := ScreenToClient( Mouse.CursorPos ).Y div RowHeight ;
    if( Column.Data_Type = CT_Selection ) then
    begin
        S := Column.Other ;
        P := TPopupMenu.Create( Application ) ;
        I := 0 ;
        while( S <> '' ) do
        begin
            M := TMenuItem.Create( Application ) ;
            P.Items.Add( M ) ;
            Dummy := pos( ',', S + ',' ) ;
            M.Caption := copy( S, 1, Dummy - 1 ) ;
            M.Tag := I ;
            inc( I ) ;
            S := copy( S, Dummy + 1, length( S ) ) ;
            M.OnClick := CB_Change_Selection ;
        end ;
        P.PopUp( Mouse.CursorPos.X, Mouse.CursorPos.Y ) ;
    end else
    if( Column.Data_Type = CT_String ) then
    begin
        S := Get_Value( _Click_Row, Column.Index ) ;
        if( inputquery( 'Treasure', 'New value', S ) ) then
        begin
            Set_Value( _Click_Row, Column.Index, S ) ;
        end ;
    end else
    if( Column.Data_Type = CT_Integer ) then
    begin
        Expression_Query.Caption := 'Treasure' ;
        Expression_Query.Value.Text := Get_Value( _Click_Row, Column.Index ) ;
        Expression_Query.ActiveControl := Expression_Query.Value ;
        if( Expression_Query.ShowModal = mrOK ) then
        begin
            Set_Value( _Click_Row, Column.Index, Expression_Query.Get_Value ) ;
        end ;
    end ;
end ; // TVTable.CB_Column_Click


procedure TVTable.CB_Column_Right_Click( Sender : TObject ; P : TPoint ;
    var B : boolean ) ;

begin
    CB_Column_Click( Sender ) ;
end ;


procedure TVTable.CB_Change_Selection( Sender : TObject ) ;

var C, I, Row, Dummy : integer ;
    Column : TColumn_Info ;
    M : TComponent ;
    S : string ;

begin
    M := TComponent( Sender ) ;
    if( ( _Click_Row > 0 ) and ( _Click_Row <= RowCount ) ) then
    begin
        I := M.Tag ;
        Column := _Click_Column ;
        S := Column.Other ;
        Dummy := pos( ',', S ) ;
        while( I > 0 ) do
        begin
            dec( I ) ;
            S := copy( S, Dummy + 1, length( S ) ) ;
            Dummy := pos( ',', S ) ;
        end ;
        Dummy := pos( ',', S + ',' ) ;
        Set_Value( _Click_Row, Column.Index, copy( S, 1, Dummy - 1 ) ) ;
    end ;
end ;


function TVTable.Get_Row( Index : integer ) : TStringList ;

begin
    if( ( Index < 0 ) or ( Index > RowCount - 1 ) ) then // Out of range
    begin
        Result := nil ;
        exit ;
    end ;
    Result := TStringList( Rows[ Index ] ) ;
end ;


// Private properties...

function TVTable.Get_Columns : TList ;

begin
    if( _Columns = nil ) then
    begin
        _Columns := TList.Create ;
    end ;
    Result := _Columns ;
end ;


function TVTable.Get_Rows : TList ;

begin
    if( _Rows = nil ) then
    begin
        _Rows := TList.Create ;
    end ;
    Result := _Rows ;
end ;


// Overrides...

procedure TVTable.Resizing( State : TWindowState ) ;

begin
    inherited Resizing( State ) ;

    _Main_Panel.Width := ClientWidth ;
end ;


// API...

procedure TVTable.Add_Column( Name, Heading : string ;
    Data_Type : TColumn_Types ; Default_Value, Other : string ) ;

var B : TPanel ;
    Info : TColumn_Info ;
    L : TLabel ;
    P, P1 : TPanel ;
    S : TSplitter ;

begin
    Info := TColumn_Info.Create ;
    Info.Index := ColumnCount ;
    Info.Name := Name ;
    Info.Heading := Heading ;
    Info.Data_Type := Data_Type ;
    Info.Default_Value := Default_Value ;
    Info.Other := Other ;
    P := TPanel.Create( self ) ;
    P.Parent := _Inner_Panel ;
    P.BorderStyle := bsNone ;
    P.BevelInner := bvNone ;
    P.BevelOuter := bvNone ;
    P.Align := alLeft ;
    P.OnResize := CB_Column_Resize ;
    Info.Panel := P ;
    S := TSplitter.Create( self ) ;
    S.Parent := _Inner_Panel ;
    S.Align := alLeft ;
    S.Color := clBlack ;
    S.Width := 2 ;
    P1 := TPanel.Create( self ) ;
    P1.Parent := _Inner_Panel ;
    P1.BorderStyle := bsNone ;
    P1.BevelInner := bvNone ;
    P1.BevelOuter := bvNone ;
    P1.Align := alClient ;
    _Inner_Panel := P1 ;
    L := TLabel.Create( self ) ;
    L.Parent := P ;
    L.Caption := ' ' + Heading ;
    L.Color := clLtGray ;
    L.AutoSize := True ;
    if( _Main_Panel.Height < L.Height ) then
    begin
        _Main_Panel.Height := L.Height ;
    end ;
    if( L.Width < 16 ) then
    begin
        P.Width := 16 ;
    end else
    begin
        P.Width := L.Width ;
    end ;
    L.AutoSize := False ;
    L.Align := alTop ;
    RowHeight := L.Height + 2 ;
    Columns.Add( Info ) ;
    B := TPanel.Create( self ) ;
    B.Parent := P ;
    B.BorderStyle := bsNone ;
    B.BevelInner := bvNone ;
    B.BevelOuter := bvNone ;
    B.Top := RowHeight - 2 ;
    B.Color := clBlack ;
    B.Height := 2 ;
end ; // TVTable.Add_Column


procedure TVTable.Add_Row( S : TStringList ) ;

var B : TPanel ;
    C : integer ;
    CB : TCheckBox ;
    Column : TColumn_Info ;
    L : TLabel ;
    SS : string ;
    SL : TStringList ;

begin
    SL := TStringList.Create ;
    for C := 0 to ColumnCount - 1 do // For each column
    begin
        Column := TColumn_Info( Columns[ C ] ) ;
        if( C >= S.Count ) then
        begin
            SS := '' ;
        end else
        begin
            SS := S[ C ] ;
        end ;
        if( SS = '' ) then
        begin
            SS := Column.Default_Value ;
        end ;
        if( Column.Data_Type = CT_Boolean ) then
        begin
            CB := TCheckBox.Create( self ) ;
            CB.Parent := Column.Panel ;
            if( RowHeight = 0 ) then
            begin
                RowHeight := CB.Height ;
            end ;
            CB.Top := RowHeight * ( RowCount + 1 ) ;
            CB.Height := RowHeight - 2 ;
            CB.Checked := ( trim( lowercase( SS ) ) <> 'false' ) ;
            CB.ParentColor := True ;
            CB.Height := RowHeight - 2 ;
            CB.Width := Column.Panel.Clientwidth ;
            CB.Tag := integer( Column ) ;
            CB.OnClick := CB_CB_Change ;
        end else
        begin
            L := TLabel.Create( self ) ;
            L.Parent := Column.Panel ;
            L.Top := RowHeight * ( RowCount + 1 ) ;
            L.Caption := ' ' + SS ;
            L.ParentColor := True ;
            L.AutoSize := False ;
            L.Width := Column.Panel.Clientwidth ;
            L.Tag := integer( Column ) ;
            L.OnClick := CB_Column_Click ;
            L.OnContextPopup := CB_Column_Right_Click ;
        end ;
        B := TPanel.Create( self ) ;
        B.Parent := Column.Panel ;
        B.BorderStyle := bsNone ;
        B.BevelInner := bvNone ;
        B.BevelOuter := bvNone ;
        B.Top := RowHeight * ( RowCount + 1 ) - 2 ;
        B.Color := clBlack ;
        B.Height := 2 ;
    end ; // for C := 0 to ColumnCount - 1
    _Main_Panel.Height := _Main_Panel.Height + RowHeight ;
    Rows.Add( SL ) ;
end ; // TVTable.Add_Row


function TVTable.ColumnCount : integer ;

begin
    if( _Columns = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := _Columns.Count ;
    end ;
end ;


function TVTable.RowCount : integer ;

begin
    if( _Rows = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := _Rows.Count ;
    end ;
end ;


function TVTable.Get_Value( Row, Column : integer ) : string ;

var C : TCheckBox ;
    Info : TColumn_Info ;
    L : TLabel ;
    Panel : TPanel ;

begin
    if( ( Row < 1 ) or ( Column < 0 ) or ( Row > RowCount ) or ( Column >= ColumnCount ) ) then
    begin
        exit ; // Out of range
    end ;
    Info := TColumn_Info( _Columns[ Column ] ) ;
    Panel := Info.Panel ;
    case Info.Data_Type of
        CT_Boolean :
          begin
              C := TCheckBox( Panel.Controls[ Row * 2 ] ) ;
              if( C.Checked ) then
              begin
                  Result := 'true' ;
              end else
              begin
                  Result := 'false' ;
              end ;
          end ;
        else
          begin
              L := TLabel( Panel.Controls[ Row * 2 ] ) ;
              Result := L.Caption ;
              Result := copy( Result, 2, length( Result ) ) ;
          end ;
    end ; // case Info.Data_Type
end ;


procedure TVTable.Set_Value( Row, Column : integer ; S : string ) ;

var C : TCheckBox ;
    Info : TColumn_Info ;
    L : TLabel ;
    Panel : TPanel ;
    Saved : TModify_Event ;

begin
    if( ( Row < 1 ) or ( Column < 0 ) or ( Row > RowCount ) or ( Column >= ColumnCount ) ) then
    begin
        exit ; // Out of range
    end ;
    Saved := On_Modify ;
    try
        On_Modify := nil ;
        Info := TColumn_Info( _Columns[ Column ] ) ;
        if( Get_Value( Row, Column ) = S ) then // No change
        begin
            exit ;
        end ;
        Panel := Info.Panel ;
        case Info.Data_Type of
            CT_Boolean :
              begin
                  C := TCheckBox( Panel.Controls[ Row * 2 ] ) ;
                  C.Checked := trim( lowercase( S ) ) <> 'false' ;
              end ;
            else
              begin
                  L := TLabel( Panel.Controls[ Row * 2 ] ) ;
                  L.Caption := ' ' + S ;
              end ;
        end ; // case Info.Data_Type
    finally
        On_Modify := Saved ;
    end ;
    if( assigned( On_Modify ) ) then
    begin
        On_Modify( Row, Column, S ) ;
    end ;
end ;



end.

