{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
        Program Name : _DevMan
        Package Name : Subroutine Library - Layer 1
        Purpose      : Device Manager class definitions
        Institution  : Conroy & Conroy Co.
        Date Written : 7-Jul-2006
        Written By   : Alan Conroy
        Version      : 1.1

        Copyright (C) 2006-2009 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON                         

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TDevice_Manager is a class for managing access to devices for
        Operating system objects.
}

unit _DevMan ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

uses { Layer 0... }
     _Common, { TObject }
     COMInter, { TCommon_COM_Interface }
     _Types ; { int64 }

{ Device Media IDs: }
const DMI_Unknown = -1 ; { Unknown media type }
const DMI_Remote = 4 ; { Remote (network) drive. }
const DMI_CDROM = 5 ; { CD or DVD. }
const DMI_RAM = 6 ; { RAM disk. }
const DMI_Removable = $F0 ; { 3.5" DS, 18 sectors (or other removable) }
const DMI_Fixed = $F8 ; { Fixed disk }
const DMI_Floppy1 = $F9 ; { 5.25", DS, 15 sectors or 3.5" DS, 9 sectors }
const DMI_Floppy2 = $FC ; { 5.25" SS, 9 sectors }
const DMI_Floppy3 = $FD ; { 5.25" DS, 9 sectors }
const DMI_Floppy4 = $FE ; { 5.25" SS, 8 sectors }
const DMI_Floppy5 = $FF ; { 5.25" DS, 8 sectors }

{ Device File System Flags: }
const DFSF_Case_Sensitive = 1 ; { Specifies that searches are case-sensitive. }
const DFSF_Preserve_Case = 2 ; { Preserves case in directory entries. }
const DFSF_UNICODE = 4 ; { Uses Unicode characters in file and directory names. }
const DFSF_Long_Filenames = $4000 ; { Supports new long filename functions. }
const DFSF_Volume_Compressed = $8000 ; { The volume is compressed. }

{ Device class: }
const DC_Unknown = 0 ;
const DC_Disk = 1 ;
const DC_Tape = 2 ;
const DC_Serial = 3 ; { Includes NUL }
const DC_Network = 4 ;
const DC_System = 5 ; { Includes clock }
const DC_Printer = 6 ;
const DC_Optical = 7 ; { CD, DVD, etc. }
const DC_Other = 8 ; { Everything else }

{$IFDEF DYNAMIC_OBJECTS}
type TCOM_Device_Info = class( TCommon_COM_Interface )
                            public { API... }
                                function Media_ID : longint ;
                                    virtual ; stdcall ; abstract ;
                                { Return type of media in/on device.  -1 if
                                  error or unknown. }

                                function File_System : PAnsiChar ;
                                    virtual ; stdcall ; abstract ;
                                { Returns the name of the file system used on
                                  the device.  Returns a null string if the
                                  device is invalid or the file system is
                                  unknown. }

                                function File_System_Flags : longint ;
                                    virtual ; stdcall ; abstract ;
                                { Returns the flags for the file system.  See
                                  DFSF_*. }

                                function Fixed : boolean ;
                                    virtual ; stdcall ; abstract ;
                                { Returns true if the passed device is fixed
                                  (non-removable) }

                                function Max_Filename_Length : longint ;
                                    virtual ; stdcall ; abstract ;
                                { Returns the maximum filename length for this
                                  device's file system. }

                                function Version : longint ;
                                    virtual ; stdcall ; abstract ;
                                { Returns version number of this object.  11 = 1.1 }
                        end ;

type TCOM_Device_Manager = class( TCommon_COM_Interface )
                               public { API... }
                                 function Device_Info( Device : PAnsiChar ) : TCOM_Device_Info ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns an information object for the
                                   requested device. }

                                 function Disk_Clusters( Drive : PAnsiChar ) : int64 ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns number of clusters on specified disk
                                   drive.  Returns -1 if an error or unknown. }

                                 function Disk_Sector_Size( Drive : PAnsiChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns size of disk sectors, in bytes.
                                   Returns -1 if an error or unknown. }

                                 function File_System( Device : PAnsiChar ) : PAnsiChar ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the name of the file system used on
                                   the device.  Returns a null string if the
                                   device is invalid or the file system is
                                   unknown. }

                                 function File_System_Flags( Device : PAnsiChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the flags for the file system.  See
                                   DFSF_*. }

                                 function Fixed_Device( Device : PAnsiChar ) : boolean ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns true if the passed device is fixed
                                   (non-removable) }

                                 function Max_Filename_Length( Device : PAnsiChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the maximum filename length for this
                                   device's file system. }

                                 function Media_ID( Device : PAnsiChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Return type of media in/on device.  -1 if
                                   error or unknown. }

                                 function Valid_Device( Device : PAnsiChar ) : boolean ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns true if the passed device is valid. }

                                 function Device_Name( Clas, Index : integer ) : PAnsiChar ;
                                     virtual ; stdcall ; abstract ;
                                 { Return name of Indexth device of given class.
                                   if Class is unknown, all device classes are
                                   assumed. }

                                 function Current_Path( Device : PAnsiChar ) : PAnsiChar ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the current path (or default path if
                                   no current path) for the specified device.
                                   Returns null if device is not a hierarchical
                                   file system device. }

                                 function Version : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns version number of this object.  11 = 1.1 }

                                 function Media_Present( Device : PAnsiChar ) : boolean ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns true if media is present in the device }
                           end ; { TCOM_Device_Manager }

type TDevice_Info = class( TCommon_COM_Interface )
                        public { API... }
                            function Media_ID : longint ; virtual ; abstract ;
                            { Return type of media in/on device.  -1 if
                              error or unknown. }

                            function File_System : PAnsiChar ; virtual ; abstract ;
                            { Returns the name of the file system used on
                              the device.  Returns a null string if the
                              device is invalid or the file system is
                              unknown. }

                            function File_System_Flags : longint ;
                                virtual ; abstract ;
                            { Returns the flags for the file system.  See
                              DFSF_*. }

                            function Fixed : boolean ; virtual ; abstract ;
                            { Returns true if the passed device is fixed
                              (non-removable) }

                            function Max_Filename_Length : longint ;
                                virtual ; abstract ;
                            { Returns the maximum filename length for this
                              device's file system. }

                            function Version : longint ;
                                virtual ; abstract ;
                            { Returns version number of this object.  11 = 1.1 }
                    end ;

type TDevice_Manager = class( TDelphi_Object )
                           public { API... }
                                 function Device_Info( Device : Ansistring ) : TDevice_Info ;
                                     virtual ; abstract ;
                                 { Returns an information object for the
                                   requested device. }

                                 function Disk_Clusters( Drive : Ansistring ) : int64 ;
                                     virtual ; abstract ;
                                 { Returns number of clusters on specified disk
                                   drive.  Returns -1 if an error or unknown. }

                                 function Disk_Sector_Size( Drive : Ansistring ) : longint ;
                                     virtual ; abstract ;
                                 { Returns size of disk sectors, in bytes.
                                   Returns -1 if an error or unknown. }

                                 function File_System( Device : Ansistring ) : Ansistring ;
                                     virtual ; abstract ;
                                 { Returns the name of the file system used on
                                   the device.  Returns a null string if the
                                   device is invalid or the file system is
                                   unknown. }

                                 function File_System_Flags( Device : Ansistring ) : longint ;
                                     virtual ; abstract ;
                                 { Returns the flags for the file system.  See
                                   DFSF_*. }

                                 function Fixed_Device( Device : Ansistring ) : boolean ;
                                     virtual ; abstract ;
                                 { Returns true if the passed device is fixed
                                   (non-removable) }

                                 function Max_Filename_Length( Device : Ansistring ) : longint ;
                                     virtual ; abstract ;
                                 { Returns the maximum filename length for this
                                   device's file system. }

                                 function Media_ID( Device : Ansistring ) : longint ;
                                     virtual ; abstract ;
                                 { Return type of media in/on device.  -1 if
                                   error or unknown. }

                                 function Valid_Device( Device : Ansistring ) : boolean ;
                                     virtual ; abstract ;
                                 { Returns true if the passed device is valid. }

                                 function Device_Name( Clas, Index : integer ) : Ansistring ;
                                     virtual ; abstract ;
                                 { Return name of Indexth device of given class.
                                   if Class is unknown, all device classes are
                                   assumed. }

                                 function Current_Path( Device : Ansistring ) : Ansistring ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the current path (or default path if
                                   no current path) for the specified device.
                                   Returns null if device is not a hierarchical
                                   file system device. }

                                 function Version : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns version number of this object.  11 = 1.1 }

                                 function Media_Present( Device : Ansistring ) : boolean ;
                                     virtual ; abstract ;
                                 { Returns true if media is present in the device }
                       end ; { TDevice_Manager }
{$ENDIF}

type TSDevice_Info = object( TObject )
                         public { API... }
                             function Media_ID : longint ; virtual ;
                             { Return type of media in/on device.  -1 if
                               error or unknown. }

                             function File_System : Ansistring ; virtual ;
                             { Returns the name of the file system used on
                               the device.  Returns a null string if the
                               device is invalid or the file system is
                               unknown. }

                             function File_System_Flags : longint ;
                                 virtual ;
                             { Returns the flags for the file system.  See
                               DFSF_*. }

                             function Fixed : boolean ; virtual ;
                             { Returns true if the passed device is fixed
                               (non-removable) }

                             function Max_Filename_Length : longint ;
                                 virtual ;
                             { Returns the maximum filename length for this
                               device's file system. }

                             function Version : longint ;
                                 virtual ;
                             { Returns version number of this object.  11 = 1.1 }
                     end ;
     PSDevice_Info = ^TSDevice_Info ;

type TSDevice_Manager = object( TObject )
                            public { API... }
                                 function Device_Info( Device : Ansistring ) : PSDevice_Info ;
                                     virtual ;
                                 { Returns an information object for the
                                   requested device. }

                                 function Disk_Clusters( Drive : Ansistring ) : int64 ;
                                     virtual ;
                                 { Returns number of clusters on specified disk
                                   drive.  Returns -1 if an error or unknown. }

                                 function Disk_Sector_Size( Drive : Ansistring ) : longint ;
                                     virtual ;
                                 { Returns size of disk sectors, in bytes.
                                   Returns -1 if an error or unknown. }

                                 function File_System( Device : Ansistring ) : Ansistring ;
                                     virtual ;
                                 { Returns the name of the file system used on
                                   the device.  Returns a null string if the
                                   device is invalid or the file system is
                                   unknown. }

                                 function File_System_Flags( Device : Ansistring ) : longint ;
                                     virtual ;
                                 { Returns the flags for the file system.  See
                                   DFSF_*. }

                                 function Fixed_Device( Device : Ansistring ) : boolean ;
                                     virtual ;
                                 { Returns true if the passed device is fixed
                                   (non-removable) }

                                 function Max_Filename_Length( Device : Ansistring ) : longint ;
                                     virtual ;
                                 { Returns the maximum filename length for this
                                   device's file system. }

                                 function Media_ID( Device : Ansistring ) : longint ;
                                     virtual ;
                                 { Return type of media in/on device.  -1 if
                                   error or unknown.  See Device Media Ids. }

                                 function Valid_Device( Device : Ansistring ) : boolean ;
                                     virtual  ;
                                 { Returns true if the passed device is valid. }

                                 function Device_Name( Clas, Index : integer ) : Ansistring ;
                                     virtual ;
                                 { Return name of Indexth device of given class.
                                   if Class is unknown, all device classes are
                                   assumed. }

                                 function Current_Path( Device : Ansistring ) : Ansistring ;
                                     virtual ;
                                 { Returns the current path (or default path if
                                   no current path) for the specified device.
                                   Returns null if device is not a hierarchical
                                   file system device. }

                                 function Version : longint ; virtual ;
                                 { Returns version number of this object.  11 = 1.1 }

                                 function Media_Present( Device : Ansistring ) : boolean ;
                                     virtual ;
                                 { Returns true if media is present in the device }
                        end ; { TSDevice_Manager }
     PSDevice_Manager = ^TSDevice_Manager ;

implementation

{ TSDevice_Info methods... }

function TSDevice_Info.Media_ID : longint ;

begin
    Media_ID := -1 ;
end ;


function TSDevice_Info.File_System : Ansistring ;

begin
    File_System := '' ;
end ;


function TSDevice_Info.File_System_Flags : longint ;

begin
    File_System_Flags := 0 ;
end ;


function TSDevice_Info.Fixed : boolean ;

begin
    Fixed := False ;
end ;


function TSDevice_Info.Max_Filename_Length : longint ;

begin
    Max_Filename_Length := 1 ;
end ;


function TSDevice_Info.Version : longint ;

begin
    Version := 11 ;
end ;



{ TSDevice_Manager methods... }

function TSDevice_Manager.Device_Info( Device : Ansistring ) : PSDevice_Info ;

begin
    Device_Info := nil ;
end ;


function TSDevice_Manager.Disk_Clusters( Drive : Ansistring ) : int64 ;

begin
    Disk_Clusters := -1 ;
end ;


function TSDevice_Manager.Disk_Sector_Size( Drive : Ansistring ) : longint ;

begin
    Disk_Sector_Size := -1 ;
end ;


function TSDevice_Manager.File_System( Device : Ansistring ) : Ansistring ;

begin
    File_System := '' ;
end ;


function TSDevice_Manager.File_System_Flags( Device : Ansistring ) : longint ;

begin
    File_System_Flags := 0 ;
end ;


function TSDevice_Manager.Fixed_Device( Device : Ansistring ) : boolean ;

begin
    Fixed_Device := False ;
end ;


function TSDevice_Manager.Max_Filename_Length( Device : Ansistring ) : longint ;

begin
    Max_Filename_Length := 1 ;
end ;


function TSDevice_Manager.Media_ID( Device : Ansistring ) : longint ;

begin
    Media_ID := -1 ;
end ;


function TSDevice_Manager.Valid_Device( Device : Ansistring ) : boolean ;

begin
    Valid_Device := False ;
end ;


function TSDevice_Manager.Device_Name( Clas, Index : integer ) : Ansistring ;

begin
    Device_Name := '' ;
end ;


function TSDevice_Manager.Current_Path( Device : Ansistring ) : Ansistring ;

begin
    Current_Path := '' ;
end ;


function TSDevice_Manager.Version : longint ;

begin
    Version := 11 ;
end ;


function TSDevice_Manager.Media_Present( Device : Ansistring ) : boolean ;

begin
    Media_Present := False ;
end ;


end.
