/*
              Program Name : _FiP
              Package Name : Subroutine Library - Layer 1
              Purpose      : File Processor abstract class definition
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-08
              Written By   : Alan Conroy
              Version      : 2.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2008 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

            DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the abstract File_Processor class and the classes
        for registering file system servers, etc.
*/

#ifndef _FIPH
    #define _FIPH

#include "eDefines.h"
#include "cominter.h"
#include "_filesys.h"
#include "uec.h"

// File Open modes...
const FSOM_Default = 0 ; // Open file if there, create it if not.
const FSOM_Append = 1 ; // Append mode (file must already exist)
const FSOM_Input = 2 ; // File must already exist
const FSOM_Output = 3 ; // File is created (deleted if existing)
const FSOM_Text = 4 ; // File is text mode
      // None of the following is currently supported...
const FSOM_Read_Only = 8 ; // File cannot be written to
const FSOM_Share_Read = 16 ; // File can be opened read-only by others
const FSOM_Share_Write = 32 ; // File can be opened writable by others
// FS Open Mode 64+ are reserved for future use

// Filename query flags...
const TFQF_Read_Only = 1 ; // Open as read-only
const TFQF_Overwrite = 2 ; // Prompt for overwrite on existing files
const TFQF_No_Validation = 4 ; // Do not validate filename
const TFQF_Multiple_Files = 8 ; // Allow multiple filename to be returned
const TFQF_Path_Must_Exist = 16 ; // Path must exist
const TFQF_File_Must_Exist = 32 ; // File must exist
const TFQF_Prompt_On_Creation = 64 ; // Prompt if file doesn't exist
const TFQF_No_Read_Only = 128 ; // Don't allow read-only files
const TFQF_No_Test = 256 ; // Disables checking for network file protection and inaccessibility of disk drives (when the user tries to save a file in a create-no-modify shared network directory)
const TFQF_No_Dereference = 512 ; // Disable dereferencing of file links
const TFQF_Show_All = 1024 ; // Show all files
	  //2048+ = reserved

	  // Filename query modes...
enum TFilename_Query_Mode { TFNQM_Open, TFNQM_Save } ;

typedef void* PSFile_System_Server ; // TODO: Implement COM version of this
typedef void* PSStore_Server ; // TODO: Implement COM version of this
typedef void* PSFile ; // TODO: Implement COM version of this

class TCOM_File_Processor : public TCommon_COM_Interface
{
	public: // API...
	  virtual char* Lookup( char* Filename, TFind_Record& Find_Record ) = 0;
	  virtual PSFile Open( char* Name, int Mode, TUEC& UEC ) = 0;
	  virtual void __stdcall _Initialize( char* Nam, char* ID, int Flags,
		  TUEC& UEC ) = 0;
	  virtual int64 DeviceSize( char* Nam ) = 0;
	  virtual int64 DeviceFree( char* Nam ) = 0;
	  virtual int MKDir( char* Nam ) = 0;
	  virtual void __stdcall Close_File_System( char* Name ) = 0;
      virtual PSFile_System Open_File_System( char* Name,
		  int Flags, TUEC& Res ) = 0;
	  virtual bool File_System_Open( char* Nam ) = 0;
      // Scan a file specification and fill the Info record.  Separator is the path separator to use in parsing.
      virtual void __stdcall Filespec_String_Scan( char* Name, char* Separator,
		  TFSSS& Info ) = 0;
      virtual char* Query_Filename( char* Default, char* Prompt,
		  TFilename_Query_Mode Mode, int Flags ) = 0;
	  virtual void __stdcall Register_File_System_Server( PSFile_System_Server FSS ) = 0;
	  virtual void __stdcall Register_Store_Server( PSStore_Server SS ) = 0;
	  virtual int64 __stdcall File_Size( char* Nam ) = 0;
} ; // TCOM_File_Processor

#endif
