{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{$A-}
{
              Program Name : _FileSys
              Package Name : Subroutine Library - Layer 1
              Purpose      : Abstract File System rlass
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-88
              Written By   : Alan Conroy
              Version      : 1.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2008 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

            DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the abstract TSFile_System, and related, objects.
}

unit _FileSys ;

interface

{$I EDEFINES.INC}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$IFDEF DELPHI}
  {$DEFINE WINDOWS}
{$ENDIF}

uses { Layer 0... }
     CommBase, { TBase_Object }
     _Types, { Ansistring }

     { Layer 1... }
     _Stores, { TStore_Address }

{$IFDEF WINDOWS}
  {$IFDEF FPC_OR_DELPHI}
     SysUtils,
  {$ELSE}
     Windos,
     strings,
     Objects,
  {$ENDIF}

{$ELSE}
     { Borland... }
     Dos,
     Windos,
     Objects,
     Strings,
{$ENDIF}
     { Layer 1... }
     _UE ; { TUnified_Exception }

type FS_Header_Information = packed record
                                 Header_Version : smallint ; { Version of FS_Header_Information }
{$IFDEF MOBILE}
                                 ID : array[ 0..127 ] of byte ; { Label for this file system device }
{$ELSE}
                                 ID : string[ 127 ] ; { Label for this file system device }
{$ENDIF}
                                 Root : TStore_Address ; { Pointer to root of file system/directories }
                                 Version : smallint ; { File system structure version }
                                 Default_Cluster_Size : smallint ; { Default cluster size for allocations (0 is invalid) }
                                 Flags : smallint ; { FSD_* }
                             end ;

{ Used to hold results of Filespec_String_Scan }
type TFSSS = record
                 Device : string ; { Device, server, or protocol specification, including colon }
                 Path : string ; { Path, including final separator }
                 Filename : string ; { Full filename }
                 Name : string ; { Name, minus (last) extension }
                 Extension : string ; { (Last) extension (not including dot) }
                 Password : string ; { Embedded password }
                 Switches : string ; { Any switches at end of filename }
                 Flags : longint ; { Result flags (see FSSS_*) }
             end ;

      { File System File Flags... }
const FSFF_Directory = 1 ; { A directory file }
      FSFF_Journaling = 2 ; { A journalled file }
      FSFF_File_System = 4 ; { A file-system file }
      FSFF_Compressed = 8 ; { A compressed file }
      FSFF_Erase = 16 ; { Erase file upon deletion }
      FSFF_Statistics = 32 ; { Keep file access statistics }
      FSFF_Read_Verify = 64 ; { Verify read accesses }
      FSFF_Write_Verify = 128 ; { Verify write accesses }
      FSFF_Hidden = 256 ; { A hidden file }
      FSFF_Contiguous = 512 ; { File is contiguous on file system device }
      FSFF_Deleted = 1024 ;
      { File is deleted but still open (marked for deletion) }
      FSFF_Protected = 2048 ; { for DOS compatibility only }
      FSFF_System = 4096 ; { for DOS compatibility only }
      FSFF_VolumeID = 8192 ; { for DOS compatibility only }
      FSFF_Archive = 16384 ; { for DOS compatibility only }
      { FS File Flags 32768+ are reserved for future use }

const { File System Flags... }
      FSF_Fixed_Size = 1 ; { The file system is on a fixed-size device }
      FSF_Read_Only = 2 ; { The file system is opened in read-only mode }
      FSF_NoOwnStore = 4 ; { The file system doesn't own the store - so don't close it when file system closes }
      { 8-32768 = reserved }

{ Flags for TFSSS.Flags }
const FSSS_Null_Extension = 1 ; { Filename had no dot for an extension }

type PSWildcard = ^TSWildcard ;

     PSFile_System = ^TSFile_System ;

     TFind_Record = record { TODO: Make this a class }
                        Initialized : longint ; { 0 if this record was initialized }
                        Context : PSWildcard ; { Top-level Wildcard object }
                        Separator : Ansistring ; { File list separator (NULL means none) }
                        Specification : string ; { Original specification }
                        File_Name : string ; { Full name and path of found file }
                        File_List : string ; { comma-delimited list of paths }
                        FS : PSFile_System ; { File system }
                        Wildcard_Context : string ; { Used by FindFirst/FindNext }
                        Wildcard_Attributes : integer ; { Used by FindFirst/FindNext }
                        Last_Name : string ; { Used by FindFirst/FindNext }
                        Found_Name : string ; { Name (not path) of found file }
                        Found_Attributes : longint ; { Attributes of found file }
{$IFDEF CC_DEBUG}
                        Sanity : string[ 13 ] ;
                        Serial_Number : integer ;
{$ENDIF}
                        FR : TSearchRec ; { Special data for DOS/Windows }
                        Data : array[ 0..511 ] of byte ; { Context information - file system dependant }
                    end ;
     PFind_Record = ^TFind_Record ;

     TSWildcard = object( TBase_Object )
                     public { Constructors and destructors... }
                         destructor Done ; virtual ;

                     public { API... }
                         function Lookup( const Name, Path : Ansistring ;
                             var Context : TFind_Record ) : Ansistring ; virtual ;
                 end ; { TSWildcard }

     TSFile_System = object( TBase_Object )   { Extract to it's own unit }
                        public { Constructors and destructors... }
                            destructor Done ; virtual ;

                        public
                            IO_Error : PSUnified_Exception ;

                        public { API... }
                            procedure Filespec_String_Scan( const Name, _Separator : Ansistring ;
                                var Info : TFSSS ) ; virtual ;

                            procedure Set_Error( X : PSUnified_Exception ) ;
                                virtual ;

                            { Increment reference count to file system. }
                            procedure Attach ; virtual ;

                            { Decrement reference count to file system. }
                            procedure Detach ; virtual ;

                            { Close an open file via handle. }
                            procedure Close( Handle : integer ) ; virtual ;

                            { Close an open find record. }
                            procedure Close_Find( var Find_Record : TFind_Record ) ;
                                virtual ;

                            { Create file and open with specified mode.  If
                              file already exists, it is deleted and re-created.
                              Returns the file handle (0 if an error). }
                            function Create( const Name : Ansistring ;
                                Mode : integer ) : integer ; virtual ;

                            { Deletes the specified file. }
                            procedure Erase( const Name : Ansistring ) ; virtual ;

                            { Returns name associated with specified file. }
                            function File_Name( Handle : integer ) : Ansistring ;
                                 virtual ;

                            function First_Directory( const Name : Ansistring ) : integer ;
                                virtual ;

                            { Flush any buffered output for specified file. }
                            procedure Flush( Handle : integer ) ; virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_Attributes( Handle : integer ) : integer ;
                                 virtual ;

                            { Returns size of specified file, in bytes. }
                            function Get_Size( Handle : integer ) : TStore_Address ;
                                 virtual ;

                            function Get_Time( Handle, Typ : integer ) : Longint ;
                                 virtual ;

                            { Normalize the passed path: (device:)\path }
                            function Normalize_Path( Name : Ansistring ) : Ansistring ;
                                virtual ;

                            { Open existing file with specified mode.  Returns
                              the file handle (0 if an error). }
                            function Open( const Name : Ansistring ;
                                Mode : integer ) : integer ; virtual ;

                            { Read from specified file at specified position
                              into Buffer, up to Len bytes.  Actual bytes read
                              is returned in Count.  Count is ignored on call. }
                            procedure Read( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ; virtual ;

                            { Renames Old file to New file name. }
                            procedure Rename( const Old, New : Ansistring ) ;
                                 virtual ;

                            procedure Set_Attributes( Handle : integer ;
                                Attr : longint ) ; virtual ;

                            procedure Set_Time( Handle : integer ;
                                Typ, Time : Longint ) ; virtual ;

                            { Sets the size of the specified file to the
                              specified size, in bytes. }
                            procedure Set_Size( Handle : integer ;
                                New_Size : TStore_Address ) ; virtual ;

                            { Write to specified file at specified position from
                              Buffer, up to Len bytes.  Actual bytes written is
                              returned in Count.  Count is ignored on call. }
                            procedure Write( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ;
                                virtual ;

                            { Returns last error code. }
                            function Last_Error : PSUnified_Exception ;
                                virtual ;

                            function DeviceSize( Nam : Ansistring ) : TStore_Address ;
                                virtual ;

                            function DeviceFree( Nam : Ansistring ) : TStore_Address ;
                                virtual ;

                            function MkDir( Nam : Ansistring ) : longint ;
                                virtual ;

                            procedure Get_Header( const _Path : Ansistring ;
                                var Header : FS_Header_Information ;
                                 var Res : PSUnified_Exception ) ; virtual ;

                            procedure Set_Header( const _Path : Ansistring ;
                                var Header : FS_Header_Information ;
                                var Res : PSUnified_Exception ) ; virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_File_Attributes( const Name : Ansistring ) : longint ;
                                 virtual ;

                            { Returns the attributes of the specified file. }
                            procedure Set_File_Attributes( const Name : Ansistring ;
                                Value : longint ) ; virtual ;

                            function FindFirst( Nam : Ansistring ; Attributes : longint ;
                                var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            function FindNext( var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            procedure Initialize( Nam, ID : Ansistring ;
                                Flags : longint ; var UEC : PSUnified_Exception ) ; virtual ;

                            procedure Initialize_Store( S : PSStore ; ID : Ansistring ;
                                Flags : longint ; var UEC : PSUnified_Exception ) ; virtual ;

                            procedure Verify( Name : Ansistring ; var Res : PSUnified_Exception ) ;
                                virtual ;

                            function Path : Ansistring ; virtual ;

                            { What character or string is used to separate
                              paths from each other and the filename.  If the
                              file system doesn't support paths, this returns a
                              null string. }
                            function Separator : Ansistring ; virtual ;

                            function Facility : longint ; virtual ;

                            function Version : longint ; virtual ;

                            { Locks the file system for synchronous I/O.
                              Returns True if this process was the previously
                              locking process. }
                            function Lock( Nam : Ansistring ) : boolean ; virtual ;

                            { Unlocks the file system. }
                            procedure Unlock( Nam : Ansistring ) ; virtual ;
                     end ; { TSFile_System }

type TSDefault_File_System_Server = object( TBase_Object )
                                        public { API... }
                                            function Get_File_System_For_Store( Store : PSStore ;
                                                Own : boolean ) : PSFile_System ;
                                                virtual ;
                                            { Creates and returns a file system for the
                                              passed store.  If there is no known file
                                              system on the store, nil is returned. }

                                            function Get_File_System( const Nam : Ansistring ) : PSFile_System ;
                                                virtual ;
                                            { Creates and returns a file system for the
                                              passed file specification.  If there is no
                                              known file system on the store, nil is
                                              returned. }

                                            procedure Initialize( const Nam, ID : Ansistring ;
                                                Flags : integer ; var UEC : PSUnified_Exception ) ;
                                                virtual ;
                                            { Initializes a file system on the specified
                                              device, with the specified ID and flags.
                                              Returns status in UEC. }
                                        end ; { TSFile_System_Server }
     PSDefault_File_System_Server = ^TSDefault_File_System_Server ;


procedure Initialize_Find_Record( var F : TFind_Record ) ;
procedure Uninitialize_Find_Record( var F : TFind_Record ) ;

implementation

uses { Subroutine Library - Layer 0... }
     _Common, { Abstract_Error }

     { Subroutine Library - Layer 1... }
     _Applic ; { Application_Manager }


{ Unit API... }

procedure Initialize_Find_Record( var F : TFind_Record ) ;

begin
    fillchar( F, sizeof( F ), 0 ) ;
{$IFDEF CC_DELPHI}
    F.Initialized := 0 ;
    F.Context := nil ;
    F.FS := nil ;
    fillchar( F.Data, sizeof( F.Data ), 0 ) ;
{$ELSE}
    fillchar( F, sizeof( F ), 0 ) ;
{$ENDIF}
    F.Separator := ',' ; { Default file separator }
{$IFDEF CC_DEBUG}
    F.Sanity := 'TFind_Record' ;
    F.Serial_Number := Application_Manager^.Get_Serial_Number( 0 ) ;
{$ENDIF}
end ;


procedure Uninitialize_Find_Record( var F : TFind_Record ) ;

begin
    if( F.Context <> nil ) then
    begin
        dispose( F.Context, Done ) ;
        F.Context := nil ;
    end ;
    if( F.FS <> nil ) then
    begin
        F.FS^.Close_Find( F ) ;
    end ;
    F.FS := nil ;
{$IFDEF CC_DEBUG}
    F.Sanity := ' ' ;
{$ENDIF}
end ; { Uninitialize_Find_Record }



{ TSWildcard methods... }

{ Constructors and destructors... }

destructor TSWildcard.Done ;

begin
    inherited Done ;
end ;


{ API... }

function TSWildcard.Lookup( const Name, Path : Ansistring ;
    var Context : TFind_Record ) : Ansistring ;

begin
    Lookup := '' ;
    Abstract_Error ;
end ;



{ TSFile_System methods... }

{ Constructors and destructors... }

destructor TSFile_System.Done ;

begin
    inherited Done ;
end ;


{ API... }

procedure TSFile_System.Filespec_String_Scan( const Name, _Separator : Ansistring ;
    var Info : TFSSS ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Set_Error( X : PSUnified_Exception ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Attach ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Detach ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Close( Handle : integer ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Close_Find( var Find_Record : TFind_Record ) ;

begin
    Abstract_Error ;
end ;


function TSFile_System.Create( const Name : Ansistring ;
    Mode : integer ) : integer ;

begin
    Create := 0 ;
    Abstract_Error ;
end ;


procedure TSFile_System.Erase( const Name : Ansistring ) ;

begin
    Abstract_Error ;
end ;


function TSFile_System.File_Name( Handle : integer ) : Ansistring ;

begin
    File_Name := '' ;
    Abstract_Error ;
end ;


function TSFile_System.First_Directory( const Name : Ansistring ) : integer ;

begin
    First_Directory := 0 ;
    Abstract_Error ;
end ;


procedure TSFile_System.Flush( Handle : integer ) ;

begin
    Abstract_Error ;
end ;


function TSFile_System.Get_Attributes( Handle : integer ) : integer ;

begin
    Get_Attributes := 0 ;
    Abstract_Error ;
end ;


function TSFile_System.Get_Size( Handle : integer ) : TStore_Address ;

begin
    Get_Size := 0 ;
    Abstract_Error ;
end ;


function TSFile_System.Get_Time( Handle, Typ : integer ) : Longint ;

begin
    Get_Time := 0 ;
    Abstract_Error ;
end ;


function TSFile_System.Normalize_Path( Name : Ansistring ) : Ansistring ;

begin
    Normalize_Path := '' ;
    Abstract_Error ;
end ;


function TSFile_System.Open( const Name : Ansistring ;
    Mode : integer ) : integer ;

begin
    Open := 0 ;
    Abstract_Error ;
end ;


procedure TSFile_System.Read( Handle : integer ;
    Position : TStore_Address ; Len : TStore_Address ;
    var Buffer ; var Count : TStore_Address ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Rename( const Old, New : Ansistring ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Set_Attributes( Handle : integer ;
    Attr : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Set_Time( Handle : integer ;
    Typ, Time : Longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Set_Size( Handle : integer ;
    New_Size : TStore_Address ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_System.Write( Handle : integer ;
    Position : TStore_Address ; Len : TStore_Address ;
    var Buffer ; var Count : TStore_Address ) ;

begin
    Abstract_Error ;
end ;


function TSFile_System.Last_Error : PSUnified_Exception ;

begin
    Last_Error := nil ;
    Abstract_Error ;
end ;


function TSFile_System.DeviceSize( Nam : Ansistring ) : TStore_Address ;

begin
    DeviceSize := -1 ;
    Abstract_Error ;
end ;


function TSFile_System.DeviceFree( Nam : Ansistring ) : TStore_Address ;

begin
    DeviceFree := -1 ;
    Abstract_Error ;
end ;


function TSFile_System.MkDir( Nam : Ansistring ) : longint ;

begin
    MkDir := -1 ;
    Abstract_Error ;
end ;


procedure TSFile_System.Get_Header( const _Path : Ansistring ;
    var Header : FS_Header_Information ; var Res : PSUnified_Exception ) ;

begin
    fillchar( Res, sizeof( Res ), 0 ) ;
    Abstract_Error ;
end ;


procedure TSFile_System.Set_Header( const _Path : Ansistring ;
    var Header : FS_Header_Information ;
    var Res : PSUnified_Exception ) ;

begin
    fillchar( Res, sizeof( Res ), 0 ) ;
    Abstract_Error ;
end ;


function TSFile_System.Get_File_Attributes( const Name : Ansistring ) : longint ;

begin
    Get_File_Attributes := 0 ;
    Abstract_Error ;
end ;


procedure TSFile_System.Set_File_Attributes( const Name : Ansistring ;
    Value : longint ) ;

begin
    Abstract_Error ;
end ;


function TSFile_System.FindFirst( Nam : Ansistring ; Attributes : longint ;
    var Find_Record : TFind_Record ) : longint ;

begin
    FindFirst := -1 ;
    Abstract_Error ;
end ;


function TSFile_System.FindNext( var Find_Record : TFind_Record ) : longint ;

begin
    FindNext := -1 ;
    Abstract_Error ;
end ;


procedure TSFile_System.Initialize( Nam, ID : Ansistring ;
    Flags : longint ; var UEC : PSUnified_Exception ) ;

begin
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Abstract_Error ;
end ;


procedure TSFile_System.Initialize_Store( S : PSStore ; ID : Ansistring ;
    Flags : longint ; var UEC : PSUnified_Exception ) ;

begin
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Abstract_Error ;
end ;


procedure TSFile_System.Verify( Name : Ansistring ; var Res : PSUnified_Exception ) ;

begin
    fillchar( Res, sizeof( Res ), 0 ) ;
    Abstract_Error ;
end ;


function TSFile_System.Path : Ansistring ;

begin
    Path := '' ;
    Abstract_Error ;
end ;


function TSFile_System.Separator : Ansistring ;

begin
    Separator := '\' ;
    Abstract_Error ;
end ;


function TSFile_System.Facility : longint ;

begin
    Facility := -1 ;
    Abstract_Error ;
end ;


function TSFile_System.Version : longint ;

begin
    Version := 0 ;
    Abstract_Error ;
end ;


function TSFile_System.Lock( Nam : Ansistring ) : boolean ;

begin
    Lock := False ;
    Abstract_Error ;
end ;


procedure TSFile_System.Unlock( Nam : Ansistring ) ;

begin
    Abstract_Error ;
end ;


function TSDefault_File_System_Server.Get_File_System_For_Store( Store : PSStore ;
    Own : boolean ) : PSFile_System ;

begin
    Get_File_System_For_Store := nil ;
    Abstract_Error ;
end ;


function TSDefault_File_System_Server.Get_File_System( const Nam : Ansistring ) : PSFile_System ;

begin
    Get_File_System := nil ;
    Abstract_Error ;
end ;


procedure TSDefault_File_System_Server.Initialize( const Nam, ID : Ansistring ;
    Flags : integer ; var UEC : PSUnified_Exception ) ;

begin
    Abstract_Error ;
end ;


end.
