{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
unit _Stores ;
{
        Program Name : _Stores
        Package Name : Subroutine Library - Layer 1
        Purpose      : Storage object definitions
        Institution  : Conroy & Conroy Co.
        Date Written : 1999
        Written By   : Alan Conroy
        Version      : 1.1B

        Copyright (C) 1999-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

          1-Nov-2002    EAC         Use TObject from Common.


        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TStore is the base class for all storage-related classes.  It is an
        unmanaged store, meaning that it serves as a simple holder for data.  A
        store is assumed to be a contiguous storage area of a specific size.
        It has a base address of 0 (relative to store - not to memory).  A
        resize can be requested, but may not be supported.  It can be read/write
        read-only, or write-only.  This is commonly referred to as a BLOb
        (Binary Large OBject).
}

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

{$IFDEF FPC_OR_DELPHI}
uses Classes,
{$ELSE}
uses Objects,
{$ENDIF}
     { Layer 0... }
     ComInter, { TCommon_COM_Interface }
     CommBase, { TBase_Object }
{$IFNDEF MOBILE}
     _DebugIn, { PSDebug_Interface }
{$ENDIF}
     _Types, { int64 }

     { Layer 1... }
     _Cache, { PSCache }
     _UE ; { TUnified_Exception }

{$IFDEF _64BIT}
type TStore_Address = int64 ;
type TStore_Size = int64 ;
{$ELSE}
type TStore_Address = longint ;
type TStore_Size = longint ;
{$ENDIF}
type TStore_Address64 = int64 ;
type TStore_Size64 = int64 ;

const Null : TStore_Address = 0 ;
{$IFDEF _32BIT}
const Null64 : int64 = 0 ;
{$ELSE}
  {$IFDEF _64BIT}
const Null64 : int64 = 0 ;
  {$ELSE}
const Null64 : int64 = 0.0 ;
  {$ENDIF}
{$ENDIF}

type TSStore = object( TBase_Object )
                   public { API... }
                       function Read_Data( var Data ; Address, _Size : TStore_Address ;
                           var UEC : PSUnified_Exception ) : TStore_Address ; virtual ;
                       { Read data from specified Address in store to Data.
                         Read _Size bytes.  Returns success of operation in UEC.
                         Function returns the number of bytes actually read. }

                       function Write_Data( var Data ; Address, _Size : TStore_Address ;
                           var UEC : PSUnified_Exception ) : TStore_Address ; virtual ;
                       { Write data from Data to specified Address in store.
                         Write _Size bytes.  Returns success of operation in UEC.
                         Function returns the number of bytes actually written. }

                       function Max_Storage : TStore_Size ; virtual ;
                       { Returns the maximum (current) size of the store, in
                         bytes.  In other words, this is the highest valid
                         address (plus 1). }

                       function Min_Storage : TStore_Address ; virtual ;
                       { Returns the minimum atomic size of the store.  For
                         instance, this would be the sector size for a disk. }

                       function Extend( Amount : TStore_Address ) : TStore_Address ;
                           virtual ;
                       { Requests extension of store by the amount specified.
                         Returns the amount actually added to store.  If Amount
                         is 0, function returns 0 if store is extendable, and -1
                         if it is not. }

                       function Get_Read_Only : boolean ; virtual ;
                       { Indicates that the store can only be read from. }

                       function Get_Write_Only : boolean ; virtual ;
                       { Indicates that the store can only be written to. }

                       procedure Initialize ; virtual ;
                       { Initializes the TStore object. }

                       procedure Terminate ; virtual ;
                       { Terminates the TStore object. }

                       procedure Format ; virtual ;
                       { Does a low-level format of the store. }

                       function Get_Name : Ansistring ; virtual ;
                       { Returns name of store. }

{$IFNDEF MOBILE}
                       function Debugger : PSDebug_Interface ; virtual ;
                       { Returns a debugger object. }
{$ENDIF}

                       function Get_Cache : PSCache ; virtual ;
                       { Returns current cache object. }

                       procedure Set_Cache( Value : PSCache ) ; virtual ;
                       { Sets the current cache object. }

                       function Contiguous_Store : boolean ; virtual ;
                       { Returns True if the store's addresses are contiguous
                         (non-sparse). }

                       procedure Set_Max_Storage( Value : TStore_Address ;
                           var Res : PSUnified_Exception ) ; virtual ;
                       { Sets the maximum (current) size of the store, in bytes.
                         In other words, this is the new highest valid address
                         (plus 1).  If the store is not resizable, the call
                         fails and an error code is returned in Res.  This should
                         only be used to shrink the size of the store, not to
                         extend it.  To extend, use the Extend method. }

                       function Extended_Size : TStore_Address ; virtual ;
                       { Returns the maximum theoretical size that the store can
                         be extended to.  If the store is not extendable, this
                         returns the same value as Max_Storage. }

                       function Get_Bytes_Read : longint ;
                           virtual ; { Data bytes read }
                       function Get_Bytes_Written : longint ;
                           virtual ; { Data bytes written }
                       function Get_Reads : longint ;
                           virtual ; { Data read count }
                       function Get_Writes : longint ;
                           virtual ; { Data write count }
                       function Get_Error_Count : longint ;
                           virtual ; { Error count }
                       procedure Set_Bytes_Read( Value : longint ) ;
                           virtual ; { Data bytes read }
                       procedure Set_Bytes_Written( Value : longint ) ;
                           virtual ; { Data bytes written }
                       procedure Set_Reads( Value : longint ) ;
                           virtual ; { Data read count }
                       procedure Set_Writes( Value : longint ) ;
                           virtual ; { Data write count }
                       procedure Set_Error_Count( Value : longint ) ;
                           virtual ; { Error count }
                       procedure Set_Read_Only( Value : boolean ) ;
                           virtual ; { Sets the store read-only value. }
                       procedure Set_Write_Only( Value : boolean ) ;
                           virtual ;

{$IFDEF PROPERTIES}
                       property Read_Only : boolean
                           read Get_Read_Only
                           write Set_Read_Only ;

                       property Write_Only : boolean
                           read Get_Write_Only
                           write Set_Write_Only ;

                       property Bytes_Read : longint
                           read Get_Bytes_Read
                           write Set_Bytes_Read ;

                       property Bytes_Written : longint
                           read Get_Bytes_Written
                           write Set_Bytes_Written ;

                       property Reads : longint
                           read Get_Reads
                           write Set_Reads ;

                       property Writes : longint
                           read Get_Writes
                           write Set_Writes ;

                       property Error_Count : longint
                           read Get_Error_Count
                           write Set_Error_Count ;
{$ENDIF}
               end ; { TSStore }
     PSStore = ^TSStore ;

type TSStore64 = object( TBase_Object )
                   public { API... }
                       function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                           var UEC : PSUnified_Exception ) : TStore_Address64 ; virtual ;
                       { Read data from specified Address in store to Data.
                         Read _Size bytes.  Returns success of operation in UEC.
                         Function returns the number of bytes actually read. }

                       function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                           var UEC : PSUnified_Exception ) : TStore_Address64 ; virtual ;
                       { Write data from Data to specified Address in store.
                         Write _Size bytes.  Returns success of operation in UEC.
                         Function returns the number of bytes actually written. }

                       function Max_Storage : TStore_Size64 ; virtual ;
                       { Returns the maximum (current) size of the store, in
                         bytes.  In other words, this is the highest valid
                         address (plus 1). }

                       function Min_Storage : TStore_Address64 ; virtual ;
                       { Returns the minimum atomic size of the store.  For
                         instance, this would be the sector size for a disk. }

                       function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                           virtual ;
                       { Requests extension of store by the amount specified.
                         Returns the amount actually added to store.  If Amount
                         is 0, function returns 0 if store is extendable, and -1
                         if it is not. }

                       function Get_Read_Only : boolean ; virtual ;
                       { Indicates that the store can only be read from. }

                       function Get_Write_Only : boolean ; virtual ;
                       { Indicates that the store can only be written to. }

                       procedure Initialize ; virtual ;
                       { Initializes the TStore object. }

                       procedure Terminate ; virtual ;
                       { Terminates the TStore object. }

                       procedure Format ; virtual ;
                       { Does a low-level format of the store. }

                       function Get_Name : Ansistring ; virtual ;
                       { Returns name of store. }

{$IFNDEF MOBILE}
                       function Debugger : PSDebug_Interface ; virtual ;
                       { Returns a debugger object. }
{$ENDIF}

                       function Get_Cache : PSCache64 ; virtual ;
                       { Returns current cache object. }

                       procedure Set_Cache( Value : PSCache64 ) ; virtual ;
                       { Sets the current cache object. }

                       function Contiguous_Store : boolean ; virtual ;
                       { Returns True if the store's addresses are contiguous
                         (non-sparse). }

                       procedure Set_Max_Storage( Value : TStore_Address64 ;
                           var Res : PSUnified_Exception ) ; virtual ;
                       { Sets the maximum (current) size of the store, in bytes.
                         In other words, this is the new highest valid address
                         (plus 1).  If the store is not resizable, the call
                         fails and an error code is returned in Res.  This should
                         only be used to shrink the size of the store, not to
                         extend it.  To extend, use the Extend method. }

                       function Extended_Size : TStore_Address64 ; virtual ;
                       { Returns the maximum theoretical size that the store can
                         be extended to.  If the store is not extendable, this
                         returns the same value as Max_Storage. }

                       function Get_Bytes_Read : longint ;
                           virtual ; { Data bytes read }
                       function Get_Bytes_Written : longint ;
                           virtual ; { Data bytes written }
                       function Get_Reads : longint ;
                           virtual ; { Data read count }
                       function Get_Writes : longint ;
                           virtual ; { Data write count }
                       function Get_Error_Count : longint ;
                           virtual ; { Error count }
                       procedure Set_Bytes_Read( Value : longint ) ;
                           virtual ; { Data bytes read }
                       procedure Set_Bytes_Written( Value : longint ) ;
                           virtual ;{ Data bytes written }
                       procedure Set_Reads( Value : longint ) ;
                           virtual ; { Data read count }
                       procedure Set_Writes( Value : longint ) ;
                           virtual ; { Data write count }
                       procedure Set_Error_Count( Value : longint ) ;
                           virtual ; { Error count }
                       procedure Set_Read_Only( Value : boolean ) ;
                           virtual ; { Sets the store read-only value. }
                       procedure Set_Write_Only( Value : boolean ) ;
                           virtual ;

{$IFDEF PROPERTIES}
                       property Read_Only : boolean
                           read Get_Read_Only
                           write Set_Read_Only ;

                       property Write_Only : boolean
                           read Get_Write_Only
                           write Set_Write_Only ;

                       property Bytes_Read : longint
                           read Get_Bytes_Read
                           write Set_Bytes_Read ;

                       property Bytes_Written : longint
                           read Get_Bytes_Written
                           write Set_Bytes_Written ;

                       property Reads : longint
                           read Get_Reads
                           write Set_Reads ;

                       property Writes : longint
                           read Get_Writes
                           write Set_Writes ;

                       property Error_Count : longint
                           read Get_Error_Count
                           write Set_Error_Count ;
{$ENDIF}
               end ;
     PSStore64 = ^TSStore64 ;

{$IFDEF DYNAMIC_OBJECTS}
type TStore = class( TBase_Delphi_Object )
                  public { API... }
                      function Read_Data( var Data ; Address, _Size : TStore_Address ;
                          var UEC : TUnified_Exception ) : TStore_Address ; virtual ; abstract ;
                      { Read data from specified Address in store to Data.
                        Read _Size bytes.  Returns success of operation in UEC.
                        Function returns the number of bytes actually read. }

                      function Write_Data( var Data ; Address, _Size : TStore_Address ;
                          var UEC : TUnified_Exception ) : TStore_Address ; virtual ; abstract ;
                      { Write data from Data to specified Address in store.
                        Write _Size bytes.  Returns success of operation in UEC.
                        Function returns the number of bytes actually written. }

                      function Max_Storage : TStore_Size ; virtual ; abstract ;
                      { Returns the maximum (current) size of the store, in
                        bytes.  In other words, this is the highest valid
                        address (plus 1). }

                      function Min_Storage : TStore_Address ;
                          virtual ; abstract ;
                      { Returns the minimum atomic size of the store.  For
                        instance, this would be the sector size for a disk. }

                      function Extend( Amount : TStore_Address ) : TStore_Address ;
                          virtual ; abstract ;
                      { Requests extension of store by the amount specified.
                        Returns the amount actually added to store.  If Amount
                        is 0, function returns 0 if store is extendable, and -1
                        if it is not. }

                      function Get_Read_Only : boolean ; virtual ; abstract ;
                      { Indicates that the store can only be read from. }

                      function Get_Write_Only : boolean ; virtual ; abstract ;
                      { Indicates that the store can only be written to. }

                      procedure Initialize ; virtual ; abstract ;
                      { Initializes the TStore object. }

                      procedure Terminate ; virtual ; abstract ;
                      { Terminates the TStore object. }

                      procedure Format ; virtual ; abstract ;
                      { Does a low-level format of the store. }

                      function Get_Name : Ansistring ; virtual ; abstract ;
                      { Returns name of store. }

                      function Get_Cache : TCache ; virtual ; abstract ;
                      { Returns current cache object. }

                      procedure Set_Cache( Value : TCache ) ;
                          virtual ; abstract ;
                      { Sets the current cache object. }

                      function Contiguous_Store : boolean ;
                          virtual ; abstract ;
                      { Returns True if the store's addresses are contiguous
                        (non-sparse). }

                      procedure Set_Max_Storage( Value : TStore_Address ;
                          var Res : TUnified_Exception ) ; virtual ; abstract ;
                      { Sets the maximum (current) size of the store, in bytes.
                        In other words, this is the new highest valid address
                        (plus 1).  If the store is not resizable, the call
                        fails and an error code is returned in Res.  This should
                        only be used to shrink the size of the store, not to
                        extend it.  To extend, use the Extend method. }

                      function Extended_Size : TStore_Address ;
                          virtual ; abstract ;
                      { Returns the maximum theoretical size that the store can
                        be extended to.  If the store is not extendable, this
                        returns the same value as Max_Storage. }

                      procedure Set_Read_Only( Value : boolean ) ;
                          virtual ; abstract ;
                      { Sets the store read-only value. }

                      procedure Set_Write_Only( Value : boolean ) ;
                          virtual ; abstract ;
                      { Sets the store write-only value. }

                      function Get_Bytes_Read : longint ;
                          virtual ; abstract ; { Data bytes read }
                      function Get_Bytes_Written : longint ;
                          virtual ; abstract ; { Data bytes written }
                      function Get_Reads : longint ;
                          virtual ; abstract ; { Data read count }
                      function Get_Writes : longint ;
                          virtual ; abstract ; { Data write count }
                      function Get_Error_Count : longint ;
                          virtual ; abstract ; { Error count }
                      procedure Set_Bytes_Read( Value : longint ) ;
                          virtual ; abstract ; { Data bytes read }
                      procedure Set_Bytes_Written( Value : longint ) ;
                          virtual ; abstract ; { Data bytes written }
                      procedure Set_Reads( Value : longint ) ;
                          virtual ; abstract ; { Data read count }
                      procedure Set_Writes( Value : longint ) ;
                          virtual ; abstract ; { Data write count }
                      procedure Set_Error_Count( Value : longint ) ;
                          virtual ; abstract ; { Error count }

                      property Read_Only : boolean
                          read Get_Read_Only
                          write Set_Read_Only ;

                      property Write_Only : boolean
                          read Get_Write_Only
                          write Set_Write_Only ;

                      property Bytes_Read : longint
                          read Get_Bytes_Read
                          write Set_Bytes_Read ;

                      property Bytes_Written : longint
                          read Get_Bytes_Written
                          write Set_Bytes_Written ;

                      property Reads : longint
                          read Get_Reads
                          write Set_Reads ;

                      property Writes : longint
                          read Get_Writes
                          write Set_Writes ;

                      property Error_Count : longint
                          read Get_Error_Count
                          write Set_Error_Count ;
               end ;

type TStore64 = class( TBase_Delphi_Object )
                    public { API... }
                        function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                            var UEC : TUnified_Exception ) : TStore_Address64 ;
                            virtual ; abstract ;
                        { Read data from specified Address in store to Data.
                          Read _Size bytes.  Returns success of operation in UEC.
                          Function returns the number of bytes actually read. }

                        function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                            var UEC : TUnified_Exception ) : TStore_Address64 ;
                            virtual ; abstract ;
                        { Write data from Data to specified Address in store.
                          Write _Size bytes.  Returns success of operation in UEC.
                          Function returns the number of bytes actually written. }

                        function Max_Storage : TStore_Size64 ;
                            virtual ; abstract ;
                        { Returns the maximum (current) size of the store, in
                          bytes.  In other words, this is the highest valid
                          address (plus 1). }

                        function Min_Storage : TStore_Address64 ;
                            virtual ; abstract ;
                        { Returns the minimum atomic size of the store.  For
                          instance, this would be the sector size for a disk. }

                        function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                            virtual ; abstract ;
                        { Requests extension of store by the amount specified.
                          Returns the amount actually added to store.  If Amount
                          is 0, function returns 0 if store is extendable, and -1
                          if it is not. }

                        function Get_Read_Only : boolean ; virtual ; abstract ;
                        { Indicates that the store can only be read from. }

                        function Get_Write_Only : boolean ; virtual ; abstract ;
                        { Indicates that the store can only be written to. }

                        procedure Initialize ; virtual ; abstract ;
                        { Initializes the TStore object. }

                        procedure Terminate ; virtual ; abstract ;
                        { Terminates the TStore object. }

                        procedure Format ; virtual ; abstract ;
                        { Does a low-level format of the store. }

                        function Get_Name : Ansistring ; virtual ; abstract ;
                        { Returns name of store. }

                        function Get_Cache : TCache64 ; virtual ; abstract ;
                        { Returns current cache object. }

                        procedure Set_Cache( Value : TCache64 ) ;
                            virtual ; abstract ;
                        { Sets the current cache object. }

                        function Contiguous_Store : boolean ;
                            virtual ; abstract ;
                        { Returns True if the store's addresses are contiguous
                          (non-sparse). }

                        procedure Set_Max_Storage( Value : TStore_Address64 ;
                            var Res : TUnified_Exception ) ; virtual ; abstract ;
                        { Sets the maximum (current) size of the store, in bytes.
                          In other words, this is the new highest valid address
                          (plus 1).  If the store is not resizable, the call
                          fails and an error code is returned in Res.  This should
                          only be used to shrink the size of the store, not to
                          extend it.  To extend, use the Extend method. }

                        function Extended_Size : TStore_Address64 ;
                            virtual ; abstract ;
                        { Returns the maximum theoretical size that the store can
                          be extended to.  If the store is not extendable, this
                          returns the same value as Max_Storage. }

                        procedure Set_Read_Only( Value : boolean ) ;
                            virtual ; abstract ;
                        { Sets the store read-only value. }

                        procedure Set_Write_Only( Value : boolean ) ;
                            virtual ; abstract ;
                        { Sets the store write-only value. }

                        function Get_Bytes_Read : longint ;
                            virtual ; abstract ; { Data bytes read }
                        function Get_Bytes_Written : longint ;
                            virtual ; abstract ; { Data bytes written }
                        function Get_Reads : longint ;
                            virtual ; abstract ; { Data read count }
                        function Get_Writes : longint ;
                            virtual ; abstract ; { Data write count }
                        function Get_Error_Count : longint ;
                            virtual ; abstract ; { Error count }
                        procedure Set_Bytes_Read( Value : longint ) ;
                            virtual ; abstract ; { Data bytes read }
                        procedure Set_Bytes_Written( Value : longint ) ;
                            virtual ; abstract ; { Data bytes written }
                        procedure Set_Reads( Value : longint ) ;
                            virtual ; abstract ; { Data read count }
                        procedure Set_Writes( Value : longint ) ;
                            virtual ; abstract ; { Data write count }
                        procedure Set_Error_Count( Value : longint ) ;
                            virtual ; abstract ; { Error count }

                        property Read_Only : boolean
                            read Get_Read_Only
                            write Set_Read_Only ;

                        property Write_Only : boolean
                            read Get_Write_Only
                            write Set_Write_Only ;

                        property Bytes_Read : longint
                            read Get_Bytes_Read
                            write Set_Bytes_Read ;

                        property Bytes_Written : longint
                            read Get_Bytes_Written
                            write Set_Bytes_Written ;

                        property Reads : longint
                            read Get_Reads
                            write Set_Reads ;

                        property Writes : longint
                            read Get_Writes
                            write Set_Writes ;

                        property Error_Count : longint
                            read Get_Error_Count
                            write Set_Error_Count ;
                end ; { TStore64 }

type TCOM_Store = class( TCommon_COM_Interface )
                      public { API... }
                          function Read_Data( var Data ;
                              Address, _Size : TStore_Address ;
                              var UEC : TUnified_Exception ) : TStore_Address ;
                              virtual ; stdcall ; abstract ;
                          { Read data from specified Address in store to Data.
                            Read _Size bytes.  Returns success of operation in UEC.
                            Function returns the number of bytes actually read. }

                          function Write_Data( var Data ;
                              Address, _Size : TStore_Address ;
                              var UEC : TUnified_Exception ) : TStore_Address ;
                                  virtual ; stdcall ; abstract ;
                          { Write data from Data to specified Address in store.
                            Write _Size bytes.  Returns success of operation in UEC.
                            Function returns the number of bytes actually written. }

                          function Max_Storage : TStore_Size ;
                              virtual ; stdcall ; abstract ;
                          { Returns the maximum (current) size of the store, in
                            bytes.  In other words, this is the highest valid
                            address (plus 1). }

                          function Min_Storage : TStore_Address ;
                              virtual ; stdcall ; abstract ;
                          { Returns the minimum atomic size of the store.  For
                            instance, this would be the sector size for a disk. }

                          function Extend( Amount : TStore_Address ) : TStore_Address ;
                              virtual ; stdcall ; abstract ;
                          { Requests extension of store by the amount specified.
                            Returns the amount actually added to store.  If Amount
                            is 0, function returns 0 if store is extendable, and -1
                            if it is not. }

                          function Get_Read_Only : boolean ;
                              virtual ; stdcall ; abstract ;
                          { Indicates that the store can only be read from. }

                          function Get_Write_Only : boolean ;
                              virtual ; stdcall ; abstract ;
                          { Indicates that the store can only be written to. }

                          procedure Format ; virtual ; stdcall ; abstract ;
                          { Does a low-level format of the store. }

                          function Get_Name : PAnsiChar ;
                              virtual ; stdcall ; abstract ;
                          { Returns name of store. }

                          function Get_Cache : TCOM_Cache ;
                              virtual ; stdcall ; abstract ;
                          { Returns current cache object. }

                          procedure Set_Cache( Value : TCOM_Cache ) ;
                              virtual ; stdcall ; abstract ;
                          { Sets the current cache object. }

                          function Contiguous_Store : boolean ;
                              virtual ; stdcall ; abstract ;
                          { Returns True if the store's addresses are contiguous
                            (non-sparse). }

                          procedure Set_Max_Storage( Value : TStore_Address ;
                              var Res : TUnified_Exception ) ; virtual ; stdcall ; abstract ;
                          { Sets the maximum (current) size of the store, in bytes.
                            In other words, this is the new highest valid address
                            (plus 1).  If the store is not resizable, the call
                            fails and an error code is returned in Res.  This should
                            only be used to shrink the size of the store, not to
                            extend it.  To extend, use the Extend method. }

                          function Extended_Size : TStore_Address ;
                              virtual ; stdcall ; abstract ;
                          { Returns the maximum theoretical size that the store can
                            be extended to.  If the store is not extendable, this
                            returns the same value as Max_Storage. }

                          function Get_Bytes_Read : longint ;
                              virtual ; stdcall ; abstract ; { Data bytes read }
                          function Get_Bytes_Written : longint ;
                              virtual ; stdcall ; abstract ; { Data bytes written }
                          function Get_Reads : longint ;
                              virtual ; stdcall ; abstract ; { Data read count }
                          function Get_Writes : longint ;
                              virtual ; stdcall ; abstract ; { Data write count }
                          function Get_Error_Count : longint ;
                              virtual ; stdcall ; abstract ; { Error count }
                          procedure Set_Bytes_Read( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Data bytes read }
                          procedure Set_Bytes_Written( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Data bytes written }
                          procedure Set_Reads( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Data read count }
                          procedure Set_Writes( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Data write count }
                          procedure Set_Error_Count( Value : longint ) ;
                              virtual ; stdcall ; abstract ; { Error count }
                          procedure Set_Read_Only( Value : boolean ) ;
                              virtual ; stdcall ; abstract ; { Sets the store read-only value. }
                          procedure Set_Write_Only( Value : boolean ) ;
                              virtual ; stdcall ; abstract ; 

                          property Read_Only : boolean
                              read Get_Read_Only
                              write Set_Read_Only ;

                          property Write_Only : boolean
                              read Get_Write_Only
                              write Set_Write_Only ;

                          property Bytes_Read : longint
                              read Get_Bytes_Read
                              write Set_Bytes_Read ;

                          property Bytes_Written : longint
                              read Get_Bytes_Written
                              write Set_Bytes_Written ;

                          property Reads : longint
                              read Get_Reads
                              write Set_Reads ;

                          property Writes : longint
                              read Get_Writes
                              write Set_Writes ;

                          property Error_Count : longint
                              read Get_Error_Count
                              write Set_Error_Count ;
                   end ;

type TCOM_Managed_Store64_Debugger = class( TCommon_COM_Interface )
                                         public // API...
                                             procedure Log_Allocation( Address, Length : TStore_Address64 ; Special : boolean ) ;
                                                 virtual ; stdcall ; abstract ;
                                             procedure Log_Deallocation( Address, Length : TStore_Address64 ; Special : boolean ) ;
                                                 virtual ; stdcall ; abstract ;
                                             procedure Log_Read( Address, Length : TStore_Address64 ) ;
                                                 virtual ; stdcall ; abstract ;
                                             procedure Log_Write( Address, Length : TStore_Address64 ) ;
                                                 virtual ; stdcall ; abstract ;
                                             procedure Clear ;
                                                 virtual ; stdcall ; abstract ;
                                     end ;

type TCOM_Store64 = class( TCommon_COM_Interface )
                        public { API... }
                            function Read_Data( var Data ; Address, _Size : TStore_Address64 ;
                                var UEC : TUnified_Exception ) : TStore_Address64 ;
                                virtual ; stdcall ; abstract ;
                            { Read data from specified Address in store to Data.
                              Read _Size bytes.  Returns success of operation in UEC.
                              Function returns the number of bytes actually read. }

                            function Write_Data( var Data ; Address, _Size : TStore_Address64 ;
                                var UEC : TUnified_Exception ) : TStore_Address64 ;
                                virtual ; stdcall ; abstract ;
                            { Write data from Data to specified Address in store.
                              Write _Size bytes.  Returns success of operation in UEC.
                              Function returns the number of bytes actually written. }

                            function Max_Storage : TStore_Size64 ;
                                virtual ; stdcall ; abstract ;
                            { Returns the maximum (current) size of the store, in
                              bytes.  In other words, this is the highest valid
                              address (plus 1). }

                            function Min_Storage : TStore_Address64 ;
                                virtual ; stdcall ; abstract ;
                            { Returns the minimum atomic size of the store.  For
                              instance, this would be the sector size for a disk. }

                            function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                virtual ; stdcall ; abstract ;
                            { Requests extension of store by the amount specified.
                              Returns the amount actually added to store.  If Amount
                              is 0, function returns 0 if store is extendable, and -1
                              if it is not. }

                            function Get_Read_Only : boolean ;
                                virtual ; stdcall ; abstract ;
                            { Indicates that the store can only be read from. }

                            function Get_Write_Only : boolean ;
                                virtual ; stdcall ; abstract ;
                            { Indicates that the store can only be written to. }

                            procedure Format ; virtual ; stdcall ; abstract ;
                            { Does a low-level format of the store. }

                            function Get_Name : PAnsiChar ;
                                virtual ; stdcall ; abstract ;
                            { Returns name of store. }

                            function Get_Cache : TCOM_Cache64 ;
                                virtual ; stdcall ; abstract ;
                            { Returns current cache object. }

                            procedure Set_Cache( Value : TCOM_Cache64 ) ;
                                virtual ; stdcall ; abstract ;
                            { Sets the current cache object. }

                            function Contiguous_Store : boolean ;
                                virtual ; stdcall ; abstract ;
                            { Returns True if the store's addresses are contiguous
                              (non-sparse). }

                            procedure Set_Max_Storage( Value : TStore_Address64 ;
                                var Res : TUnified_Exception ) ;
                                virtual ; stdcall ; abstract ;
                            { Sets the maximum (current) size of the store, in bytes.
                              In other words, this is the new highest valid address
                              (plus 1).  If the store is not resizable, the call
                              fails and an error code is returned in Res.  This should
                              only be used to shrink the size of the store, not to
                              extend it.  To extend, use the Extend method. }

                            function Extended_Size : TStore_Address64 ;
                                virtual ; stdcall ; abstract ;
                            { Returns the maximum theoretical size that the store can
                              be extended to.  If the store is not extendable, this
                              returns the same value as Max_Storage. }

                            function Get_Bytes_Read : longint ;
                                virtual ; stdcall ; abstract ; { Data bytes read }
                            function Get_Bytes_Written : longint ;
                                virtual ; stdcall ; abstract ; { Data bytes written }
                            function Get_Reads : longint ;
                                virtual ; stdcall ; abstract ; { Data read count }
                            function Get_Writes : longint ;
                                virtual ; stdcall ; abstract ; { Data write count }
                            function Get_Error_Count : longint ;
                                virtual ; stdcall ; abstract ; { Error count }
                            procedure Set_Bytes_Read( Value : longint ) ;
                                virtual ; stdcall ; abstract ; { Data bytes read }
                            procedure Set_Bytes_Written( Value : longint ) ;
                                virtual ; stdcall ; abstract ; { Data bytes written }
                            procedure Set_Reads( Value : longint ) ;
                                virtual ; stdcall ; abstract ; { Data read count }
                            procedure Set_Writes( Value : longint ) ;
                                virtual ; stdcall ; abstract ; { Data write count }
                            procedure Set_Error_Count( Value : longint ) ;
                                virtual ; stdcall ; abstract ; { Error count }
                            procedure Set_Read_Only( Value : boolean ) ;
                                virtual ; stdcall ; abstract ; { Sets the store read-only value. }
                            procedure Set_Write_Only( Value : boolean ) ;
                                virtual ; stdcall ; abstract ; { Sets the store write-only value. }
                            function Get_Verifier : TCOM_Managed_Store64_Debugger ;
                                virtual ; stdcall ; abstract ;
                            procedure Set_Verifier( Value : TCOM_Managed_Store64_Debugger ) ;
                                virtual ; stdcall ; abstract ;

                            property Verifier : TCOM_Managed_Store64_Debugger
                                read Get_Verifier
                                write Set_Verifier ;

                            property Read_Only : boolean
                                read Get_Read_Only
                                write Set_Read_Only ;

                            property Write_Only : boolean
                                read Get_Write_Only
                                write Set_Write_Only ;

                            property Bytes_Read : longint
                                read Get_Bytes_Read
                                write Set_Bytes_Read ;

                            property Bytes_Written : longint
                                read Get_Bytes_Written
                                write Set_Bytes_Written ;

                            property Reads : longint
                                read Get_Reads
                                write Set_Reads ;

                            property Writes : longint
                                read Get_Writes
                                write Set_Writes ;

                            property Error_Count : longint
                                read Get_Error_Count
                                write Set_Error_Count ;
                    end ; { TCOM_Store64 }
{$ENDIF}

type TSStore_Server = object( TBase_Object )
                    public { API... }
                        function Get_Store_ID( Name : Ansistring ) : Ansistring ;
                            virtual ;
                        { Returns the portion of Name which is the ID of
                          the store.  For instance, the ID might be the
                          device name (eg C:) for a device. }

                        function Get_Store( var Name : Ansistring ) : PSStore ;
                            virtual ;
                        { Creates and returns a TStore object for the
                          passed specification.  On return, Name has the
                          device ID stripped from it.  If there is no store
                          available, nil is returned. }
                end ; { TSStore_Server }
     PSStore_Server = ^TSStore_Server ;


implementation

uses { Layer 0... }
     _Common ; { Abstract_Error }


{ TSStore methods... }

{ API... }

function TSStore.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : PSUnified_Exception ) : TStore_Address ;

begin
    Read_Data := Null ;
    UEC := nil ;
    Abstract_Error ;
end ;


function TSStore.Write_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : PSUnified_Exception ) : TStore_Address ;

begin
    Write_Data := Null ;
    UEC := nil ;
    Abstract_Error ;
end ;


function TSStore.Max_Storage : TStore_Size ;

begin
    Max_Storage := Null ;
    Abstract_Error ;
end ;


function TSStore.Min_Storage : TStore_Address ;

begin
    Min_Storage := Null ;
    Abstract_Error ;
end ;


function TSStore.Extend( Amount : TStore_Address ) : TStore_Address ;

begin
    Extend := Null ;
    Abstract_Error ;
end ;


function TSStore.Get_Read_Only : boolean ;

begin
    Get_Read_Only := False ;
    Abstract_Error ;
end ;


function TSStore.Get_Write_Only : boolean ;

begin
    Get_Write_Only := False ;
    Abstract_Error ;
end ;


procedure TSStore.Initialize ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Terminate ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Format ;

begin
    Abstract_Error ;
end ;


function TSStore.Get_Name : Ansistring ;

begin
    Get_Name := '' ;
    Abstract_Error ;
end ;


{$IFNDEF MOBILE}
function TSStore.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
    Abstract_Error ;
end ;
{$ENDIF}


function TSStore.Get_Cache : PSCache ;

begin
    Get_Cache := nil ;
    Abstract_Error ;
end ;


procedure TSStore.Set_Cache( Value : PSCache ) ;

begin
    Abstract_Error ;
end ;


function TSStore.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ;
    Abstract_Error ;
end ;


procedure TSStore.Set_Max_Storage( Value : TStore_Address ; var Res : PSUnified_Exception ) ;

begin
    Abstract_Error ;
end ;


function TSStore.Extended_Size : TStore_Address ;

begin
    Extended_Size := Null ;
    Abstract_Error ;
end ;


function TSStore.Get_Bytes_Read : longint ;

begin
    Get_Bytes_Read := 0 ;
    Abstract_Error ;
end ;


function TSStore.Get_Bytes_Written : longint ;

begin
    Get_Bytes_Written := 0 ;
    Abstract_Error ;
end ;


function TSStore.Get_Reads : longint ;

begin
    Get_Reads := 0 ;
    Abstract_Error ;
end ;


function TSStore.Get_Writes : longint ;

begin
    Get_Writes := 0 ;
    Abstract_Error ;
end ;


function TSStore.Get_Error_Count : longint ;

begin
    Get_Error_Count := 0 ;
    Abstract_Error ;
end ;


procedure TSStore.Set_Bytes_Read( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Set_Bytes_Written( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Set_Reads( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Set_Writes( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Set_Error_Count( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Set_Read_Only( Value : boolean ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore.Set_Write_Only( Value : boolean ) ;

begin
    Abstract_Error ;
end ;



{ TSStore64 method... }

{ API... }

function TSStore64.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : PSUnified_Exception ) : TStore_Address64 ;

begin
    Read_Data := Null64 ;
    Abstract_Error ;
end ;


function TSStore64.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : PSUnified_Exception ) : TStore_Address64 ;

begin
    Write_Data := Null64 ;
    Abstract_Error ;
end ;


function TSStore64.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := Null64 ;
    Abstract_Error ;
end ;


function TSStore64.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := Null64 ;
    Abstract_Error ;
end ;


function TSStore64.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

begin
    Extend := Null64 ;
    Abstract_Error ;
end ;


function TSStore64.Get_Read_Only : boolean ;

begin
    Get_Read_Only := False ;
    Abstract_Error ;
end ;


function TSStore64.Get_Write_Only : boolean ;

begin
    Get_Write_Only := False ;
    Abstract_Error ;
end ;


procedure TSStore64.Initialize ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Terminate ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Format ;

begin
    Abstract_Error ;
end ;


function TSStore64.Get_Name : Ansistring ;

begin
    Get_Name := '' ;
    Abstract_Error ;
end ;


{$IFNDEF MOBILE}
function TSStore64.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
    Abstract_Error ;
end ;
{$ENDIF}


function TSStore64.Get_Cache : PSCache64 ;

begin
    Get_Cache := nil ;
    Abstract_Error ;
end ;


procedure TSStore64.Set_Cache( Value : PSCache64 ) ;

begin
    Abstract_Error ;
end ;


function TSStore64.Contiguous_Store : boolean ;

begin
    Contiguous_Store := False ;
    Abstract_Error ;
end ;


procedure TSStore64.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : PSUnified_Exception ) ;

begin
    Abstract_Error ;
end ;


function TSStore64.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Null64 ;
    Abstract_Error ;
end ;


function TSStore64.Get_Bytes_Read : longint ;

begin
    Get_Bytes_Read := 0 ;
    Abstract_Error ;
end ;


function TSStore64.Get_Bytes_Written : longint ;

begin
    Get_Bytes_Written := 0 ;
    Abstract_Error ;
end ;


function TSStore64.Get_Reads : longint ;

begin
    Get_Reads := 0 ;
    Abstract_Error ;
end ;


function TSStore64.Get_Writes : longint ;

begin
    Get_Writes := 0 ;
    Abstract_Error ;
end ;


function TSStore64.Get_Error_Count : longint ;

begin
    Get_Error_Count := 0 ;
    Abstract_Error ;
end ;


procedure TSStore64.Set_Bytes_Read( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Set_Bytes_Written( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Set_Reads( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Set_Writes( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Set_Error_Count( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Set_Read_Only( Value : boolean ) ;

begin
    Abstract_Error ;
end ;


procedure TSStore64.Set_Write_Only( Value : boolean ) ;

begin
    Abstract_Error ;
end ;



{ TSStore_Server methods... }

{ API... }

function TSStore_Server.Get_Store_ID( Name : Ansistring ) : Ansistring ;

begin
    Get_Store_ID := '' ;
end ;


function TSStore_Server.Get_Store( var Name : Ansistring ) : PSStore ;

begin
    Get_Store := nil ;
end ;



end.
