/*
        Program Name : _TPM
        Package Name : Subroutine Library - Level 1
        Purpose      : Text Presentation Manager class definitions
        Institution  : Conroy & Conroy Co.
        Date Written : 27-Jun-2007
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2007-2009 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON                         

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The text presentation manager is used to provide text for use in
        various circumstances.
*/

#ifndef _TPMH
    #define _TPMH

class TCOM_Numeric_Text_Presentation_Manager : public TCommon_COM_Interface
{
    public: // API...
        virtual char* __stdcall Decimal() = 0;
        // Character(s) to use for decimal point.

        virtual char* __stdcall Group() = 0;
        // Character(s) to use for thounsand separator.

        virtual char* __stdcall Negative() = 0;
        // Character(s) to use for negative sign.

        virtual char* __stdcall Positive() = 0;
        // Character(s) to use for positive sign.
} ;


class TCOM_Monetary_Text_Presentation_Manager : public TCommon_COM_Interface
{
    public: // API...
        virtual char* __stdcall Decimal() = 0;
        // Character(s) to use for decimal point.

        virtual char* __stdcall Group() = 0;
        // Character(s) to use for thounsand separator.

        virtual char* __stdcall Money_Symbol() = 0;
        // Character(s) used for money.

        virtual char* __stdcall Internationl_Money_Symbol() = 0;
        // Character(s) used for international money.
} ;


class TCOM_Date_Text_Presentation_Manager : public TCommon_COM_Interface
{
    public: // API...
        virtual char* __stdcall AM() = 0;
        // Character used to indicate AM.

        virtual char* __stdcall Date_Separator() = 0;
        // Character(s) used to separate date components.

        virtual char* __stdcall Days( int M ) = 0;
        // Name of day with passed ordinal (1-7).

        virtual char* __stdcall Months( int M ) = 0;
        // Name of month with passed ordinal (1-13).

        virtual char* __stdcall PM() = 0;
        // Character used to indicate PM.

        virtual char* __stdcall Short_Days( int M ) = 0;
        // Name of day with passed ordinal (1-7), abbreviated.

        virtual char* __stdcall Short_Months( int M ) = 0;
        // Name of month with passed ordinal (1-13), abbreviated.

        virtual char* __stdcall Time_Separator() = 0;
        // Character(s) used to separate time components.
} ;


class TCOM_Text_Presentation_Manager : public TCommon_COM_Interface
{
    public: // API...
		virtual TCOM_Numeric_Text_Presentation_Manager* __stdcall Numeric_Text_Presentation_Manager() = 0;
        // Returns a presentation manager for numeric text.

		virtual TCOM_Monetary_Text_Presentation_Manager* __stdcall Monetary_Text_Presentation_Manager() = 0;
        // Returns a presentation manager for monetary text.

		virtual TCOM_Date_Text_Presentation_Manager* __stdcall Date_Text_Presentation_Manager() = 0;
        // Returns a presentation manager for date/time text.
} ;

#endif
