{
        Program Name : _UOS_File_System
        Package Name : Subroutine Library Layer 2
        Purpose      : UOS File System class definition
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Nov-2015
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Abstract UOS File System class definition.
}

unit _UOS_File_System ;

interface

uses // C&C Subroutine Library - Layer 0...
     ComInter, // TCommon_COM_Interface
     Heaps, // TCOM_Heap64
     _Stores, // TStore_Address64
     _Types, // Ansistring
     _UE, // TUnified_Exception

     // Subroutine Library - Layer 1...
     MStores ; // TCOM_Managed_Store64

{$I EDEFINES.INC}

const Max_Header_Data_Stream = 4 ;
      Max_Header_Cluster = 4 ;

type TData_Stream = packed record
                        Name : int64 ;
                        Pointer : TStore_Address64 ;
                    end ;

     TUOS_File_Header = packed record
                                // Name and sizes...
                                Name : longint ;
                                Size : int64 ; // Size on disk
                                EOF : int64 ; // Logical size
                                Uncompressed_Size : int64 ;
                                Clustersize : cardinal ;
                                Record_Size : cardinal ;

                                // Dates...
                                Creation : int64 ; // Creation date
                                Last_Modified : int64 ; // Last write date
                                Last_Backup : int64 ;
                                Last_Access : int64 ; // Last read/open date
                                Expiration : int64 ;

                                // Security...
                                Creator : cardinal ;
                                Owner : cardinal ;
                                ACL : int64 ;

                                // Misc...
                                Flags : int64 ;
                                Version_Limit : longint ;
                                Extension : TStore_Address64 ;

                                // Data streams...
                                Streams : array[ 0..Max_Header_Data_Stream ] of TData_Stream ;
                                Data_Stream : TStore_Address64 ; // Overflow pointer for additional streams
                                Clusters : array[ 0..Max_Header_Cluster ] of TStore_Address64 ;

                                // File-system data...
                                Parent : int64 ;
                                File_System : int64 ; // Reserved for file system
                        end ;

      // File Attribute Flags...
const FAF_DSM_MASK = 3 ; // Mask for data security mode
      FAF_PLACED = 4 ; // Placed at a specific location
      FAF_CONTIGUOUS = 8 ; // Contiguous file data
      FAF_DELETED = 16 ; // Deleted file
      FAF_READONLY = 32 ; // Read-only data
      FAF_LINK = 64 ; // Link to another file
      FAF_SYSTEM = 128 ; // System file
      FAF_HIDDEN = 256 ; // Hidden file
      FAF_DIRECTORY = 512 ; // Directory file
      FAF_PERMANENT = 1024 ; // Non-deletable file
      FAF_COPYING = 2048 ; // In the process of copying
      // 4096-32768 // Reserved

      // Protection codes...
      FAF_PROTECTION_OWNER_READ = $10000 ;
      FAF_PROTECTION_OWNER_WRITE = $20000 ;
      FAF_PROTECTION_OWNER_DELETE = $40000 ;
      FAF_PROTECTION_OWNER_EXECUTE = $80000 ;
      FAF_PROTECTION_GROUP_READ = $100000 ;
      FAF_PROTECTION_GROUP_WRITE = $200000 ;
      FAF_PROTECTION_GROUP_DELETE = $400000 ;
      FAF_PROTECTION_GROUP_EXECUTE = $800000 ;
      FAF_PROTECTION_SYSTEM_READ = $1000000 ;
      FAF_PROTECTION_SYSTEM_WRITE = $2000000 ;
      FAF_PROTECTION_SYSTEM_DELETE = $4000000 ;
      FAF_PROTECTION_SYSTEM_EXECUTE = $8000000 ;
      FAF_PROTECTION_WORLD_READ = $10000000 ;
      FAF_PROTECTION_WORLD_WRITE = $20000000 ;
      FAF_PROTECTION_WORLD_DELETE = $40000000 ;
      FAF_PROTECTION_WORLD_EXECUTE = $80000000 ;


type FS_Boot_Record = packed record // First 60 bytes of store header
                          Bootstrap1 : array[ 0..15 ] of byte ;
                          Header : TStore_Address64 ; // Address of FS header
                          HAL_Position : TStore_Address64 ;
                          HAL_Length : longint ;
                          FS_Position : TStore_Address64 ;
                          FS_Length : longint ;
                          Bootstrap_Position : TStore_Address64 ;
                          Bootstrap_Length : longint ;
                      end ;

     FS_Header = packed record // 240 bytes
                     ID : smallint ; // 255                                     // 0-1
                     ID1 : byte ; // 135                                        // 2
                     Version : byte ; { UOSFS format version 10 = V1.0 }        // 3
                     AT_Offset : TStore_Address64 ; { Current AT location }     // 4-11
                     AT_Size : TStore_Size64 ; { AT size }                      // 12-19
                     Flags : int64 ;                                            // 20-27
                     Clustersize : cardinal ; // Volume clustersize             // 28-31
                     Folder_Clustersize : cardinal ; // Default folder file clustersize // 32-35
                     Root : TStore_Address64 ; // Root folder header            // 36-43
{$IFDEF MOBILE}
                     Volume_Label : array[ 0..127 ] of byte ;                   // 44-171
{$ELSE}
                     Volume_Label : string[ 127 ] ;                             // 44-171
{$ENDIF}
{$IFDEF MOBILE}
                     Password : array[ 0..31 ] of byte ;                        // 172-203
{$ELSE}
                     Password : string[ 31 ] ;                                  // 172-203
{$ENDIF}
                     OS_Position : TStore_Address64 ;                           // 204-211
                     OS_Length : longint ;                                      // 212-215
                     MMC_Position : TStore_Address64 ;                          // 216-223
                     MMC_Length : longint ;                                     // 224-227
                     HMC_Position : TStore_Address64 ;                          // 228-235
                     HMC_Length : longint ;                                     // 236-239
                 end ;

     TUOS_File_Info = packed record
                          Size : int64 ; // Size on disk
                          EOF : int64 ; // Logical size
                          Uncompressed_Size : int64 ;
                          Clustersize : cardinal ;
                          Record_Size : cardinal ;
                          Creation : int64 ; // Creation date
                          Last_Modified : int64 ; // Last write date
                          Last_Backup : int64 ;
                          Last_Access : int64 ; // Last read/open date
                          Expiration : int64 ;
                          Creator : cardinal ;
                          Owner : cardinal ;
                          Flags : int64 ;
                          Version_Limit : longint ;
                          Location : int64 ;
                      end ; // TUOS_File_Info


      // File System Flags...

const UOSFSF_Dirty = 1 ; // 0=clean, 1=dirty
      UOSFSF_Private = 2 ; // 0=private volume, 2=public volume

      // File system mount flags...
const UOSFSM_Force = 1 ; // 1=force mount, 0=enfore password/label


type TScan_Notify = procedure( Sender : pointer ; Progress, Count : cardinal ;
         var Abort : boolean ) of object ;


type TUOS_File = class( TCommon_COM_Interface )
                     public // API...
                         // Add new stream...
                         function Create_Stream( Name : int64 ) : longint ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         // Delete stream...
                         procedure Delete_Stream( Name : int64 ; Index : longint ) ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         // Stearm information...
                         function Max_Stream : longint ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         function Stream_Name( Index : longint ) : int64 ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         // I/O...
                         function Read( Stream : longint ;
                             Position : TStore_Address64 ;
                             Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         function Write( Stream : longint ;
                             Position : TStore_Address64 ;
                             Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                	 	     function Get_Stream_Size( Stream : longint ) : TStore_Size64 ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                	 	     procedure Set_Stream_Size( Stream : longint ;
                             Value : TStore_Size64 ) ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         function Get_Contiguous : boolean ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         procedure Set_Contiguous( Value : boolean ) ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         function Get_File_Size : int64 ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         procedure Set_File_Size( Value : int64 ) ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         function Read_Only : boolean ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         function Write_Only : boolean ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                         function XSpaceAvail : int64 ;
                             virtual ;
{$IFDEF Use_STDCALL}
                             stdcall ;
{$ENDIF}
                             abstract ;

                     public // properties...
                         property Contiguous : boolean
                             read Get_Contiguous
                             write Set_Contiguous ;
                         property File_Size : int64
                             read Get_File_Size
                             write Set_File_Size ;
                 end ; // TUOS_File


     TUOS_File_System = class( TCommon_COM_Interface )
                            public // API...
                                function Get_Store : TCOM_Managed_Store64 ;
                                    virtual ; stdcall ; abstract ;
                                procedure Set_Store( Value : TCOM_Managed_Store64 ) ;
                                    virtual ; stdcall ; abstract ;
                                function Get_Heap : TCOM_Heap64 ;
                                    virtual ; stdcall ; abstract ;
                                procedure Set_Heap( Value : TCOM_Heap64 ) ;
                                    virtual ; stdcall ; abstract ;
                                function Get_Flags : int64 ;
                                    virtual ; stdcall ; abstract ;

                                procedure Init( FSLabel, Password : PAnsiChar ;
                                    Patterns, Clustersize, Folder_Clustersize, Flags : longint ;
                                    var AT_Offset : TStore_Address64 ;
                                    Notifier : TScan_Notify ) ;
                                    virtual ; stdcall ; abstract ;
                                function Rebuild : boolean ;
                                    virtual ; stdcall ; abstract ;
                                function Mount( _Label, Password : PAnsiChar ;
                                    Flags : longint ) : TUnified_Exception ;
                                    virtual ; stdcall ; abstract ;
                                function Dismount : TUnified_Exception ;
                                    virtual ; stdcall ; abstract ;

                                function Get_File_Info( Name : PAnsiChar ;
                                    Stream : longint ) : TUOS_File_Info ;
                                    virtual ; stdcall ; abstract ;
                                function Set_File_Info( Name : PAnsiChar ;
                                    Stream : longint ;
                                    Info, Flags : TUOS_File_Info ) : TUnified_Exception ;
                                    virtual ; stdcall ; abstract ;
                                function Lookup( Spec : PAnsiChar ;
                                    var Context : pointer ) : PAnsiChar ;
                                    virtual ; stdcall ; abstract ;
                                procedure Clear_Context( var Context : pointer ) ;
                                    virtual ; stdcall ; abstract ;
                                function Create_File( Filespec : PAnsiChar ;
                                    Info : TUOS_File_Info ) : TUnified_Exception ;
                                    virtual ; stdcall ; abstract ;
                                function Delete_File( Filespec : PAnsiChar ) : TUnified_Exception ;
                                    virtual ; stdcall ; abstract ;
                                function Rename( Filespec, New_Name : PAnsiChar ) : TUnified_Exception ;
                                    virtual ; stdcall ; abstract ;
                                function Move_File( Old_Filespec, New_Filespec : PAnsiChar ) : TUnified_Exception ;
                                    virtual ; stdcall ; abstract ;
                                function Valid_File_System( var P : TStore_Address64 ) : boolean ;
                                    virtual ; stdcall ; abstract ;

                                function Get_File( Name : PAnsiChar ) : TUOS_File ;
                                    virtual ; stdcall ; abstract ;
                                function Translate_Error : longint ;
                                    virtual ; stdcall ; abstract ;
                                function Normalize( Name : PAnsiChar ) : PAnsiChar ;
                                    virtual ; stdcall ; abstract ;

                            public // Properties...
                                property Flags : int64
                                    read Get_Flags ;
                                property Heap : TCOM_Heap64
                                    read Get_Heap
                                    write Set_Heap ;
                                property Store : TCOM_Managed_Store64
                                    read Get_Store
                                    write Set_Store ;
                        end ; // TUOS_File_System

implementation


begin
    if( sizeof( TUOS_File_Header ) <> 256 ) then
    begin
        halt ;
    end ;
end.
