{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
        Program Name : DevMan
        Package Name : Subroutine Library - Layer 1
        Purpose      : Device Manager wrappers
        Institution  : Conroy & Conroy Co.
        Date Written : 7-Jul-2008
        Written By   : Alan Conroy
        Version      : 1.1

        Copyright (C) 2006-2009 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit provides wrappers to the Device Manager classes.
}

unit DevMan ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

uses { Subroutine Library, Layer 0... }
     _DevMan { TCOM_Device_Manager }
{$IFNDEF FPC}
     ,_Types { int64 }
{$ENDIF}
     ;

{$IFDEF DYNAMIC_OBJECTS}
function COM_Device_Manager( DM : PSDevice_Manager ) : TCOM_Device_Manager ;
function COM_Device_Info( DM : PSDevice_Info ) : TCOM_Device_Info ;
{$ENDIF}

implementation

{$IFDEF DYNAMIC_OBJECTS}
uses SysUtils ; // Lowercase

type TCOM_Device_Info_Wrapper = class( TCOM_Device_Info )
                                    public // Constructors and destructors...
                                        constructor Create( D : PSDevice_Info ) ;

                                    private // Instance data...
                                        DI : PSDevice_Info ;
                                        Temp : Ansistring ;
                                        
                                    public { API... }
                                        function Is_Class( Name : PAnsiChar ) : boolean ;
                                            override ;

                                        function Media_ID : longint ;
                                            override ;
                                        { Return type of media in/on device.  -1 if
                                          error or unknown. }

                                        function File_System : PAnsiChar ;
                                            override ;
                                        { Returns the name of the file system used on
                                          the device.  Returns a null string if the
                                          device is invalid or the file system is
                                          unknown. }

                                        function File_System_Flags : longint ;
                                            override ;
                                        { Returns the flags for the file system.  See
                                          DFSF_*. }

                                        function Fixed : boolean ; override ;
                                        { Returns true if the passed device is fixed
                                          (non-removable) }

                                        function Max_Filename_Length : longint ;
                                            override ;
                                        { Returns the maximum filename length for this
                                          device's file system. }

                                        function Version : longint ; override ;
                                        { Returns version number of this object.  11 = 1.1 }
                                end ;

// Constructors and destructors...

constructor TCOM_Device_Info_Wrapper.Create( D : PSDevice_Info ) ;

begin
    inherited Create ;

    DI := D ;
end ;


{ API... }

function TCOM_Device_Info_Wrapper.Is_Class( Name : PAnsiChar ) : boolean ; stdcall ;

var S : Ansistring ;

begin
    S := lowercase( Ansistring( Name ) ) ;
    Result := S = 'tcom_device_info_wrapper' ;
end ;


function TCOM_Device_Info_Wrapper.Media_ID : longint ; stdcall ;

begin
    Result := DI^.Media_ID ;
end ;


function TCOM_Device_Info_Wrapper.File_System : PAnsiChar ; stdcall ;

begin
    Temp := DI^.File_System ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_Device_Info_Wrapper.File_System_Flags : longint ; stdcall ;

begin
    Result := DI^.File_System_Flags ;
end ;


function TCOM_Device_Info_Wrapper.Fixed : boolean ; stdcall ;

begin
    Result := DI^.Fixed ;
end ;


function TCOM_Device_Info_Wrapper.Max_Filename_Length : longint ; stdcall ;

begin
    Result := DI^.Max_Filename_Length ;
end ;


function TCOM_Device_Info_Wrapper.Version : longint ; stdcall ;

begin
    Result := DI^.Version ;
end ;



type TCOM_Device_Manager_Wrapper = class( TCOM_Device_Manager )
                                       public // Constructors and destructors...
                                           constructor Create( D : PSDevice_Manager ) ;

                                       private // Instance data...
                                           DM : PSDevice_Manager ;
                                           Temp : Ansistring ;

                                       public { API... }
                                         function Is_Class( Name : PAnsiChar ) : boolean ;
                                             override ;

                                         function Device_Info( Device : PAnsiChar ) : TCOM_Device_Info ;
                                             override ;
                                         { Returns an information object for the
                                           requested device. }

                                         function Disk_Clusters( Drive : PAnsiChar ) : int64 ;
                                             override ;
                                         { Returns number of clusters on specified disk
                                           drive.  Returns -1 if an error or unknown. }

                                         function Disk_Sector_Size( Drive : PAnsiChar ) : longint ;
                                             override ;
                                         { Returns size of disk sectors, in bytes.
                                           Returns -1 if an error or unknown. }

                                         function File_System( Device : PAnsiChar ) : PAnsiChar ;
                                             override ;
                                         { Returns the name of the file system used on
                                           the device.  Returns a null string if the
                                           device is invalid or the file system is
                                           unknown. }

                                         function File_System_Flags( Device : PAnsiChar ) : longint ;
                                             override ;
                                         { Returns the flags for the file system.  See
                                           DFSF_*. }

                                         function Fixed_Device( Device : PAnsiChar ) : boolean ;
                                             override ;
                                         { Returns true if the passed device is fixed
                                           (non-removable) }

                                         function Max_Filename_Length( Device : PAnsiChar ) : longint ;
                                             override ;
                                         { Returns the maximum filename length for this
                                           device's file system. }

                                         function Media_ID( Device : PAnsiChar ) : longint ;
                                             override ;
                                         { Return type of media in/on device.  -1 if
                                           error or unknown. }

                                         function Valid_Device( Device : PAnsiChar ) : boolean ;
                                             override ;
                                         { Returns true if the passed device is valid. }

                                         function Device_Name( Clas, Index : integer ) : PAnsiChar ;
                                             override ;
                                         { Return name of Indexth device of given class.
                                           if Class is unknown, all device classes are
                                           assumed. }

                                         function Current_Path( Device : PAnsiChar ) : PAnsiChar ;
                                             override ;
                                         { Returns the current path (or default path if
                                           no current path) for the specified device.
                                           Returns null if device is not a hierarchical
                                           file system device. }

                                         function Version : longint ; override ;
                                         { Returns version number of this object.  11 = 1.1 }

                                         function Media_Present( Device : PAnsiChar ) : boolean ;
                                             override ;
                                         { Returns true if media is present in the device }
                                   end ; { TCOM_Device_Manager_Wrapper }

// Constructors and destructors...

function TCOM_Device_Manager_Wrapper.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := lowercase( Ansistring( Name ) ) ;
    Result := S = 'tcom_device_manager_wrapper' ;
end ;


constructor TCOM_Device_Manager_Wrapper.Create( D : PSDevice_Manager ) ;

begin
    inherited Create ;

    DM := D ;
end ;


{ API... }

function TCOM_Device_Manager_Wrapper.Device_Info( Device : PAnsiChar ) : TCOM_Device_Info ; stdcall ;

begin
    Result := COM_Device_Info( DM^.Device_Info( Ansistring( Device ) ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.Disk_Clusters( Drive : PAnsiChar ) : int64 ; stdcall ;

begin
    Result := DM^.Disk_Clusters( Ansistring( Drive ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.Disk_Sector_Size( Drive : PAnsiChar ) : longint ; stdcall ;

begin
    Result := DM^.Disk_Sector_Size( Ansistring( Drive ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.File_System( Device : PAnsiChar ) : PAnsiChar ; stdcall ;

begin
    Temp := DM^.File_System( Ansistring( Device ) ) ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_Device_Manager_Wrapper.File_System_Flags( Device : PAnsiChar ) : longint ; stdcall ;

begin
    Result := DM^.File_System_Flags( Ansistring( Device ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.Fixed_Device( Device : PAnsiChar ) : boolean ; stdcall ;

begin
    Result := DM^.Fixed_Device( Ansistring( Device ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.Max_Filename_Length( Device : PAnsiChar ) : longint ; stdcall ;

begin
    Result := DM^.Max_Filename_Length( Ansistring( Device ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.Media_ID( Device : PAnsiChar ) : longint ; stdcall ;

begin
    Result := DM^.Media_ID( Ansistring( Device ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.Valid_Device( Device : PAnsiChar ) : boolean ; stdcall ;

begin
    Result := DM^.Valid_Device( Ansistring( Device ) ) ;
end ;


function TCOM_Device_Manager_Wrapper.Device_Name( Clas, Index : integer ) : PAnsiChar ; stdcall ;

begin
    Temp := DM^.Device_Name( Clas, Index ) ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_Device_Manager_Wrapper.Current_Path( Device : PAnsiChar ) : PAnsiChar ; stdcall ;

begin
    Temp := DM^.Current_Path( Ansistring( Device ) ) ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_Device_Manager_Wrapper.Version : longint ; stdcall ;

begin
    Result := DM^.Version ;
end ;


function TCOM_Device_Manager_Wrapper.Media_Present( Device : PAnsiChar ) : boolean ; stdcall ;

begin
    Result := DM^.Media_Present( Ansistring( Device ) ) ;
end ;



function COM_Device_Manager( DM : PSDevice_Manager ) : TCOM_Device_Manager ;

begin
    Result := TCOM_Device_Manager_Wrapper.Create( DM ) ;
end ;


function COM_Device_Info( DM : PSDevice_Info ) : TCOM_Device_Info ;

begin
    Result := TCOM_Device_Info_Wrapper.Create( DM ) ;
end ;

{$ENDIF}


end.
