{
        Program Name : EBCDICs
        Package Name : Subroutine Library - Layer 1
        Purpose      : EBCDIC support
        Institution  : Conroy & Conroy Co.
        Date Written : 27-March-2001
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2001-2015 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit supports EBCDIC character formatting.
}

unit EBCDICs ;

interface

var EBCDIC : array[ 0..255 ] of Char ;

function To_EBCDIC( const S : string ) : string ; { Convert an ASCII string to EBCDIC }
function From_EBCDIC( const S : string ) : string ; { Convert an EBCDIC string to ASCII }
function Valid_EBCDIC( const S : string ) : boolean ;

function Decode_BCD( S : string ; _Packed : boolean ) : string ;
function Encode_BCD( S : string ; _Packed : boolean ) : string ;


implementation

var EBCDIC_String : string ;

function From_EBCDIC( const S : string ) : string ;

var Loop : integer ;
    Res : string ;

begin
    Res := '' ;
    for Loop := 1 to length( S ) do
    begin
        Res := Res + EBCDIC[ ord( S[ Loop ] ) ] ;
    end ;
    From_EBCDIC := Res ;
end ;


function To_EBCDIC( const S : string ) : string ; { Convert an ASCII string to EBCDIC }

var Loop : integer ;
    Res : string ;

begin
    Res := '' ;
    for Loop := 1 to length( S ) do
    begin
        Res := Res + chr( pos( S[ Loop ], EBCDIC_String ) ) ;
    end ;
    To_EBCDIC := Res ;
end ;


function Valid_EBCDIC( const S : string ) : boolean ;

var Loop : integer ;

begin
    Valid_EBCDIC := True ;
    for Loop := 1 to length( S ) do
    begin
        if( pos( S[ Loop ], EBCDIC_String ) = 0 ) then
        begin
            Valid_EBCDIC := False ;
            exit ;
        end ;
    end ;
end ;


function Encode_BCD( S : string ; _Packed : boolean ) : string ;

    function Get_Code( C : string ; S : integer ) : integer ;

    begin
        if( C = '' ) then
        begin
            Result := 0 ;
        end else
        if( C[ 1 ] = '-' ) then
        begin
            Result := $D ;
        end else
        if( C[ 1 ] = '+' ) then
        begin
            Result := $C ;
        end else
        begin
            Result := ord( C[ 1 ] ) - 48 ;
        end ;
        Result := Result shl S ;
    end ;

var Loop, X : integer ;

begin
    Result := '' ;
    if( _Packed ) then
    begin
        Loop := 1 ;
        while( Loop <= length( S ) ) do
        begin
            X := Get_Code( S[ Loop ], 0 ) ;
            inc( Loop ) ;
            X := X or ( Get_Code( copy( S, Loop, 1 ), 4 ) ) ;
            Result := Result + chr( X ) ;
            inc( Loop ) ;
        end ;
    end else
    begin
        for Loop := 1 to length( S ) do
        begin
            Result := Result + chr( Get_Code( S[ Loop ], 0 ) ) ;
        end ;
    end ;
end ; // Encode_BCD


function Decode_BCD( S : string ; _Packed : boolean ) : string ;

    procedure Process( C : integer ) ;

    begin
        case C and $F of
            0..9 : Result := Result + chr( C + 48 ) ;
            10, 12, 14 : Result := Result + '+' ;
            else Result := Result + '-' ;
        end ;
    end ;

var C, Loop : integer ;

begin
    Result := '' ;
    if( _Packed ) then
    begin
        for Loop := 1 to length( S ) do
        begin
            C := ord( S[ Loop ] ) ;
            Process( C and $F ) ;
            Process( C shr 4 ) ;
        end ;
    end else
    begin
        for Loop := 1 to length( S ) do
        begin
            Process( ord( S[ Loop ] ) ) ;
        end ;
    end ;
end ;



procedure Init_Unit ;

var A : integer ;

begin
    { Build EBCDIC array }
    fillchar( EBCDIC, sizeof( EBCDIC ), 32 ) ;
    EBCDIC_String := '[.<(+!&         ]$*);^-\         ,%_>?         :#@' + Chr( 39 ) +
        '=" ABCDEFGHI       JKLMNOPQR       STUVWXYZ                       ' +
        'ABCDEFGHI       JKLMNOPQR     \  STUVWXYZ      0123456789' ;
    for A := 1 to 174 do EBCDIC[ A + 73 ] := EBCDIC_String[ A ] ;
    setlength( EBCDIC_String, 256 ) ;
    move( EBCDIC, EBCDIC_String[ 1 ], 256 ) ;
end ;


begin
    Init_Unit ;
end.
