{
              Program Name : Format
              Package Name : Subroutine Library - Layer 1
              Purpose      : Various Formatting routines
              Institution  : Conroy & Conroy Co.
              Date Written : 05-Jul-1993
              Written By   : EAC
              Version      : 1.0

        Copyright (C) 1993 by Alan Conroy.  Released to the Public Domain.

	TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

	  This unit consists of functions to perform various text
	formatting functions.
}

unit Format ;

interface

function Format_String( Format : Ansistring ; var Start : integer ; Sub : Ansistring ;
    Options : integer ) : Ansistring ; {UNTESTED}
function Float_To_String( S : extended ; Decimal : Ansistring ) : Ansistring ; {UNTESTED}
function Try_String_To_Float( S : Ansistring ; var E : extended ) : boolean ;
function String_To_Float( S : Ansistring ) : extended ;
function Tab2Space( Count : Integer ; const X : String ) : String ;

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}


implementation

uses { Subroutine Library - Layer 0... }
     _ASCII, { HT }
     O_S ; { OS }

function Float_To_String( S : extended ; Decimal : Ansistring ) : Ansistring ; {UNTESTED}

var D : Ansistring ;
    Dummy : integer ;
    Res : Ansistring ;

begin
    Res := Num1( S ) ;
    if( Decimal <> '' ) then
    begin
        D := OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal ;
        Dummy := pos( D, Res ) ;
        if( Dummy > 0 ) then
        begin
            Res := copy( Res, 1, Dummy - 1 ) + Decimal +
                copy( Res, Dummy + 1, length( Res ) ) ;
        end ;
    end ;
    Float_To_String := Res ;
end ;


function String_To_Float( S : Ansistring ) : extended ;

var Dummy : integer ;

begin
    Dummy := pos( '.', S ) ;
    if( Dummy > 0 ) then
    begin
        S := copy( S, 1, Dummy - 1 ) +
            OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal +
            copy( S, Dummy + 1, length( S ) ) ;
    end ;
    Result := XVal( S, Dummy ) ;
end ;


function Try_String_To_Float( S : Ansistring ; var E : extended ) : boolean ;

var Dummy : integer ;

begin
    Dummy := pos( '.', S ) ;
    if( Dummy > 0 ) then
    begin
        S := copy( S, 1, Dummy - 1 ) +
            OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal +
            copy( S, Dummy + 1, length( S ) ) ;
    end ;
    E := XVal( S, Dummy ) ;
    Result := ( Dummy = 0 ) ;
end ;


function Format_String( Format : Ansistring ; var Start : integer ; Sub : Ansistring ;
    Options : integer ) : Ansistring ; {UNTESTED}
{ Format a string via another string.  Format = formatting control string,
  Start = which position in Format to start at (1 = first), Sub = string to
  format, Options = formatting control options.  Note that the function formats
  one field at a time, advancing Start to the next position to process.  Thus,
  multiple items can be formatted with one formatting control string by making
  multiple calls to the function.  Options: and 1 = Allow underscores as special }

    function Get_Numeric_Format( Decimal : boolean ) : Ansistring ;

    var C : Ansichar ;

    begin
        while( Start <= length( Format ) ) do
        begin
            C := Format[ Start ] ;
            case C of
                '-', '+' : // Trailing sign
                           begin
                               inc( Start ) ;
                               exit ; // End of format field
                           end ;
                '.' : begin
                          if( Decimal ) then // Second decimal indicates end of field
                          begin
                              exit ;
                          end ;
                          Decimal := True ;
                      end ;
                '#', ',' : ;
                else exit ; // Field ends on any other character
            end ;
        end ;
    end ;


    function Format_Number( S : Ansistring ) : Ansistring ;

    var D, D1 : integer ;
        E, E1 : extended ;
        Commas : boolean ;
        Res, Whole : Ansistring ;

    begin
        { Setup... }
        Res := '' ;
        E := XVal( Sub, D ) ;

        { Validation... }
        if( D <> 0 ) then { Not a valid number }
        begin
            while( length( Res ) < length( S ) ) do
            begin
               Res := Res + '%' ; { Indicate a numeric error }
            end ;
            exit ;
        end ;

        { Handle sign... }
        case S[ 1 ] of
            '-' : begin
                      if( E < 0 ) then
                      begin
                          Res := '-' ;
                          Sub := copy( Sub, 2, length( Sub ) ) ;
                      end else
                      if( E > 0 ) then
                      begin
                          Res := '+' ;
                      end else
                      begin
                          Res := ' ' ;
                      end ;
                      S := copy( S, 2, length( S ) ) ;
                  end ;
            '+' : begin
                      if( E < 0 ) then
                      begin
                          Res := '-' ;
                          Sub := copy( Sub, 2, length( Sub ) ) ;
                      end else
                      begin
                          Res := ' ' ;
                      end ;
                      S := copy( S, 2, length( S ) ) ;
                  end ;
        end ;

        { Process whole number side... }
        Commas := pos( ',', S ) > 0 ;
        if( ( pos( '.', S ) = 0 ) and ( pos( '.', Sub ) > 0 ) ) then
        begin
            { Have a decimal in the passed string, but no decimal in the format field, so round up }
            E := int( E + 0.5 ) ;
            Sub := Num1( E ) ;
        end ;
        D1 := pos( '.', Sub + '.' ) ;
        D := pos( '.', S + '.' ) ;
        Whole := copy( Sub, 1, D1 - 1 ) ;
        if( Commas ) then
        begin
            while( D1 > 3 ) do
            begin
                D1 := D1 - 3 ;
                Insert( ',', Whole, D1 ) ;
            end ;
        end ;
        if( length( Whole ) >= D ) then // Too large for field
        begin
            Whole := '%' ;
            while( length( Whole ) < D - 1 ) do
            begin
                Whole := Whole + '%' ;
            end ;
            exit ;
        end ;
        while( length( Whole ) < D ) do
        begin
            Whole := ' ' + Whole ;
        end ;
        Res := Res + Whole ;
        Sub := copy( Sub, D1, length( Sub ) ) ; { Strip off whole portion }

        { Handle decimal point... }
        if( copy( S, D, 1 ) = '.' ) then
        begin
            Res := Res + '.' ;
        end ;
        S := copy( S, D + 1, length( S ) ) ; { Trim decimal and whole portion from format control }
        if( pos( '.', Sub ) = 1 ) then
        begin
            Sub := copy( Sub, 2, length( Sub ) ) ; { Trim decimal from string }
        end ;

        { Handle fractional portion... }
        D := 1 ;
        Whole := '0.' ;
        while( ( D <= length( S ) ) and ( S[ D ] = '#' ) ) do
        begin
            inc( D ) ;
            Whole := Whole + '0' ;
        end ;
        Whole := Whole + '5' ; { Value to use for rounding }
        if( length( Sub ) > length( Whole ) - 2 ) then { Rounding needed }
        begin
            E1 := XVal( Whole, D1 ) ;
            E := E + E1 ; { Round }
            Sub := Num1( E ) ;
            D1 := pos( '.', Sub ) ;
            Sub := copy( Sub, D1 + 1, length( Sub ) ) ;
        end ;
        Sub := copy( Sub, 1, D - 1 ) ; { Trim excess digits }
        while( length( Sub ) < D - 1 ) do
        begin
            Sub := Sub + '0' ; { Zero fill if too short }
        end ;
        Res := Res + Sub ;

        { Handle trailing stuff... }
        S := copy( S, D, length( S ) ) ; { Trim fractional part }
        if( S = '-' ) then
        begin
            if( E < 0 ) then
            begin
                Res := Res + '-' ;
            end else
            if( E > 0 ) then
            begin
                Res := Res + '+' ;
            end else
            begin
                Res := Res + ' ' ;
            end ;
        end else
        if( S = '+' ) then
        begin
            if( E < 0 ) then
            begin
                Res := Res + '-' ;
            end else
            begin
                Res := Res + ' ' ;
            end ;
        end ;
        Format_Number := Res ;
    end ; { Format_String.Format_Number }


var C : Ansichar ;
    Res : Ansistring ;
    Origin : integer ;
    S : Ansistring ;

begin { Format_String }
    { Setup... }
    if( Start < 1 ) then
    begin
        Start := 1 ;
    end ;
    Res := '' ;

    { Process format string... }
    while( Start <= length( Format ) ) do
    begin
        C := Format[ Start ] ;
        inc( Start ) ;
        case C of
            '_' : begin
                      if( ( Options and 1 ) = 0 ) then
                      begin
                          Res := Res + C ; { Treat all underscores as literal }
                      end else
                      begin
                          Res := Res + copy( Sub, Start, 1 ) ;
                          inc( Start ) ;
                      end ;
                  end ;
            '!' : begin
                      Origin := Start - 1 ;
                      while( copy( Format, Start, 1 ) = '!' ) do
                      begin
                          inc( Start ) ;
                      end ;
                      S := copy( Sub, 1, Start - Origin ) ;
                      while( length( S ) < Start - Origin ) do
                      begin
                          S := S + ' ' ; { Space-fill string to fill rest of range }
                      end ;
                      Res := Res + S ;
                      exit ;
                  end ;
            '\' : begin
                      Origin := Start - 1 ;
                      while(
                             ( copy( Format, Start, 1 ) <> ' ' )
                             and
                             ( copy( Format, Start, 1 ) <> '\' )
                           ) do
                      begin
                          inc( Start ) ;
                      end ;
                      S := copy( Sub, 1, Start - Origin ) ;
                      while( length( S ) < Start - Origin ) do
                      begin
                          S := S + ' ' ; { Space-fill string to fill rest of range }
                      end ;
                      Res := Res + S ;
                      exit ;
                  end ;
            '#' : begin
                      S := Get_Numeric_Format( False ) ;
                      Res := Res + Format_Number( S ) ;
                      exit ;
                  end ;
            '.' : begin
                      S := Get_Numeric_Format( True ) ;
                      Res := Res + Format_Number( S ) ;
                      exit ;
                  end ;
            '-' : begin
                      S := Get_Numeric_Format( False ) ;
                      Res := Res + Format_Number( S ) ;
                      exit ;
                  end ;
            '+' : begin
                      S := Get_Numeric_Format( False ) ;
                      Res := Res + Format_Number( S ) ;
                      exit ;
                  end ;
            else Res := Res + C ;
        end ;
        Format_String := Res ;
    end ;
end ; { Format_String }


function Tab2Space( Count : Integer ; const X : String ) : String ;
{ Convert string with tabs to string with appropriate spaces. }

var C, Dummy : Integer ;
    Y : String ; { Place to hold result }

begin
    Y := '' ;
    C := 1 ;
    for Dummy := 1 to Length( X ) do
    begin
	if( X[ Dummy ] = HT ) then { tab }
        begin
            while( C <= Count ) do
            begin
                Y := Y + ' ' ;
                Inc( C ) ;
            end ;
            C := 1 ;
        end else
        begin
            Y := Y + X[ Dummy ] ;
            Inc( C ) ; { Count position within tab region }
            if( C > Count ) then
            begin
                C := 1 ;
            end ;
        end ;
    end ;
    Tab2Space := Y ;
end ;


{$IFDEF CC_Regression_Test}
procedure Test_Unit ;

    procedure Error ;

    begin
        writeln( 'Invalid result' ) ;
        halt ;
    end ;


    procedure Test_Tab2Space ;

    var S : string ;

    begin
        S := Tab2Space( 8, '1234567'#9'9' ) ;
        if( S <> '1234567 9' ) then
        begin
            Error ;
        end ;
        S := Tab2Space( 8, '12345678'#9'9' ) ;
        if( S <> '12345678        9' ) then
        begin
            Error ;
        end ;
        S := Tab2Space( 4, '1234567'#9'9' ) ;
        if( S <> '1234567 9' ) then
        begin
            Error ;
        end ;
        S := Tab2Space( 3, #9'1234567'#9'9' ) ;
        if( S <> '   1234567  9' ) then
        begin
            Error ;
        end ;
        S := Tab2Space( 1, '1234567'#9'9' ) ;
        if( S <> '1234567 9' ) then
        begin
            Error ;
        end ;
    end ;

begin
    Test_Tab2Space ;
end ;
{$ENDIF}

end.
