unit SCU32Srv ;

interface

uses // Borland...
     Classes, // TList
     Forms, // Application
     SysUtils, // inttostr
     Windows, // CreateNamedPipe

     // Other...
     DebugInt, // PDebug_Interface
     _Loggers, // PLogger,
     MailSlot, // TMessenger
     RLoggers, // TRemote_Logger
     _SCU, // T_SCU
     SCU32Int, // SCU_Operation_
     SCULib, // TSCU
     Standard, // Standard
     _UEHDefs ; // TUEC

type TSCU_Server = class
                       private // Instance data...
                           Mail_Slot : TMail_Slot ;
                           List, Names : TStringList ;

                       public // Constructors and destructors...
                           constructor Create( S : string ) ;
                           destructor Destroy ; override ;

                       private // Internal utility methods...
                           procedure CB_Command( Sender : TObject ;
                               const Origin, Name : string ;
                               var Data : string ) ;
                           procedure CB_Message( Sender : TObject ;
                               const Origin, Name, Data : string ) ;
                           procedure Disconnect( const Name : string ) ;
                           function IndexOf( const Name : string ) : integer ;
                           function Initialize_Access( const Name : string ) : integer ;
                           procedure Shutdown ;
                   end ;

var Free_Server : boolean = false ;


implementation

// Constructors and destructors...

constructor TSCU_Server.Create( S : string ) ;

begin
    inherited Create ;
    if( S = '' ) then
    begin
        S := 'SCU' ;
    end ;

    List := TStringList.Create ;
    Names := TStringList.Create ;

    Mail_Slot := TMail_Slot.Create( S, True ) ;
    Mail_Slot.Activate ;
    Mail_Slot.On_Message := CB_Message ;
    Mail_Slot.On_Command := CB_Command ;
end ;


destructor TSCU_Server.Destroy ;

begin
    Mail_Slot.Free ;
    Mail_Slot := nil ;

    List.Free ;
    List := nil ;

    Names.Free ;
    Names := nil ;

    inherited Destroy ;
end ;


function Decode_Integer( const S : string ; Index : longint ) : longint ;

begin
    move( PChar( S )[ Index - 1 ], Result, sizeof( Result ) ) ;
end ;


function Encode_Integer( I : longint ) : string ;

begin
    setlength( Result, sizeof( longint ) ) ;
    move( I, PChar( Result )^, length( Result ) ) ;
end ;


function Decode_String( const S : string ; Index : integer ) : string ;

var Dummy : integer ;

begin
    if( ( Index < 1 ) or ( Index + sizeof( longint ) > length( S ) ) ) then
    begin
        Decode_String := '' ; {UNTESTED}
        exit ;
    end ;
    Dummy := 0 ;
    move( PChar( S )[ Index - 1 ], Dummy, sizeof( longint ) ) ;
    setlength( Result, Dummy ) ;
    move( PChar( S )[ Index - 1 + sizeof( longint ) ], PChar( Result )^, Dummy ) ;
end ;


function Encode_String( const S : string ) : string ;

var Dummy : integer ;

begin
    { This function returns a string whose first 4 bytes are an integer length
      and where that many bytes follow, starting at offset 4, containing the
      contents of S. }
    Dummy := length( S ) ;
    setlength( Result, Dummy + sizeof( integer ) ) ;
    move( Dummy, PChar( Result )^, sizeof( integer ) ) ;
    move( PChar( S )^, PChar( Result )[ sizeof( integer ) ], Dummy ) ;
end ;


function Encode_UEC( Facility, Code : longint ) : string ;

var UEC : TUEC ;

begin
    UEC.Facility := Facility ;
    UEC.Code := Code ;
    setlength( Result, sizeof( UEC ) ) ;
    move( UEC, PChar( Result )^, length( Result ) ) ;
end ;


// Internal utility methods...

procedure TSCU_Server.CB_Command( Sender : TObject ;
    const Origin, Name : string ; var Data : string ) ;

begin
    if( Data = 'OFFLINE' ) then
    begin
        Disconnect( Origin + '\' + Name ) ;
    end ;
end ;


procedure TSCU_Server.CB_Message( Sender : TObject ;
    const Origin, Name, Data : string ) ;

var S : string ;
    Header : TSCU_Message_Header ;
    Index : integer ;
    DT, Size : integer ;
    _Logger : TRemote_Logger ;
    PC : PChar ;
    SCU : TSCU ;
    Temp, Work : string ;
    UEC : TUEC ;

begin
    Size := sizeof( Header ) ;
    if( Size > length( Data ) ) then
    begin
        Size := length( Data ) ;
    end ;
    fillchar( Header, sizeof( Header ), 0 ) ;
    move( Data[ 1 ], Header, Size ) ;
    if( Header.Format <> 0 ) then // Invalid format
    begin
        S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Format ) ;
        Mail_Slot.Send( Origin + '\' + Name, S ) ;
        exit ;
    end ;
    case Header.Operation of
        SCU_Operation_Null: S := Encode_UEC( 0, 0 ) ; {UNTESTED}
        SCU_Operation_Init:
            begin
                Disconnect( Origin + '\' + Name ) ;
                Index := Initialize_Access( Decode_String( Data, sizeof( Header ) + 1 ) ) ;
                SCU := TSCU( List.Objects[ Index ] ) ;
                SCU.Attach ;
                S := Encode_UEC( 0, 0 ) + Encode_Integer( Index ) ;
                Names.AddObject( Origin + '\' + Name, pointer( Index ) ) ;
            end ;
        SCU_Operation_Version:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( 0 ) ;
                end else
                begin
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( TSCU( List.Objects[ Header.Context ] ).Version ) ;
                end ;
            end ;
        SCU_Operation_Write:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    UEC := SCU.Write( PChar( S ), PChar( Data ) + sizeof( Header ) + length( S ) + sizeof( longint ) * 2, length( Data ) - length( S ) - sizeof( longint ) * 2 - sizeof( Header ) ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Read:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Size := Decode_Integer( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    setlength( Temp, Size ) ;
                    UEC := SCU.Read( PChar( S ), PChar( Temp ), Size ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) + Temp ;
                end ;
            end ;
        SCU_Operation_Read_Size:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( -1 ) ;
                end else
                begin
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( TSCU( List.Objects[ Header.Context ] ).Read_Size( PChar( Decode_String( Data, sizeof( Header ) + 1 ) ) ) ) ;
                end ;
            end ;
        SCU_Operation_Read_Type:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( -1 ) ;
                end else
                begin
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( TSCU( List.Objects[ Header.Context ]).Read_Type( PChar( Decode_String( Data, sizeof( Header ) + 1 ) ) ) ) ;
                end ;
            end ;
        SCU_Operation_Exists:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( 0 ) ;
                end else
                begin
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( ord( TSCU( List.Objects[ Header.Context ] ).Exists( PChar( Decode_String( Data, sizeof( Header ) + 1 ) ) ) ) ) ;
                end ;
            end ;
        SCU_Operation_Key_Exists:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( 0 ) ;
                end else
                begin
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( ord( TSCU( List.Objects[ Header.Context ] ).Key_Exists( PChar( Decode_String( Data, sizeof( Header ) + 1 ) ) ) ) ) ;
                end ;
            end ;
        SCU_Operation_Lookup:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Size := Decode_Integer( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    S := Encode_UEC( 0, 0 ) + string( SCU.Lookup( PChar( S ), Size ) ) ;
                end ;
            end ;
        SCU_Operation_Delete:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    UEC := TSCU( List.Objects[ Header.Context ] ).Delete(  PChar( Decode_String( Data, sizeof( Header ) + 1 ) ) ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Create:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Size := Decode_Integer( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    UEC := SCU.Create_( PChar( S ), Size ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Create_And_Write:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    Size := Decode_Integer( Data, sizeof( Header ) + length( S ) + length( Temp ) + sizeof( longint ) * 2 + 1 ) ;
                    UEC := SCU.Create_And_Write( PChar( S ), Size, PChar( Temp ), length( Temp ) ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Rename:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    UEC := SCU.Rename( PChar( S ), PChar( Temp ) ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Set_Value:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    Work := Decode_String( Data, sizeof( Header ) + length( S ) + length( Temp ) + sizeof( longint ) * 2 + 1 ) ;
                    UEC := SCU.Set_Value( PChar( S ), PChar( Temp ), PChar( Work ), length( Work ) ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Get_Value:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    Size := Decode_Integer( Data, sizeof( Header ) + length( S ) + length( Temp ) + sizeof( longint ) * 2 + 1 ) ;
                    setlength( Work, Size ) ;
                    UEC := SCU.Get_Value( PChar( S ), PChar( Temp ), PChar( Work ), Size ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) + Work ;
                end ;
            end ;
        SCU_Operation_Value_Size:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( -1 ) ;
                end else
                begin
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( TSCU( List.Objects[ Header.Context ] ).Value_Size( PChar( S ), PChar( Temp ) ) ) ;
                end ;
            end ;
        SCU_Operation_Value_Type:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( -1 ) ;
                end else
                begin
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( TSCU( List.Objects[ Header.Context ] ).Value_Type( PChar( S ), PChar( Temp ) ) ) ;
                end ;
            end ;
        SCU_Operation_Value_Exists:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) + Encode_Integer( 0 ) ;
                end else
                begin
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( ord( TSCU( List.Objects[ Header.Context ] ).Value_Exists( PChar( S ), PChar( Temp ) ) ) ) ;
                end ;
            end ;
        SCU_Operation_Delete_Value:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    UEC := TSCU( List.Objects[ Header.Context ] ).Delete_Value( PChar( S ), PChar( Temp ) ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Create_Value:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    Size := Decode_Integer( Data, sizeof( Header ) + length( S ) + length( Temp ) + sizeof( longint ) * 2 + 1 ) ;
                    UEC := SCU.Create_Value( PChar( S ), PChar( Temp ), Size ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Verify:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    UEC := TSCU( List.Objects[ Header.Context ] ).Verify ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                end ;
            end ;
        SCU_Operation_Debugger: {UNTESTED}
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    S := Encode_UEC( 0, 0 ) + Encode_Integer( integer( TSCU( List.Objects[ Header.Context ] ).Debugger ) ) ;
                end ;
            end ;
        SCU_Operation_Read_Value:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    SCU := TSCU( List.Objects[ Header.Context ] ) ;
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    Temp := Decode_String( Data, sizeof( Header ) + length( S ) + sizeof( longint ) + 1 ) ;
                    UEC := SCU.Read_Value( PChar( S ), PChar( Temp ), DT, Size, PC ) ;
                    setlength( S, Size ) ;
                    move( PC^, PChar( S )^, Size ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) + Encode_Integer( DT ) + Encode_Integer( Size ) + S ;
                end ;
            end ;
        SCU_Operation_Init_And_Extend:
            begin
                Disconnect( Origin + '\' + Name ) ;
                S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                Index := IndexOf( S ) ;
                if( Index = -1 ) then
                begin
                    Index := List.Count ;
                    SCU := TSCU.Construct ;
                    SCU.Initialize ;
                    SCU.Singleton := True ;
                    UEC := SCU.Init_And_Extend( PChar( S ), Size ) ;
                    List.AddObject( S, SCU ) ;
                end else
                begin
                    fillchar( UEC, sizeof( UEC ), 0 ) ;
                end ;
                SCU := TSCU( List.Objects[ Index ] ) ;
                SCU.Attach ;
                S := Encode_UEC( UEC.Facility, UEC.Code ) + Encode_Integer( Index ) ;
                Names.AddObject( Origin + '\' + Name, pointer( Index ) ) ;
            end ;
        SCU_Operation_Delete_With_Callback:
            begin
                if( ( Header.Context < 0 ) or ( Header.Context >= List.Count ) ) then
                begin
                    S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Handle ) ;
                end else
                begin
                    S := Decode_String( Data, sizeof( Header ) + 1 ) ;
                    _Logger := TRemote_Logger.Create ;
                    _Logger.Initialize_Logger( Decode_String( Data, length( S ) + sizeof( Header ) + sizeof( longint ) + 1 ), False, nil ) ;
                    UEC := TSCU( List.Objects[ Header.Context ] ).Delete_With_Callback( PChar( S ), _Logger ) ;
                    S := Encode_UEC( UEC.Facility, UEC.Code ) ;
                    _Logger.Free ;
                end ;
            end ;
        SCU_Operation_Shutdown:
            begin
                Shutdown ;
            end ;
        else S := Encode_UEC( SCU_ServerFacility, SCU_ServerErr_Invalid_Operation ) ; {UNTESTED}
    end ;
    Mail_Slot.Send( Origin + '\' + Name, S ) ;
end ; // TSCU_Server.CB_Message


procedure TSCU_Server.Shutdown ;

var Index, Size : integer ;

begin
    Size := 0 ;
    for Index := 0 to List.Count - 1 do
    begin
        if( List.Objects[ Index ] <> nil ) then
        begin
            inc( Size ) ;
            break ;
        end ;
    end ;
    if( Size = 0 ) then { No open SCU files }
    begin
        Free_Server := True ;
        exit ;
    end ;
end ;


procedure TSCU_Server.Disconnect( const Name : string ) ;

var C : integer ;
    Index : integer ;

begin
    Index := Names.Indexof( Name ) ;
    if( Index <> -1 ) then
    begin
        Index := integer( Names.Objects[ Index ] ) ;
        C := TSCU( List.Objects[ Index ] ).Reference_Count ;
        TSCU( List.Objects[ Index ] ).Detach ;
        if( C < 2 ) then // Object is no longer referenced
        begin
            List.Objects[ Index ] := nil ;
            List[ Index ] := '' ;
        end ;
    end ;
end ;


function TSCU_Server.IndexOf( const Name : string ) : integer ;

var Index : integer ;
    S : string ;

begin
    Result := -1 ;
    for Index := 0 to List.Count - 1 do
    begin
        S := List[ Index ] ;
        if(
            ( copy( Name, 1, length( S ) ) = S )
            and
            (
              ( length( Name ) = length( S ) )
              or
              ( copy( Name, length( S ) + 1, 1 ) = '\' )
            )
          ) then
        begin
            Result := Index ;
            exit ;
        end ;
    end ;
end ;


function TSCU_Server.Initialize_Access( const Name : string ) : integer ;

var SCU : TSCU ;

begin
    Result := IndexOf( Name ) ;
    if( Result = -1 ) then
    begin
        Result := List.Count ;
        SCU := TSCU.Construct ;
        SCU.Initialize ;
        SCU.Singleton := True ;
        SCU.Init( PChar( Name ) ) ;
        List.AddObject( Name, SCU ) ;
    end ;
end ;


end.
