{
        Program Name : TPanel
        Package Name : Subroutine Library - Layer 5
        Purpose      : Transparent panel component
        Institution  : Conroy & Conroy Co.
        Date Written : 10-July-2005
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TTransparent_Panel is a Delphi TPanel which doesn't erase it's
        background on paint, provides a paint (and other) delegations, and makes
        its canvas available for painting.
}

unit TPanels ;

interface

uses // Borland...
     Classes, // TNotifyEvent
     Messages, // TMessage
     ExtCtrls ; // TPanel

type TTransparent_Panel = class( TPanel )
                              public // Constructors and destructors...
                                  constructor Create( AOwner : TComponent ) ;
                                      override ;

                              private // Instance data...
                                  _OnPaint : TNotifyEvent ;
                                  _Transparent : boolean ;
                                  _OnSetFocus : TNotifyEvent ;
                                  _OnKillFocus : TNotifyEvent ;

                              public // Message handlers...
                                  procedure WMSetFocus( var Msg : TMessage ) ;
                                      message WM_SetFocus ;
                                  procedure WMKillFocus( var Msg : TMessage ) ;
                                      message WM_KillFocus ;
                                  procedure wmerasebkgnd( var Msg : TMessage ) ;
                                      message WM_Erasebkgnd ;

                              public // API...
                                  procedure Paint ; override ;

                              published // Properties...
                                  property OnPaint : TNotifyEvent
                                      read _OnPaint
                                      write _OnPaint ;

                                  property OnSetFocus : TNotifyEvent
                                      read _OnSetFocus
                                      write _OnSetFocus ;
                                  property OnKillFocus : TNotifyEvent
                                      read _OnKillFocus
                                      write _OnKillFocus ;
                                  property Transparent : boolean
                                      read _Transparent
                                      write _Transparent ;

                                   property Canvas ;
                                   property Color ;
                                   property Font ;
                                   property OnClick ;
                                   property OnKeyDown ;
                          end ;

implementation

uses // VCL...
     Graphics ; // psDash

// TTransparent_Panel methods...

// Constructors and destructors...

constructor TTransparent_Panel.Create( AOwner : TComponent ) ;

begin
    inherited Create( AOwner ) ;

    Transparent := True ;
end ;


// Message handlers...

procedure TTransparent_Panel.wmerasebkgnd( var Msg : TMessage ) ;

begin
    if( Transparent or ( csDesigning in ComponentState ) ) then
    begin
        Msg.Result := 1 ;
    end ;
end ;


// API...

procedure TTransparent_Panel.Paint ;

begin
    Canvas.Font := Font ;
    Canvas.Brush.Color := Color ;
    if csDesigning in ComponentState then
    begin
        // Show as a dashed rectangle in design mode...
        with Canvas do
        begin
            Pen.Style := psDash ;
            Brush.Style := bsClear ;
            Rectangle( 0, 0, Width, Height ) ;
        end ;
    end ;
    if Assigned( _OnPaint ) then
    begin
        _OnPaint( Self ) ;
    end else
    if( not Transparent ) then
    begin
        inherited Paint ;
    end ;
end ;


procedure TTransparent_Panel.WMSetFocus( var Msg : TMessage ) ;

begin
    inherited ;

    if( assigned( OnSetFocus ) ) then
    begin
        OnSetFocus( self ) ;
    end ;
end ;


procedure TTransparent_Panel.WMKillFocus( var Msg : TMessage ) ;

begin
    if( assigned( OnKillFocus ) ) then
    begin
        OnKillFocus( self ) ;
    end ;

    inherited ;
end ;



end.

