{
        Program Name : UE
        Package Name : Subroutine Library - Layer 1
        Purpose      : Unified Exceptions
        Institution  : Conroy & Conroy Co.
        Date Written : 10-Jan-1999
        Last Modified: 19-May-2002
        Written By   : Alan Conroy
        Version      : 1.2

        Copyright (C) 2014-2015 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides Unified Exception handling.
}

unit UE ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

uses { C&C Subroutine Library - Layer 0 }
{$IFDEF CC_Track_Resources}
     ResMon,
{$ENDIF}
     _Common, { TMemory_Type }
{$IFNDEF MOBILE}
     _DebugIn, { TDebug_Interface }
{$ENDIF}
     _Types, { Ansistring }
     _UE ; { tUnified_Exception }

{$IFDEF FPC_OR_DELPHI}
type TStandard_Error_Interface = class( tUnified_Exception )
  {$IFDEF CC_Track_Resources}
                                     public { Constructors and destructors... }
                                         constructor Create ;
                                         destructor Destroy ; override ;
  {$ENDIF}

                                     protected // Instance data...
                                         _Reference_Count : integer ;
                                         _Additional_Text : Ansistring ;
                                         Temp : Ansistring ;

                                     public
                                         function Initialize : TUnified_Exception ;
                                             override ;
                                         procedure Terminate ; override ;
                                         procedure Attach ; override ;
                                         procedure Detach ; override ;
                                         function Is_Class( Name : PAnsiChar ) : boolean ;
                                             override ;
                                         function Interface_Version : longint ;
                                             override ;
                                         function Get_Facility : longint ;
                                             override ;
                                         function Get_Facility_Version : longint ;
                                             override ;
                                         function Version : longint ; override ;
{$IFNDEF MOBILE}
                                         function Debugger : TDebug_Interface ;
                                             override ;
{$ENDIF}
                                         function Severity : longint ;
                                             override ;
                                         function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                             override ;
                                         function Get_Error : longint ;
                                             override ;
                                         function Get_Previous : tUnified_Exception ;
                                             override ;
                                 end ; { TStandard_Error_Interface }

function Create_Simple_UE( F, FV, C, S : integer ; ET, Extra : Ansistring ) : TUnified_Exception ;
function Create_MSDOS_UE( C : integer ; Extra : Ansistring ) : TUnified_Exception ;

type TSimple_Error_Interface = class( TStandard_Error_Interface )
                                     private
                                         _Facility, _Facility_Version : integer ;
                                         _Code, _Severity : integer ;
                                         _Error_Text : Ansistring ;

                                     public
                                         function Get_Facility : longint ;
                                             override ;
                                         function Get_Facility_Version : longint ;
                                             override ;
                                         function Severity : longint ;
                                             override ;
                                         function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                             override ;
                                         function Get_Error : longint ;
                                             override ;
                                 end ;

type TMSDOS_Error_Interface = class( TStandard_Error_Interface )
                                     private
                                         _Code : integer ;

                                     public
                                         function Get_Facility : longint ;
                                             override ;
                                         function Get_Facility_Version : longint ;
                                             override ;
                                         function Severity : longint ;
                                             override ;
                                         function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                             override ;
                                         function Get_Error : longint ;
                                             override ;
                                 end ;
{$ENDIF}

type TSStandard_Error_Interface = object( tSUnified_Exception )
                                      private
                                          _Memory_Type : TMemory_Type ;

                                      public
                                          function Get_Facility : longint ;
                                              virtual ;
                                          function Get_Facility_Version : longint ;
                                              virtual ;
                                          function Severity : longint ;
                                              virtual ;
                                          function Get_Error : longint ;
                                              virtual ;
                                          function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                              virtual ;
                                          function Get_Previous : psUnified_Exception ;
                                              virtual ;
                                          procedure Terminate ;
                                              virtual ;
                                          procedure Set_Memory_Type( MT : TMemory_Type ) ;
                                              virtual ;
                                          function Get_Memory_Type : TMemory_Type ;
                                              virtual ;
                                  end ;
     PSStandard_Error_Interface = ^TSStandard_Error_Interface ;

type TSSimple_Error_Interface = object( TSStandard_Error_Interface )
                                     private
                                         _Facility, _Facility_Version : integer ;
                                         _Code, _Severity : integer ;
                                         _Error_Text : Ansistring ;

                                     public
                                         function Get_Facility : longint ;
                                             virtual ;
                                         function Get_Facility_Version : longint ;
                                             virtual ;
                                         function Severity : longint ;
                                             virtual ;
                                         function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                             virtual ;
                                         function Get_Error : longint ;
                                             virtual ;
                                 end ;
     PSSimple_Error_Interface = ^TSSimple_Error_Interface ;

function Create_PSSimple_UE( F, FV, C, S : integer ; ET, Extra : Ansistring ) : PSUnified_Exception ;
function Create_PSMSDOS_UE( C : integer ; Extra : Ansistring ) : PSUnified_Exception ;
function Create_PSPascal_UE( C : integer ; Extra : Ansistring ) : PSUnified_Exception ;

function Get_Abstract_Error_Exception : PSUnified_Exception ;
{$IFDEF DYNAMIC_OBJECTS}
function Create_Abstract_Error_Exception : TUnified_Exception ;
function Create_UE_Wrapper( U : PSUnified_Exception ) : TUnified_Exception ;
{$ENDIF}

function DOS_ERT( Code : integer ) : Ansistring ;
function Ert( Err : Integer ) : Ansistring ; { Return Pascal error text }

type TSMSDOS_Error_Interface = object( TSStandard_Error_Interface )
                                     public { Instance data... }
                                         _Code : integer ;
                                         _Additional_Text : Ansistring ;

                                     public
                                         function Get_Facility : longint ;
                                             virtual ;
                                         function Get_Facility_Version : longint ;
                                             virtual ;
                                         function Severity : longint ;
                                             virtual ;
                                         function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                             virtual ;
                                         function Get_Error : longint ;
                                             virtual ;
                                 end ;
     PSMSDOS_Error_Interface = ^TSMSDOS_Error_Interface ;


implementation

{$IFDEF FPC_OR_DELPHI}
uses { Borland... }
     Sysutils, { strpas }
{$ELSE}
uses Strings { strlen }
{$ENDIF}
     { Subroutine Library - Layer 0... }
     L10n ; { Text_* }

function Pascal_ERT( Code : integer ) : Ansistring ;

var Err_String : Ansistring ;

begin
    case Code of
        { Runtime errors... }
        0 : Err_String := Text_Success ;
        100 : Err_String := Text_Pascal_Err_String_100 ;
        101 : Err_String := Text_Pascal_Err_String_101 ;
        102 : Err_String := Text_Pascal_Err_String_102 ;
        103 : Err_String := Text_Pascal_Err_String_103 ;
        104 : Err_String := Text_Pascal_Err_String_104 ;
        105 : Err_String := Text_Pascal_Err_String_105 ;
        106 : Err_String := Text_Pascal_Err_String_106 ;
        150 : Err_String := 'Disk is write-protected' ;
        151 : Err_String := 'Unknown unit' ;
        152 : Err_String := 'Drive not ready' ;
        153 : Err_String := 'Unknown command' ;
        154 : Err_String := 'CRC error in data' ;
        155 : Err_String := 'Bad request structure length' ;
        156 : Err_String := 'Disk seek error' ;
        157 : Err_String := 'Unknown media type' ;
        168 : Err_String := 'Sector not found' ;
        159 : Err_String := 'Printer out of paper' ;
        160 : Err_String := 'Device write fault' ;
        161 : Err_String := 'Device read fault' ;
        162 : Err_String := Text_Pascal_Err_String_162 ;

        { Fatal errors... }
        200 : Err_String := Text_Pascal_Err_String_200 ;
        201 : Err_String := Text_Pascal_Err_String_201 ;
        202 : Err_String := Text_Pascal_Err_String_202 ;
        203 : Err_String := Text_Pascal_Err_String_203 ;
        204 : Err_String := Text_Pascal_Err_String_204 ;
        205 : Err_String := Text_Pascal_Err_String_205 ;
        206 : Err_String := Text_Pascal_Err_String_206 ;
        207 : Err_String := Text_Pascal_Err_String_207 ;
        208 : Err_String := Text_Pascal_Err_String_208 ;
        209 : Err_String := Text_Pascal_Err_String_209 ;
        210 : Err_String := Text_Pascal_Err_String_210 ;
        211 : Err_String := Text_Pascal_Err_String_211 ;
        212 : Err_String := Text_Pascal_Err_String_212 ;
        213 : Err_String := Text_Pascal_Err_String_213 ;
        214 : Err_String := Text_Pascal_Err_String_214 ;
        215 : Err_String := Text_Pascal_Err_String_215 ;
        216 : Err_String := Text_Pascal_Err_String_216 ;
        217 : Err_String := Text_Pascal_Err_String_217 ;
        218 : Err_String := Text_Pascal_Err_String_218 ;
        219 : Err_String := Text_Pascal_Err_String_219 ;
        220 : Err_String := Text_Pascal_Err_String_220 ;
        221 : Err_String := Text_Pascal_Err_String_221 ;
        222 : Err_String := Text_Pascal_Err_String_222 ;
        223 : Err_String := Text_Pascal_Err_String_223 ;
        224 : Err_String := Text_Pascal_Err_String_224 ;
        225 : Err_String := Text_Pascal_Err_String_225 ;
        226 : Err_String := Text_Pascal_Err_String_226 ;
        227 : Err_String := Text_Pascal_Err_String_227 ;
        228 : Err_String := Text_Pascal_Err_String_228 ;
        229 : Err_String := Text_Pascal_Err_String_229 ;
        230 : Err_String := Text_Pascal_Err_String_230 ;
        231 : Err_String := Text_Pascal_Err_String_231 ;
        232 : Err_String := Text_Pascal_Err_String_232 ;
        else Err_String := 'Unused error number' ;
    end ;
    Pascal_ERT := Err_String ;
end ; { Pascal_ERT }


function DOS_ERT( Code : integer ) : Ansistring ;

var Err_String : Ansistring ;

begin
    case Code of
        0 : Err_String := Text_DOS_Error_Success ;
        1 : Err_String := Text_DOS_Error_1 ;
        2 : Err_String := Text_DOS_Error_2 ;
        3 : Err_String := Text_DOS_Error_3 ;
        4 : Err_String := Text_DOS_Error_4 ;
        5 : Err_String := Text_DOS_Error_5 ;
        6 : Err_String := Text_DOS_Error_6 ;
        7 : Err_String := Text_DOS_Error_7 ;
        8 : Err_String := Text_DOS_Error_8 ;
        9 : Err_String := Text_DOS_Error_9 ;
        10 : Err_String := Text_DOS_Error_10 ;
        11 : Err_String := Text_DOS_Error_11 ;
        12 : Err_String := Text_DOS_Error_12 ;
        13 : Err_String := Text_DOS_Error_13 ;
        14 : Err_String := Text_DOS_Error_14 ;
        15 : Err_String := Text_DOS_Error_15 ;
        16 : Err_String := Text_DOS_Error_16 ;
        17 : Err_String := Text_DOS_Error_17 ;
        18 : Err_String := Text_DOS_Error_18 ;
        19 : Err_String := Text_DOS_Error_19 ;
        20 : Err_String := Text_DOS_Error_20 ;
        21 : Err_String := Text_DOS_Error_21 ;
        22 : Err_String := Text_DOS_Error_22 ;
        23 : Err_String := Text_DOS_Error_23 ;
        24 : Err_String := Text_DOS_Error_24 ;
        25 : Err_String := Text_DOS_Error_25 ;
        26 : Err_String := Text_DOS_Error_26 ;
        27 : Err_String := Text_DOS_Error_27 ;
        28 : Err_String := Text_DOS_Error_28 ;
        29 : Err_String := Text_DOS_Error_29 ;
        30 : Err_String := Text_DOS_Error_30 ;
        31 : Err_String := Text_DOS_Error_31 ;
        32 : Err_String := Text_DOS_Error_32 ;
        33 : Err_String := Text_DOS_Error_33 ;
        34 : Err_String := Text_DOS_Error_34 ;
        35 : Err_String := Text_DOS_Error_35 ;
        36 : Err_String := Text_DOS_Error_36 ;
        38 : Err_String := Text_DOS_Error_38 ;
        50 : Err_String := Text_DOS_Error_50 ;
        51 : Err_String := Text_DOS_Error_51 ;
        52 : Err_String := Text_DOS_Error_52 ;
        53 : Err_String := Text_DOS_Error_53 ;
        54 : Err_String := Text_DOS_Error_54 ;
        55 : Err_String := Text_DOS_Error_55 ;
        56 : Err_String := Text_DOS_Error_56 ;
        57 : Err_String := Text_DOS_Error_57 ;
        58 : Err_String := Text_DOS_Error_58 ;
        59 : Err_String := Text_DOS_Error_59 ;
        60 : Err_String := Text_DOS_Error_60 ;
        61 : Err_String := Text_DOS_Error_61 ;
        62 : Err_String := Text_DOS_Error_62 ;
        63 : Err_String := Text_DOS_Error_63 ;
        64 : Err_String := Text_DOS_Error_64 ;
        65 : Err_String := Text_DOS_Error_65 ;
        66 : Err_String := Text_DOS_Error_66 ;
        67 : Err_String := Text_DOS_Error_67 ;
        68 : Err_String := Text_DOS_Error_68 ;
        69 : Err_String := Text_DOS_Error_69 ;
        70 : Err_String := Text_DOS_Error_70 ;
        71 : Err_String := Text_DOS_Error_71 ;
        72 : Err_String := Text_DOS_Error_72 ;
        80 : Err_String := Text_DOS_Error_80 ;
        81 : Err_String := Text_DOS_Error_81 ;
        82 : Err_String := Text_DOS_Error_82 ;
        83 : Err_String := Text_DOS_Error_83 ;
        84 : Err_String := Text_DOS_Error_84 ;
        85 : Err_String := Text_DOS_Error_85 ;
        86 : Err_String := Text_DOS_Error_86 ;
        87 : Err_String := Text_DOS_Error_87 ;
        88 : Err_String := Text_DOS_Error_88 ;
        89 : Err_String := Text_DOS_Error_89 ;
        90 : Err_String := Text_DOS_Error_90 ;
        91 : Err_String := Text_DOS_Error_91 ;
        92 : Err_String := Text_DOS_Error_92 ;
        93 : Err_String := Text_DOS_Error_93 ;
        95 : Err_String := Text_DOS_Error_95 ;
        100 : Err_String := Text_DOS_Error_100 ;
        101 : Err_String := Text_DOS_Error_101 ;
        102 : Err_String := Text_DOS_Error_102 ;
        103 : Err_String := Text_DOS_Error_103 ;
        104 : Err_String := Text_DOS_Error_104 ;
        105 : Err_String := Text_DOS_Error_105 ;
        106 : Err_String := Text_DOS_Error_106 ;
        107 : Err_String := Text_DOS_Error_107 ;
        108 : Err_String := Text_DOS_Error_108 ;
        109 : Err_String := Text_DOS_Error_109 ;
        110 : Err_String := Text_DOS_Error_110 ;
        111 : Err_String := Text_DOS_Error_111 ;
        112 : Err_String := Text_DOS_Error_112 ;
        113 : Err_String := Text_DOS_Error_113 ;
        114 : Err_String := Text_DOS_Error_114 ;
        115 : Err_String := Text_DOS_Error_115 ;
        116 : Err_String := Text_DOS_Error_116 ;
        117 : Err_String := Text_DOS_Error_117 ;
        118 : Err_String := Text_DOS_Error_118 ;
        119 : Err_String := Text_DOS_Error_119 ;
        120 : Err_String := Text_DOS_Error_120 ;
        121 : Err_String := Text_DOS_Error_121 ;
        122 : Err_String := Text_DOS_Error_122 ;
        123 : Err_String := Text_DOS_Error_123 ;
        124 : Err_String := Text_DOS_Error_124 ;
        125 : Err_String := Text_DOS_Error_125 ;
        126 : Err_String := Text_DOS_Error_126 ;
        127 : Err_String := Text_DOS_Error_127 ;
        128 : Err_String := Text_DOS_Error_128 ;
        129 : Err_String := Text_DOS_Error_129 ;
        130 : Err_String := Text_DOS_Error_130 ;
        131 : Err_String := Text_DOS_Error_131 ;
        132 : Err_String := Text_DOS_Error_132 ;
        133 : Err_String := Text_DOS_Error_133 ;
        134 : Err_String := Text_DOS_Error_134 ;
        135 : Err_String := Text_DOS_Error_135 ;
        136 : Err_String := Text_DOS_Error_136 ;
        137 : Err_String := Text_DOS_Error_137 ;
        138 : Err_String := Text_DOS_Error_138 ;
        139 : Err_String := Text_DOS_Error_139 ;
        140 : Err_String := Text_DOS_Error_140 ;
        141 : Err_String := Text_DOS_Error_141 ;
        142 : Err_String := Text_DOS_Error_142 ;
        143 : Err_String := Text_DOS_Error_143 ;
        144 : Err_String := Text_DOS_Error_144 ;
        145 : Err_String := Text_DOS_Error_145 ;
        146 : Err_String := Text_DOS_Error_146 ;
        147 : Err_String := Text_DOS_Error_147 ;
        148 : Err_String := Text_DOS_Error_148 ;
        149 : Err_String := Text_DOS_Error_149 ;
        150 : Err_String := Text_DOS_Error_150 ;
        151 : Err_String := Text_DOS_Error_151 ;
        152 : Err_String := Text_DOS_Error_152 ;
        153 : Err_String := Text_DOS_Error_153 ;
        154 : Err_String := Text_DOS_Error_154 ;
        155 : Err_String := Text_DOS_Error_155 ;
        156 : Err_String := Text_DOS_Error_156 ;
        157 : Err_String := Text_DOS_Error_157 ;
        158 : Err_String := Text_DOS_Error_158 ;
        159 : Err_String := Text_DOS_Error_159 ;
        160 : Err_String := Text_DOS_Error_160 ;
        161 : Err_String := Text_DOS_Error_161 ;
        162 : Err_String := Text_DOS_Error_162 ;
        163 : Err_String := Text_DOS_Error_163 ;
        164 : Err_String := Text_DOS_Error_164 ;
        165 : Err_String := Text_DOS_Error_165 ;
        180 : Err_String := Text_DOS_Error_180 ;
        181 : Err_String := Text_DOS_Error_181 ;
        182 : Err_String := Text_DOS_Error_182 ;
        183 : Err_String := Text_DOS_Error_183 ;
        184 : Err_String := Text_DOS_Error_184 ;
        185 : Err_String := Text_DOS_Error_185 ;
        186 : Err_String := Text_DOS_Error_186 ;
        187 : Err_String := Text_DOS_Error_187 ;
        188 : Err_String := Text_DOS_Error_188 ;
        189 : Err_String := Text_DOS_Error_189 ;
        190 : Err_String := Text_DOS_Error_190 ;
        191 : Err_String := Text_DOS_Error_191 ;
        192 : Err_String := Text_DOS_Error_192 ;
        193 : Err_String := Text_DOS_Error_193 ;
        194 : Err_String := Text_DOS_Error_194 ;
        195 : Err_String := Text_DOS_Error_195 ;
        196 : Err_String := Text_DOS_Error_196 ;
        197 : Err_String := Text_DOS_Error_197 ;
        198 : Err_String := Text_DOS_Error_198 ;
        199 : Err_String := Text_DOS_Error_199 ;
        200 : Err_String := Text_DOS_Error_200 ;
        201 : Err_String := Text_DOS_Error_201 ;
        202 : Err_String := Text_DOS_Error_202 ;
        203 : Err_String := Text_DOS_Error_203 ;
        204 : Err_String := Text_DOS_Error_204 ;
        205 : Err_String := Text_DOS_Error_205 ;
        206 : Err_String := Text_DOS_Error_206 ;
        207 : Err_String := Text_DOS_Error_207 ;
        208 : Err_String := Text_DOS_Error_208 ;
        209 : Err_String := Text_DOS_Error_209 ;
        210 : Err_String := Text_DOS_Error_210 ;
        211 : Err_String := Text_DOS_Error_211 ;
        212 : Err_String := Text_DOS_Error_212 ;
        213 : Err_String := Text_DOS_Error_213 ;
        214 : Err_String := Text_DOS_Error_214 ;
        215 : Err_String := Text_DOS_Error_215 ;
        216 : Err_String := Text_DOS_Error_216 ;

        { RAS error codes... }
        600 : Err_String := Text_DOS_Error_600 ;
        601 : Err_String := Text_DOS_Error_601 ;
        602 : Err_String := Text_DOS_Error_602 ;
        603 : Err_String := Text_DOS_Error_603 ;
        604 : Err_String := Text_DOS_Error_604 ;
        605 : Err_String := Text_DOS_Error_605 ;
        606 : Err_String := Text_DOS_Error_606 ;
        607 : Err_String := Text_DOS_Error_607 ;
        608 : Err_String := Text_DOS_Error_608 ;
        609 : Err_String := Text_DOS_Error_609 ;
        610 : Err_String := Text_DOS_Error_610 ;
        611 : Err_String := Text_DOS_Error_611 ;
        612 : Err_String := Text_DOS_Error_612 ;
        613 : Err_String := Text_DOS_Error_613 ;
        614 : Err_String := Text_DOS_Error_614 ;
        615 : Err_String := Text_DOS_Error_615 ;
        616 : Err_String := Text_DOS_Error_616 ;
        617 : Err_String := Text_DOS_Error_617 ;
        618 : Err_String := Text_DOS_Error_618 ;
        619 : Err_String := Text_DOS_Error_619 ;
        620 : Err_String := Text_DOS_Error_620 ;
        621 : Err_String := Text_DOS_Error_621 ;
        622 : Err_String := Text_DOS_Error_622 ;
        623 : Err_String := Text_DOS_Error_623 ;
        624 : Err_String := Text_DOS_Error_624 ;
        625 : Err_String := Text_DOS_Error_625 ;
        626 : Err_String := Text_DOS_Error_626 ;
        627 : Err_String := Text_DOS_Error_627 ;
        628 : Err_String := Text_DOS_Error_628 ;
        629 : Err_String := Text_DOS_Error_629 ;
        630 : Err_String := Text_DOS_Error_630 ;
        631 : Err_String := Text_DOS_Error_631 ;
        632 : Err_String := Text_DOS_Error_632 ;
        633 : Err_String := Text_DOS_Error_633 ;
        634 : Err_String := Text_DOS_Error_634 ;
        635 : Err_String := Text_DOS_Error_635 ;
        636 : Err_String := Text_DOS_Error_636 ;
        637 : Err_String := Text_DOS_Error_637 ;
        638 : Err_String := Text_DOS_Error_638 ;
        639 : Err_String := Text_DOS_Error_639 ;
        640 : Err_String := Text_DOS_Error_640 ;
        641 : Err_String := Text_DOS_Error_641 ;
        642 : Err_String := Text_DOS_Error_642 ;
        643 : Err_String := Text_DOS_Error_643 ;
        644 : Err_String := Text_DOS_Error_644 ;
        645 : Err_String := Text_DOS_Error_645 ;
        646 : Err_String := Text_DOS_Error_646 ;
        647 : Err_String := Text_DOS_Error_647 ;
        648 : Err_String := Text_DOS_Error_648 ;
        649 : Err_String := Text_DOS_Error_649 ;
        650 : Err_String := Text_DOS_Error_650 ;
        651 : Err_String := Text_DOS_Error_651 ;
        652 : Err_String := Text_DOS_Error_652 ;
        653 : Err_String := Text_DOS_Error_653 ;
        654 : Err_String := Text_DOS_Error_654 ;
        655 : Err_String := Text_DOS_Error_655 ;
        656 : Err_String := Text_DOS_Error_656 ;
        657 : Err_String := Text_DOS_Error_657 ;
        658 : Err_String := Text_DOS_Error_658 ;
        659 : Err_String := Text_DOS_Error_659 ;
        660 : Err_String := Text_DOS_Error_660 ;
        661 : Err_String := Text_DOS_Error_661 ;
        662 : Err_String := Text_DOS_Error_662 ;
        663 : Err_String := Text_DOS_Error_663 ;
        664 : Err_String := Text_DOS_Error_664 ;
        665 : Err_String := Text_DOS_Error_665 ;
        666 : Err_String := Text_DOS_Error_666 ;
        667 : Err_String := Text_DOS_Error_667 ;
        668 : Err_String := Text_DOS_Error_668 ;
        669 : Err_String := Text_DOS_Error_669 ;
        670 : Err_String := Text_DOS_Error_670 ;
        671 : Err_String := Text_DOS_Error_671 ;
        672 : Err_String := Text_DOS_Error_672 ;
        673 : Err_String := Text_DOS_Error_673 ;
        674 : Err_String := Text_DOS_Error_674 ;
        675 : Err_String := Text_DOS_Error_675 ;
        676 : Err_String := Text_DOS_Error_676 ;
        677 : Err_String := Text_DOS_Error_677 ;
        678 : Err_String := Text_DOS_Error_678 ;
        679 : Err_String := Text_DOS_Error_679 ;
        680 : Err_String := Text_DOS_Error_680 ;
        681 : Err_String := Text_DOS_Error_681 ;
        682 : Err_String := Text_DOS_Error_682 ;
        683 : Err_String := Text_DOS_Error_683 ;
        684 : Err_String := Text_DOS_Error_684 ;
        685 : Err_String := Text_DOS_Error_685 ;
        686 : Err_String := Text_DOS_Error_686 ;
        687 : Err_String := Text_DOS_Error_687 ;
        688 : Err_String := Text_DOS_Error_688 ;
        689 : Err_String := Text_DOS_Error_689 ;
        690 : Err_String := Text_DOS_Error_690 ;
        691 : Err_String := Text_DOS_Error_691 ;
        692 : Err_String := Text_DOS_Error_692 ;
        693 : Err_String := Text_DOS_Error_693 ;
        694 : Err_String := Text_DOS_Error_694 ;
        695 : Err_String := Text_DOS_Error_695 ;
        696 : Err_String := Text_DOS_Error_696 ;
        697 : Err_String := Text_DOS_Error_697 ;
        698 : Err_String := Text_DOS_Error_698 ;
        699 : Err_String := Text_DOS_Error_699 ;
        700 : Err_String := Text_DOS_Error_700 ;
        701 : Err_String := Text_DOS_Error_701 ;
        702 : Err_String := Text_DOS_Error_702 ;
        703 : Err_String := Text_DOS_Error_703 ;
        704 : Err_String := Text_DOS_Error_704 ;
        705 : Err_String := Text_DOS_Error_705 ;
        706 : Err_String := Text_DOS_Error_706 ;
        707 : Err_String := Text_DOS_Error_707 ;
        708 : Err_String := Text_DOS_Error_708 ;
        709 : Err_String := Text_DOS_Error_709 ;
        710 : Err_String := Text_DOS_Error_710 ;
        711 : Err_String := Text_DOS_Error_711 ;
        712 : Err_String := Text_DOS_Error_712 ;
        713 : Err_String := Text_DOS_Error_713 ;
        714 : Err_String := Text_DOS_Error_714 ;
        715 : Err_String := Text_DOS_Error_715 ;
        716 : Err_String := Text_DOS_Error_716 ;
        717 : Err_String := Text_DOS_Error_717 ;
        718 : Err_String := Text_DOS_Error_718 ;
        719 : Err_String := Text_DOS_Error_719 ;
        720 : Err_String := Text_DOS_Error_720 ;
        721 : Err_String := Text_DOS_Error_721 ;
        722 : Err_String := Text_DOS_Error_722 ;
        723 : Err_String := Text_DOS_Error_723 ;
        724 : Err_String := Text_DOS_Error_724 ;
        725 : Err_String := Text_DOS_Error_725 ;
        726 : Err_String := Text_DOS_Error_726 ;
        727 : Err_String := Text_DOS_Error_727 ;
        728 : Err_String := Text_DOS_Error_728 ;
        729 : Err_String := Text_DOS_Error_729 ;
        730 : Err_String := Text_DOS_Error_730 ;
        731 : Err_String := Text_DOS_Error_731 ;
        732 : Err_String := Text_DOS_Error_732 ;
        733 : Err_String := Text_DOS_Error_733 ;
        734 : Err_String := Text_DOS_Error_734 ;
        735 : Err_String := Text_DOS_Error_735 ;
        736 : Err_String := Text_DOS_Error_736 ;
        737 : Err_String := Text_DOS_Error_737 ;
        738 : Err_String := Text_DOS_Error_738 ;
        739 : Err_String := Text_DOS_Error_739 ;
        740 : Err_String := Text_DOS_Error_740 ;
        741 : Err_String := Text_DOS_Error_741 ;
        742 : Err_String := Text_DOS_Error_742 ;
        743 : Err_String := Text_DOS_Error_743 ;
        744 : Err_String := Text_DOS_Error_744 ;
        752 : Err_String := Text_DOS_Error_752 ;
        else
        Begin
            Str( Code, Err_String ) ;
            Err_String := Substitute1( Text_Error_Unused_Error_Number, Err_String ) ;
        end ;
    end ; { Case Code of }
    DOS_ERT := Err_String ;
end ; { DOS_ERT }


function Ert( Err : Integer ) : Ansistring ; { Return Pascal error text }

begin
    ERT := Pascal_ERT( Err ) ;
end ;


type TSCC_Unified_Exception = object( TSUnified_Exception )
                                  public
                                      function Get_Facility : longint ; virtual ;
                                      function Get_Facility_Version : longint ; virtual ;
                                      function Version : longint ; virtual ;
                                      function Severity : longint ;
                                          virtual ;
                                      function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                          virtual ;
                                      function Get_Error : longint ; virtual ;
                                      function Get_Previous : PsUnified_Exception ;
                                          virtual ;
                                      procedure Terminate ; virtual ;
                              end ;
     PSCC_Unified_Exception = ^TSCC_Unified_Exception ;

function TSCC_Unified_Exception.Get_Facility : longint ;

begin
    Get_Facility := 129 ;
end ;


function TSCC_Unified_Exception.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 10 ;
end ;


function TSCC_Unified_Exception.Version : longint ;

begin
    Version := 10 ;
end ;


function TSCC_Unified_Exception.Severity : longint ;

begin
    Severity := UE_Fatal ;
end ;


function TSCC_Unified_Exception.Error_Text( var Size, Typ : longint ) : Ansistring ;

begin
    Typ := 0 ;
    Size := 15 ;
    Error_Text := 'Abstract method' ;
end ;


function TSCC_Unified_Exception.Get_Error : longint ;

begin
    Get_Error := 1 ;
end ;


function TSCC_Unified_Exception.Get_Previous : PsUnified_Exception ;

begin
    Get_Previous := nil ;
end ;


procedure TSCC_Unified_Exception.Terminate ;

begin
    Free ;
end ;



function Get_Abstract_Error_Exception : PSUnified_Exception ;

var E : PSCC_Unified_Exception ;

begin
    new( E, Init ) ;
    Get_Abstract_Error_Exception := E ;
end ;



{$IFDEF DYNAMIC_OBJECTS}
type TCC_Unified_Exception = class( TStandard_Error_Interface )
                                 private
                                     Temp : Ansistring ;

                                 public
                                     function Get_Facility : longint ; override ;
                                     function Get_Facility_Version : longint ;
                                         override ;
                                     function Version : longint ; override ;
                                     function Severity : longint ;
                                         override ;
                                     function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                         override ;
                                     function Get_Error : longint ; override ;
                                     function Get_Previous : TUnified_Exception ;
                                         override ;
                                     procedure Terminate ; override ;
                             end ;

function TCC_Unified_Exception.Get_Facility : longint ;

begin
    Get_Facility := 129 ;
end ;


function TCC_Unified_Exception.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 10 ;
end ;


function TCC_Unified_Exception.Version : longint ;

begin
    Version := 10 ;
end ;


function TCC_Unified_Exception.Severity : longint ;

begin
    Severity := UE_Fatal ;
end ;


function TCC_Unified_Exception.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

begin
    Typ := 0 ;
    Size := 15 ;
    Temp := 'Abstract method' ;
    Error_Text := PAnsiChar( Temp ) ;
end ;


function TCC_Unified_Exception.Get_Error : longint ;

begin
    Get_Error := 1 ;
end ;


function TCC_Unified_Exception.Get_Previous : TUnified_Exception ;

begin
    Get_Previous := nil ;
end ;


procedure TCC_Unified_Exception.Terminate ;

begin
    Free ;
end ;
{$ENDIF}



function Create_Abstract_Error_Exception : TUnified_Exception ;

begin
    Result := TCC_Unified_Exception.Create ;
end ;


type TUE_Wrapper = class( TStandard_Error_Interface )
                       private // Instance data...
                           Temp : Ansistring ;
                           
                       public
                           UEC : PSUnified_Exception ;

                           function Get_Facility : longint ; override ;
                           function Get_Facility_Version : longint ;
                               override ;
                           function Version : longint ; override ;
                           function Severity : longint ; override ;
                           function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                               override ;
                           function Get_Error : longint ; override ;
                           function Get_Previous : tUnified_Exception ;
                               override ;
                           procedure Terminate ; override ;
                   end ;


function TUE_Wrapper.Get_Facility : longint ;

begin
    Result := UEC.Get_Facility ;
end ;


function TUE_Wrapper.Get_Facility_Version : longint ;

begin
    Result := UEC.Get_Facility_Version ;
end ;


function TUE_Wrapper.Version : longint ;

begin
    Result := UEC.Version ;
end ;


function TUE_Wrapper.Severity : longint ;

begin
    Result := UEC.Severity ;
end ;


function TUE_Wrapper.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

begin
    Temp := UEC.Error_Text( Size, Typ ) ;
    Result := PAnsiChar( Temp ) ;
end ;


function TUE_Wrapper.Get_Error : longint ;

begin
    Result := UEC.Get_Error ;
end ;


function TUE_Wrapper.Get_Previous : tUnified_Exception ;

begin
    Result := nil ;
end ;


procedure TUE_Wrapper.Terminate ;

begin
    UEC.Detach ;
    Free ;
end ;



function Create_UE_Wrapper( U : PSUnified_Exception ) : TUnified_Exception ;

begin
    if( U = nil ) then
    begin
        Result := nil ;
        exit ;
    end ;
    Result := TUE_Wrapper.Create ;
    TUE_Wrapper( Result ).UEC := U ;
    U.Attach ;
end ;



{ TStandard_Error_Interface methods... }

{ API... }

function TSStandard_Error_Interface.Get_Facility : longint ;

begin
    Get_Facility := -1 ;
end ;


function TSStandard_Error_Interface.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 0 ;
end ;


function TSStandard_Error_Interface.Severity : longint ;

begin
    Severity := UE_Error ;
end ;


function TSStandard_Error_Interface.Error_Text( var Size, Typ : longint ) : Ansistring ;

var P : Ansistring ;

begin
    P := '' ;
    Size := length( P ) ;
    Typ := 0 ; { 7-bit ASCII }
    Error_Text := P ;
end ;


function TSStandard_Error_Interface.Get_Error : longint ;

begin
    Get_Error := 0 ;
end ;


function TSStandard_Error_Interface.Get_Previous : psUnified_Exception ;

begin
    Get_Previous := nil ;
end ;


procedure TSStandard_Error_Interface.Terminate ;

begin
    if( Get_Previous <> nil ) then
    begin
        Get_Previous.Free ;
    end ;
end ;


procedure TSStandard_Error_Interface.Set_Memory_Type( MT : TMemory_Type ) ;

begin
    _Memory_Type := MT ;
end ;


function TSStandard_Error_Interface.Get_Memory_Type : TMemory_Type ;

begin
    Get_Memory_Type := _Memory_Type ;
end ;


{$IFDEF FPC_OR_DELPHI}
{ TStandard_Error_Interface methods... }

{ Constructors and destructors... }

  {$IFDEF CC_Track_Resources}
constructor TStandard_Error_Interface.Create ;

begin
    inherited Create ;

    Add_Resource( RT_Object, integer( self ) ) ;
end ;


destructor TStandard_Error_Interface.Destroy ;

begin
    Remove_Resource( RT_Object, integer( self ) ) ;

    inherited Destroy ;
end ;
  {$ENDIF}


{ API... }

function TStandard_Error_Interface.Interface_Version : longint ;

begin
    Interface_Version := 0 ;
end ;


function TStandard_Error_Interface.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := Name ;
    try
        Is_Class := ( S = Classname ) ;
    except
        Is_Class := False ;
    end ;
end ;


{$IFNDEF MOBILE}
function TStandard_Error_Interface.Debugger : TDebug_Interface ;

begin
    Debugger := nil ;
end ;
{$ENDIF}


function TStandard_Error_Interface.Get_Facility : longint ;

begin
    Get_Facility := -1 ;
end ;


function TStandard_Error_Interface.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 0 ;
end ;


function TStandard_Error_Interface.Version : longint ;

begin
    Version := 10 ; { Complies with UE V1.0 specification }
end ;


function TStandard_Error_Interface.Severity : longint ;

begin
    Severity := UE_Error ;
end ;
{$ENDIF}


function TStandard_Error_Interface.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

var P : PAnsiChar ;

begin
    P := nil ;
    if( P = nil ) then
    begin
        Size := 0 ;
    end else
    begin
        Size := strlen( P ) ;
    end ;
    Typ := 0 ; { 7-bit ASCII }
    Error_Text := P ;
end ;


function TStandard_Error_Interface.Get_Error : longint ;

begin
    Result := 0 ;
end ;


function TStandard_Error_Interface.Get_Previous : tUnified_Exception ;

begin
    Result := nil ;
end ;


function TStandard_Error_Interface.Initialize : TUnified_Exception ;

begin
    Initialize := nil ;
end ;


procedure TStandard_Error_Interface.Terminate ;

begin
    if( Get_Previous <> nil ) then
    begin
        Get_Previous.Terminate ;
    end ;
    Free ;
end ;


procedure TStandard_Error_Interface.Attach ;

begin
    inc( _Reference_Count ) ;
end ;


procedure TStandard_Error_Interface.Detach ;

begin
    dec( _Reference_Count ) ;
    if( _Reference_Count <= 0 ) then
    begin
        Free ;
    end ;
end ;


function Create_Simple_UE( F, FV, C, S : integer ; ET, Extra : Ansistring ) : TUnified_Exception ;

begin
    Result := TSimple_Error_Interface.Create ;
    TSimple_Error_Interface( Result )._Facility := F ;
    TSimple_Error_Interface( Result )._Facility_Version := FV ;
    TSimple_Error_Interface( Result )._Code := C ;
    TSimple_Error_Interface( Result )._Severity := S ;
    TSimple_Error_Interface( Result )._Error_Text := ET + Extra;
end ;


function Create_PSSimple_UE( F, FV, C, S : integer ; ET, Extra : Ansistring ) : PSUnified_Exception ;

var E : PSSimple_Error_Interface ;

begin
    E := new( PSSimple_Error_Interface, Init ) ;
    E._Facility := F ;
    E._Facility_Version := FV ;
    E._Code := C ;
    E._Severity := S ;
    E._Error_Text := ET + Extra ;
    Create_PSSimple_UE := E ;
end ;


function Create_MSDOS_UE( C : integer ; Extra : Ansistring ) : TUnified_Exception ;

begin
    Result := TMSDOS_Error_Interface.Create ;
    TMSDOS_Error_Interface( Result )._Code := C ;
    TMSDOS_Error_Interface( Result )._Additional_Text := Extra ;
end ;


// TSMSDOS_Error_Interface...

function TSMSDOS_Error_Interface.Get_Facility : longint ;

begin
    Get_Facility := 1 ; // MSDOS
end ;


function TSMSDOS_Error_Interface.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 40 ; // 4.0 (or later)
end ;


function TSMSDOS_Error_Interface.Severity : longint ;

begin
    Severity := UE_Error ;
end ;


function TSMSDOS_Error_Interface.Error_Text( var Size, Typ : longint ) : Ansistring ;

begin
    Result := DOS_ERT( _Code ) ;
    Size := length( Result ) ;
    Typ := 0 ;
end ;


function TSMSDOS_Error_Interface.Get_Error : longint ;

begin
    Get_Error := _Code ;
end ;



function Create_PSMSDOS_UE( C : integer ; Extra : Ansistring ) : PSUnified_Exception ;

var E : PSUnified_Exception ;

begin
    if( C = 0 ) then
    begin
        Create_PSMSDOS_UE := nil ;
        exit ;
    end ;
    E := new( PSMSDOS_Error_Interface, Init ) ;
    PSMSDOS_Error_Interface( E )._Code := C ;
    PSMSDOS_Error_Interface( E )._Additional_Text := Extra ;
    Create_PSMSDOS_UE := E ;
end ;


type TSPascal_Error_Interface = object( TSUnified_Exception )
                                     private
                                         _Code : integer ;
                                         _Additional_Text : Ansistring ;

                                     public
                                         function Get_Facility : longint ;
                                             virtual ;
                                         function Get_Facility_Version : longint ;
                                             virtual ;
                                         function Severity : longint ;
                                             virtual ;
                                         function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                             virtual ;
                                         function Get_Error : longint ;
                                             virtual ;
                                 end ;
     PSPascal_Error_Interface = ^TSPascal_Error_Interface ;

function TSPascal_Error_Interface.Get_Facility : longint ;

begin
    Get_Facility := 2 ; // Delphi
end ;


function TSPascal_Error_Interface.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 10 ; // 1.0 (or later)
end ;


function TSPascal_Error_Interface.Severity : longint ;

begin
    Severity := UE_Error ;
end ;


function TSPascal_Error_Interface.Error_Text( var Size, Typ : longint ) : Ansistring ;

begin
    Result := Pascal_ERT( _Code ) ;
    Size := length( Result ) ;
    Typ := 0 ;
end ;


function TSPascal_Error_Interface.Get_Error : longint ;

begin
    Get_Error := _Code ;
end ;



function Create_PSPascal_UE( C : integer ; Extra : Ansistring ) : PSUnified_Exception ;

var E : PSUnified_Exception ;

begin
    if( C = 0 ) then
    begin
        Create_PSPascal_UE := nil ;
        exit ;
    end ;
    E := new( PSPascal_Error_Interface, Init ) ;
    PSPascal_Error_Interface( E )._Code := C ;
    PSPascal_Error_Interface( E )._Additional_Text := Extra ;
    Create_PSPascal_UE := E ;
end ;



// TSimple_Error_Interface methods...

function TSimple_Error_Interface.Get_Facility : longint ;

begin
    Result := _Facility ;
end ;


function TSimple_Error_Interface.Get_Facility_Version : longint ;

begin
    Result := _Facility_Version ;
end ;


function TSimple_Error_Interface.Severity : longint ;

begin
    Result := _Severity ;
end ;


function TSimple_Error_Interface.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

begin
    Result := PAnsichar( _Error_Text ) ;
end ;


function TSimple_Error_Interface.Get_Error : longint ;

begin
    Result := _Code ;
end ;


// TMSDOS_Error_Interface methods...

function TMSDOS_Error_Interface.Get_Facility : longint ;

begin
    Result := 1 ; // MSDOS
end ;


function TMSDOS_Error_Interface.Get_Facility_Version : longint ;

begin
    Result := 4 ; // 4.0 (or later)
end ;


function TMSDOS_Error_Interface.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function TMSDOS_Error_Interface.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

begin
    Temp := DOS_ERT( _Code ) ;
    Error_Text := PAnsichar( Temp ) ;
end ;


function TMSDOS_Error_Interface.Get_Error : longint ;

begin
    Result := _Code ;
end ;


// TSSimple_Error_Interface methods...

function TSSimple_Error_Interface.Get_Facility : longint ;

begin
    Result := _Facility ;
end ;


function TsSimple_Error_Interface.Get_Facility_Version : longint ;

begin
    Result := _Facility_Version ;
end ;


function TSSimple_Error_Interface.Severity : longint ;

begin
    Result := _Severity ;
end ;


function TSSimple_Error_Interface.Error_Text( var Size, Typ : longint ) : Ansistring ;

begin
    Result := _Error_Text ;
end ;


function TSSimple_Error_Interface.Get_Error : longint ;

begin
    Result := _Code ;
end ;



end.

