{
        Program Name : Win_UE
        Package Name : Subroutine Library - Layer 1
        Purpose      : Unified Exceptions for Windows
        Institution  : Conroy & Conroy Co.
        Date Written : 10-Jan-1999
        Last Modified: 19-May-2002
        Written By   : Alan Conroy
        Version      : 1.2

        Copyright (C) 2014-2018 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides Unified Exception handling for Windows-specific errors.
}

unit win_ue ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

uses { C&C Subroutine Library - Layer 0 }
{$IFDEF CC_Track_Resources}
     ResMon,
{$ENDIF}
     _Common, { TMemory_Type }
     _DebugIn, { TDebug_Interface }
     _UE, { tUnified_Exception }
     UE ; { TStandard_Error_Interface }

function Create_Windows_UE( C : integer ; Extra : Ansistring ) : TUnified_Exception ;

{$IFDEF FPC_OR_DELPHI}
type TWindows_Error_Interface = class( TStandard_Error_Interface )
                                     private
                                         _Code : integer ;

                                     public
                                         function Get_Facility : longint ;
                                             override ;
                                         function Get_Facility_Version : longint ;
                                             override ;
                                         function Severity : longint ;
                                             override ;
                                         function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                             override ;
                                         function Get_Error : longint ;
                                             override ;
                                 end ;
{$ENDIF}

function Create_PSWindows_UE( C : integer ; Extra : Ansistring ) : PSUnified_Exception ;


implementation

{$IFDEF FPC_OR_DELPHI}
uses { Borland... }
     Windows, { FormatMessage }
     Sysutils, { strpas }
{$ELSE}
uses Strings { strlen }
{$ENDIF}
     { Subroutine Library - Layer 0... }
     L10n ; { Text_* }

function Create_Windows_UE( C : integer ; Extra : Ansistring ) : TUnified_Exception ;

begin
    Result := TWindows_Error_Interface.Create ;
    TWindows_Error_Interface( Result )._Code := C ;
    TWindows_Error_Interface( Result )._Additional_Text := Extra ;
end ;



type TSWindows_Error_Interface = object( TSUnified_Exception )
                                     private
                                         _Code : integer ;
                                         _Additional_Text : Ansistring ;

                                     public
                                         function Get_Facility : longint ;
                                             virtual ;
                                         function Get_Facility_Version : longint ;
                                             virtual ;
                                         function Severity : longint ;
                                             virtual ;
                                         function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                             virtual ;
                                         function Get_Error : longint ;
                                             virtual ;
                                 end ;
     PSWindows_Error_Interface = ^TSWindows_Error_Interface ;

function TSWindows_Error_Interface.Get_Facility : longint ;

begin
    Get_Facility := 0 ; // Windows
end ;


function TSWindows_Error_Interface.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 10 ; // 1.0 (or later)
end ;


function TSWindows_Error_Interface.Severity : longint ;

begin
    Severity := UE_Error ;
end ;


function TSWindows_Error_Interface.Error_Text( var Size, Typ : longint ) : Ansistring ;

var S : string[ 128 ] ;

begin
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_IGNORE_INSERTS,
        nil, _Code, 0, @S[ 1 ], 128, nil ) ;
    S[ 128 ] := #0 ; // Make sure it ends with null
    Result := S + ' ' + _Additional_Text ;
    Size := length( Result ) ;
    Typ := 0 ;
end ;


function TSWindows_Error_Interface.Get_Error : longint ;

begin
    Get_Error := _Code ;
end ;



function Create_PSWindows_UE( C : integer ; Extra : Ansistring ) : PSUnified_Exception ;

var E : PSUnified_Exception ;

begin
    if( C = 0 ) then
    begin
        Create_PSWindows_UE := nil ;
        exit ;
    end ;
    E := new( PSWindows_Error_Interface, Init ) ;
    PSMSDOS_Error_Interface( E )._Code := C ;
    PSMSDOS_Error_Interface( E )._Additional_Text := Extra ;
    Create_PSWindows_UE := E ;
end ;



// TWindows_Error_Interface methods...

function TWindows_Error_Interface.Get_Facility : longint ;

begin
    Result := 0 ; // Windows
end ;


function TWindows_Error_Interface.Get_Facility_Version : longint ;

begin
    Result := 4 ; // 4.0 (or later)
end ;


function TWindows_Error_Interface.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function TWindows_Error_Interface.Error_Text( var Size, Typ : longint ) : PAnsiChar ;

var S : string[ 128 ] ;

begin
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_IGNORE_INSERTS,
        nil, _Code, 0, @S[ 1 ], 128, nil ) ;
    S[ 128 ] := #0 ; // Make sure it ends with null
    Temp := S + _Additional_Text ;
    Size := length( Temp ) ;
    Typ := 0 ;
    Error_Text := PAnsichar( Temp ) ;
end ;


function TWindows_Error_Interface.Get_Error : longint ;

begin
    Result := _Code ;
end ;



end.

